Package: R.ComDim
Type: Package
Title: Common Dimensions (ComDim) Multi-Block Analysis
Version: 1.0.0
Authors@R: person(given = "Francesc",
                  family = "Puig-Castellví",
                  role = c("aut", "cre"),
                  email = "puig.francesc@gmail.com")
Description: 
    Common Dimensions (ComDim) is a multi-block method that simultaneously
    considers multiple data tables to find latent components that are common
    to all the tables as well as those specific to each data table, along
    with the contribution of each table to each component.
    See Jouan-Rimbaud Bouveresse and Rutledge (2024)
    <doi:10.1002/cem.3454>, Boccard and Rutledge (2013)
    <doi:10.1016/j.aca.2013.01.022>, and Puig-Castellví et al. (2021)
    <doi:10.1016/j.chemolab.2021.104422>.
License: MIT + file LICENSE
Depends: methods
Imports: pracma, utils, ggplot2, ConsensusOPLS
Suggests: gridExtra, imputeLCMD, ica, MultiAssayExperiment,
        SummarizedExperiment, ropls
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Collate: 'R.ComDim-package.R' 'ComDim-classes.R' 'AddMetadata.R'
        'ComDim_Exploratory.R' 'ComDim_PCA.R' 'ComDim_PLS.R'
        'ComDim_OPLS.R' 'ComDim_y.R' 'ExpandMultiBlock.R'
        'FilterSamplesMultiBlock.R' 'NAInfRemoveMultiBlock.R'
        'MakeComDimLoadingsTable.R' 'MakeComDimScoresTable.R'
        'MultiAssayExperiment2MultiBlock.R' 'MultiBlock.R'
        'MultiBlock2MultiAssayExperiment.R' 'MultiBlock2Matrix.R'
        'NormalizeMultiBlock.R' 'PredictMultiBlock.R'
        'ProcessMultiBlock.R' 'SelectFeaturesRW.R'
        'SimulateMultiBlock.R' 'SplitRW.R'
        'SummarizedExperiment2MultiBlock.R'
        'data-KEGG_table_metabolites.R' 'data-RNAseq3.R' 'data-extra.R'
        'data-gcms.R' 'data-intra.R' 'data-lcms.R' 'data-lipids.R'
        'data-metadata_RNAseq3.R' 'data-metadata_lipids.R'
        'data-mirnaseq.R' 'data-rnaseq.R'
NeedsCompilation: no
Packaged: 2026-05-08 14:28:08 UTC; Xisco
Author: Francesc Puig-Castellví [aut, cre]
Maintainer: Francesc Puig-Castellví <puig.francesc@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-13 07:40:07 UTC
Built: R 4.5.2; ; 2026-05-13 09:56:45 UTC; unix

# Generated by roxygen2: do not edit by hand

export("blockNames<-")
export("sampleNames<-")
export("variableNames<-")
export(AddMetadata)
export(ComDim_Exploratory)
export(ComDim_OPLS)
export(ComDim_PCA)
export(ComDim_PLS)
export(ComDim_y)
export(ExpandMultiBlock)
export(FilterSamplesMultiBlock)
export(MakeComDimLoadingsTable)
export(MakeComDimScoresTable)
export(MultiAssayExperiment2MultiBlock)
export(MultiBlock)
export(MultiBlock2Matrix)
export(MultiBlock2MultiAssayExperiment)
export(NAInfRemoveMultiBlock)
export(NormalizeMultiBlock)
export(OPLS_NIPALS_DNR)
export(PredictMultiBlock)
export(ProcessMultiBlock)
export(SelectFeaturesRW)
export(SimulateMultiBlock)
export(SplitRW)
export(SummarizedExperiment2MultiBlock)
export(blockNames)
export(sampleNames)
export(variableNames)
exportMethods("blockNames<-")
exportMethods("sampleNames<-")
exportMethods("variableNames<-")
exportMethods(blockNames)
exportMethods(ncol)
exportMethods(nrow)
exportMethods(sampleNames)
exportMethods(variableNames)
importFrom(methods,hasArg)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setClassUnion)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,validObject)
importFrom(stats,median)
importFrom(stats,qnorm)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(utils,globalVariables)

#' MultiBlock2Matrix
#'
#' Combines the blocks of a MultiBlock into a single matrix by column-binding the selected blocks.
#' Useful for computing summary statistics across the whole MultiBlock (e.g. maximum value).
#' @param MB A \code{MultiBlock} object.
#' @param blocks The blocks to combine. A vector of integers (block indices) or a vector of block
#'   names. When omitted, all blocks are included.
#' @param vars The variables to keep. A list of the same length as \code{blocks}. Each element
#'   contains the indices (integer) or names (character) of the variables to retain from the
#'   corresponding block. When omitted, all variables in each block are included.
#' @details
#'   Blocks are column-bound in the order given by \code{blocks}. Row names of the output matrix
#'   are the sample names of \code{MB}. Column names are the variable names; if the same variable
#'   name appears in more than one block, duplicates are disambiguated by appending
#'   \code{.<block_name>} to all occurrences of the repeated name, and a warning is issued.
#' @return A numeric matrix with rows corresponding to samples and columns corresponding to the
#'   variables of the selected blocks concatenated in order.
#' @seealso \code{\link{MultiBlock}}, \code{\link{ProcessMultiBlock}}
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' # Combine all blocks into a matrix and compute a summary statistic:
#' mat <- MultiBlock2Matrix(mb)
#' max(mat)
#' # Combine only the first block, keeping variables 1-10:
#' mat <- MultiBlock2Matrix(mb, blocks = 1, vars = list(1:10))
#' @export
MultiBlock2Matrix <- function(MB = MB,
                              blocks = NULL,
                              vars = NULL) {
  if (!("MultiBlock" %in% class(MB))) {
    stop("'MB' is not a MultiBlock object.")
  }

  if (!(is.null(blocks))) {
    if (is.character(blocks)) {
      block.ids <- match(blocks, names(MB@Data))
      if (all(is.na(block.ids))) {
        stop("None of the block names were found in the MultiBlock.")
      }
      if (any(is.na(block.ids))) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(blocks[which(is.na(block.ids))], sep = ", ")
        )
        block.ids <- block.ids[!is.na(block.ids)]
      }
    } else if (is.numeric(blocks)) {
      block.ids <- intersect(seq_along(MB@Data), blocks)
      if (length(block.ids) == 0) {
        stop("None of the block names were found in the MultiBlock.")
      }
      if (length(blocks) != length(block.ids)) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(setdiff(seq_along(MB@Data), blocks), sep = ", ")
        )
      }
    }
  } else {
    block.ids <- seq(1, length(MB@Data))
  }

  if (!is.null(vars)) {
    if (!is.list(vars)) {
      stop("'variables' must be a list object.")
    } else if (length(vars) != length(block.ids)) {
      stop("'variables' should have the same length as 'blocks'.")
    }

    for (j in seq_along(block.ids)) {
      i <- block.ids[j]
      if (is.character(vars[[j]])) {
        xx <- match(vars[[j]], MB@Variables[[i]])
        if (any(is.na(xx))) {
          warning(sprintf(
            "Some variable names were not found in block %s: %s", i,
            paste0(vars[[j]][which(is.na(xx))], sep = ", ")
          ))
          vars[[j]] <- xx[!is.na(xx)]
        } else {
          vars[[j]] <- xx
        }
      } else if (is.numeric(vars[[j]])) {
        xx <- intersect(seq_along(MB@Variables[[i]]), vars[[j]])
        if (length(xx) != length(vars[[j]])) {
          warning(sprintf(
            "Some variables were not found in the MultiBlock object %s: %s", i,
            paste0(setdiff(vars[[j]], seq_along(MB@Variables[[i]])), sep = ", ")
          ))
        }
        vars[[j]] <- xx
      }
    }
  } else {
    vars <- lapply(block.ids, function(i) seq_along(MB@Variables[[i]]))
  }

  cnames <- vector()
  rep_names <- vector()
  block_names <- vector()
  common <- FALSE

  for (j in seq_along(block.ids)) {
    i <- block.ids[j]
    if (j == 1) {
      mat <- MB@Data[[i]][, vars[[j]], drop = FALSE]
    } else {
      mat <- cbind(mat, MB@Data[[i]][, vars[[j]], drop = FALSE])
    }

    names_new <- MB@Variables[[i]][vars[[j]]]
    if (any(names_new %in% cnames)) {
      common <- TRUE
      rep_names <- append(rep_names, names_new[which(names_new %in% cnames)])
      names_new[which(names_new %in% cnames)] <-
        paste0(
          names_new[which(names_new %in% cnames)],
          sprintf(".%s", names(MB@Data)[i])
        )
    }
    block_names <- append(block_names, rep(names(MB@Data)[i], length(names_new)))
    cnames <- append(cnames, names_new)
  }

  # Now replace the name of the variables that had a replicated name.
  if (length(rep_names) != 0) {
    unique_rep_names <- as.character(unique(rep_names))
    cnames <- as.character(cnames)
    for (i in unique_rep_names) {
      pos <- which(cnames == i)
      cnames[pos] <- paste0(i, ".", block_names[match(i, cnames)])
    }
  }

  if (common) {
    warning("Some variable names were repeated across blocks.
  The repeated variable names were renamed.")
  }

  rownames(mat) <- MB@Samples
  colnames(mat) <- cnames

  return(mat)
}

#' NAInfRemoveMultiBlock
#'
#' Remove NA and Infinite values from a MultiBlock object in a single pass.
#' Variables whose combined count of NA and Infinite values meets or exceeds
#' \code{minfrac * nrow} are discarded first. Remaining Infinite values are
#' then replaced according to \code{inf.method}; remaining NA values are
#' imputed according to \code{na.method}.
#' @param MB The MultiBlock object.
#' @param blocks Blocks to process: a vector of integers or block names
#'   (optional; all blocks are processed when omitted).
#' @param minfrac Minimum fraction of valid (non-NA, finite) values required
#'   to retain a variable. Variables at or below this threshold are discarded.
#'   Default: 0.5.
#' @param na.method Imputation method for remaining NA values after the
#'   minfrac filter. One of: \code{'none'} (keep NAs), \code{'zero'},
#'   \code{'median'}, \code{'discard'} (drop columns still containing NAs),
#'   \code{'fixed.value'}, \code{'fixed.value.all'}, \code{'fixed.noise'},
#'   \code{'random.noise'}, \code{'QRILC'}. Default: \code{'random.noise'}.
#' @param inf.method Replacement method for remaining Infinite values after
#'   the minfrac filter. One of: \code{'none'} (keep Inf),
#'   \code{'fixed.noise'}, \code{'random.noise'}.
#'   Default: \code{'random.noise'}.
#' @param constant For \code{na.method = 'fixed.value'}: value to replace NAs.
#'   For \code{na.method = 'fixed.value.all'}: value added to every element
#'   (NAs are then set to this value). Default: 0.
#' @param factor.NA Noise factor used by \code{'fixed.noise'} and
#'   \code{'random.noise'} for both NA and Inf imputation. For NA and -Inf:
#'   the replacement mean equals \code{col_min * factor.NA}. For +Inf:
#'   \code{col_max / factor.NA}. Default: 0.5.
#' @param sd.noise Standard-deviation factor for \code{'random.noise'}:
#'   \code{SD = abs(mean * sd.noise)}. Default: 0.3.
#' @param tune.sigma Tuning scalar for \code{na.method = 'QRILC'}. Default:
#'   1 (Gaussian complete-data assumption).
#' @param showWarning If \code{TRUE}, emit a warning when variables are
#'   discarded by the minfrac filter. Default: \code{TRUE}.
#' @return The processed MultiBlock object.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' b2[c(2, 3, 5), c(1, 2, 3)] <- NA
#' b2[c(1, 4), c(4, 5)] <- Inf
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' mb <- NAInfRemoveMultiBlock(mb, na.method = "zero", inf.method = "fixed.noise")
#' @export
NAInfRemoveMultiBlock <- function(MB,
                                  blocks = NULL,
                                  minfrac = 0.5,
                                  na.method = c(
                                    "none", "zero", "median", "discard",
                                    "fixed.value", "fixed.value.all",
                                    "fixed.noise", "random.noise", "QRILC"
                                  )[8],
                                  inf.method = c("none", "fixed.noise", "random.noise")[3],
                                  constant = 0,
                                  factor.NA = 0.5,
                                  sd.noise = 0.3,
                                  tune.sigma = 1,
                                  showWarning = TRUE) {
  # ---- Input validation -------------------------------------------------- #

  if (!("MultiBlock" %in% class(MB))) {
    stop("'MB' is not a MultiBlock object.")
  }

  if (is.null(na.method) || length(na.method) > 1) {
    stop(
      "Provide a single na.method: 'none', 'zero', 'median', 'discard', ",
      "'fixed.value', 'fixed.value.all', 'fixed.noise', 'random.noise', or 'QRILC'."
    )
  }
  na.method <- tolower(na.method)
  if (!(na.method %in% c(
    "none", "zero", "median", "discard", "fixed.value",
    "fixed.value.all", "fixed.noise", "random.noise", "qrilc"
  ))) {
    stop(
      "Invalid na.method. Choose from: 'none', 'zero', 'median', 'discard', ",
      "'fixed.value', 'fixed.value.all', 'fixed.noise', 'random.noise', 'QRILC'."
    )
  }
  if (na.method == "qrilc" && !requireNamespace("imputeLCMD", quietly = TRUE)) {
    stop("Package 'imputeLCMD' is required for na.method = 'QRILC'. ",
         "Please install it with: install.packages('imputeLCMD').")
  }

  if (is.null(inf.method) || length(inf.method) > 1) {
    stop("Provide a single inf.method: 'none', 'fixed.noise', or 'random.noise'.")
  }
  inf.method <- tolower(inf.method)
  if (!(inf.method %in% c("none", "fixed.noise", "random.noise"))) {
    stop("Invalid inf.method. Choose from: 'none', 'fixed.noise', 'random.noise'.")
  }

  # ---- Resolve block indices --------------------------------------------- #

  if (!is.null(blocks)) {
    if (is.character(blocks)) {
      block.ids <- match(blocks, names(MB@Data))
      missing_b <- blocks[is.na(block.ids)]
      if (length(missing_b) > 0) {
        warning(sprintf(
          "Block name(s) not found in the MultiBlock: %s",
          paste(missing_b, collapse = ", ")
        ))
        block.ids <- block.ids[!is.na(block.ids)]
      }
    } else if (is.numeric(blocks)) {
      block.ids <- intersect(seq_along(MB@Data), blocks)
      missing_b <- setdiff(blocks, seq_along(MB@Data))
      if (length(missing_b) > 0) {
        warning(sprintf(
          "Block index/indices not found in the MultiBlock: %s",
          paste(missing_b, collapse = ", ")
        ))
      }
    } else {
      stop("'blocks' must be a character or numeric vector.")
    }
  } else {
    block.ids <- seq_along(MB@Data)
  }

  if (length(block.ids) == 0) {
    warning("0 blocks were selected. Data was not modified.")
    return(MB)
  }

  # ---- Validate and convert minfrac to an absolute count ----------------- #

  if (is.null(minfrac) || minfrac < 0) {
    if (showWarning) warning("'minfrac' was set to 0.")
    minfrac <- 0
  } else if (minfrac > 1) {
    if (showWarning) warning("'minfrac' was set to 1.")
    minfrac <- 1
  }
  minfrac_n <- round(nrow(MB@Data[[block.ids[1]]]) * minfrac)

  # ---- Step 1: discard variables with too many NA / Inf values ----------- #

  n_discarded <- 0L
  to_delete <- integer(0)

  for (i in block.ids) {
    if (any(is.complex(MB@Data[[i]]))) {
      stop(sprintf(
        "Block '%s' contains complex values that cannot be processed.",
        names(MB@Data)[i]
      ))
    }
    if (any(is.logical(MB@Data[[i]]))) {
      stop(sprintf(
        "Block '%s' contains logical values that cannot be processed.",
        names(MB@Data)[i]
      ))
    }

    bad_n <- apply(MB@Data[[i]], 2, function(x) sum(is.na(x) | is.infinite(x)))
    drop_idx <- which(bad_n >= minfrac_n)

    if (length(drop_idx) > 0) {
      keep_idx <- setdiff(seq_len(ncol(MB@Data[[i]])), drop_idx)
      MB@Data[[i]] <- MB@Data[[i]][, keep_idx, drop = FALSE]
      MB@Variables[[i]] <- MB@Variables[[i]][keep_idx]
      n_discarded <- n_discarded + length(drop_idx)
    }

    if (ncol(MB@Data[[i]]) == 0) {
      to_delete <- c(to_delete, i)
    }
  }

  if (length(to_delete) > 0) {
    to_delete <- sort(to_delete, decreasing = TRUE)
    names_to_delete <- names(MB@Data)[to_delete]
    old_names <- names(MB@Data)
    for (i in to_delete) {
      nm <- names(MB@Data)[i]
      if (nm %in% names(MB@Batch)) MB@Batch[[nm]] <- NULL
      if (nm %in% names(MB@Metadata)) MB@Metadata[[nm]] <- NULL
      MB@Data[[i]] <- NULL
      MB@Variables[[i]] <- NULL
    }
    kept_names <- setdiff(old_names[block.ids], names_to_delete)
    if (length(kept_names) == 0) {
      stop("All selected blocks contained only NA/Infinite values and were removed.")
    }
    block.ids <- match(kept_names, names(MB@Data))
  }

  if (n_discarded > 0L && showWarning) {
    warning(sprintf(
      "%d variable(s) exceeded the NA/Infinite threshold and were deleted.",
      n_discarded
    ))
  }

  # ---- Step 2: replace remaining Infinite values ------------------------- #

  if (inf.method != "none") {
    for (i in block.ids) {
      for (nc in seq_len(ncol(MB@Data[[i]]))) {
        col <- MB@Data[[i]][, nc]
        if (!any(is.infinite(col))) next

        finite_vals <- col[is.finite(col) & !is.na(col)]
        col_min <- if (length(finite_vals) > 0) min(finite_vals) else 0
        col_max <- if (length(finite_vals) > 0) max(finite_vals) else 0
        neg_pos <- which(is.infinite(col) & col < 0)
        pos_pos <- which(is.infinite(col) & col > 0)

        if (inf.method == "fixed.noise") {
          if (length(neg_pos) > 0) {
            MB@Data[[i]][neg_pos, nc] <- col_min * factor.NA
          }
          if (length(pos_pos) > 0) {
            MB@Data[[i]][pos_pos, nc] <- col_max * (1 / factor.NA)
          }
        } else { # random.noise
          if (length(neg_pos) > 0) {
            m <- col_min * factor.NA
            MB@Data[[i]][neg_pos, nc] <- rnorm(length(neg_pos),
              mean = m,
              sd   = abs(m * sd.noise)
            )
          }
          if (length(pos_pos) > 0) {
            m <- col_max * (1 / factor.NA)
            MB@Data[[i]][pos_pos, nc] <- rnorm(length(pos_pos),
              mean = m,
              sd   = abs(m * sd.noise)
            )
          }
        }
      }
    }
  }

  # ---- Step 3: impute remaining NA values -------------------------------- #

  if (na.method == "none") {
    return(MB)
  }

  for (i in block.ids) {
    if (na.method == "zero") {
      MB@Data[[i]][is.na(MB@Data[[i]])] <- 0
    } else if (na.method == "median") {
      for (nc in seq_len(ncol(MB@Data[[i]]))) {
        na_pos <- which(is.na(MB@Data[[i]][, nc]))
        if (length(na_pos) > 0) {
          MB@Data[[i]][na_pos, nc] <- median(MB@Data[[i]][, nc], na.rm = TRUE)
        }
      }
    } else if (na.method == "fixed.value") {
      MB@Data[[i]][is.na(MB@Data[[i]])] <- constant
    } else if (na.method == "fixed.value.all") {
      MB@Data[[i]] <- MB@Data[[i]] + constant # adds to non-NA elements
      MB@Data[[i]][is.na(MB@Data[[i]])] <- constant # fills former NAs
    } else if (na.method == "discard") {
      has_na <- apply(MB@Data[[i]], 2, anyNA)
      keep_cols <- which(!has_na)
      if (length(keep_cols) == 0) {
        nm <- names(MB@Data)[i]
        if (nm %in% names(MB@Batch)) MB@Batch[[nm]] <- NULL
        if (nm %in% names(MB@Metadata)) MB@Metadata[[nm]] <- NULL
        MB@Data[[i]] <- NULL
        MB@Variables[[i]] <- NULL
        warning(sprintf("Block '%s' was removed because all columns still contained NA values.", nm))
      } else {
        MB@Data[[i]] <- MB@Data[[i]][, keep_cols, drop = FALSE]
        MB@Variables[[i]] <- MB@Variables[[i]][keep_cols]
      }
    } else if (na.method == "fixed.noise") {
      for (nc in seq_len(ncol(MB@Data[[i]]))) {
        na_pos <- which(is.na(MB@Data[[i]][, nc]))
        if (length(na_pos) > 0) {
          col_min <- min(MB@Data[[i]][, nc], na.rm = TRUE)
          MB@Data[[i]][na_pos, nc] <- col_min * factor.NA
        }
      }
    } else if (na.method == "random.noise") {
      for (nc in seq_len(ncol(MB@Data[[i]]))) {
        na_pos <- which(is.na(MB@Data[[i]][, nc]))
        if (length(na_pos) > 0) {
          col_min <- min(MB@Data[[i]][, nc], na.rm = TRUE)
          m <- col_min * factor.NA
          MB@Data[[i]][na_pos, nc] <- rnorm(length(na_pos),
            mean = m,
            sd   = abs(m * sd.noise)
          )
        }
      }
    } else if (na.method == "qrilc") {
      for (nc in seq_len(ncol(MB@Data[[i]]))) {
        if (anyNA(MB@Data[[i]][, nc])) {
          MB@Data[[i]][, nc] <- imputeLCMD::impute.QRILC(
            as.matrix(MB@Data[[i]][, nc]),
            tune.sigma = tune.sigma
          )[[1]]
        }
      }
    }
  }

  return(MB)
}

#' SimulateMultiBlock
#'
#' Generate a synthetic \code{MultiBlock} dataset built from a known number of
#' orthogonal latent sources plus Gaussian noise.  Useful for benchmarking and
#' testing ComDim functions.
#'
#' The dataset is constructed as follows:
#' \enumerate{
#'   \item \code{n_sources} score vectors (\eqn{n \times \text{n\_sources}}) are
#'     drawn from a standard normal distribution and orthonormalised by QR
#'     decomposition.
#'   \item Loading vectors (\eqn{\text{n\_sources} \times p}) are built so that
#'     each source loads primarily (SD = 1) on one equal-sized variable segment,
#'     with small cross-loadings (SD = 0.10) on the remaining variables.
#'   \item The true signal \eqn{X = TP} is computed.
#'   \item Gaussian noise is added such that
#'     \eqn{\text{noise\_var} = \text{signal\_var} \times
#'     \text{noise} / (1 - \text{noise})}.
#'   \item The \eqn{p} variables are split into \code{n_blocks} equal-width
#'     blocks, each assembled as a named element of the returned
#'     \code{MultiBlock}.
#' }
#'
#' @param n Number of samples. Default: \code{500}.
#' @param p Total number of variables (split evenly across blocks). Must be
#'   divisible by \code{n_blocks} and by \code{n_sources}. Default: \code{2000}.
#' @param n_sources Number of orthogonal latent sources. Default: \code{4}.
#' @param noise Fraction of total variance attributed to noise, in (0, 1).
#'   Default: \code{0.05} (5 \% noise).
#' @param n_blocks Number of blocks to split the variables into. Default:
#'   \code{2}.
#' @return A \code{\link{MultiBlock}} object with \code{n_blocks} blocks, each
#'   of size \eqn{n \times (p / \text{n\_blocks})}, named \code{"Block1"},
#'   \code{"Block2"}, etc.
#' @seealso \code{\link{ComDim_PCA}}, \code{\link{MultiBlock}}
#' @examples
#' mb <- SimulateMultiBlock(n = 100, p = 200, n_sources = 4,
#'                          noise = 0.05, n_blocks = 2)
#' mb <- NormalizeMultiBlock(mb, method = 'norm')
#' res <- ComDim_PCA(mb, ndim = 4)
#' @export
SimulateMultiBlock <- function(n         = 500L,
                               p         = 2000L,
                               n_sources = 4L,
                               noise     = 0.05,
                               n_blocks  = 2L) {
  if (p %% n_blocks != 0L)
    stop("'p' must be divisible by 'n_blocks'.")
  if (p %% n_sources != 0L)
    stop("'p' must be divisible by 'n_sources'.")
  if (noise <= 0 || noise >= 1)
    stop("'noise' must be strictly between 0 and 1.")

  # Orthonormal score matrix (n x n_sources)
  T_orth <- qr.Q(qr(matrix(rnorm(n * n_sources), nrow = n, ncol = n_sources)))

  # Loading matrix (n_sources x p): each source loads primarily on one segment
  segment <- p / n_sources
  P <- matrix(0, nrow = n_sources, ncol = p)
  for (k in seq_len(n_sources)) {
    idx   <- ((k - 1L) * segment + 1L):(k * segment)
    other <- setdiff(seq_len(p), idx)
    P[k, idx]   <- rnorm(segment, sd = 1)
    P[k, other] <- rnorm(length(other), sd = 0.10)
  }

  X_signal   <- T_orth %*% P
  signal_var <- var(as.vector(X_signal))
  noise_var  <- signal_var * noise / (1 - noise)
  X          <- X_signal + matrix(rnorm(n * p, sd = sqrt(noise_var)),
                                  nrow = n, ncol = p)

  rownames(X) <- paste0("S", seq_len(n))
  colnames(X) <- paste0("V", seq_len(p))

  # Split into blocks
  block_width <- p / n_blocks
  blocks <- vector("list", n_blocks)
  names(blocks) <- paste0("Block", seq_len(n_blocks))
  for (b in seq_len(n_blocks)) {
    cols <- ((b - 1L) * block_width + 1L):(b * block_width)
    blocks[[b]] <- X[, cols, drop = FALSE]
  }

  MultiBlock(Data = blocks)
}

#' miRNAseq data on cell differentiation
#'
#'
#' @name mirnaseq
#'
#' @docType data
#'
#' @usage data(dataset3)
#'
#' @keywords dataset
#'
#' @author  Cabrero et al. (2019) Scientific data 6:256
#' (\doi{10.1038/s41597-019-0202-7})
#'
#' @source \href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE75394}{NCBI}
#'
#' @examples
#' data(dataset3)
#' #' mb_d3 <- MultiBlock(Data = list(mirnaseq = mirnaseq))
"mirnaseq"

#' R.ComDim: ComDim Analysis in R
#'
#' Provides functions to run ComDim analysis. ComDim is an unsupervised
#' multi-block method that simultaneously considers multiple data tables to
#' find latent components common to all tables as well as those specific to
#' each table.
#'
#' @keywords internal
"_PACKAGE"

#' @importFrom stats var sd median rnorm qnorm setNames
#' @importFrom methods hasArg
#' @importFrom utils globalVariables
NULL

# Suppress NOTEs about Bioconductor functions used in optional conversion helpers.
# These packages (MultiAssayExperiment, SummarizedExperiment) are in Suggests.
utils::globalVariables(c(
  "ExperimentList", "MultiAssayExperiment",
  "assays", "colData", "experiments", "intersectColumns", "sampleMap"
))

#' ExpandMultiBlock
#'
#' Splits data into several blocks, allowing variables to appear in more than one block
#' simultaneously. Each variable is duplicated into every block to which it is assigned according
#' to the \code{metadata} mapping table.
#' @param data A data.frame or matrix with samples in rows and variables in columns.
#' @param metadata A 2-column data.frame describing how variables are assigned to blocks. The
#'   first column gives the block name; the second column gives the variable name, and must match
#'   the column names of \code{data}. A variable may appear in multiple rows (and therefore in
#'   multiple blocks).
#' @param minblock Integer. Blocks with fewer than \code{minblock} variables are discarded. Use
#'   \code{0} (default) to keep all blocks regardless of size.
#' @param loquace Logical. If \code{TRUE} (default), a warning is printed listing column names in
#'   \code{data} that have no matching entry in the second column of \code{metadata}.
#' @details
#'   For each row in \code{metadata} that matches a variable in \code{data}, the variable's
#'   values are copied into the corresponding block column. Column names in the resulting expanded
#'   matrix are formed as \code{<block>.<variable>}. Variables with all-\code{NA} values after
#'   expansion are removed. If no matches exist between \code{data} column names and
#'   \code{metadata}, \code{NULL} is returned with a warning. If \code{minblock} filtering
#'   removes all blocks, \code{NULL} is returned.
#' @return A \code{MultiBlock} object whose blocks are defined by the first column of
#'   \code{metadata}, or \code{NULL} if no valid blocks could be constructed.
#' @seealso \code{\link{MultiBlock}}, \code{\link{ProcessMultiBlock}}
#' @examples
#' data(mouse_ds)
#' lipidsMB <- ExpandMultiBlock(data = lipids, metadata = metadata_lipids,
#'   minblock = 0, loquace = FALSE)
#' @export
ExpandMultiBlock <- function(data = NULL, metadata = NULL, minblock = 0, loquace = TRUE) {
  if (is.null(data)) {
    stop("No data was provided.")
  }

  if (is.null(metadata)) {
    stop("No metadata was provided.")
  }

  if (is.numeric(metadata[, 2])) {
    metadata[, 2] <- as.character(metadata[, 2])
  }

  meta_vars <- as.character(metadata[, 2])

  if (any(duplicated(rownames(data)))) {
    stop("There are duplicated rownames. Data must have samples in rows and variables in columns.")
  }

  if (!any(colnames(data) %in% meta_vars)) {
    warning(paste(
      "0 coincidences were found between the column names of 'data' and the",
      "second column of 'metadata'. Either the input data or the metadata",
      "data.frame may be in an incorrect format."
    ))
    return(NULL)
  }

  if (any(!(colnames(data) %in% meta_vars))) {
    aa <- colnames(data)[!(colnames(data) %in% meta_vars)]
    if (loquace) {
      warning(sprintf(
        "%d column names are not present in metadata: %s",
        length(aa),
        paste0(aa, collapse = ", ")
      ))
    }
    data <- data[, colnames(data) %in% meta_vars, drop = FALSE]
  }

  if (any(!(meta_vars %in% colnames(data)))) {
    metadata <- metadata[meta_vars %in% colnames(data), ]
    meta_vars <- as.character(metadata[, 2])
  }

  if (minblock != 0) {
    xx <- as.data.frame(table(metadata[, 1]))
    xx <- xx[xx$Freq >= minblock, ]
    metadata  <- metadata[metadata[, 1] %in% xx$Var1, ]
    meta_vars <- as.character(metadata[, 2])
    data <- data[, colnames(data) %in% meta_vars, drop = FALSE]
  }

  expanded_data <- matrix(ncol = nrow(metadata), nrow = nrow(data))

  for (i in seq_len(ncol(data))) {
    pos <- which(meta_vars == colnames(data)[i])
    expanded_data[, pos] <- kronecker(matrix(1, 1, length(pos)), as.matrix(data[, i]))
  }

  colnames(expanded_data) <- paste(metadata[, 1], meta_vars, sep = ".")
  rownames(expanded_data) <- rownames(data)

  # Remove columns with all NAs
  pos_na <- which(apply(expanded_data, 2, function(x) all(is.na(x))))
  if (length(pos_na) != 0) {
    expanded_data <- expanded_data[, -pos_na]
    metadata <- metadata[-pos_na, ]
    meta_vars <- as.character(metadata[, 2])
  }

  block_assignments <- as.character(metadata[, 1])
  unique_blocks <- unique(block_assignments)

  data_list <- list()
  for (bn in unique_blocks) {
    idx <- which(block_assignments == bn)
    mat <- expanded_data[, idx, drop = FALSE]
    colnames(mat) <- meta_vars[idx]
    data_list[[bn]] <- mat
  }

  if (length(data_list) == 0) {
    warning("0 blocks were made. Check whether the minblock threshold was set too high.")
    return(NULL)
  }

  MB <- MultiBlock(Data = data_list)
  return(MB)
}

#' ComDim_PCA
#'
#' Finding common dimensions in multi-block datasets.
#' @param MB A MultiBlock object.
#' @param ndim Number of Common Dimensions.
#' @param normalise To apply normalisation. FALSE == no (default), TRUE == yes.
#' @param threshold The threshold limit to stop the iterations. If the "difference of fit" < threshold (1e-10 as default).
#' @param loquace To display the calculation times. TRUE == yes, FALSE == no (default).
#' @param CompMethod To speed-up the analysis for really big MultiBlocks. 'Normal' (default), 'Kernel', 'PCT', 'Tall' or 'Wide'.
#' @param Partitions To speed-up the analysis for really big MultiBlocks. This parameter is used if CompMethod is 'Tall' or 'Wide'.
#' @return A \code{ComDim} object. Slots for supervised analysis
#'   (\code{R2Y}, \code{Q2}, \code{DQ2}, \code{VIP}, \code{VIP.block},
#'   \code{PLS.model}, \code{cv}, \code{Prediction}) are empty. The
#'   populated slots are:
#' \describe{
#'   \item{\code{Method}}{\code{"PCA"}.}
#'   \item{\code{ndim}}{Number of Common Dimensions extracted.}
#'   \item{\code{Q.scores}}{Global scores matrix (\eqn{n \times ndim}).
#'     Column names are \code{CC1}, \code{CC2}, etc.; row names are sample
#'     names.  Each column \eqn{\mathbf{q}_a} is a unit-norm consensus score,
#'     the dominant left singular vector of the salience-weighted concatenated
#'     blocks
#'     \eqn{\mathbf{W} = [\sqrt{\lambda_1}\mathbf{X}_1 \mid \cdots \mid
#'     \sqrt{\lambda_B}\mathbf{X}_B]}.}
#'   \item{\code{T.scores}}{Named list of block-specific local scores matrices
#'     (\eqn{n \times ndim} each).  For block \eqn{b} and component \eqn{a}:
#'     local loading \eqn{\mathbf{p}_{ba} = \mathbf{X}_b'\mathbf{q}_a} and
#'     local score
#'     \eqn{\mathbf{t}_{ba} =
#'     \mathbf{X}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
#'   \item{\code{P.loadings}}{Global loadings matrix (\eqn{p_{tot} \times
#'     ndim}).  Column \eqn{a} is
#'     \eqn{\mathbf{P}_a = \mathbf{X}'\mathbf{q}_a},
#'     where \eqn{\mathbf{X}} is the mean-centred (and optionally normalised)
#'     concatenated blocks.}
#'   \item{\code{Saliences}}{Block salience (weight) matrix (\eqn{ntable
#'     \times ndim}, row names = block names).  Entry \eqn{(b,a)} is
#'     \eqn{\lambda_{ba} =
#'     \mathbf{q}_a'\mathbf{X}_b\mathbf{X}_b'\mathbf{q}_a},
#'     the variance of block \eqn{b} captured by global score \eqn{a}.}
#'   \item{\code{R2X}}{Proportion of multi-block inertia captured by each
#'     component (named vector, length \eqn{ndim}).  Let
#'     \eqn{d_a} be the leading singular value of \eqn{\mathbf{W}} for
#'     component \eqn{a} (stored as \eqn{Singular_a = d_a^2}); then
#'     \deqn{R2X_a = Singular_a^2 \big/ \sum_k Singular_k^2 = d_a^4 \big/
#'     \sum_k d_k^4.}}
#'   \item{\code{Singular}}{Squared leading singular values of
#'     \eqn{\mathbf{W}}, one per component:
#'     \eqn{Singular_a = d_a^2}.}
#'   \item{\code{Mean}}{List with \code{MeanMB}: named list of column-mean
#'     vectors per block, used for mean-centring.}
#'   \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms used for
#'     block normalisation (all ones when \code{normalise = FALSE}).}
#'   \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
#'     indicating the block name of each row in \code{P.loadings}.}
#'   \item{\code{runtime}}{Total computation time in seconds.}
#' }
#' @examples
#' # Example 1: two data blocks.
#' b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
#' b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' results <- ComDim_PCA(mb, 2)
#'
#' # Example 2: two data blocks, each with different replicate number
#' b1 <- matrix(rnorm(500), 10, 50)
#' batch_b1 <- rep(1, 10)
#' b2 <- matrix(rnorm(2400), 30, 80)
#' batch_b2 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
#' mb <- MultiBlock(
#'   Samples = list(
#'     b1 = paste0("samples_", 1:10),
#'     b2 = rep(paste0("samples_", 1:10), 3)
#'   ),
#'   Data = list(b1 = b1, b2 = b2),
#'   Batch = list(b1 = batch_b1, b2 = batch_b2),
#'   ignore.size = TRUE
#' )
#' rw <- SplitRW(mb)
#' results <- ComDim_PCA(rw, 2)
#' @references Jouan-Rimbaud Bouveresse D, Rutledge DN (2024).
#'   A synthetic review of some recent extensions of ComDim.
#'   \emph{Journal of Chemometrics}, 38(5), e3454.
#'   \doi{10.1002/cem.3454}
#'
#'   Original MATLAB implementation: \url{https://github.com/DNRutledge/ComDim/}
#' @export
ComDim_PCA <- function(MB = MB, ndim = NULL,
                       normalise = FALSE, threshold = 1e-10,
                       loquace = FALSE,
                       CompMethod = "Normal", Partitions = 1) {
  # INITIALISATION

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")

  start_ini_time <- Sys.time() # To start counting calculation time for the initialization


  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is not a MultiBlock.")
  }

  ntable <- length(blockNames(MB)) # number of blocks
  nrowMB <- length(sampleNames(MB)) # number of samples
  variable_number <- ncol(MB) # Number of variables per block

  give_error <- 0

  # Remove all variables with 0 variance. If there are no remaining variables, give_error
  numvar <- 0
  numblock0 <- vector()
  for (i in 1:ntable) {
    var0 <- apply(MB@Data[[i]], 2, function(x) {
      var(x)
    })
    var0 <- which(var0 == 0)

    if (length(var0) != 0) {
      numvar <- numvar + length(var0)
      if (length(var0) == length(MB@Variables[[i]])) {
        numblock0 <- append(numblock0, i)
      }
      MB@Data[[i]] <- MB@Data[[i]][, setdiff(1:variable_number[i], var0)]
      MB@Variables[[i]] <- MB@Variables[[i]][setdiff(1:variable_number[i], var0)]
      variable_number[i] <- length(MB@Variables[[i]])
    }
  }
  if (numvar != 0) {
    warning(sprintf("Number of variables excluded from the analysis because of their zero variance: %d", numvar))
  }
  if (length(numblock0) != 0) {
    numblock0 <- rev(numblock0) # To sort in decreasing order.
    for (i in seq_along(numblock0)) {
      MB@Data[[numblock0[i]]] <- NULL
      MB@Variables[[numblock0[i]]] <- NULL
      if (numblock0[[i]] %in% MB@Batch) {
        MB@Batch[[numblock0[i]]] <- NULL
      }
      if (numblock0[[i]] %in% MB@Metadata) {
        MB@Metadata[[numblock0[i]]] <- NULL
      }
    }
    warning(sprintf("Number of blocks excluded from the analysis because of their zero variance: %d", length(numblock0)))
  }

  ntable <- length(blockNames(MB)) # number of blocks
  variable_number <- ncol(MB) # In case the number of blocks has changed.

  if (length(MB@Data) == 0) { # In case all blocks had 0 variance...
    warning("All blocks had 0 variance")
    give_error <- 1
  }

  # Check for infinite or missing values (they cannot be handled with svd)
  for (i in 1:ntable) {
    if (any(is.na(MB@Data[[i]]))) {
      stop("The MB contains NA values. They must be removed first for ComDim analysis.")
    }
    if (any(is.infinite(MB@Data[[i]]))) {
      stop("The MB contains infinite values. They must be removed first for ComDim analysis.")
    }
  }

  if (give_error) {
    stop("The data is not ready for ComDim.")
  } else {
    message("The data can be used for ComDim.")
  }


  if (is.null(ndim)) {
    ndim <- ntable # If the number of components is not defined,
    # the number of components to extract is equal to the number of blocks.
  }

  pieceBar <- 4 + 2 * ntable + ndim # Number of updates in the progress bar.
  pieceBar <- 80 / pieceBar
  total_progress <- pieceBar

  DimLabels <- paste0("CC", 1:ndim) # One label per component.
  TableLabels <- blockNames(MB) # One label per block.

  # isT <- grepl(output, 'T')
  # isL <- grepl(output, 'L')
  # isP <- grepl(output, 'P')

  if (CompMethod == "Tall" && any(variable_number > 1000)) {
    message("The method 'Tall' is not recommeded for large blocks (> 1000 columns).")
    message("Do you still want to continue with the analysis?")
    message("You can change now the CompMethod if you prefer: (yes/no)")
    ans1 <- tolower(readline("Type your choice: (y/n)"))
    if (ans1 == "y" || ans1 == "yes") {
      message("Which method do you want to use instead (Normal/Kernel/PCT/Wide)?")
      ans2 <- tolower(readline("Type your choice: (n/k/p/w)"))
      if (ans2 == "n" || ans2 == "normal") {
        CompMethod <- "Normal"
      } else if (ans2 == "k" || ans2 == "kernel") {
        CompMethod <- "Kernel"
      } else if (ans2 == "p" || ans2 == "pct") {
        CompMethod <- "PCT"
      } else if (ans2 == "w" || ans2 == "wide") {
        CompMethod <- "Wide"
      } else {
        stop("Wrong answer typed.")
      }
    } else if (ans1 == "n" || ans1 == "no") {
      message("The method 'Tall' will be used")
    } else {
      stop("Wrong answer typed.")
    }
  }

  if (CompMethod == "Kernel" && nrowMB >= min(variable_number)) {
    stop(sprintf(
      paste0("CompMethod = 'Kernel' requires n < p for every block ",
             "(n = %d, smallest block p = %d). ",
             "Use 'Normal', 'PCT', or 'Wide' instead."),
      nrowMB, min(variable_number)
    ))
  }

  end_ini_time <- Sys.time() # To end the count of the calculation time.

  if (loquace) {
    message(sprintf("Initialisation finished after : %s millisecs", (end_ini_time - start_ini_time) * 1000))
  }

  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # NORMALISATION

  # start_norm_time <- Sys.time()  # To start counting calculation time for the initialization

  X_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))
  Xnorm_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))

  res_calib <- list()
  temp_tabCalib <- list()
  s_n <- list()
  res_calib$SingVal <- as.vector(NULL)
  res_calib$NormMB <- as.vector(NULL)
  res_calib$MeanMB <- list()

  for (i in 1:ntable) {
    res_calib$MeanMB[[TableLabels[i]]] <- colMeans(MB@Data[[i]])
    names(res_calib$MeanMB[[TableLabels[i]]]) <- MB@Variables[[i]]
    # res_calib$MeanMB$i <- TableLabels

    if (normalise) {
      # Normalise original blocks

      X_mean <- MB@Data[[i]] - matrix(data = rep(1, nrowMB), ncol = 1, nrow = nrowMB) %*% res_calib$MeanMB[[i]]
      # In a previous version (that worked), I had used as.matrix instead of matrix.
      XX <- X_mean * X_mean

      Norm_X <- sqrt(sum(XX))
      X_Normed <- X_mean / Norm_X

      res_calib$NormMB[i] <- Norm_X

      temp_tabCalib[[i]] <- X_Normed
      s_n[[i]] <- X_Normed
    } else {
      res_calib$NormMB[[TableLabels[i]]] <- rep(1, length(MB@Variables[[i]]))
      names(res_calib$NormMB[[TableLabels[i]]]) <- MB@Variables[[i]]
      # res_calib$MeanMB$Variables <- MB@Variables[[i]]
      # res_calib$MeanMB$i <- TableLabels

      temp_tabCalib[[i]] <- MB@Data[[i]]
      s_n[[i]] <- MB@Data[[i]]
    }

    if (i == 1) {
      X_mat[, 1:variable_number[1]] <- MB@Data[[i]]
      Xnorm_mat[, 1:variable_number[1]] <- temp_tabCalib[[i]]
    } else {
      beg <- sum(variable_number[1:(i - 1)]) + 1
      ending <- sum(variable_number[1:i])
      X_mat[, beg:ending] <- MB@Data[[i]]
      Xnorm_mat[, beg:ending] <- temp_tabCalib[[i]]
    }

    norm_comdim <- Sys.time()
    if (loquace) {
      message(sprintf("Normalization of block %s finished after : %s millisecs", i, (norm_comdim - start_ini_time) * 1000))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  # res_calib$NormMB$i <- 'Norm'
  names(res_calib$NormMB) <- TableLabels

  nR <- nrow(Xnorm_mat)
  nC <- ncol(Xnorm_mat)

  # COMPRESSION

  s_r <- Compress_Data_2020(s_n, CompMethod, Partitions)

  end_comp_time <- Sys.time()

  if (loquace) {
    message(sprintf("Compression finished after : %s millisecs", (end_comp_time - start_ini_time) * 1000))
  }

  total_progress <- total_progress + pieceBar * ntable
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  ## DO ComDim
  # ini_comdim <- Sys.time()

  # saliences <- list()
  saliences <- matrix(, ncol = ndim, nrow = ntable)
  # Q <- list()
  Q <- matrix(, ncol = ndim, nrow = nrowMB)
  unexplained <- ntable # Warning!: true only if the tables were set to norm=1
  varexp <- as.vector(NULL)

  for (dims in 1:ndim) {
    previousfit <- 10000
    deltafit <- 1000000
    lambda <- matrix(rep(1, ntable), nrow = ntable, ncol = 1)

    qini <- as.vector(s_r[[1]][, 1])
    qini <- qini / sqrt(as.vector(qini %*% qini)) # random initialization of t + set to unit length
    qold <- 100

    iters <- 0
    while (norm(qold - qini, "2") > threshold && iters < 100) {
      iters <- iters + 1
      qold <- qini
      q <- 0

      W <- matrix(, nrow = nrowMB, ncol = 0)
      for (j in 1:ntable) {
        W <- cbind(W, sqrt(lambda[j]) * s_r[[j]])
      }

      usv <- svd(W)
      qtmp <- usv$u[, 1]

      for (j in 1:ntable) {
        # takes each table into account for lambda after PCA
        lambda[j] <- t(qtmp) %*% (s_r[[j]] %*% t(s_r[[j]])) %*% qtmp
        q <- q + lambda[j] * qtmp
      }

      q <- q / sqrt(as.vector(t(q) %*% q)) # standardizes t
      if (abs(min(q)) > abs(max(q))) {
        q <- -q
      }
      qini <- q
    } # deltafit>threshold

    saliences[, dims] <- lambda
    Q[, dims] <- q


    # SingVal is square of the SVD singular value
    # because here it is based on X, not X.X'
    # Does NOT work for Kernel & Wide & Tall
    if (CompMethod == "Normal" || CompMethod == "PCT") {
      res_calib$SingVal[dims] <- usv$d[1] * usv$d[1] # from SVD on W
      varexp[dims] <- res_calib$SingVal[dims]^2
    } else {
      varexp[dims] <- 0 # This will be overwritten at the end
      res_calib$SingVal[dims] <- 0 # This will be overwritted at the end
    }


    # Deflate blocks
    aux <- diag(nrowMB) - as.matrix(q) %*% t(as.matrix(q))
    for (j in 1:ntable) {
      s_r[[j]] <- aux %*% s_r[[j]]
    }

    iter_comdim <- Sys.time()
    if (loquace) {
      message(sprintf("Component %s determined after : %s millisecs", dims, (iter_comdim - start_ini_time) * 1000))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  # res_calib$SingVal$i <- 'Singular value'
  names(res_calib$SingVal) <- DimLabels # Dimensions

  # Q$i <- data[[1]]$Samples
  # Q$v <- DimLabels
  colnames(Q) <- DimLabels
  rownames(Q) <- MB@Samples
  res_calib$Q <- Q
  rm(Q)

  ## Adding metadata. Metadata extracted from the first block
  res_calib$metadata <- list()
  if (length(MB@Metadata) != 0) {
    res_calib$metadata[[1]] <- MB@Metadata
  }


  end_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Scores finished after : %s millisecs", (end_comdim - start_ini_time) * 1000))
  }
  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  meanW <- as.vector(NULL)
  for (j in 1:ntable) {
    meanW <- c(meanW, mean(s_r[[j]]))
  }
  # res_calib$s_n <- s_n
  # res_calib$s_r <- s_r
  rm(s_n)
  rm(s_r)

  ##
  # explained <- list()
  explained <- varexp / sum(varexp)
  names(explained) <- DimLabels
  # explained$i <-'% explained'
  # explained$v <- DimLabels #Dimensions
  res_calib$explained <- explained
  rm(varexp)
  rm(explained)

  ## Calculate Sums of saliences for each Dim
  # Sum_saliences_Dim <- list()
  Sum_saliences_Dim <- colSums(saliences)
  names(Sum_saliences_Dim) <- DimLabels
  # Sum_saliences_Dim$i <-'Sum Dim Saliences'
  # Sum_saliences_Dim$v <- DimLabels #Dimensions

  res_calib$Sum_saliences_Dim <- Sum_saliences_Dim
  rm(Sum_saliences_Dim)

  # Calculate Sums of saliences for each Table
  # Sum_saliences_Tab <- list()
  Sum_saliences_Tab <- rowSums(saliences)
  # Sum_saliences_Tab$i <- 'Sum Tab Saliences'
  names(Sum_saliences_Tab) <- TableLabels
  # Sum_saliences_Tab$v <- TableLabels #tables

  res_calib$Sum_saliences_Tab <- Sum_saliences_Tab
  rm(Sum_saliences_Tab)

  # saliences$i <- TableLabels #tables
  # saliences$v <- DimLabels #Dimensions
  res_calib$saliences <- saliences
  colnames(res_calib$saliences) <- DimLabels
  rownames(res_calib$saliences) <- TableLabels

  ## Calculate Normalised concatenated Xs ('Calib') from col
  # Calculate concatenated CD loadings
  # Reorganise Loadings - 1 matrix / LV

  ## If Output = NULL, nothing else is calculated
  # if(!(is.null(output))){
  L_CD_Vec <- NULL
  L_X_Vec <- NULL
  Calib <- NULL

  nCalib <- nrowMB
  # }

  ## Already defined in during ComDim initiallization
  # isT <- grepl(output, 'T')
  # isL <- grepl(output, 'L')
  # isP <- grepl(output, 'P')
  # isLP <- c(isP,isL)
  # isTLP <- c(isT,isP,isL)

  # tic
  # Calculate concatenated CD loadings
  # Reorganise Loadings - 1 matrix / LV
  # L_CD <- list()
  # L_X <- list()
  T_Loc <- list()
  for (i in 1:ntable) { # Prepare lists
    # L_CD[[TableLabels[i]]] <- matrix(,ncol = ndim, nrow = ncol(temp_tabCalib[[i]]))
    # L_X[[TableLabels[i]]] <- matrix(,ncol = ndim, nrow = ncol(temp_tabCalib[[i]]))
    T_Loc[[TableLabels[i]]] <- matrix(, ncol = ndim, nrow = nrowMB)
  }
  b <- matrix(, ncol = ndim, nrow = ntable) # Used to calculate R2X in Kernel and Tall

  if (!requireNamespace("pracma", quietly = TRUE)) {
    stop("The package pracma is needed.")
  }

  for (j in 1:ndim) {
    T_mat <- matrix(, nrow = nrowMB, ncol = 0)

    for (i in 1:ntable) {
      # Q.d are orthonormal in ICA & PCA

      temp <- t(temp_tabCalib[[i]]) %*% res_calib$Q[, j] # Scaled CD 'local' Loadings

      # if  (!(is.null(isP))){
      # L_CD[[TableLabels[i]]][,j] <- temp
      # }

      # if  (!(is.null(isL))){
      #  #Q.d are orthonormal in ICA & PCA
      #  L_X[[TableLabels[i]]][,j] <- t(MB@Data[[i]])%*%Q$Data[,j] #Unscaled X 'local' Loadings;
      # }

      # T_mat <- cbind(T_mat,temp_tabCalib[[i]]%*%(temp/as.vector(t(temp)%*%temp))) #local Scores
      T_mat <- cbind(T_mat, temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp))) # local Scores

      # if(!(is.null(isT))){
      # T_Loc[[TableLabels[i]]][,j] <- temp_tabCalib[[i]]%*%(temp/as.vector(t(temp)%*%temp)) # local Scores
      T_Loc[[TableLabels[i]]][, j] <- temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp)) # local Scores
      # }


      # Deflate each temp_tabCalib
      temp_tabCalib[[i]] <- temp_tabCalib[[i]] - res_calib$Q[, j] %*% t(temp)
    }


    # For each CC
    # MLR b-coefficients between Local and Global Scores
    # [b0,b[,j]] <- mlr_DB(T_mat,Q$d[,j],0)
    # b=pinv(X'*X)*X'*Y;
    b[, j] <- pracma::pinv(t(T_mat) %*% T_mat) %*% t(T_mat) %*% res_calib$Q[, j]
    # b0 <- 0
  }

  for (i in 1:ntable) {
    rownames(T_Loc[[TableLabels[i]]]) <- rownames(res_calib$Q)
    colnames(T_Loc[[TableLabels[i]]]) <- colnames(res_calib$Q)
  }

  # If Output==[], nothing else is calculated
  # if(!(is.null(output))){
  # Calculate Global Loadings
  # if(!(is.null(isP))){
  L_CD_Vec <- t(Xnorm_mat) %*% res_calib$Q # Scaled CD 'global' Loadings
  # }

  # if(!(is.null(isL))){
  #  L_X_Vec <- t(X_mat)%*%Q$Data # Unscaled X 'global' Loadings
  # }
  # }

  #### If Output==[], nothing else is calculated
  load_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Loadings finished after : %s millisecs", (load_comdim - start_ini_time) * 1000))
  }

  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  rm(X_mat)
  rm(temp_tabCalib)
  rm(temp)

  ## Output
  end_function <- Sys.time()

  # res_calib$b <- b
  # colnames(res_calib$b) <- DimLabels
  # rownames(res_calib$b) <- TableLabels
  # res_calib$b$i <- TableLabels # tables
  # res_calib$b$v <- DimLabels  # Dimensions

  # if(!(is.null(isT))){
  res_calib$T_Loc <- T_Loc
  # res_calib$T_Loc$i <- res_calib$Q$i # samples
  # res_calib$T_Loc$v <- DimLabels # Dimensions
  # }

  # T_Loc is no longer needed
  rm(T_Loc)


  # if(!(is.null(isP))){

  varnames <- vector()
  for (i in 1:ntable) {
    varnames <- append(varnames, MB@Variables[[i]])
  }
  # if(length(which(duplicated(varnames))) != 0){
  #  warning('There are at least two variables with the same name. Their name in the rownames of P has been altered.\n')
  # }

  res_calib$P <- L_CD_Vec
  # res_calib$P$i <- varnames # numbers of all variables;
  # res_calib$P$v <- DimLabels # Dimensions
  colnames(res_calib$P) <- DimLabels
  rownames(res_calib$P) <- varnames

  # res_calib$P_Loc <- L_CD
  # res_calib$P_Loc$i <- TableLabels # Tables
  # res_calib$P_Loc$v <- DimLabels # Dimensions
  # for(i in 1:ntable){
  #  rownames(res_calib$P_Loc[[TableLabels[i]]]) <- MB@Variables[[i]]
  #  colnames(res_calib$P_Loc[[TableLabels[i]]]) <- DimLabels
  # }
  # }

  rm(L_CD_Vec)
  # rm(L_CD)

  # if(!(is.null(isL))){
  #   res_calib$Lx$Data <- L_X_Vec
  #   res_calib$Lx$i <- t(c(1:nC)) # numbers of all variables;
  #   res_calib$Lx$v <- DimLabels # Dimensions
  #   colnames(res_calib$Lx$Data) <- DimLabels
  #   rownames(res_calib$Lx$Data) <- t(c(1:nC))
  #   #res_calib$Lx_Loc$i <- TableLabels # Tables
  #   #res_calib$Lx_Loc$v <- DimLabels # Dimensions
  #   res_calib$Lx_Loc$Data <- L_X
  #   for (i in 1:ntable){
  #     colnames(res_calib$Lx_Loc$Data[[TableLabels[i]]]) <- DimLabels
  #     rownames(res_calib$Lx_Loc$Data[[TableLabels[i]]]) <- MB@Variables[[i]]
  #   }
  # }
  # rm(L_X_Vec)
  # rm(L_X)

  ##
  # Singular value calculated from b and saliences
  # Since :
  # b_T_Q(:,j)=saliences.d(:,j).*saliences.d(:,j)/SingVal.d(1,j);
  if (CompMethod == "Kernel" || CompMethod == "Tall" || CompMethod == "Wide") {
    for (dims in 1:ndim) {
      res_calib$SingVal[dims] <- t(b[, dims]) %*% ((saliences[, dims]^2) / as.vector(t(b[, dims]) %*% b[, dims]))
    }
    names(res_calib$SingVal) <- DimLabels
    varexp <- res_calib$SingVal^2
    res_calib$explained <- varexp / sum(varexp)
    names(res_calib$explained) <- DimLabels
  }

  rm(saliences)

  # Define block length
  belong_block <- rep(0, nrow(res_calib$P))
  k <- 1
  for (i in 1:ntable) {
    belong_block[k:(k + variable_number[i] - 1)] <- TableLabels[i]
    k <- k + variable_number[i]
  }

  #### If Output==[], nothing else is calculated
  # if(normalise == 1){
  #  res_calib$Xnorm_mat$i <- res_calib$Q$i # samples
  #  res_calib$Xnorm_mat$v <- t(c(1:nC)) # numbers of all variables;
  #  res_calib$Xnorm_mat$Data <- Xnorm_mat
  # }
  # rm(Xnorm_mat)
  #### If Output==[], nothing else is calculated

  end_output <- Sys.time()
  running_time <- (end_output - start_ini_time)
  if (loquace) {
    message(sprintf("Analysis finished after : %s seconds", running_time))
  }
  res_calib$runtime <- running_time # Total time of analysis

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
  utils::setTxtProgressBar(progress_bar, value = 80)

  close(progress_bar)

  # Save data in a ComDim structure
  res_calib <- new("ComDim",
    Method = "PCA",
    ndim = ndim,
    Q.scores = res_calib$Q,
    T.scores = res_calib$T_Loc,
    P.loadings = res_calib$P,
    Saliences = res_calib$saliences,
    Orthogonal = list(),
    R2X = res_calib$explained,
    R2Y = vector(),
    Q2 = vector(),
    DQ2 = vector(),
    Singular = res_calib$SingVal,
    Mean = list(MeanMB = res_calib$MeanMB, MeanY = NULL),
    Norm = list(NormMB = res_calib$NormMB),
    PLS.model = list(),
    cv = list(),
    Prediction = list(),
    Metadata = res_calib$metadata,
    variable.block = belong_block,
    runtime = as.numeric(res_calib$runtime)
  )
  return(res_calib)
}


#' Compress_Data_2020
#'
#' Compress large multi-block objects. Internal function of ComDim_PCA().
#' @param s_n The multi-block object.
#' @param CompMethod It can be 'Normal' (default), 'Kernel', 'PCT', 'Tall' or 'Wide'.
#' @param Partitions The number of partitions.
#' @return The compressed multi-block.
#' @references Original MATLAB implementation:
#'   \url{https://github.com/DNRutledge/ComDim/blob/main/Compress_Data_2020.m}
#' @keywords internal
Compress_Data_2020 <- function(s_n = s_n, CompMethod = CompMethod, Partitions = Partitions) {
  s_r <- list()
  ntable <- length(s_n)
  nR <- nrow(s_n[[1]])

  if (!is.null(CompMethod)) {
    if (CompMethod == "Tall") {
      for (i in 1:ntable) {
        MaxRank <- min(c(nrow(s_n[[i]]), ncol(s_n[[i]])))
        # Tall segmented PCT
        s_r[[i]] <- PCA_Tall_PCT_DNR(Xn = s_n[[i]], PCs = MaxRank, Partitions = Partitions)
      }
    } else if (CompMethod == "Wide") {
      for (i in 1:ntable) {
        Xi <- ColumnsPartition(Xn = s_n[[i]], Partitions = Partitions)
        T_all <- matrix(, nrow = nrow(s_n[[i]]), ncol = 0)
        for (p in 1:Partitions) {
          usv <- svd(Xi[[p]])
          T_in <- usv$u %*% diag(usv$d)
          T_all <- cbind(T_all, T_in)
        }
        s_r[[i]] <- T_all
        # print(s_r[[i]][1:3,1:4])
      }
    } else if (CompMethod == "Kernel") {
      for (i in 1:ntable) {
        Kern <- (s_n[[i]] %*% t(s_n[[i]])) / nR
        usv <- svd(Kern)
        s_r[[i]] <- sqrt(nR) * usv$u %*% diag(sqrt(usv$d))
      }
    } else if (CompMethod == "PCT") {
      for (i in 1:ntable) {
        usv <- svd(s_n[[i]])
        s_r[[i]] <- usv$u %*% diag(usv$d)
      }
    } else if (CompMethod == "Normal") {
      for (i in 1:ntable) {
        s_r[[i]] <- s_n[[i]]
      }
    }
  }
  return(s_r)
}


#' PCA_Tall_PCT_DNR
#'
#' Compress a block using the PCT method. Internal function of ComDim_PCA().
#' @param Xn The block to compress.
#' @param PCs Number of PCs to keep.
#' @param Partitions The number of partitions.
#' @return The compressed block.
#' @keywords internal
PCA_Tall_PCT_DNR <- function(Xn = Xn, PCs = PCs, Partitions = Partitions) {
  W <- RowsPartition(Xn, Partitions)
  cols <- ncol(Xn)

  D <- t(W[[1]]) %*% W[[1]]

  if (Partitions > 1) {
    for (par in 2:Partitions) {
      D <- D + (t(W[[par]]) %*% W[[par]])
    }
  }

  vs <- eigen(D)
  vs_flipped <- vs$vectors # eigen() in R returns eigenvalues in decreasing order; columns are eigenvectors

  Tm <- matrix(, nrow = nrow(Xn), ncol = 0)
  Taux <- matrix(, nrow = 0, ncol = PCs)

  for (par in 1:Partitions) {
    to <- W[[par]] %*% vs_flipped

    Taux <- rbind(Taux, to[, 1:PCs])
  }

  Tm <- cbind(Tm, Taux)
  return(Tm)
}

#' ColumnsPartition
#'
#' Calculate vertical partitions in a block. Internal function of ComDim_PCA().
#' @param Xn A block.
#' @param Partitions The number of partitions.
#' @return A list of column-partitioned sub-matrices.
#' @keywords internal
ColumnsPartition <- function(Xn = Xn, Partitions = Partitions) {
  cols <- ncol(Xn)
  stride <- floor(cols / Partitions)
  remaining <- cols %% Partitions

  count <- 1
  W <- list()

  for (i in 1:Partitions) {
    step <- count + stride - 1
    if (i == Partitions) {
      step <- step + remaining
    }
    W[[i]] <- Xn[, count:step]
    count <- count + stride
  }
  return(W)
}

#' RowsPartition
#'
#' Calculate horizontal partitions in a block. Internal function of ComDim_PCA().
#' @param Xn A block.
#' @param Partitions The number of partitions.
#' @return A list of row-partitioned sub-matrices.
#' @keywords internal
RowsPartition <- function(Xn = Xn, Partitions = Partitions) {
  rows <- nrow(Xn)
  stride <- floor(rows / Partitions)
  remaining <- rows %% Partitions

  count <- 1
  W <- list()

  for (i in 1:Partitions) {
    step <- count + stride - 1
    if (i == Partitions) {
      step <- step + remaining
    }
    W[[i]] <- Xn[count:step, ]
    count <- count + stride
  }
  return(W)
}

#' MultiBlock2MultiAssayExperiment
#'
#' Converts a MultiBlock into a MultiAssayExperiment. Each block becomes one experiment; sample
#' names, batch information, and metadata are carried over into the \code{colData} of the result.
#' @param MB A \code{MultiBlock} object.
#' @param MSEmetadata An optional list of unstructured metadata describing the overall content of
#'   the MultiAssayExperiment (stored in its \code{metadata} slot). Pass \code{NULL} (default) to
#'   omit experiment-level metadata.
#' @details
#'   Each block in \code{MB} is transposed (features x samples) and stored as a named matrix in
#'   the \code{ExperimentList}. Row names are set to the variable names and column names to the
#'   sample names of the MultiBlock. The \code{colData} is constructed from \code{MB@@Samples};
#'   any \code{Metadata} and \code{Batch} information present in the MultiBlock is appended as
#'   additional columns. A \code{sampleMap} is generated mapping every sample to every experiment
#'   using the same primary and column names.
#' @return A \code{MultiAssayExperiment} object with one experiment per block in \code{MB}.
#' @seealso \code{\link{MultiBlock}}, \code{\link{MultiAssayExperiment2MultiBlock}}
#' @examples
#' \donttest{
#' if (requireNamespace("MultiAssayExperiment", quietly = TRUE)) {
#' library(MultiAssayExperiment)
#' b1 <- matrix(rnorm(50), 5, 10, dimnames = list(paste0("s", 1:5), paste0("v", 1:10)))
#' b2 <- matrix(rnorm(30), 5,  6, dimnames = list(paste0("s", 1:5), paste0("w", 1:6)))
#' mb <- MultiBlock(Data = list(block1 = b1, block2 = b2))
#' mae <- MultiBlock2MultiAssayExperiment(mb)
#' mae <- MultiBlock2MultiAssayExperiment(mb, MSEmetadata = list(study = "example"))
#' }
#' }
#' @export
MultiBlock2MultiAssayExperiment <- function(MB,
                                            MSEmetadata = NULL) { # List of unstructured metadata describing the overall content of the object.

  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is  not a MultiBlock.")
  }

  ## Create ExperimentList
  exp <- list()
  for (i in seq_along(MB@Data)) {
    exp[[names(MB@Data)[i]]] <- t(MB@Data[[i]])
    colnames(exp[[names(MB@Data)[i]]]) <- MB@Samples
    rownames(exp[[names(MB@Data)[i]]]) <- MB@Variables[[names(MB@Data)[i]]]
  }
  exp <- ExperimentList(exp)

  ## Create colData
  colDat <- matrix(MB@Samples, nrow = length(MB@Samples), ncol = 1)
  for (i in seq_along(MB@Data)) {
    if (names(MB@Data)[i] %in% names(MB@Metadata)) {
      colDat <- cbind.data.frame(colDat, MB@Metadata[[names(MB@Metadata)[i]]])
    }
    if (names(MB@Data)[i] %in% names(MB@Batch)) {
      colDat <- cbind.data.frame(colDat, MB@Batch[[names(MB@Data)[i]]])
    }
  }
  rownames(colDat) <- MB@Samples

  # SampleMap
  for (i in seq_along(MB@Data)) {
    if (i == 1) {
      SEsampleMap <- data.frame(
        assay = rep(names(MB@Data)[i], length(MB@Samples)),
        primary = MB@Samples,
        colname = MB@Samples
      )
    } else {
      SEsampleMap <- rbind.data.frame(
        SEsampleMap,
        data.frame(
          assay = rep(
            names(MB@Data)[i],
            length(MB@Samples)
          ),
          primary = MB@Samples,
          colname = MB@Samples
        )
      )
    }
  }


  if (is.null(MSEmetadata)) {
    se <- MultiAssayExperiment(
      experiments = exp,
      colData = colDat,
      sampleMap = SEsampleMap
    )
  } else {
    se <- MultiAssayExperiment(
      experiments = exp,
      colData = colDat,
      sampleMap = SEsampleMap,
      metadata = MSEmetadata
    )
  }

  return(se)
}

#' ProcessMultiBlock
#'
#' Apply a custom function to transform a MultiBlock and/or select variables or blocks.
#' When multiple operations are supplied, the order of execution is:
#' \code{vars} subsetting, then \code{FUN}, then \code{FUN.SelectVars}, then \code{FUN.SelectBlocks}.
#' @param MB A MultiBlock object.
#' @param blocks The blocks to process. A vector of integers or block names. When omitted, all
#'   blocks are processed.
#' @param vars The variables to keep. A list with the same length as \code{blocks}. Each element
#'   contains the indices (integer) or names (character) of the variables to retain in that block.
#'   Subsetting is applied before any of the \code{FUN} arguments.
#' @param FUN A function applied to each selected block's data matrix (samples x variables). It
#'   receives the matrix as its sole argument and must return a matrix of the same dimensions.
#' @param FUN.SelectVars A function applied to each selected block's data matrix to determine which
#'   variables to keep. It receives the matrix as its sole argument and must return a logical vector
#'   of length equal to \code{ncol} of the matrix (\code{TRUE} = keep).
#' @param FUN.SelectBlocks A function applied to each selected block's data matrix to determine
#'   whether the block should be retained. It receives the matrix as its sole argument and must
#'   return a single \code{TRUE} or \code{FALSE}.
#' @return The processed MultiBlock object, with data matrices transformed and/or blocks/variables
#'   removed according to the supplied arguments. Blocks not listed in \code{blocks} are left
#'   unchanged.
#' @seealso \code{\link{MultiBlock}}, \code{\link{MultiBlock2Matrix}}
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' # Normalize each block to 0-100 range
#' BY100 <- function(x) 100 * (x - min(x)) / (max(x) - min(x))
#' mb <- ProcessMultiBlock(mb, FUN = BY100)
#' # Keep only variables with non-zero variance
#' mb <- ProcessMultiBlock(mb, FUN.SelectVars = function(x) apply(x, 2, var) > 0)
#' # Remove blocks where all values are below 0.5
#' mb <- ProcessMultiBlock(mb, FUN.SelectBlocks = function(x) max(x) >= 0.5)
#' @export
ProcessMultiBlock <- function(MB = MB,
                              blocks = NULL,
                              vars = NULL,
                              FUN = NULL,
                              FUN.SelectVars = NULL,
                              FUN.SelectBlocks = NULL) {
  if (is.null(FUN) && is.null(FUN.SelectBlocks) &&
    is.null(FUN.SelectVars) && is.null(blocks) &&
    is.null(vars)) {
    warning("'FUN','FUN.SelectBlocks' and 'FUN.SelectVars' arguments were missing. 'MB' was not processed.")
    return(MB)
  }

  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is not a MultiBlock object.")
  }

  if (!(is.null(blocks))) {
    if (is.character(blocks)) {
      block.ids <- match(blocks, names(MB@Data))
      if (any(is.na(block.ids))) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(blocks[which(is.na(block.ids))], sep = ", ")
        )
        block.ids <- block.ids[!is.na(block.ids)]
      }
    } else if (is.numeric(blocks)) {
      block.ids <- intersect(seq_along(MB@Data), blocks)
      if (length(blocks) != length(block.ids)) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(setdiff(seq_along(MB@Data), blocks), sep = ", ")
        )
      }
    }
  } else {
    block.ids <- seq(1, length(MB@Data))
  }
  block.fixed <- setdiff(seq(1, length(MB@Data)), block.ids)

  if (!is.null(vars)) {
    if (!is.list(vars)) {
      stop("'variables' must be a list object.")
    } else if (length(vars) != length(block.ids)) {
      stop("'variables' should have the same length as 'blocks'.")
    }

    for (j in seq_along(vars)) {
      i <- block.ids[j]
      if (is.character(vars[[j]])) {
        xx <- match(vars[[j]], MB@Variables[[i]])
        if (any(is.na(xx))) {
          warning(sprintf(
            "Some variable names were not found in block %s: %s", names(MB@Data)[i],
            paste0(vars[[j]][which(is.na(xx))], collapse = ", ")
          ))
          xx <- xx[!is.na(xx)]
        }
      } else if (is.numeric(vars[[j]])) {
        xx <- intersect(seq_along(MB@Variables[[i]]), vars[[j]])
        if (length(xx) != length(vars[[j]])) {
          warning(sprintf(
            "Some variable indices were out of range for block %s: %s", names(MB@Data)[i],
            paste0(setdiff(vars[[j]], seq_along(MB@Variables[[i]])), collapse = ", ")
          ))
        }
      }
      MB@Data[[i]]      <- MB@Data[[i]][, xx, drop = FALSE]
      MB@Variables[[i]] <- MB@Variables[[i]][xx]
    }
  }

  if (!is.null(FUN)) {
    for (i in block.ids) {
      MB@Data[[i]] <- FUN(MB@Data[[i]])
    }
  }

  if (!is.null(FUN.SelectVars)) {
    for (i in block.ids) {
      pos <- which(FUN.SelectVars(MB@Data[[i]]))

      if (length(pos) != 0) {
        MB@Data[[i]] <- MB@Data[[i]][, pos]
        MB@Variables[[i]] <- MB@Variables[[i]][pos]
      }
    }
  }

  if (!is.null(FUN.SelectBlocks)) {
    to_remove <- integer(0)
    for (i in block.ids) {
      if (!FUN.SelectBlocks(MB@Data[[i]])) {
        to_remove <- c(to_remove, i)
      }
    }
    if (length(to_remove) > 0) {
      to_remove <- sort(to_remove, decreasing = TRUE)
      for (i in to_remove) {
        MB@Data[[i]] <- NULL
        MB@Variables[[i]] <- NULL
      }
    }
  }

  return(MB)
}

#' MakeComDimLoadingsTable
#'
#' Creates a long (tidy) data frame with the local P-loadings from a ComDim model, suitable for
#' use with \pkg{ggplot2}.
#' @param model The output from a ComDim analysis (a \code{ComDim} object).
#' @param blocks The blocks from which loadings will be extracted. A vector of integers (block
#'   indices) or block names. When omitted, all blocks are included.
#' @param dim Integer vector of predictive component indices to include. When omitted, all
#'   predictive components in the model are included.
#' @param dim.ort Integer vector of orthogonal component indices to include. When \code{NULL},
#'   all orthogonal components present in the model are included. When \code{0}, no orthogonal
#'   components are included.
#' @return A long data frame with one row per variable–component combination, containing the
#'   following columns:
#'   \describe{
#'     \item{\code{variable.id}}{Variable name (factor).}
#'     \item{\code{variable.id.number}}{Position of the variable across all blocks (factor).}
#'     \item{\code{block.id}}{Integer index of the block.}
#'     \item{\code{block.name}}{Name of the block (factor).}
#'     \item{\code{dim}}{Component number.}
#'     \item{\code{value}}{Loading value (from \code{P.loadings} or \code{Orthogonal$P.loadings.ort}).}
#'   }
#' @seealso \code{\link{MakeComDimScoresTable}}, \code{\link{ComDim_PCA}}
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50) # 10 rows and 50 columns
#' b2 <- matrix(rnorm(800), 10, 80) # 10 rows and 80 columns
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' model <- ComDim_PCA(mb, ndim = 2)
#' tbl <- MakeComDimLoadingsTable(model)
#' @export
MakeComDimLoadingsTable <- function(model,
                                    blocks = NULL,
                                    dim = NULL,
                                    dim.ort = NULL) {
  if (!inherits(model, "ComDim")) {
    stop("'model' is  not the output of a ComDim analysis.")
  }

  if (is.null(dim)) {
    dim <- 1:model@ndim
  }

  if (is.null(dim.ort) && length(model@Orthogonal) != 0) {
    if (model@Orthogonal$nort != 0) {
      dim.ort <- 1:model@Orthogonal$nort
    }
  } else if (is.null(dim.ort) && length(model@Orthogonal) == 0) {
    dim.ort <- 0
  }

  block_names <- names(model@T.scores)

  if (!(is.null(blocks))) {
    if (is.character(blocks)) {
      block.ids <- match(blocks, block_names)
      if (any(is.na(block.ids))) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(blocks[which(is.na(block.ids))], sep = ", ")
        )
        block.ids <- block.ids[!is.na(block.ids)]
      }
    } else if (is.numeric(blocks)) {
      block.ids <- intersect(seq_along(block_names), blocks)
      if (length(blocks) != length(block.ids)) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(setdiff(seq_along(block_names), blocks), sep = ", ")
        )
      }
    }
  } else {
    block.ids <- seq(1, length(block_names))
  }


  if (length(block.ids) == 0) {
    stop("0 blocks were selected. The table cannot be built.")
  }


  for (i in block.ids) { # For each of the dimensions

    posvars <- which(model@variable.block == block_names[i])

    for (v in dim) {
      df2 <- data.frame(
        variable.id = rownames(model@P.loadings[posvars, ]),
        variable.id.number = posvars,
        block.id = rep(i, length(posvars)),
        block.name = rep(block_names[i], length(posvars)),
        dim = rep(v, length(posvars)),
        value = model@P.loadings[posvars, v]
      )
      if (v == dim[1] && i == block.ids[1]) {
        df <- df2
      } else {
        df <- rbind.data.frame(df, df2)
      }
    }

    if (dim.ort != 0) {
      for (v in dim.ort) {
        df2 <- data.frame(
          variable.id = rownames(model@Orthogonal$P.loadings.ort[posvars, ]),
          variable.id.number = posvars,
          block.id = rep(i, length(posvars)),
          block.name = rep(block_names[i], length(posvars)),
          dim = rep(v, length(posvars)),
          value = model@Orthogonal$P.loadings.ort[posvars, v]
        )
        df <- rbind.data.frame(df, df2)
      }
    }
  }
  rownames(df) <- NULL


  # Add factors to data
  df$variable.id <- factor(df$variable.id, levels = unique(df$variable.id))
  df$variable.id.number <- factor(df$variable.id.number)
  df$block.name <- factor(df$block.name, levels = block_names)

  return(df)
}

# For the DA situation I should include the formula of the VIPs.
# It requires pracma::inv, which is invoked in the package.
# Need to change all matrix divisions by pracma:inv * ...
# Check how the saliences are computed.
# See if the B values make sense (on June DNR was skeptical on how to deal with this situation in the multi-block scenareo due to the different scaling...)


#' ComDim_PLS
#'
#' Finding common dimensions in multi-block datasets using PLS.
#' @param MB A MultiBlock object.
#' @param y The Y-block to use in the PLS model as dependent data. A class vector or a dummy matrix.
#' @param method PLS-DA or PLS-R.
#' @param decisionRule Only used if method is set to PLS-DA. If 'fixed', samples are assigned to the class with Y-hat above 1/nclasses. If 'max', samples are assigned to the class with the highest Y-hat.
#' @param ndim Number of Common Dimensions.
#' @param normalise To apply block normalisation. FALSE == no (default), TRUE == yes.
#'   When TRUE each block is mean-centred and then divided by its Frobenius norm so
#'   that all blocks have unit total inertia entering the ComDim loop.
#'   Has no effect on the Y-block; use \code{scale.y} for that.
#' @param scale.y Logical (default FALSE). When TRUE and \code{method = 'PLS-R'},
#'   each column of Y is mean-centred and scaled to unit variance before the PLS
#'   fit. The stored B, B0, and Ypred are back-transformed to the original Y scale,
#'   so downstream outputs (R2Y, Q2, predictions) are always in the original units.
#'   Ignored when \code{method = 'PLS-DA'} (dummy Y should not be scaled).
#' @param threshold The threshold limit to stop the iterations. If the "difference of fit" < threshold (1e-10 as default).
#' @param loquace To display the calculation times. TRUE == yes, FALSE == no (default).
#' @param CompMethod To speed up the analysis for really big multi-blocks. 'Normal' (default), 'Kernel', 'PCT', 'Tall' or 'Wide'.
#' @param Partitions To speed up the analysis for really big multi-blocks. This parameter is used if CompMethod is 'Tall' or 'Wide'.
#' @param cv.k Number of folds for k-fold cross-validation (default 7). Set to 0 to skip CV. When cv.k >= 2, Q2 and DQ2 in the output reflect cross-validated predictive ability; otherwise they reflect training-set fit (R2).
#' @return A \code{ComDim} object. All slots are populated. Key slots:
#' \describe{
#'   \item{\code{Method}}{\code{"PLS-DA"} or \code{"PLS-R"}.}
#'   \item{\code{ndim}}{Number of Common Dimensions extracted.}
#'   \item{\code{Q.scores}}{Global consensus PLS scores (\eqn{n \times ndim}).
#'     Each column \eqn{\mathbf{q}_a} (unit-norm) is the dominant left
#'     singular vector from the NIPALS PLS applied to the salience-weighted
#'     concatenated blocks.}
#'   \item{\code{T.scores}}{Named list of block-specific local scores
#'     (\eqn{n \times ndim} each).  Local loading
#'     \eqn{\mathbf{p}_{ba} = \mathbf{X}_b'\mathbf{q}_a}; local score
#'     \eqn{\mathbf{t}_{ba} =
#'     \mathbf{X}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
#'   \item{\code{P.loadings}}{Global loadings (\eqn{p_{tot} \times ndim}):
#'     \eqn{\mathbf{P} = \mathbf{X}'\mathbf{Q}}, where \eqn{\mathbf{X}} is
#'     the mean-centred (and optionally normalised) concatenated blocks.}
#'   \item{\code{Saliences}}{Block salience matrix (\eqn{ntable \times ndim}):
#'     \eqn{\lambda_{ba} =
#'     \mathbf{q}_a'\mathbf{X}_b\mathbf{X}_b'\mathbf{q}_a}.}
#'   \item{\code{R2X}}{Proportion of X variance captured by each component
#'     (named vector, length \eqn{ndim}).  Let \eqn{\mathbf{t}_a} be the
#'     NIPALS PLS X-score vector for component \eqn{a} on the salience-
#'     weighted blocks; then
#'     \deqn{R2X_a = \|\mathbf{t}_a\|^4 \big/ \sum_k \|\mathbf{t}_k\|^4.}}
#'   \item{\code{R2Y}}{Cumulative Y-variance explained (named vector, length
#'     \eqn{ndim}).  \eqn{R2Y_a} is the \eqn{R^2} from an OLS regression of
#'     \eqn{\mathbf{Y}} on the first \eqn{a} global scores with an intercept:
#'     \deqn{R2Y_a = 1 - RSS_a / TSS_Y,}
#'     where \eqn{RSS_a} is the residual SS when predicting \eqn{\mathbf{Y}}
#'     from \eqn{[1, \mathbf{q}_1, \ldots, \mathbf{q}_a]}.
#'     \strong{Note:} \eqn{R2Y_a} is cumulative — it reflects the total
#'     Y-variance explained by the first \eqn{a} components together,
#'     not the marginal contribution of component \eqn{a} alone.}
#'   \item{\code{Q2}}{Predictive Q2 per response column (PLS-R) or per class
#'     (PLS-DA), named accordingly:
#'     \deqn{Q2 = 1 - PRESS / TSS_Y,}
#'     where \eqn{PRESS = \sum_i (\hat{y}_i - y_i)^2} and
#'     \eqn{TSS_Y = \sum_i (y_i - \bar{y})^2}.
#'     When \code{cv.k >= 2}, \eqn{\hat{y}_i} are out-of-sample
#'     cross-validated predictions; otherwise training-set predictions are
#'     used (i.e. Q2 = R2Y for the full model).}
#'   \item{\code{DQ2}}{(PLS-DA only) Discriminant Q2 per class.  Only
#'     penalising residuals contribute to the sum:
#'     \deqn{DQ2 = 1 - PRESSD / TSS_Y,}
#'     where \eqn{PRESSD} sums \eqn{\hat{y}_i^2} for class-0 samples with
#'     \eqn{\hat{y}_i > 0}, and \eqn{(\hat{y}_i - 1)^2} for class-1 samples
#'     with \eqn{\hat{y}_i < 1}.  Same cross-validation logic as \code{Q2}.}
#'   \item{\code{Singular}}{Squared L2 norm of the NIPALS PLS score vector
#'     per component (\eqn{\|\mathbf{t}_a\|^2}), used to derive
#'     \code{R2X}.}
#'   \item{\code{VIP}}{Global VIP scores (named numeric vector, length
#'     \eqn{p_{tot}}) using the Wold formula:
#'     \deqn{VIP_j = \sqrt{p_{tot}
#'       \cdot \frac{\sum_a s_a \tilde{w}_{j,a}^2}{\sum_a s_a}},}
#'     where \eqn{s_a = \|\mathbf{t}_a\|^2 \|\mathbf{q}_a\|^2},
#'     \eqn{\tilde{w}_{j,a} = w_{j,a} / \|\mathbf{w}_a\|} is the
#'     L2-normalised \eqn{j}-th element of the \eqn{a}-th NIPALS weight
#'     vector, and \eqn{p_{tot}} is the total number of variables.}
#'   \item{\code{VIP.block}}{Named list (one \code{data.frame} per block)
#'     with columns \code{p} (per-block predictive VIP computed with block
#'     size \eqn{p_b} instead of \eqn{p_{tot}}) and \code{tot} (= \code{p}
#'     for PLS; included for consistency with OPLS output).  Row names are
#'     variable names.}
#'   \item{\code{PLS.model}}{List with: \code{W} (NIPALS X weight matrix,
#'     \eqn{p_{tot} \times ndim}); \code{B} (regression coefficients,
#'     \eqn{p_{tot} \times ncol(Y)},
#'     \eqn{\mathbf{B} =
#'     \mathbf{W}(\mathbf{P}'\mathbf{W})^{-1}\mathbf{Q}'},
#'     in original Y units); \code{B0} (intercept vector, length
#'     \eqn{ncol(Y)},
#'     \eqn{\mathbf{B}_0 = \bar{\mathbf{y}} - \bar{\mathbf{x}}\mathbf{B}});
#'     \code{Y} (original response matrix as supplied).
#'     Training-set Y predictions:
#'     \eqn{\hat{\mathbf{Y}} = \mathbf{X}\mathbf{B} + \mathbf{B}_0}.}
#'   \item{\code{cv}}{Cross-validation results when \code{cv.k >= 2} (empty
#'     list otherwise): \code{k} (number of folds), \code{fold}
#'     (sample-to-fold vector), \code{Ypred} (\eqn{n \times ncol(Y)} matrix
#'     of out-of-sample predictions), \code{Q2} (CV Q2 per class/response),
#'     \code{DQ2} (mean CV DQ2 across classes, PLS-DA only),
#'     \code{DQ2.perclass} (CV DQ2 per class, PLS-DA only).}
#'   \item{\code{Prediction}}{Training-set predictions: \code{Y.pred}
#'     (\eqn{n \times ncol(Y)}); for PLS-DA also \code{decisionRule},
#'     \code{trueClass} (character vector), \code{predClass} (data.frame),
#'     \code{Sensitivity} and \code{Specificity} (named per class),
#'     \code{confusionMatrix} (named list of 2x2 matrices, one per class).}
#'   \item{\code{Mean}}{List with \code{MeanMB} (column means per block),
#'     \code{MeanY} (column means of Y before any scaling), and
#'     \code{ScaleY} (column SDs of Y; all ones when \code{scale.y = FALSE}).}
#'   \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms for block
#'     normalisation.}
#'   \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
#'     mapping each row of \code{P.loadings} and each element of \code{VIP}
#'     to its block.}
#'   \item{\code{runtime}}{Total computation time in seconds.}
#' }
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' batch_b1 <- rep(1, 10)
#' b2 <- matrix(rnorm(2400), 30, 80)
#' batch_b2 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
#' mb <- MultiBlock(
#'   Samples = list(
#'     b1 = paste0("samples_", 1:10),
#'     b2 = rep(paste0("samples_", 1:10), 3)
#'   ),
#'   Data = list(b1 = b1, b2 = b2),
#'   Batch = list(b1 = batch_b1, b2 = batch_b2),
#'   ignore.size = TRUE
#' )
#' rw <- SplitRW(mb)
#' y <- scale(1:10, center = TRUE)
#' results.plsr <- ComDim_PLS(rw, y, 2, method = "PLS-R")
#' groups <- c(rep("A", 5), rep("B", 5))
#' results.plsda <- ComDim_PLS(rw, y = groups, 2, method = "PLS-DA")
#' @references Jouan-Rimbaud Bouveresse D, Rutledge DN (2024).
#'   A synthetic review of some recent extensions of ComDim.
#'   \emph{Journal of Chemometrics}, 38(5), e3454.
#'   \doi{10.1002/cem.3454}
#' @export
ComDim_PLS <- function(MB = MB, y = y, ndim = NULL,
                       method = c("PLS-DA", "PLS-R"),
                       decisionRule = c("fixed", "max")[2],
                       normalise = FALSE, scale.y = FALSE,
                       threshold = 1e-10,
                       loquace = FALSE,
                       CompMethod = "Normal", Partitions = 1,
                       cv.k = 7) {
  # INITIALISATION

  if (!(method == "PLS-DA" || method == "PLS-R")) {
    stop("'method' must be 'PLS-DA' or 'PLS-R'.")
  }

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")

  start_ini_time <- Sys.time() # To start counting calculation time for the initialization

  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is  not a MultiBlock.")
  }

  ntable <- length(blockNames(MB)) # number of blocks
  nrowMB <- length(sampleNames(MB)) # number of samples
  variable_number <- ncol(MB) # Number of variables per block

  give_error <- 0

  # Remove all variables with 0 variance. If there are no remaining variables, give_error
  numvar <- 0
  numblock0 <- vector()
  for (i in 1:ntable) {
    var0 <- apply(MB@Data[[i]], 2, function(x) {
      var(x)
    })
    var0 <- which(var0 == 0)

    if (length(var0) != 0) {
      numvar <- numvar + length(var0)
      if (length(var0) == length(MB@Variables[[i]])) {
        numblock0 <- append(numblock0, i)
      }
      MB@Data[[i]] <- MB@Data[[i]][, setdiff(1:variable_number[i], var0)]
      MB@Variables[[i]] <- MB@Variables[[i]][setdiff(1:variable_number[i], var0)]
      variable_number[i] <- length(MB@Variables[[i]])
    }
  }
  if (numvar != 0) {
    warning(sprintf("Number of variables excluded from the analysis because of their zero variance: %d", numvar))
  }
  if (length(numblock0) != 0) {
    numblock0 <- rev(numblock0) # To sort in decreasing order.
    for (i in seq_along(numblock0)) {
      MB@Data[[numblock0[i]]] <- NULL
      MB@Variables[[numblock0[i]]] <- NULL
      if (numblock0[[i]] %in% MB@Batch) {
        MB@Batch[[numblock0[i]]] <- NULL
      }
      if (numblock0[[i]] %in% MB@Metadata) {
        MB@Metadata[[numblock0[i]]] <- NULL
      }
    }
    warning(sprintf("Number of blocks excluded from the analysis because of their zero variance: %d", length(numblock0)))
  }

  ntable <- length(blockNames(MB)) # number of blocks
  variable_number <- ncol(MB) # In case the number of blocks has changed.

  if (length(MB@Data) == 0) { # In case all blocks had 0 variance...
    warning("All blocks had 0 variance")
    give_error <- 1
  }

  # Check for infinite or missing values (they cannot be handled with svd)
  for (i in 1:ntable) {
    if (any(is.na(MB@Data[[i]]))) {
      stop("The MB contains NA values. They must be removed first for ComDim analysis.")
    }
    if (any(is.infinite(MB@Data[[i]]))) {
      stop("The MB contains infinite values. They must be removed first for ComDim analysis.")
    }
  }

  if (give_error) {
    stop("The data is not ready for ComDim.")
  } else {
    message("The data can be used for ComDim.")
  }


  if (is.null(ndim)) {
    ndim <- ntable # If the number of components is not defined,
    # the number of components to extract is equal to the number of blocks.
  }

  ## Check y matrix (convert to dummy matrix if it is not already.)
  if (method == "PLS-DA") {
    tmp <- unique(as.vector(y))
    if (all(tmp %in% c(0, 1))) { # Is it dummy?

      if (!is.matrix(y)) { # If it's a vector and the method is PLS-R
        y <- as.matrix(y)
      }

      classVect <- rep(NA, nrow(y))

      if (ncol(y) == 1) {
        classVect <- as.vector(y)
      } else {
        if (any(rowSums(y) != 1)) {
          if (any(rowSums(y) == 0)) {
            stop("At least one sample was not assigned to a class.")
          } else if (any(colSums(y) > 1)) {
            stop("At least one sample was assigned to more than one class.")
          }
        }
      }
      if (is.null(colnames(y))) {
        colnames(y) <- paste0("Class", seq_len(ncol(y)))
      }
      for (i in seq_len(ncol(y))) {
        classVect[which(y[, i] == 1)] <- colnames(y)[i]
      }
    } else { # If not dummy
      if ((is.matrix(y) || is.data.frame(y)) && ncol(y) > 1) {
        stop("ComDim-PLS can only be applied if Y is a dummy matrix or a class vector")
      }
      classVect <- as.vector(y)
      classVect_sorted <- sort(unique(classVect))

      # Now generate the dummy matrix from y
      y <- matrix(rep(0, length(classVect_sorted) * length(classVect)),
        ncol = length(classVect_sorted), nrow = length(classVect)
      )
      for (i in seq_along(classVect_sorted)) {
        y[which(classVect == classVect_sorted[i]), i] <- 1
      }
      colnames(y) <- classVect_sorted
      rm(classVect_sorted)
    }
  }

  if (!is.matrix(y)) { # If it's a vector and the method is PLS-R
    y <- as.matrix(y)
  }

  # Store original y (after validation / dummy conversion) for use in R2Y, Q2,
  # and as the response saved in PLS.model$Y.
  y_raw <- y

  # Scale Y for PLS-R if requested.
  # For PLS-DA the dummy matrix must not be scaled; emit a warning and skip.
  if (scale.y && method == "PLS-DA") {
    warning("'scale.y = TRUE' is ignored when method = 'PLS-DA': dummy Y matrices should not be centred or scaled.")
    scale.y <- FALSE
  }
  if (scale.y) {
    y_center <- colMeans(y)
    y_sd <- apply(y, 2, sd)
    y_sd[y_sd < .Machine$double.eps] <- 1 # guard against constant columns
    y <- sweep(sweep(y, 2, y_center, "-"), 2, y_sd, "/")
  } else {
    y_center <- colMeans(y)
    y_sd <- rep(1, ncol(y))
  }

  pieceBar <- 4 + 2 * ntable + ndim # Number of updates in the progress bar.
  pieceBar <- 80 / pieceBar
  total_progress <- pieceBar

  DimLabels <- paste0("CC", 1:ndim) # One label per component.
  TableLabels <- blockNames(MB) # One label per block.

  # isT <- grepl(output, 'T')
  # isL <- grepl(output, 'L')
  # isP <- grepl(output, 'P')

  if (CompMethod == "Tall" && any(variable_number > 1000)) {
    message("The method 'Tall' is not recommeded for large blocks (> 1000 columns).")
    message("Do you still want to continue with the analysis?")
    message("You can change now the CompMethod if you prefer: (yes/no)")
    ans1 <- tolower(readline("Type your choice: (y/n)"))
    if (ans1 == "y" || ans1 == "yes") {
      message("Which method do you want to use instead (Normal/Kernel/PCT/Wide)?")
      ans2 <- tolower(readline("Type your choice: (n/k/p/w)"))
      if (ans2 == "n" || ans2 == "normal") {
        CompMethod <- "Normal"
      } else if (ans2 == "k" || ans2 == "kernel") {
        CompMethod <- "Kernel"
      } else if (ans2 == "p" || ans2 == "pct") {
        CompMethod <- "PCT"
      } else if (ans2 == "w" || ans2 == "wide") {
        CompMethod <- "Wide"
      } else {
        stop("Wrong answer typed.")
      }
    } else if (ans1 == "n" || ans1 == "no") {
      message("The method 'Tall' will be used")
    } else {
      stop("Wrong answer typed.")
    }
  }

  if (CompMethod == "Kernel" && nrowMB >= min(variable_number)) {
    stop(sprintf(
      paste0("CompMethod = 'Kernel' requires n < p for every block ",
             "(n = %d, smallest block p = %d). ",
             "Use 'Normal', 'PCT', or 'Wide' instead."),
      nrowMB, min(variable_number)
    ))
  }

  end_ini_time <- Sys.time() # To end the count of the calculation time.

  if (loquace) {
    message(sprintf("Initialisation finished after : %s millisecs",
                    (end_ini_time - start_ini_time) * 1000))
  }

  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # NORMALISATION

  # start_norm_time <- Sys.time()  # To start counting calculation time for the initialization

  X_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))
  Xnorm_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))

  res_calib <- list()
  temp_tabCalib <- list()
  s_n <- list()
  res_calib$SingVal <- as.vector(NULL)
  res_calib$NormMB <- as.vector(NULL)
  res_calib$MeanMB <- list()

  for (i in 1:ntable) {
    res_calib$MeanMB[[TableLabels[i]]] <- colMeans(MB@Data[[i]])
    names(res_calib$MeanMB[[TableLabels[i]]]) <- MB@Variables[[i]]
    # res_calib$MeanMB$i <- TableLabels

    if (normalise) {
      # Normalise original blocks

      X_mean <- MB@Data[[i]] - matrix(data = rep(1, nrowMB), ncol = 1, nrow = nrowMB) %*% res_calib$MeanMB[[i]]
      # In a previous version (that worked), I had used as.matrix instead of matrix.
      XX <- X_mean * X_mean

      Norm_X <- sqrt(sum(XX))
      X_Normed <- X_mean / Norm_X

      res_calib$NormMB[i] <- Norm_X

      temp_tabCalib[[i]] <- X_Normed
      s_n[[i]] <- X_Normed
    } else {
      res_calib$NormMB[[TableLabels[i]]] <- rep(1, length(MB@Variables[[i]]))
      names(res_calib$NormMB[[TableLabels[i]]]) <- MB@Variables[[i]]
      # res_calib$MeanMB$Variables <- MB@Variables[[i]]
      # res_calib$MeanMB$i <- TableLabels

      temp_tabCalib[[i]] <- MB@Data[[i]]
      s_n[[i]] <- MB@Data[[i]]
    }

    if (i == 1) {
      X_mat[, 1:variable_number[1]] <- MB@Data[[i]]
      Xnorm_mat[, 1:variable_number[1]] <- temp_tabCalib[[i]]
    } else {
      beg <- sum(variable_number[1:(i - 1)]) + 1
      ending <- sum(variable_number[1:i])
      X_mat[, beg:ending] <- MB@Data[[i]]
      Xnorm_mat[, beg:ending] <- temp_tabCalib[[i]]
    }

    # PLS addition
    meanX <- colMeans(Xnorm_mat)
    meanY <- colMeans(y)

    # End PLS addition

    norm_comdim <- Sys.time()
    if (loquace) {
      message(sprintf("Normalization of block %s finished after : %s millisecs", i, (norm_comdim - start_ini_time) * 1000))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  # res_calib$NormMB$i <- 'Norm'
  # names(res_calib$NormMB$Data) <- TableLabels

  nR <- nrow(Xnorm_mat)
  nC <- ncol(Xnorm_mat)

  # COMPRESSION

  s_r <- Compress_Data_2020(s_n, CompMethod, Partitions)

  end_comp_time <- Sys.time()

  if (loquace) {
    message(sprintf("Compression finished after : %s millisecs", (end_comp_time - start_ini_time) * 1000))
  }

  total_progress <- total_progress + pieceBar * ntable
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # PLS addition
  ICs <- ndim
  # End of PLS addition

  ## DO ComDim
  # ini_comdim <- Sys.time()

  # saliences <- list()
  saliences <- matrix(, ncol = ndim, nrow = ntable)
  # Q <- list()
  Q <- matrix(, ncol = ndim, nrow = nrowMB)
  unexplained <- ntable # Warning!: true only if the tables were set to norm=1
  varexp <- as.vector(NULL)
  varexp.y <- as.vector(NULL)
  PLS2_Scores <- matrix(, ncol = ndim, nrow = nrowMB)
  PLS2_P <- matrix(, ncol = ndim, nrow = sum(variable_number))
  PLS2_W <- matrix(, ncol = ndim, nrow = sum(variable_number))
  PLS2_Q <- matrix(, ncol = ndim, nrow = ncol(as.matrix(y)))
  PLS2_U <- matrix(, ncol = ndim, nrow = nrowMB)

  for (dims in 1:ndim) {
    previousfit <- 10000
    deltafit <- 1000000
    lambda <- matrix(rep(1, ntable), nrow = ntable, ncol = 1)

    qini <- as.vector(s_r[[1]][, 1])
    qini <- qini / sqrt(as.vector(qini %*% qini)) # random initialization of t + set to unit length
    qold <- 100

    iters <- 0
    while (norm(qold - qini, "2") > threshold && iters < 100) {
      iters <- iters + 1
      qold <- qini
      q <- 0

      W <- matrix(, nrow = nrowMB, ncol = 0)
      for (j in 1:ntable) {
        W <- cbind(W, sqrt(lambda[j]) * s_r[[j]])
      }

      # usv <-svd(W) This was for ComDim_PCA

      # Addition for PLS
      PLS2_Res <- PLS2_DNR_2018(X = W, Y = y, lvs = ICs)

      PLS2_Scores[, dims] <- PLS2_Res$Scores[, 1]
      PLS2_P[, dims] <- PLS2_Res$P[, 1]
      PLS2_W[, dims] <- PLS2_Res$W[, 1]
      PLS2_Q[, dims] <- PLS2_Res$Q[, 1]
      PLS2_U[, dims] <- PLS2_Res$U[, 1]

      sv <- sqrt(t(PLS2_Scores[, dims]) %*% PLS2_Scores[, dims]) # For R2X
      varexp.y[dims] <- (t(PLS2_U[, dims]) %*% PLS2_U[, dims])^2 # For R2Y (see varexp)
      qtmp <- PLS2_Scores[, dims] / as.vector(sv)
      # End of addition for PLS

      # qtmp <- usv$u[,1]

      for (j in 1:ntable) {
        # takes each table into account for lambda after PCA
        lambda[j] <- t(qtmp) %*% (s_r[[j]] %*% t(s_r[[j]])) %*% qtmp
        q <- q + lambda[j] * qtmp
      }

      q <- q / sqrt(as.vector(t(q) %*% q)) # standardizes t
      if (abs(min(q)) > abs(max(q))) {
        q <- -q
      }
      qini <- q
    } # deltafit>threshold

    ICs <- ICs - 1

    saliences[, dims] <- lambda
    Q[, dims] <- q


    # SingVal is square of the SVD singular value
    # because here it is based on X, not X.X'
    # Does NOT work for Kernel & Wide & Tall
    if (CompMethod == "Normal" || CompMethod == "PCT") {
      res_calib$SingVal[dims] <- sv^2 # Calculated from the scores (new for PLS)
      varexp[dims] <- res_calib$SingVal[dims]^2
    } else {
      varexp[dims] <- 0 # This will be overwritten at the end
      res_calib$SingVal[dims] <- 0 # This will be overwritten at the end
    }


    # Deflate blocks
    aux <- diag(nrowMB) - as.matrix(q) %*% t(as.matrix(q))
    for (j in 1:ntable) {
      s_r[[j]] <- aux %*% s_r[[j]]
    }

    iter_comdim <- Sys.time()
    if (loquace) {
      message(sprintf("Component %s determined after : %s millisecs", dims, (iter_comdim - start_ini_time) * 1000))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  # res_calib$SingVal$i <- 'Singular value'
  names(res_calib$SingVal) <- DimLabels # Dimensions

  # Q$i <- data[[1]]$Samples
  # Q$v <- DimLabels
  colnames(Q) <- DimLabels
  rownames(Q) <- MB@Samples
  res_calib$Q <- Q

  ## Adding metadata. Metadata extracted from the first block
  res_calib$metadata <- list()
  if (length(MB@Metadata) != 0) {
    res_calib$metadata[[1]] <- MB@Metadata
  }

  end_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Scores finished after : %s millisecs", (end_comdim - start_ini_time) * 1000))
  }
  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # res_calib$s_n <- s_n
  # res_calib$s_r <- s_r
  rm(s_n)
  rm(s_r)

  ##
  # explained <- list()
  explained <- varexp / sum(varexp)
  names(explained) <- DimLabels
  # explained$i <-'% explained'
  # explained$v <- DimLabels #Dimensions
  res_calib$explained <- explained
  rm(varexp)
  rm(explained)

  ## Calculate Sums of saliences for each Dim
  # Sum_saliences_Dim <- list()
  Sum_saliences_Dim <- colSums(saliences)
  names(Sum_saliences_Dim) <- DimLabels
  # Sum_saliences_Dim$i <-'Sum Dim Saliences'
  # Sum_saliences_Dim$v <- DimLabels #Dimensions

  res_calib$Sum_saliences_Dim <- Sum_saliences_Dim
  rm(Sum_saliences_Dim)

  # Calculate Sums of saliences for each Table
  # Sum_saliences_Tab <- list()
  Sum_saliences_Tab <- rowSums(saliences)
  # Sum_saliences_Tab$i <- 'Sum Tab Saliences'
  names(Sum_saliences_Tab) <- TableLabels
  # Sum_saliences_Tab$v <- TableLabels #tables

  res_calib$Sum_saliences_Tab <- Sum_saliences_Tab
  rm(Sum_saliences_Tab)

  # saliences$i <- TableLabels #tables
  # saliences$v <- DimLabels #Dimensions
  res_calib$saliences <- saliences
  colnames(res_calib$saliences) <- DimLabels
  rownames(res_calib$saliences) <- TableLabels

  ## Calculate Normalised concatenated Xs ('Calib') from col
  # Calculate concatenated CD loadings
  # Reorganise Loadings - 1 matrix / LV

  ## If Output = NULL, nothing else is calculated
  # if(!(is.null(output))){
  L_CD_Vec <- NULL
  L_X_Vec <- NULL
  Calib <- NULL

  nCalib <- nrowMB
  # }

  ## Already defined in during ComDim initiallization
  # isT <- grepl(output, 'T')
  # isL <- grepl(output, 'L')
  # isP <- grepl(output, 'P')
  # isLP <- c(isP,isL)
  # isTLP <- c(isT,isP,isL)

  # tic
  # Calculate concatenated CD loadings
  # Reorganise Loadings - 1 matrix / LV
  # L_CD <- list()
  L_X <- list()
  T_Loc <- list()
  for (i in 1:ntable) { # Prepare lists
    # L_CD[[TableLabels[i]]] <- matrix(,ncol = ndim, nrow = ncol(temp_tabCalib[[i]]))
    # L_X[[TableLabels[i]]] <- matrix(,ncol = ndim, nrow = ncol(temp_tabCalib[[i]]))
    T_Loc[[TableLabels[i]]] <- matrix(, ncol = ndim, nrow = nrowMB)
  }
  b <- matrix(, ncol = ndim, nrow = ntable) # Used to calculate R2X in Kernel and Tall

  if (!requireNamespace("pracma", quietly = TRUE)) {
    stop("The package pracma is needed.")
  }

  for (j in 1:ndim) {
    T_mat <- matrix(, nrow = nrowMB, ncol = 0)

    for (i in 1:ntable) {
      temp <- t(temp_tabCalib[[i]]) %*% res_calib$Q[, j] # Scaled CD 'local' Loadings

      # if  (!(is.null(isP))){
      # L_CD[[TableLabels[i]]][,j] <- temp
      # }

      # if  (!(is.null(isL))){
      #  L_X[[TableLabels[i]]][,j] <- t(data[[i]]$Data)%*%res_calib$Q[,j] #Unscaled X 'local' Loadings;
      # }

      # T_mat <- cbind(T_mat,temp_tabCalib[[i]]%*%(temp/as.vector(t(temp)%*%temp))) #local Scores
      T_mat <- cbind(T_mat, temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp))) # local Scores

      # if(!(is.null(isT))){
      # T_Loc[[TableLabels[i]]][,j] <- temp_tabCalib[[i]]%*%(temp/as.vector(t(temp)%*%temp)) # local Scores
      T_Loc[[TableLabels[i]]][, j] <- temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp)) # local Scores
      # }


      # Deflate each temp_tabCalib
      temp_tabCalib[[i]] <- temp_tabCalib[[i]] - res_calib$Q[, j] %*% t(temp)
    }

    # For each CC
    # MLR b-coefficients between Local and Global Scores
    # [b0,b[,j]] <- mlr_DB(T_mat,Q$d[,j],0)
    # b=pinv(X'*X)*X'*Y;
    b[, j] <- pracma::pinv(t(T_mat) %*% T_mat) %*% t(T_mat) %*% res_calib$Q[, j]

    # b0 <- 0
  }

  # If Output==[], nothing else is calculated
  # if(!(is.null(output))){
  # Calculate Global Loadings
  # if(!(is.null(isP))){
  L_CD_Vec <- t(Xnorm_mat) %*% res_calib$Q # Scaled CD 'global' Loadings
  # }

  # if(!(is.null(isL))){
  #  L_X_Vec <- t(X_mat)%*%res_calib$Q # Unscaled X 'global' Loadings
  # }
  # }

  #### If Output==[], nothing else is calculated
  load_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Loadings finished after : %s millisecs", (load_comdim - start_ini_time) * 1000))
  }

  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  rm(X_mat)
  rm(temp_tabCalib)
  rm(temp)

  ## Output
  end_function <- Sys.time()

  ## Addition for PLS
  ComDimPLS2_B <- PLS2_W %*% pracma::pinv(t(PLS2_P) %*% PLS2_W) %*% t(PLS2_Q)
  PLS2_Res$ComDimPLS2_B <- ComDimPLS2_B
  ComDimPLS2_B0 <- meanY - colMeans(Xnorm_mat) %*% ComDimPLS2_B
  PLS2_Res$ComDimPLS2_B0 <- ComDimPLS2_B0
  res_calib$PLS2_Res <- PLS2_Res

  # Back-transform B and B0 to the original Y scale when scale.y = TRUE.
  # B_raw[,j]  = B_scaled[,j]  * y_sd[j]
  # B0_raw[j]  = B0_scaled[j]  * y_sd[j] + y_center[j]
  # After this the existing Ypred = Xnorm_mat %*% B + B0 produces predictions
  # in the original (unscaled) Y units.
  if (scale.y) {
    PLS2_Res$ComDimPLS2_B <- sweep(PLS2_Res$ComDimPLS2_B, 2, y_sd, "*")
    PLS2_Res$ComDimPLS2_B0 <- PLS2_Res$ComDimPLS2_B0 * y_sd + y_center
    res_calib$PLS2_Res <- PLS2_Res
    y <- y_raw # restore original y for all downstream computations
  }
  # End of addition for PLS

  res_calib$b <- b
  colnames(res_calib$b) <- DimLabels
  rownames(res_calib$b) <- TableLabels
  # res_calib$b$i <- TableLabels # tables
  # res_calib$b$v <- DimLabels  # Dimensions

  # if(!(is.null(isT))){
  # res_calib$T_Loc$i <- res_calib$Q$i # samples
  # res_calib$T_Loc$v <- DimLabels # Dimensions
  res_calib$T_Loc <- T_Loc

  # }

  # T_Loc is no longer needed
  rm(T_Loc)

  # if(!(is.null(isP))){

  varnames <- vector()
  for (i in 1:ntable) {
    varnames <- append(varnames, MB@Variables[[i]])
  }
  # if(length(which(duplicated(varnames))) != 0){
  #  warning('There are at least two variables with the same name. Their name in the rownames of P has been altered.\n')
  # }

  res_calib$P <- L_CD_Vec
  # res_calib$P$i <- varnames # numbers of all variables;
  # res_calib$P$v <- DimLabels # Dimensions
  colnames(res_calib$P) <- DimLabels
  rownames(res_calib$P) <- varnames
  rownames(PLS2_Res$ComDimPLS2_B) <- varnames
  if (!is.null(colnames(y))) colnames(PLS2_Res$ComDimPLS2_B) <- colnames(y)

  # res_calib$P_Loc <- L_CD
  # res_calib$P_Loc$i <- TableLabels # Tables
  # res_calib$P_Loc$v <- DimLabels # Dimensions
  # for(i in 1:ntable){
  #  rownames(res_calib$P_Loc[[TableLabels[i]]]) <- MB@Variables[[i]]
  #  colnames(res_calib$P_Loc[[TableLabels[i]]]) <- DimLabels
  # }
  # }

  rm(L_CD_Vec)
  # rm(L_CD)

  ## -------------------------------------------------------------------------
  ## VIP (Variable Importance in Projection) — Wold formula.
  ##
  ##   W_j  = L2-normalised column of the PLS weight matrix W (per block)
  ##   s_a  = colSums(T^2) * colSums(Q^2)   [length ndim]
  ##   VIP_j = sqrt( p * sum_a(W_ja^2 * s_a) / sum_a(s_a) )
  ##
  ## where p = p_b (block size) for per-block VIP, p_total for global VIP.
  ## Per-block result: data.frame with columns "p" (pred VIP) and "tot" (= "p").
  ## -------------------------------------------------------------------------

  s_vip     <- colSums(PLS2_Scores^2) * colSums(PLS2_Q^2)
  sum_s_vip <- sum(s_vip)

  if (sum_s_vip < .Machine$double.eps) {
    warning("Sum of explained SS is zero; VIP scores cannot be computed.")
    VIP_global <- setNames(rep(NA_real_, length(varnames)), varnames)
    VIP_blocks <- setNames(
      lapply(seq_len(ntable), function(i) {
        v <- rep(NA_real_, variable_number[i])
        data.frame(p = v, tot = v, row.names = MB@Variables[[i]])
      }),
      TableLabels
    )
  } else {
    # Global VIP: L2-normalise W columns across all variables
    W_norm_all <- apply(PLS2_W, 2, function(w) {
      nrm <- sqrt(sum(w^2)); if (nrm > 1e-14) w / nrm else w
    })
    p_total    <- nrow(PLS2_W)
    VIP_global <- setNames(
      as.vector(sqrt(p_total * (W_norm_all^2 %*% s_vip) / sum_s_vip)),
      varnames
    )

    # Per-block VIP: L2-normalise W columns using only the block's rows
    VIP_blocks <- setNames(vector("list", ntable), TableLabels)
    k_vip <- 1L
    for (i in seq_len(ntable)) {
      p_b  <- variable_number[i]
      idx  <- k_vip:(k_vip + p_b - 1L)
      W_b_norm <- apply(PLS2_W[idx, , drop = FALSE], 2, function(w) {
        nrm <- sqrt(sum(w^2)); if (nrm > 1e-14) w / nrm else w
      })
      vip_i <- as.vector(sqrt(p_b * (W_b_norm^2 %*% s_vip) / sum_s_vip))
      VIP_blocks[[TableLabels[i]]] <- data.frame(
        p = vip_i, tot = vip_i, row.names = MB@Variables[[i]]
      )
      k_vip <- k_vip + p_b
    }
  }
  ## -------------------------------------------------------------------------

  # if(!(is.null(isL))){
  #   res_calib$Lx$i <- t(c(1:nC)) # numbers of all variables;
  #   res_calib$Lx$v <- DimLabels # Dimensions
  #   res_calib$Lx$Data <- L_X_Vec
  #   colnames(res_calib$Lx$Data) <- DimLabels
  #   rownames(res_calib$Lx$Data) <- t(c(1:nC))
  #   #res_calib$Lx_Loc$i <- TableLabels # Tables
  #   res_calib$Lx_Loc$v <- DimLabels # Dimensions
  #   res_calib$Lx_Loc$Data <- L_X
  #   for (i in 1:ntable){
  #     colnames(res_calib$Lx_Loc$Data[[TableLabels[i]]]) <- DimLabels
  #     rownames(res_calib$Lx_Loc$Data[[TableLabels[i]]]) <- data[[i]]$Variables
  #   }
  # }
  # rm(L_X_Vec)
  # rm(L_X)

  ##
  # Singular value calculated from b and saliences
  # Since :
  # b_T_Q(:,j)=saliences.d(:,j).*saliences.d(:,j)/SingVal.d(1,j);
  if (CompMethod == "Kernel" || CompMethod == "Tall" || CompMethod == "Wide") {
    for (dims in 1:ndim) {
      res_calib$SingVal[dims] <- t(b[, dims]) %*% ((saliences[, dims]^2) / as.vector(t(b[, dims]) %*% b[, dims]))
    }
    names(res_calib$SingVal) <- DimLabels
    varexp <- res_calib$SingVal^2
    res_calib$explained <- varexp / sum(varexp)
    names(res_calib$explained) <- DimLabels
  }

  rm(saliences)

  # Define block length
  belong_block <- rep(0, nrow(res_calib$P))
  k <- 1
  for (i in 1:ntable) {
    belong_block[k:(k + variable_number[i] - 1)] <- TableLabels[i]
    k <- k + variable_number[i]
  }

  #### If Output==[], nothing else is calculated
  # if(normalise == 1){
  #  res_calib$Xnorm_mat$i <- res_calib$Q$i # samples
  #  res_calib$Xnorm_mat$v <- t(c(1:nC)) # numbers of all variables;
  #  res_calib$Xnorm_mat$Data <- Xnorm_mat
  # }
  # rm(Xnorm_mat)
  #### If Output==[], nothing else is calculated

  rm(Q)

  ## 07/08/2022 Calculate Y predictive and rescale.
  Ypred <- Xnorm_mat %*% PLS2_Res$ComDimPLS2_B

  for (i in seq_len(ncol(y))) {
    Ypred[, i] <- unlist(Ypred[, i]) + PLS2_Res$ComDimPLS2_B0[i]
  }
  rownames(Ypred) <- MB@Samples
  if (!is.null(colnames(y))) colnames(Ypred) <- colnames(y)

  ## R2Y — cumulative OLS of Y on [1 | Q_1..j] for j = 1..ndim.
  ## r2y[j] reports the fraction of Y variance explained by the first j global
  ## ComDim scores. An explicit intercept is included so that Y with non-zero
  ## mean is handled correctly.
  Y_c_r2   <- sweep(y, 2, colMeans(y))
  sstot_r2 <- sum(Y_c_r2^2)
  if (sstot_r2 < .Machine$double.eps) {
    r2y <- setNames(rep(0, ndim), paste0("CC", seq_len(ndim)))
  } else {
    r2y <- setNames(numeric(ndim), paste0("CC", seq_len(ndim)))
    for (.j in seq_len(ndim)) {
      T_r2_j     <- cbind(1, res_calib$Q[, seq_len(.j), drop = FALSE])
      B_r2_j     <- pracma::pinv(crossprod(T_r2_j)) %*% crossprod(T_r2_j, y)
      Ypred_r2_j <- T_r2_j %*% B_r2_j
      r2y[.j]    <- 1 - sum((Ypred_r2_j - y)^2) / sstot_r2
    }
  }
  rm(Y_c_r2)

  ## 07/08/2022 Predict classes
  if (method == "PLS-DA") {
    predClass <- rep(NA, nrow(y))

    if (decisionRule == "max") {
      for (i in seq_len(nrow(Ypred))) {
        xx <- which(Ypred[i, ] == max(Ypred[i, ], na.rm = TRUE))
        if (length(xx) == 1) {
          predClass[i] <- colnames(y)[xx] # Y stores the Class names
        } else {
          predClass[i] <- NaN # There is a tie.
          warning(sprintf("Class for sample %s could not be predicted.", i))
        }
      }
    } else if (decisionRule == "fixed") {
      for (i in seq_len(nrow(Ypred))) {
        xx <- which(Ypred[, i] > 1 / ncol(y))
        if (length(xx) == 1) {
          predClass[i] <- colnames(y)[xx] # Y stores the Class names
        } else {
          predClass[i] <- NaN # There is a tie.
          warning(sprintf("Class for sample %s could not be predicted.", i))
        }
      }
    }

    if (is.matrix(y)) {
      meanY <- colMeans(y)
    } else {
      meanY <- mean(y)
    }

    classy <- colnames(y)
    Q2 <- DQ2 <- specvec <- sensvec <- rep(NA, length(classy))
    confusionMatrix <- list()

    for (i in seq_along(classy)) {
      pos <- which(classVect == classy[i])
      neg <- which(classVect != classy[i])
      tp <- length(which(classVect == classy[i] & predClass == classy[i]))
      tn <- length(which(classVect != classy[i] & predClass != classy[i]))
      fp <- length(which(classVect != classy[i] & predClass == classy[i]))
      fn <- length(which(classVect == classy[i] & predClass != classy[i]))

      if (length(tp) == 0) {
        tp <- 0
      }

      if (length(tn) == 0) {
        tn <- 0
      }

      if (length(fp) == 0) {
        fp <- 0
      }

      if (length(fn) == 0) {
        fn <- 0
      }

      sensvec[i] <- tp / length(pos)
      specvec[i] <- tn / length(neg)

      cm <- matrix(c(tp, fp, fn, tn), ncol = 2, nrow = 2)
      rownames(cm) <- c("predClass1", "predClass0")
      colnames(cm) <- c("trueClass1", "trueClass0")
      confusionMatrix[[classy[i]]] <- cm

      ## DQ2
      E0 <- Ypred[neg, i] - y[neg, i] # Calculate Residuals of Class0 samples
      E1 <- Ypred[pos, i] - y[pos, i] # Calculate Residuals of Class1 samples

      E0count <- which(E0 > 0) # Find predictions for Class0 samples larger than 0
      E1count <- which(E1 < 0) # Find predictions for Class1 samples larger than 1

      SSE0_all <- t(E0) %*% E0 # Calculate SSE for those samples of Class0
      SSE1_all <- t(E1) %*% E1 # Calculate SSE for those samples of Class1

      SSE0 <- t(E0[E0count]) %*% E0[E0count] # Calculate SSE for those samples of Class0
      SSE1 <- t(E1[E1count]) %*% E1[E1count] # Calculate SSE for those samples of Class1

      PRESSD <- SSE0 + SSE1 # Calculate total SSE for all samples = PRESSD
      PRESS <- SSE0_all + SSE1_all

      Ym <- y[, i] - mean(y[, i]) # Calculate Total sum of squares (over all samples)
      TSS <- t(Ym) %*% Ym

      DQ2[i] <- 1 - (PRESSD / TSS) # Calculate DQ2
      Q2[i] <- 1 - PRESS / TSS
    }
    names(Q2)      <- classy
    names(DQ2)     <- classy
    names(sensvec) <- classy
    names(specvec) <- classy
  } else if (method == "PLS-R") {
    PRESS_r <- colSums((Ypred - y)^2)
    TSS_r <- colSums(sweep(y, 2, colMeans(y))^2)
    Q2 <- setNames(1 - PRESS_r / TSS_r, colnames(y))
  }

  # Ypred <-Ypred*meanY # Re-scale Ypred. I think this is not needed.

  ## -------------------------------------------------------------------------
  ## k-fold cross-validation: Q2 and DQ2
  ## Fold assignment is sequential:
  ##   test indices for fold k = seq(k, n, by = cv.k)
  ## -------------------------------------------------------------------------
  if (cv.k >= 2) {
    # Build the fold-id vector from sequential assignment (reproducible,
    fold_ids <- integer(nrowMB)
    for (.f in seq_len(cv.k)) fold_ids[seq.int(.f, nrowMB, by = cv.k)] <- .f

    Ypred_cv <- matrix(NA_real_, nrow = nrowMB, ncol = ncol(y))
    colnames(Ypred_cv) <- colnames(y)

    for (fold in seq_len(cv.k)) {
      test_idx  <- seq.int(fold, nrowMB, by = cv.k)
      train_idx <- setdiff(seq_len(nrowMB), test_idx)

      # Subset MB to training samples
      MB_cv <- MB
      MB_cv@Data <- lapply(MB@Data, function(x) x[train_idx, , drop = FALSE])
      MB_cv@Samples <- MB@Samples[train_idx]

      # Fit inner model (cv.k = 0 prevents nested CV; output suppressed).
      # scale.y is forwarded so each fold independently centres/scales its
      # training Y and stores back-transformed B / B0 in original Y units.
      invisible(utils::capture.output(
        res_cv_fold <- ComDim_PLS(
          MB = MB_cv,
          y = y[train_idx, , drop = FALSE],
          ndim = ndim, method = method,
          normalise = normalise,
          scale.y = scale.y,
          threshold = threshold,
          loquace = FALSE,
          CompMethod = CompMethod,
          Partitions = Partitions,
          cv.k = 0
        )
      ))

      # Build raw test X (concatenated blocks, same order as Xnorm_mat)
      X_test <- matrix(0, nrow = length(test_idx), ncol = sum(variable_number))
      k_idx <- 1
      for (i in seq_len(ntable)) {
        vi <- variable_number[i]
        X_test[, k_idx:(k_idx + vi - 1)] <- MB@Data[[i]][test_idx, , drop = FALSE]
        k_idx <- k_idx + vi
      }

      # Apply training normalisation to test set (only when normalise = TRUE)
      if (normalise) {
        k_idx <- 1
        for (i in seq_len(ntable)) {
          vi <- variable_number[i]
          mn <- res_cv_fold@Mean$MeanMB[[i]] # training col means
          nrm <- res_cv_fold@Norm$NormMB[i] # training Frobenius norm
          X_test[, k_idx:(k_idx + vi - 1)] <-
            (X_test[, k_idx:(k_idx + vi - 1)] -
              matrix(mn, nrow = length(test_idx), ncol = vi, byrow = TRUE)) / nrm
          k_idx <- k_idx + vi
        }
      }

      # Predict Y for test fold
      B_cv <- res_cv_fold@PLS.model$B
      B0_cv <- res_cv_fold@PLS.model$B0
      Ypred_cv[test_idx, ] <-
        X_test %*% B_cv +
        matrix(B0_cv, nrow = length(test_idx), ncol = ncol(y), byrow = TRUE)
    }

    # Q2 per response column (or class)
    TSS_cv <- colSums(sweep(y, 2, colMeans(y))^2)
    PRESS_cv <- colSums((Ypred_cv - y)^2)
    Q2_cv <- setNames(1 - PRESS_cv / TSS_cv, colnames(y))

    # DQ2 per class then averaged across classes (PLS-DA only).
    # Per-class values are stored in cv$DQ2.perclass;
    if (method == "PLS-DA") {
      DQ2_per_class <- setNames(rep(NA_real_, ncol(y)), colnames(y))
      for (i in seq_len(ncol(y))) {
        pos <- which(classVect == colnames(y)[i])
        neg <- which(classVect != colnames(y)[i])
        E0 <- Ypred_cv[neg, i] - y[neg, i]
        E1 <- Ypred_cv[pos, i] - y[pos, i]
        PRESSD <- sum(E0[E0 > 0]^2) + sum(E1[E1 < 0]^2)
        DQ2_per_class[i] <- 1 - PRESSD / TSS_cv[i]
      }
      DQ2_cv <- mean(DQ2_per_class)   # single averaged value (matches ComDim_OPLS)
    } else {
      DQ2_per_class <- vector()
      DQ2_cv <- vector()
    }

    # Override training R2 with CV-based Q2/DQ2
    Q2 <- Q2_cv
    if (method == "PLS-DA") DQ2 <- DQ2_cv

    cv_result <- list(
      k              = cv.k,
      fold           = fold_ids,
      Ypred          = Ypred_cv,
      Q2             = Q2_cv,
      DQ2            = if (method == "PLS-DA") DQ2_cv        else NULL,
      DQ2.perclass   = if (method == "PLS-DA") DQ2_per_class else NULL
    )
  } else {
    if (cv.k != 0) {
      warning("'cv.k' must be >= 2 or 0 (skip CV). No cross-validation performed.")
    }
    cv_result <- list()
  }
  ## -------------------------------------------------------------------------

  # The way R2X and R2Y are calculated must be verified.
  # TSS <- sum(diag(Xnorm_mat))
  # R2X <- matrix(rep(NA, ndim),
  #              ncol = 1,
  #              nrow = ndim)
  # R2Y <- vector()
  #
  # #for (i in 1:ndim){
  #   #rss <- sum(diag(Xnorm_mat - (T_mat[,i] %*% t(T_mat[,i]))))
  #   #R2X[i,1] <- 1 - rss/TSS
  #   for(j in 1:ncol(y)){
  #     PRESS <- sum((y[,j] - Ypred[,j])^2)
  #     TSSY <- sum((y[,j] - mean(y[,j]))^2)
  #     R2Y[j] <- 1 - PRESS/TSSY
  #   }
  #
  # #}
  # #rownames(R2X) <- paste0("CC",1:ndim)
  # names(R2Y) <- paste0("Y",1:ncol(y))


  # ssqX = sum(sum(Xnorm_mat*Xnorm_mat))
  # ssqY = sum(sum(y*y))
  #
  #
  # R2X <- R2Y <- vector()
  # for (i in 1:ndim){
  #
  #   for(j in 1:ncol(y)){
  #      # PRESS for Q2
  #     R2X[i] = ((t(T_mat[,i]) %*% T_mat[,i] %*% t(PLS2_P[,i]) %*% PLS2_P[,i])/ssqX)*100
  #     R2Y[i] = ((t(T_mat[,i]) %*% T_mat[,i] %*% t(PLS2_Q[i,1])%*% PLS2_Q[i,1])/ssqY)*100
  #
  #     if(ncol(y) == 2){
  #
  #     }
  #   }
  #
  # }
  #
  # PLS2_Res$R2X <- R2X
  # PLS2_Res$R2Y <- R2Y
  rm(T_mat)

  end_output <- Sys.time()
  running_time <- (end_output - start_ini_time) # Total time of analysis
  if (loquace) {
    message(sprintf("Analysis finished after : %s millisecs", running_time * 1000))
  }

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
  utils::setTxtProgressBar(progress_bar, value = 80)

  close(progress_bar)

  # Save data in a ComDim structure

  res_calib <- new("ComDim",
    Method = method, # PLS-DA or PLS-R
    ndim = ndim,
    Q.scores = res_calib$Q,
    T.scores = res_calib$T_Loc,
    P.loadings = res_calib$P,
    Saliences = res_calib$saliences,
    Orthogonal = list(),
    VIP = VIP_global,
    VIP.block = VIP_blocks,
    R2X = res_calib$explained,
    R2Y = r2y,
    Q2 = Q2,
    DQ2 = if (method == "PLS-DA") {
      DQ2
    } else {
      vector()
    },
    Singular = res_calib$SingVal,
    Mean = list(
      MeanMB = res_calib$MeanMB,
      MeanY = y_center, # original Y column means
      ScaleY = y_sd
    ), # original Y column SDs (1 if scale.y=FALSE)
    Norm = list(NormMB = res_calib$NormMB),
    PLS.model = list(
      W = PLS2_W,
      B = PLS2_Res$ComDimPLS2_B, # in original Y units
      B0 = PLS2_Res$ComDimPLS2_B0, # in original Y units
      Y = y_raw
    ), # original (unscaled) Y
    cv = cv_result,
    Prediction = if (method == "PLS-DA") { # To do?
      list(
        Y.pred = Ypred,
        decisionRule = decisionRule,
        trueClass = classVect,
        predClass = data.frame(predClass = predClass, row.names = MB@Samples),
        Sensitivity = sensvec,
        Specificity = specvec,
        confusionMatrix = confusionMatrix
      )
    } else {
      list(Y.pred = Ypred)
    },
    Metadata = res_calib$metadata,
    variable.block = belong_block,
    runtime = as.numeric(running_time)
  )

  return(res_calib)
}


#' Compress_Data_2020
#'
#' Compress large multi-block objects. Internal function of ComDim_PLS().
#' @param s_n The multi-block object.
#' @param CompMethod It can be 'Normal' (default), 'Kernel', 'PCT', 'Tall' or 'Wide'.
#' @param Partitions The number of partitions.
#' @return The compressed multi-block.
#' @references Original MATLAB implementation:
#'   \url{https://github.com/DNRutledge/ComDim/blob/main/Compress_Data_2020.m}
#' @keywords internal
Compress_Data_2020 <- function(s_n = s_n, CompMethod = CompMethod, Partitions = Partitions) {
  s_r <- list()
  ntable <- length(s_n)
  nR <- nrow(s_n[[1]])

  if (!is.null(CompMethod)) {
    if (CompMethod == "Tall") {
      for (i in 1:ntable) {
        MaxRank <- min(c(nrow(s_n[[i]]), ncol(s_n[[i]])))
        # Tall segmented PCT
        s_r[[i]] <- PCA_Tall_PCT_DNR(Xn = s_n[[i]], PCs = MaxRank, Partitions = Partitions)
      }
    } else if (CompMethod == "Wide") {
      for (i in 1:ntable) {
        Xi <- ColumnsPartition(Xn = s_n[[i]], Partitions = Partitions)
        T_all <- matrix(, nrow = nrow(s_n[[i]]), ncol = 0)
        for (p in 1:Partitions) {
          usv <- svd(Xi[[p]])
          T_in <- usv$u %*% diag(usv$d)
          T_all <- cbind(T_all, T_in)
        }
        s_r[[i]] <- T_all
      }
    } else if (CompMethod == "Kernel") {
      for (i in 1:ntable) {
        Kern <- (s_n[[i]] %*% t(s_n[[i]])) / nR
        usv <- svd(Kern)
        s_r[[i]] <- sqrt(nR) * usv$u %*% diag(sqrt(usv$d))
      }
    } else if (CompMethod == "PCT") {
      for (i in 1:ntable) {
        usv <- svd(s_n[[i]])
        s_r[[i]] <- usv$u %*% diag(usv$d)
      }
    } else if (CompMethod == "Normal") {
      for (i in 1:ntable) {
        s_r[[i]] <- s_n[[i]]
      }
    }
  }
  return(s_r)
}


#' PCA_Tall_PCT_DNR
#'
#' Compress a block using the PCT method. Internal function of ComDim_PLS().
#' @param Xn The block to compress.
#' @param PCs Number of PCs to keep.
#' @param Partitions The number of partitions.
#' @return The compressed block.
#' @keywords internal
PCA_Tall_PCT_DNR <- function(Xn = Xn, PCs = PCs, Partitions = Partitions) {
  W <- RowsPartition(Xn, Partitions)
  cols <- ncol(Xn)

  D <- t(W[[1]]) %*% W[[1]]

  if (Partitions > 1) {
    for (par in 2:Partitions) {
      D <- D + (t(W[[par]]) %*% W[[par]])
    }
  }

  vs <- eigen(D)
  vs_flipped <- vs$vectors # eigen() in R returns eigenvalues in decreasing order; columns are eigenvectors

  Tm <- matrix(, nrow = nrow(Xn), ncol = 0)
  Taux <- matrix(, nrow = 0, ncol = PCs)

  for (par in 1:Partitions) {
    to <- W[[par]] %*% vs_flipped

    Taux <- rbind(Taux, to[, 1:PCs])
  }

  Tm <- cbind(Tm, Taux)
  return(Tm)
}

#' ColumnsPartition
#'
#' Calculate vertical partitions in a block. Internal function of ComDim_PLS().
#' @param Xn A block.
#' @param Partitions The number of partitions.
#' @return A list of column-partitioned sub-matrices.
#' @keywords internal
ColumnsPartition <- function(Xn = Xn, Partitions = Partitions) {
  cols <- ncol(Xn)
  stride <- floor(cols / Partitions)
  remaining <- cols %% Partitions

  count <- 1
  W <- list()

  for (i in 1:Partitions) {
    step <- count + stride - 1
    if (i == Partitions) {
      step <- step + remaining
    }
    W[[i]] <- Xn[, count:step]
    count <- count + stride
  }
  return(W)
}

#' RowsPartition
#'
#' Calculate horizontal partitions in a block. Internal function of ComDim_PLS().
#' @param Xn A block.
#' @param Partitions The number of partitions.
#' @return A list of row-partitioned sub-matrices.
#' @keywords internal
RowsPartition <- function(Xn = Xn, Partitions = Partitions) {
  rows <- nrow(Xn)
  stride <- floor(rows / Partitions)
  remaining <- rows %% Partitions

  count <- 1
  W <- list()

  for (i in 1:Partitions) {
    step <- count + stride - 1
    if (i == Partitions) {
      step <- step + remaining
    }
    W[[i]] <- Xn[count:step, ]
    count <- count + stride
  }
  return(W)
}

PLS2_DNR_2018 <- function(X = X, Y = Y, lvs = lvs) {
  # Store original X
  X_old <- X
  px <- ncol(X)
  rows <- nrow(X)
  # if(is.vector(Y)){
  #   u_old <- as.matrix(Y)
  #   py <- 1
  # } else {
  u_old <- Y[, 1] # Any column of Y can be used to initialize u.
  py <- ncol(Y)
  # }

  # Initialize matrices
  W <- matrix(data = rep(0, px * lvs), nrow = px, ncol = lvs)
  P <- W
  Q <- matrix(data = rep(0, py * lvs), nrow = py, ncol = lvs)
  Te <- matrix(data = rep(0, rows * lvs), nrow = rows, ncol = lvs)
  U <- Te

  if (is.matrix(X)) {
    meanX <- colMeans(X)
  } else {
    meanX <- mean(X)
  }

  if (is.matrix(Y)) {
    meanY <- colMeans(Y)
  } else {
    meanY <- mean(Y)
  }

  # Necessary for PLS_ICA where nLVs decreases to 1
  B0_mat <- B_mat <- list(NULL) # Each item in the list is a matrix with px nrow and py ncol.

  for (i in 1:lvs) {
    B_mat[[i]] <- matrix(rep(0, px * py), nrow = px, ncol = py)
    B0_mat[[i]] <- matrix(rep(0, lvs * py), nrow = lvs, ncol = py)
  }

  for (i in 1:lvs) {
    limite <- 0
    while (TRUE) { # Repeat until breaking the loop
      w <- t(u_old) %*% X
      w <- w / norm(w, type = "2")
      ti <- X %*% t(w)
      ti <- ti / as.vector(w %*% t(w))
      q <- t(ti) %*% Y
      q <- as.vector(q) / as.vector(t(ti) %*% ti)
      u <- (Y %*% q) / as.vector(t(q) %*% q)
      if (norm(u - u_old, type = "2") < 1e-6) {
        break
      }
      limite <- limite + 1
      if (limite > 1000) {
        break
      }
      u_old <- u
    }
    p <- t(ti) %*% X
    p <- p / as.vector(t(ti) %*% ti)

    W[, i] <- w
    P[, i] <- p
    Q[, i] <- q
    Te[, i] <- ti
    U[, i] <- u
    X <- X - ti %*% p
    Y <- Y - ti %*% q

    i_nums <- c(1:i)
    # print(dim(pracma::pinv(t(P[,i_nums]) %*% W[,i_nums])))
    if (ncol(Y) == 1) {
      B_mat[[i]] <- (W[, i_nums] %*% pracma::pinv(t(P[, i_nums]) %*% W[, i_nums])) %*% as.matrix(Q[, i_nums])
    } else {
      B_mat[[i]] <- (W[, i_nums] %*% pracma::pinv(t(P[, i_nums]) %*% W[, i_nums])) %*% t(Q[, i_nums])
    }

    B0_mat[[i]] <- meanY - meanX %*% B_mat[[i]]
  }

  B <- (W %*% pracma::pinv(t(P) %*% W)) %*% t(Q)
  B0 <- meanY - meanX %*% B

  PLS2_Res <- list()
  PLS2_Res$B <- B
  PLS2_Res$B0 <- B0
  PLS2_Res$Scores <- Te
  PLS2_Res$P <- P
  PLS2_Res$W <- W
  PLS2_Res$U <- U
  PLS2_Res$Q <- Q

  PLS2_Res$B_mat <- B_mat
  PLS2_Res$B0_mat <- B0_mat
  PLS2_Res$Yhat <- X_old %*% B + rep(1, rows) %*% B0

  return(PLS2_Res)
}

#' MultiBlock
#'
#' Creates a MultiBlock object from a named list of data blocks.
#' @param Samples A vector of sample names shared across all blocks (optional).
#'   When omitted, sample names are taken from the row names of each block.
#'   If no row names exist and all blocks have the same number of rows,
#'   samples are numbered as integers. Use \code{ignore.names} or
#'   \code{ignore.size} for more flexible behaviour.
#' @param Data A named list of matrices or data.frames (one entry per block).
#' @param Variables A named list of variable-name vectors, one per block
#'   (optional). When omitted, column names are taken from each block; if
#'   absent, variables are numbered as integers.
#' @param Batch A named list of batch vectors, one per block (optional).
#' @param Metadata A named list of metadata data.frames, one per block
#'   (optional).
#' @param ignore.names If TRUE, sample names are not checked across blocks.
#'   All blocks must have the same number of rows unless \code{ignore.size}
#'   is also TRUE.
#' @param ignore.size If TRUE (only meaningful when \code{ignore.names = TRUE}),
#'   blocks with different row counts are accepted. The resulting MultiBlock
#'   stores a per-block Samples list and is not directly suitable for ComDim
#'   (use \code{SplitRW()} first).
#' @return A MultiBlock object.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
#' b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
#' # Minimal call: Samples and Variables are filled in automatically.
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#'
#' # With explicit sample names (enables cross-block alignment):
#' rownames(b1) <- paste0("s", 1:10)
#' rownames(b2) <- paste0("s", 1:10)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#'
#' # Blocks with different row counts (replicate design):
#' b3 <- matrix(rnorm(800), 30, 80)
#' batch_b3 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
#' mb3 <- MultiBlock(
#'   Data = list(b3 = b3), Batch = list(b3 = batch_b3),
#'   ignore.names = TRUE, ignore.size = TRUE
#' )
#' @references Puig-Castellví F, Jouan-Rimbaud Bouveresse D, Mazéas L, Chapleur O, Rutledge DN (2021).
#'   Rearrangement of incomplete multi-omics datasets combined with ComDim for evaluating replicate
#'   cross-platform variability and batch influence.
#'   \emph{Chemometrics and Intelligent Laboratory Systems}, 218, 104422.
#'   \doi{10.1016/j.chemolab.2021.104422}
#' @export
MultiBlock <- function(Samples = NULL,
                       Data,
                       Variables = NULL,
                       Batch = NULL,
                       Metadata = NULL,
                       ignore.names = FALSE,
                       ignore.size = FALSE) {
  if (is.null(Batch)) Batch <- list()
  if (is.null(Metadata)) Metadata <- list()

  # -------------------------------------------------------------------------
  # Expand any MultiBlock objects passed inside Data
  # -------------------------------------------------------------------------
  if (any(vapply(Data, is, logical(1), "MultiBlock"))) {
    expanded <- list()
    for (nm in names(Data)) {
      item <- Data[[nm]]
      if (is(item, "MultiBlock")) {
        for (bn in names(item@Data)) {
          data_block <- item@Data[[bn]]
          # Restore row names from the Samples slot so sample labels survive
          # the dimname-strip in the main processing loop.
          if (!is.list(item@Samples)) {
            rownames(data_block) <- as.character(item@Samples)
          } else if (!is.null(item@Samples[[bn]])) {
            rownames(data_block) <- as.character(item@Samples[[bn]])
          }
          expanded[[bn]] <- data_block
        }
        for (bn in names(item@Batch)) {
          if (is.null(Batch[[bn]])) Batch[[bn]] <- item@Batch[[bn]]
        }
        for (bn in names(item@Metadata)) {
          if (is.null(Metadata[[bn]])) Metadata[[bn]] <- item@Metadata[[bn]]
        }
      } else {
        expanded[[nm]] <- item
      }
    }
    Data <- expanded
  }

  # -------------------------------------------------------------------------
  # Validate Data input
  # -------------------------------------------------------------------------
  if (!is.list(Data) || length(Data) == 0) {
    stop("'Data' must be a non-empty named list of matrices or data.frames.")
  }
  if (is.null(names(Data)) || any(names(Data) == "")) {
    stop("All elements of 'Data' must be named.")
  }

  # -------------------------------------------------------------------------
  # Process each block: convert to matrix, capture row/col names
  # -------------------------------------------------------------------------
  detected_samples <- vector("list", length(Data))
  detected_vars <- vector("list", length(Data))
  block_nrow <- integer(length(Data))

  for (i in seq_along(Data)) {
    block <- Data[[i]]
    nm <- names(Data)[i]

    if (is.data.frame(block)) {
      if (any(sapply(block, is.character))) {
        stop(sprintf("Block '%s' contains character columns.", nm))
      }
      if (!is.null(rownames(block))) detected_samples[[i]] <- rownames(block)
      detected_vars[[i]] <- if (!is.null(colnames(block))) colnames(block) else seq_len(ncol(block))
      Data[[i]] <- as.matrix(block)
      dimnames(Data[[i]]) <- NULL
    } else if (is.matrix(block)) {
      if (!is.null(rownames(block))) detected_samples[[i]] <- rownames(block)
      detected_vars[[i]] <- if (!is.null(colnames(block))) colnames(block) else seq_len(ncol(block))
      dimnames(Data[[i]]) <- NULL
    } else {
      stop(sprintf("Block '%s' must be a matrix or a data.frame.", nm))
    }
    block_nrow[i] <- nrow(Data[[i]])
  }

  # -------------------------------------------------------------------------
  # Build Variables list when not provided by the user
  # -------------------------------------------------------------------------
  if (is.null(Variables)) {
    Variables <- setNames(detected_vars, names(Data))
  }

  # -------------------------------------------------------------------------
  # Build Samples when not provided by the user
  # -------------------------------------------------------------------------
  if (!is.null(Samples)) {
    # User supplied Samples explicitly — use as-is.
    # The validator will catch any size mismatches.
  } else if (ignore.names) {
    # -- ignore.names = TRUE -------------------------------------------------
    # Do not use names for cross-block alignment; only check/enforce row counts.
    if (!ignore.size && length(unique(block_nrow)) != 1) {
      stop(paste(
        "Blocks do not have the same number of rows.",
        "Provide sample names for all blocks to enable matching,",
        "or set 'ignore.size = TRUE'."
      ))
    }

    if (length(unique(block_nrow)) == 1) {
      # All blocks same size: assign integer indices
      Samples <- seq_len(block_nrow[1])
    } else {
      # ignore.size = TRUE: store a per-block Samples list
      Samples <- setNames(
        lapply(seq_along(Data), function(i) {
          if (!is.null(detected_samples[[i]])) {
            detected_samples[[i]]
          } else {
            seq_len(block_nrow[i])
          }
        }),
        names(Data)
      )
      warning(paste(
        "The resulting MultiBlock is not directly suitable for ComDim because",
        "sample numbers are not common across blocks.",
        "If data contains replicates, use SplitRW().",
        "Otherwise provide sample names and use ignore.names = FALSE."
      ))
    }
  } else {
    # -- ignore.names = FALSE (default) --------------------------------------
    # ignore.size is overridden to FALSE when ignore.names is FALSE
    ignore.size <- FALSE

    has_names <- !sapply(detected_samples, is.null)

    if (!any(has_names)) {
      # No block has row names: require equal sizes, assign integers
      if (length(unique(block_nrow)) != 1) {
        stop(paste(
          "Blocks do not have the same number of rows.",
          "Provide sample names for all blocks to enable matching,",
          "or set 'ignore.names = TRUE'."
        ))
      }
      Samples <- seq_len(block_nrow[1])
    } else if (any(has_names) && !all(has_names)) {
      # Some blocks have row names and others do not: ambiguous
      stop(paste(
        "Some blocks have row names and others do not.",
        "Either provide row names for all blocks, or none."
      ))
    } else {
      # All blocks have row names: align on common samples
      common <- detected_samples[[1]]
      for (i in seq_along(detected_samples)[-1]) {
        common <- intersect(common, detected_samples[[i]])
      }
      if (length(common) == 0) {
        stop("There are no common sample names across blocks.")
      }
      if (length(common) != max(block_nrow)) {
        warning("Some samples are not common across blocks. They will be discarded.")
      }
      # Reorder / subset each block to the common sample set
      for (i in seq_along(Data)) {
        idx <- match(common, detected_samples[[i]])
        Data[[i]] <- Data[[i]][idx, , drop = FALSE]
      }
      Samples <- common
    }
  }

  # -------------------------------------------------------------------------
  # Build and validate the MultiBlock object
  # -------------------------------------------------------------------------
  MB <- new("MultiBlock",
    Samples   = Samples,
    Data      = Data,
    Variables = Variables,
    Batch     = Batch,
    Metadata  = Metadata
  )

  if (validObject(MB)) {
    return(MB)
  } else {
    stop("The 'MultiBlock' could not be built.")
  }
}

#' SelectFeaturesRW
#'
#' Finds the important variables presenting a coordinated response across all specified replicate-blocks
#' for a given ComDim component.
#'
#' @param RW The object used as input in the ComDim analysis.
#' @param results The output object obtained in the ComDim analysis.
#' @param ndim The number of the component for which the important variables are to be identified.
#' @param blocks A vector with the indices or the names for the replicate blocks of the same data type.
#' @param threshold_cor The "times" parameter used to calculate the threshold in the following formula: cor(variable) > times * sd(cor(variables)). Minimal value that can be assigned to threshold_cor is 1.
#' @param threshold_cov The "times" parameter used to calculate the threshold in the following formula: cov(variable) > times * sd(cov(variables)). Minimal value that can be assigned to threshold_cov is 1.
#' @param mean.RW Logical value to indicate whether the RW data must be mean-centered (TRUE) or not (FALSE).
#' @param plots Parameter to indicate whether S-plots (covariance vs. correlation with the Q scores)
#'   must be produced. Possible values are \code{"NO"} for no plots, \code{"separated"} for displaying
#'   one S-plot per block individually (pausing between plots), and \code{"together"} to arrange all
#'   S-plots side by side in a single grid. In all plot variants, variables selected as important are
#'   highlighted in red and labelled with their variable name. Requires the \code{ggplot2} package;
#'   \code{"together"} additionally requires \code{gridExtra}.
#' @details
#'   The function applies an S-plot approach to identify variables that are both strongly covarying
#'   and strongly correlated with the Q scores of the chosen component. For each block in
#'   \code{blocks}, covariance (\code{s1}) and correlation (\code{s2}) of every variable with the
#'   (pseudo-inverse-scaled) Q scores are computed. A variable is considered important in a block if
#'   its absolute covariance exceeds \code{threshold_cov * sd(s1)} \emph{and} its absolute correlation
#'   exceeds \code{threshold_cor * sd(s2)}. Only variables that satisfy both criteria in \emph{all}
#'   specified blocks simultaneously are returned. The sign of the local P-loadings is used to
#'   separate variables into positive and negative groups.
#' @return A named list with two elements:
#'   \describe{
#'     \item{\code{$positive}}{Integer indices (named by variable name) of the important variables
#'       presenting a positive relationship with the Q scores (positive covariance, positive
#'       correlation, and positive local P-loading) across all specified blocks.}
#'     \item{\code{$negative}}{Integer indices (named by variable name) of the important variables
#'       presenting a negative relationship with the Q scores (negative covariance, negative
#'       correlation, and negative local P-loading) across all specified blocks.}
#'   }
#'   When \code{plots} is not \code{"NO"}, S-plots are also displayed as a side effect: each plot
#'   shows covariance on the x-axis and correlation on the y-axis, with selected variables
#'   highlighted in red.
#' @seealso \code{\link{SplitRW}}, \code{\link{ComDim_PCA}}
#' @examples
#' \donttest{
#' b1 <- matrix(rnorm(500), 10, 50)
#' batch_b1 <- rep(1, 10)
#' b2 <- matrix(rnorm(800), 30, 80)
#' batch_b2 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
#' mb <- MultiBlock(
#'   Data = list(b1 = b1, b2 = b2),
#'   Batch = list(b1 = batch_b1, b2 = batch_b2),
#'   ignore.names = TRUE, ignore.size = TRUE
#' )
#' rw <- SplitRW(mb)
#' results <- ComDim_PCA(rw, 2)
#' # Identify important variables for component 1 across replicate blocks 2, 3, and 4
#' features <- SelectFeaturesRW(RW = rw, results = results, ndim = 1, blocks = c(2, 3, 4))
#' # Use stricter thresholds and display S-plots side by side
#' features <- SelectFeaturesRW(RW = rw, results = results, ndim = 1, blocks = c(2, 3, 4),
#'                              threshold_cor = 2, threshold_cov = 2, plots = "together")
#' }
#' @export
SelectFeaturesRW <- function(RW = RW, results = results, ndim = NULL, blocks = NULL, threshold_cor = 1, threshold_cov = 1, mean.RW = TRUE, plots = "NO") {
  # Create Covariance and Correlation variables to avoid warnings during package checking.
  Covariance <- NULL
  Correlation <- NULL

  # Check data
  if (!inherits(RW, "MultiBlock")) {
    stop("'RW' is  not a MultiBlock.")
  }

  if (is.list(RW@Samples)) {
    stop("The replicate blocks to evaluate have a different sample size.")
  }
  sample_number <- length(sampleNames(RW))

  if (length(unique(lengths(variableNames(RW, blocks)))) != 1) {
    stop("The replicate blocks to evaluate have a different variable size.")
  }

  if (!("ComDim" %in% class(results))) {
    stop("'results' is not the output from a ComDim analysis.")
  }

  if (ndim > results@ndim) {
    stop("The 'ndim' value chosen for this analysis is too high.")
  }

  if (nrow(results@Q.scores) != length(sampleNames(RW))) {
    stop("The size of the local scores do not match the size of 'RW'.")
  }

  # Check if result contains the blocks of local loadings
  if (length(results@variable.block) != sum(lengths(RW@Variables))) {
    stop("The total variable number of the ComDim model does not match the total variable number of 'RW'.")
  }

  if (!(is.logical(mean.RW))) {
    stop("The parameter 'mean.RW' must be logical.")
  }

  if (!(is.character(plots))) {
    if (is.logical(plots)) {
      if (plots == FALSE) {
        plots <- "NO"
      } else {
        plots <- "separated"
      }
    }
  }

  if (!(requireNamespace("pracma", quietly = TRUE))) {
    warning("The 'pracma' package is needed but not installed.")
  }

  if (plots != "NO" && !(requireNamespace("ggplot2", quietly = TRUE))) {
    warning("The 'ggplot2' package is needed but not installed.")
    plots <- "NO"
  }

  if (plots != "NO" && !(requireNamespace("gridExtra", quietly = TRUE))) {
    warning("The 'gridExtra' package is needed but not installed.")
    plots <- "separated"
  }

  mean_blocks <- list(NULL)
  for (i in blocks) {
    # Mean-center
    if (isTRUE(mean.RW) && requireNamespace("pracma", quietly = TRUE)) {
      mean_blocks[[i]] <- RW@Data[[i]] - pracma::repmat(colMeans(RW@Data[[i]]), unique(sample_number), 1)
    } else {
      mean_blocks[[i]] <- RW@Data[[i]]
    }
  }

  if (length(ndim) != 1) {
    stop("The length of 'ndim' must be of 1.")
  }

  if (!(is.numeric(ndim))) {
    stop("'ndim' must be numeric.")
  }

  s2 <- s1 <- matrix(, nrow = length(blocks), ncol = length(variableNames(RW, blocks[1])))

  # Scale Q in case it wasnt (if the data comes from ComDim, it was).
  if (requireNamespace("pracma", quietly = TRUE)) {
    qs <- results@Q.scores[, ndim] %*% pracma::pinv(t(results@Q.scores[, ndim]) %*% results@Q.scores[, ndim])
  } else {
    qs <- results@Q.scores[, ndim]
    warning("If 'results' was not scaled, the output from this function may be incorrect.")
  }

  # Calculate the corr and cov values.
  for (i in seq_along(blocks)) {
    # Calculate the s-plot values.
    # s1(i,:) = (T' * X_Cent) ./(N - 1); % cov
    # s2(i,:) = s1(i,:) ./ (std(T) .* std(X_Cent)); %cor
    s1[i, ] <- (t(qs) %*% mean_blocks[[blocks[i]]]) / (unique(sample_number) - 1)
    s2[i, ] <- s1[i, ] / (stats::sd(qs) * apply(mean_blocks[[blocks[i]]], 2, stats::sd))
  }

  s1[is.nan(s1)] <- 0
  s2[is.nan(s2)] <- 0
  s1_imp <- list(NULL)
  s2_imp <- list(NULL)
  for (i in seq_along(blocks)) {
    if (is.numeric(threshold_cov) && length(threshold_cov) == 1) {
      if (threshold_cov < 1) {
        threshold_cov <- 1
        warning("'threshold_cov value has been modified to 1.")
      }
      std_cov <- stats::sd(s1[i, ]) * threshold_cov
      s1_imp[[i]] <- c(which(s1[i, ] > std_cov), which(s1[i, ] < -std_cov)) # Covariate and anticovariate
    } else {
      stop("'Threshold_cov' must be a numeric object of length 1.")
    }

    if (is.numeric(threshold_cor) && length(threshold_cor) == 1) {
      if (threshold_cor < 1) {
        threshold_cor <- 1
        warning("'threshold_cor value has been modified to 1.")
      }
      std_cor <- stats::sd(s2[i, ]) * threshold_cor
      s2_imp[[i]] <- c(which(s2[i, ] > std_cor), which(s2[i, ] < -std_cor))
    } else {
      stop("'Threshold_cor' must be a numeric object of length 1.")
    }
  }

  if (is.numeric(blocks)) {
    blocks <- blockNames(RW)[blocks]
  }

  names(s1_imp) <- blocks
  names(s2_imp) <- blocks
  imp_pos <- list()
  imp_neg <- list()

  for (i in blocks) {
    imp <- intersect(s1_imp[[i]], s2_imp[[i]])
    # I only use the loadings to check their sign.
    imp_pos[[i]] <- intersect(imp, which(results@P.loadings[which(results@variable.block == i), ndim] > 0))
    imp_neg[[i]] <- intersect(imp, which(results@P.loadings[which(results@variable.block == i), ndim] < 0))
  }

  common_RW_pos <- list()
  common_RW_neg <- list()
  for (i in seq_along(blocks)) {
    if (i == 1) {
      common_RW_pos <- imp_pos[[i]]
      common_RW_neg <- imp_neg[[i]]
    } else {
      common_RW_pos <- intersect(common_RW_pos, imp_pos[[i]])
      common_RW_neg <- intersect(common_RW_neg, imp_neg[[i]])
    }
  }

  names(common_RW_pos) <- RW@Variables[[blocks[1]]][common_RW_pos]
  names(common_RW_neg) <- RW@Variables[[blocks[1]]][common_RW_neg]
  ivs <- list(positive = common_RW_pos, negative = common_RW_neg)

  if (grepl("NO", plots, ignore.case = TRUE)) {
    # Do nothing.
  } else if (grepl("separated", plots, ignore.case = TRUE) && (requireNamespace("ggplot2", quietly = TRUE))) {
    for (i in seq_along(blocks)) {
      xx <- data.frame(Covariance = s1[i, ], Correlation = s2[i, ])
      color <- rep("#000000", length(variableNames(RW, blocks[i])))
      color[c(imp_pos[[i]], imp_neg[[i]])] <- rep("#FF0000", length(c(imp_pos[[i]], imp_neg[[i]])))
      labels <- rep("", length(variableNames(RW, blocks[i])))
      labels[c(imp_pos[[i]], imp_neg[[i]])] <- RW@Variables[[blocks[i]]][c(imp_pos[[i]], imp_neg[[i]])]
      useless <- readline(prompt = sprintf("S-plot for block %d shown. Press [enter] to continue", blocks[i]))
      print(ggplot2::ggplot(xx, ggplot2::aes(x = Covariance, y = Correlation)) +
        ggplot2::geom_point(color = color) +
        ggplot2::geom_text(label = labels) +
        ggplot2::ggtitle(sprintf("S-plot for Block %s and component %i", blocks[i], ndim)))
    }
  } else if (grepl("together", plots, ignore.case = TRUE) && (requireNamespace("ggplot2", quietly = TRUE))) {
    pi <- list(NULL)
    colori <- list(NULL)
    labelsi <- list(NULL)
    xx <- list(NULL)
    for (i in seq_along(blocks)) {
      colori[[i]] <- rep("#000000", length(variableNames(RW, blocks[i])))
      colori[[i]][c(imp_pos[[i]], imp_neg[[i]])] <- rep("#FF0000", length(c(imp_pos[[i]], imp_neg[[i]])))
      xx[[i]] <- data.frame(Covariance = s1[i, ], Correlation = s2[i, ])
      labelsi[[i]] <- rep("", length(variableNames(RW, blocks[i])))
      labelsi[[i]][c(imp_pos[[i]], imp_neg[[i]])] <- RW@Variables[[blocks[i]]][c(imp_pos[[i]], imp_neg[[i]])]
      pi[[i]] <- ggplot2::ggplot(xx[[i]], ggplot2::aes(x = Covariance, y = Correlation)) +
        ggplot2::geom_point(color = colori[[i]]) +
        ggplot2::geom_text(label = labelsi[[i]]) +
        ggplot2::ggtitle(sprintf("S-plot for Block %s and component %i", blocks[i], ndim))
    }
    to_eval <- "gridExtra::grid.arrange("
    for (i in seq_along(blocks)) {
      to_eval <- paste0(to_eval, sprintf("pi[[%d]], ", i))
    }
    to_eval <- paste0(to_eval, " nrow = 1)")
    eval(parse(text = to_eval))
  } else {
    warning("The 'plots' parameter was not set correctly. Plots will not be displayed.")
  }

  return(ivs) # ivs = important variables
}

#' ComDim_OPLS
#'
#' Finding common dimensions in multi-block datasets using OPLS. Also known as ConsensusOPLS
#' (ComDim-OPLS) for multiblock structures: orthogonal components uncorrelated
#' with Y are extracted from all blocks simultaneously before the predictive
#' components are computed.
#'
#' This function is a wrapper around \code{\link[ConsensusOPLS]{ConsensusOPLS}}.
#' The core kernel-OPLS extraction is delegated to that package; all ComDim
#' output slots (local scores, loadings, VIP, sensitivity, confusion matrix,
#' etc.) are computed from the returned model objects.
#' @param MB A MultiBlock object.
#' @param y The Y-block. A class vector or dummy matrix for OPLS-DA, or a
#'   numeric matrix/vector for OPLS-R.
#' @param ndim Number of predictive Common Dimensions. Default is 1.
#' @param nort Maximum number of orthogonal Common Dimensions. Default is 1.
#'   The actual number used is determined by ConsensusOPLS cross-validation and
#'   may be less than this value.
#' @param method 'OPLS-DA' for discriminant analysis or 'OPLS-R' for regression.
#' @param decisionRule Only used if method is 'OPLS-DA'. If 'fixed', samples are
#'   assigned to the class with Y-hat above 1/nclasses. If 'max', samples are
#'   assigned to the class with the highest Y-hat.
#' @param normalise To apply block normalisation. FALSE == no (default), TRUE == yes.
#' @param loquace To display the calculation times. TRUE == yes, FALSE == no (default).
#' @param cv.k Number of folds for k-fold cross-validation (default 7). Set
#'   to 0 to skip CV output. ConsensusOPLS always performs internal CV to select
#'   the optimal number of orthogonal components; when \code{cv.k >= 2} the
#'   resulting \code{Q2} and \code{DQ2} reflect that cross-validation.
#' @return A \code{ComDim} object.  All slots are populated.  Key slots:
#' \describe{
#'   \item{\code{Method}}{\code{"OPLS-DA"} or \code{"OPLS-R"}.}
#'   \item{\code{ndim}}{Number of predictive Common Dimensions.}
#'   \item{\code{Q.scores}}{Predictive global scores matrix (\eqn{n \times ndim}).}
#'   \item{\code{T.scores}}{Named list of block-specific predictive local scores.}
#'   \item{\code{P.loadings}}{Global predictive loadings.}
#'   \item{\code{Saliences}}{Predictive block salience matrix (\eqn{ntable \times ndim}).}
#'   \item{\code{Orthogonal}}{List with orthogonal component outputs: \code{nort},
#'     \code{Q.scores}, \code{T.scores}, \code{P.loadings.ort}, \code{Saliences.ort}.}
#'   \item{\code{R2X}}{Named vector (length \eqn{ndim + nort}) of X-variance fractions.}
#'   \item{\code{R2Y}}{Named vector (length \eqn{ndim + nort}) of Y-variance fractions.}
#'   \item{\code{Q2}}{Cross-validated Q2 per class/response (when \code{cv.k >= 2};
#'     otherwise training-set fit).}
#'   \item{\code{DQ2}}{(OPLS-DA only) Cross-validated discriminant Q2 per class.}
#'   \item{\code{VIP}}{Global total VIP (named vector, length \eqn{p_{tot}}).}
#'   \item{\code{VIP.block}}{Named list (one data.frame per block) with columns
#'     \code{p}, \code{o}, \code{tot}.}
#'   \item{\code{PLS.model}}{KOPLS regression objects: \code{W}, \code{B}, \code{B0}, \code{Y}.}
#'   \item{\code{cv}}{Cross-validation results when \code{cv.k >= 2}: \code{k},
#'     \code{Ypred}, \code{Q2}, \code{DQ2}.}
#'   \item{\code{Prediction}}{Training-set predictions: \code{Y.pred}; for OPLS-DA
#'     also \code{decisionRule}, \code{trueClass}, \code{predClass},
#'     \code{Sensitivity}, \code{Specificity}, \code{confusionMatrix}.}
#'   \item{\code{Mean}}{List with \code{MeanMB} and \code{MeanY}.}
#'   \item{\code{Norm}}{List with \code{NormMB}, \code{FrobNorms}, \code{RVweights}.}
#'   \item{\code{variable.block}}{Block membership of each variable.}
#'   \item{\code{runtime}}{Total computation time in seconds.}
#' }
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' y <- rep(c("A", "B"), 5)
#' results <- ComDim_OPLS(mb, y, ndim = 1, nort = 1, method = "OPLS-DA")
#' @references Boccard J, Rutledge DN (2013).
#'   A consensus OPLS-DA strategy for multiblock Omics data fusion.
#'   \emph{Analytica Chimica Acta}, 769, 30--39.
#'   \doi{10.1016/j.aca.2013.01.022}
#' @export
ComDim_OPLS <- function(MB = MB, y = y, ndim = 1, nort = 1,
                        method = c("OPLS-DA", "OPLS-R"),
                        decisionRule = c("fixed", "max")[2],
                        normalise = FALSE, #threshold = 1e-10,
                        loquace = FALSE,
                        cv.k = 7) {

  if (!requireNamespace("ConsensusOPLS", quietly = TRUE)) {
    stop("Package 'ConsensusOPLS' is required. Install with: install.packages('ConsensusOPLS')")
  }
  if (!requireNamespace("pracma", quietly = TRUE)) {
    stop("The package pracma is needed.")
  }

  # --- Input validation ---
  if (!(method == "OPLS-DA" || method == "OPLS-R")) {
    stop("'method' must be 'OPLS-DA' or 'OPLS-R'.")
  }
  if (nort < 1) stop("'nort' must be at least 1.")
  if (ndim < 1) stop("'ndim' must be at least 1.")

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
  start_ini_time <- Sys.time()

  if (!inherits(MB, "MultiBlock")) stop("'MB' is not a MultiBlock.")

  ntable         <- length(blockNames(MB))
  nrowMB         <- length(sampleNames(MB))
  variable_number <- ncol(MB)
  give_error     <- 0

  # --- Zero-variance removal ---
  numvar   <- 0
  numblock0 <- vector()
  for (i in 1:ntable) {
    var0 <- which(apply(MB@Data[[i]], 2, var) == 0)
    if (length(var0) != 0) {
      numvar <- numvar + length(var0)
      if (length(var0) == length(MB@Variables[[i]])) numblock0 <- append(numblock0, i)
      MB@Data[[i]]      <- MB@Data[[i]][, setdiff(1:variable_number[i], var0)]
      MB@Variables[[i]] <- MB@Variables[[i]][setdiff(1:variable_number[i], var0)]
      variable_number[i] <- length(MB@Variables[[i]])
    }
  }
  if (numvar != 0)
    warning(sprintf("Number of variables excluded from the analysis because of their zero variance: %d", numvar))
  if (length(numblock0) != 0) {
    numblock0 <- rev(numblock0)
    for (i in seq_along(numblock0)) {
      MB@Data[[numblock0[i]]] <- NULL
      MB@Variables[[numblock0[i]]] <- NULL
      if (names(MB@Data)[numblock0[i]] %in% names(MB@Batch))    MB@Batch[[numblock0[i]]]    <- NULL
      if (names(MB@Data)[numblock0[i]] %in% names(MB@Metadata)) MB@Metadata[[numblock0[i]]] <- NULL
    }
    warning(sprintf("Number of blocks excluded from the analysis because of their zero variance: %d",
                    length(numblock0)))
  }

  ntable          <- length(blockNames(MB))
  variable_number <- ncol(MB)

  if (length(MB@Data) == 0) { warning("All blocks had 0 variance"); give_error <- 1 }

  for (i in 1:ntable) {
    if (any(is.na(MB@Data[[i]])))       stop("The MB contains NA values. They must be removed first for ComDim analysis.")
    if (any(is.infinite(MB@Data[[i]]))) stop("The MB contains infinite values. They must be removed first for ComDim analysis.")
  }
  if (give_error) stop("The data is not ready for ComDim.") else message("The data can be used for ComDim.")

  # --- Y preparation ---
  if (method == "OPLS-DA") {
    tmp <- unique(as.vector(y))
    if (all(tmp %in% c(0, 1))) {
      if (!is.matrix(y)) y <- as.matrix(y)
      classVect <- rep(NA, nrow(y))
      if (ncol(y) == 1) {
        classVect <- as.vector(y)
      } else {
        if (any(rowSums(y) != 1)) {
          if (any(rowSums(y) == 0))   stop("At least one sample was not assigned to a class.")
          else if (any(colSums(y) > 1)) stop("At least one sample was assigned to more than one class.")
        }
      }
      if (is.null(colnames(y))) colnames(y) <- paste0("Class", seq_len(ncol(y)))
      for (i in seq_len(ncol(y))) classVect[which(y[, i] == 1)] <- colnames(y)[i]
    } else {
      if ((is.matrix(y) || is.data.frame(y)) && ncol(y) > 1)
        stop("ComDim-OPLS can only be applied if Y is a dummy matrix or a class vector")
      classVect        <- as.vector(y)
      classVect_sorted <- sort(unique(classVect))
      y <- matrix(0, nrow = length(classVect), ncol = length(classVect_sorted))
      for (i in seq_along(classVect_sorted)) y[which(classVect == classVect_sorted[i]), i] <- 1
      colnames(y) <- classVect_sorted
      rm(classVect_sorted)
    }
  }
  if (!is.matrix(y)) y <- as.matrix(y)

  pieceBar <- 80 / (4 + 2 * ntable + ndim + nort)
  total_progress <- pieceBar

  DimLabels   <- paste0("CC",  seq_len(ndim))
  OrtLabels   <- paste0("ort", seq_len(nort))
  TableLabels <- blockNames(MB)

  if (loquace)
    message(sprintf("Initialisation finished after : %s millisecs",
                    (Sys.time() - start_ini_time) * 1000))
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # --- Block preparation (normalisation) ---
  temp_tabCalib <- vector("list", ntable)
  MeanMB  <- list()
  NormMB  <- list()
  meanY   <- colMeans(y)

  for (i in 1:ntable) {
    MeanMB[[TableLabels[i]]] <- colMeans(MB@Data[[i]])
    names(MeanMB[[TableLabels[i]]]) <- MB@Variables[[i]]

    if (normalise) {
      X_mean  <- MB@Data[[i]] - matrix(rep(1, nrowMB), ncol = 1) %*% MeanMB[[TableLabels[i]]]
      Norm_X  <- sqrt(sum(X_mean^2))
      NormMB[[TableLabels[i]]] <- setNames(rep(Norm_X, variable_number[i]), MB@Variables[[i]])
      temp_tabCalib[[i]] <- X_mean / Norm_X
    } else {
      NormMB[[TableLabels[i]]] <- setNames(rep(1, variable_number[i]), MB@Variables[[i]])
      temp_tabCalib[[i]] <- MB@Data[[i]]
    }

    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  # Frobenius norms of block kernels (needed for B_kopls)
  frob_norms <- setNames(numeric(ntable), TableLabels)
  for (i in seq_len(ntable)) {
    Xi <- temp_tabCalib[[i]]
    if (normalise) { ss <- sum(Xi^2); if (ss > 1e-14) Xi <- Xi / sqrt(ss) }
    Ki      <- tcrossprod(Xi)
    frob_Ki <- sqrt(sum(Ki^2))
    frob_norms[i] <- if (frob_Ki > 1e-14) frob_Ki else 1
  }

  # --- Build data list for ConsensusOPLS ---
  data_for_cOpls <- setNames(temp_tabCalib, TableLabels)
  for (i in seq_len(ntable)) {
    rownames(data_for_cOpls[[i]]) <- MB@Samples
    colnames(data_for_cOpls[[i]]) <- MB@Variables[[i]]
  }

  cOpls <- ConsensusOPLS::ConsensusOPLS(
    data      = data_for_cOpls,
    Y         = if (method == "OPLS-DA") classVect else y,
    maxPcomp  = ndim,
    maxOcomp  = nort,
    modelType = if (method == "OPLS-DA") "da" else "reg",
    nperm     = 0L,
    cvType    = "nfold",
    nfold     = if (cv.k >= 2L) as.integer(cv.k) else 5L,
    verbose   = loquace
  )

  # Use component counts actually selected by ConsensusOPLS
  ndim <- cOpls@nPcomp
  nort <- cOpls@nOcomp
  DimLabels <- paste0("CC",  seq_len(ndim))
  OrtLabels <- paste0("ort", seq_len(nort))

  # Extract koplsModel internals
  kModel      <- cOpls@model$koplsModel
  Tp          <- kModel$scoresP
  Q_ort       <- kModel$scoresO
  Cp          <- kModel$Cp
  Up          <- kModel$Up
  Sp          <- if (is.matrix(kModel$Sp)) kModel$Sp else diag(as.vector(kModel$Sp), nrow = ndim)
  Bt          <- kModel$Bt[[length(kModel$Bt)]]   # regression after all ort deflations
  rv_weights  <- setNames(cOpls@model$RV, TableLabels)
  normKernels <- cOpls@model$normKernels

  colnames(Tp)    <- DimLabels; rownames(Tp)    <- MB@Samples
  colnames(Q_ort) <- OrtLabels; rownames(Q_ort) <- MB@Samples
  colnames(Cp)    <- DimLabels; rownames(Cp)    <- colnames(y)

  total_progress <- total_progress + pieceBar * ntable
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # --- Saliences from normalised block kernels ---
  saliences     <- matrix(NA_real_, nrow = ntable, ncol = ndim,
                          dimnames = list(TableLabels, DimLabels))
  saliences_ort <- matrix(NA_real_, nrow = ntable, ncol = nort,
                          dimnames = list(TableLabels, OrtLabels))
  for (i in seq_len(ntable)) {
    Ki <- normKernels[[i]]
    for (j  in seq_len(ndim))  saliences[i, j]      <- as.vector(crossprod(Tp[, j],     Ki) %*% Tp[, j])
    for (jj in seq_len(nort))  saliences_ort[i, jj]  <- as.vector(crossprod(Q_ort[, jj], Ki) %*% Q_ort[, jj])
  }

  # --- Orthogonal local scores, orthogonal loadings, and ort-deflated X ---
  # Save pre-deflation blocks for VIP
  X_orig_blocks <- temp_tabCalib

  T_Loc_ort <- setNames(lapply(seq_len(ntable), function(i)
    matrix(NA_real_, nrow = nrowMB, ncol = nort,
           dimnames = list(MB@Samples, OrtLabels))), TableLabels)
  orthoP <- NULL

  for (jj in seq_len(nort)) {
    for (i in seq_len(ntable)) {
      tempo <- t(temp_tabCalib[[i]]) %*% Q_ort[, jj]
      T_Loc_ort[[TableLabels[i]]][, jj] <-
        temp_tabCalib[[i]] %*% (tempo %*% pracma::pinv(t(tempo) %*% tempo))
      temp_tabCalib[[i]] <- temp_tabCalib[[i]] - Q_ort[, jj] %*% t(tempo)
      tempo2 <- if (i == 1) tempo else append(tempo2, tempo)
    }
    orthoP <- if (jj == 1) as.matrix(tempo2) else cbind(orthoP, tempo2)
  }
  colnames(orthoP) <- OrtLabels

  # Rebuild orthogonally-deflated Xnorm_mat
  Xnorm_mat <- do.call(cbind, temp_tabCalib)

  # --- Predictive local scores ---
  T_Loc <- setNames(lapply(seq_len(ntable), function(i)
    matrix(NA_real_, nrow = nrowMB, ncol = ndim,
           dimnames = list(MB@Samples, DimLabels))), TableLabels)

  for (j in seq_len(ndim)) {
    q_j_ss   <- sum(Tp[, j]^2)
    q_j_unit <- if (q_j_ss > 1e-14) Tp[, j] / sqrt(q_j_ss) else Tp[, j]
    for (i in seq_len(ntable)) {
      temp <- t(temp_tabCalib[[i]]) %*% q_j_unit
      T_Loc[[TableLabels[i]]][, j] <-
        temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp))
      temp_tabCalib[[i]] <- temp_tabCalib[[i]] - q_j_unit %*% t(temp)
    }
  }

  varnames <- unlist(lapply(seq_len(ntable), function(i) MB@Variables[[i]]))

  L_CD_Vec <- t(Xnorm_mat) %*% Tp
  colnames(L_CD_Vec) <- DimLabels; rownames(L_CD_Vec) <- varnames
  rownames(orthoP)   <- varnames

  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # --- B_kopls (variable-space regression coefficients) ---
  UpSp   <- Up %*% Sp
  W_pred <- matrix(NA_real_, nrow = sum(variable_number), ncol = ndim)
  k_idx  <- 1L
  for (i in seq_len(ntable)) {
    vi   <- variable_number[i]
    Xb_ort <- Xnorm_mat[, k_idx:(k_idx + vi - 1L), drop = FALSE]
    W_pred[k_idx:(k_idx + vi - 1L), ] <- t(Xb_ort) %*% UpSp * rv_weights[i] / frob_norms[i]
    k_idx <- k_idx + vi
  }
  colnames(W_pred) <- DimLabels; rownames(W_pred) <- varnames
  B_kopls  <- W_pred %*% Bt %*% t(Cp)
  B0_kopls <- colMeans(y) - colMeans(Xnorm_mat) %*% B_kopls

  # --- R2X and R2Y from koplsModel cumulative vectors ---
  # kModel$R2X:  cumulative X-variance explained by predictive components
  # kModel$R2XO: cumulative X-variance explained by orthogonal components
  # kModel$R2Yhat: cumulative Y-variance explained by predictive components
  r2x_pc <- kModel$R2X[seq_len(ndim)]
  r2x_oc <- kModel$R2XO[seq_len(nort)]
  r2x <- setNames(c(c(r2x_pc[1], if (ndim > 1) diff(r2x_pc)),
                    c(r2x_oc[1], if (nort > 1) diff(r2x_oc))),
                  c(DimLabels, OrtLabels))

  r2y_pc <- kModel$R2Yhat[seq_len(ndim)]
  r2y <- setNames(c(c(r2y_pc[1], if (ndim > 1) diff(r2y_pc)),
                    rep(0, nort)),
                  c(DimLabels, OrtLabels))

  SingVal            <- setNames(colSums(Tp^2), DimLabels)
  Sum_saliences_Dim  <- setNames(colSums(saliences), DimLabels)
  Sum_saliences_Tab  <- setNames(rowSums(saliences), TableLabels)

  total_progress <- total_progress + pieceBar * (1 + ndim)
  utils::setTxtProgressBar(progress_bar, value = min(total_progress, 79))

  # --- VIP (same formula as the direct implementation) ---
  .vip_for_scores <- function(scores_mat, Y_raw, X_b, p_b) {
    ncomp <- ncol(scores_mat)
    ss_T  <- colSums(scores_mat^2)
    if (any(ss_T < .Machine$double.eps)) return(rep(NA_real_, p_b))
    Qs <- crossprod(Y_raw, scores_mat) %*% diag(1 / ss_T, nrow = ncomp)
    ss_Q <- colSums(Qs^2)
    if (any(ss_Q < .Machine$double.eps)) return(rep(NA_real_, p_b))
    Us <- Y_raw %*% Qs %*% diag(1 / ss_Q, nrow = ncomp)
    ss_U <- colSums(Us^2)
    if (any(ss_U < .Machine$double.eps)) return(rep(NA_real_, p_b))
    Ws <- crossprod(X_b, Us) %*% diag(1 / ss_U, nrow = ncomp)
    Ws <- apply(Ws, 2, function(w) { nrm <- sqrt(sum(w^2)); if (nrm > 1e-14) w / nrm else w })
    s_a   <- ss_T * ss_Q
    sum_s <- sum(s_a)
    if (sum_s < .Machine$double.eps) return(rep(NA_real_, p_b))
    as.vector(sqrt(p_b * (Ws^2 %*% s_a) / sum_s))
  }

  .vip_ort_for_loadings <- function(ort_scores_mat, ort_loadings_b) {
    p_b  <- nrow(ort_loadings_b)
    ss_T <- colSums(ort_scores_mat^2)
    if (any(ss_T < .Machine$double.eps)) return(rep(NA_real_, p_b))
    P_norm <- apply(ort_loadings_b, 2, function(pp) {
      nrm <- sqrt(sum(pp^2)); if (nrm > 1e-14) pp / nrm else pp
    })
    if (is.null(dim(P_norm))) P_norm <- matrix(P_norm, ncol = 1)
    sum_s <- sum(ss_T)
    if (sum_s < .Machine$double.eps) return(rep(NA_real_, p_b))
    as.vector(sqrt(p_b * (P_norm^2 %*% ss_T) / sum_s))
  }

  VIP_blocks <- setNames(vector("list", ntable), TableLabels)
  k_ort <- 1L
  for (i in seq_len(ntable)) {
    p_b     <- variable_number[i]
    vip_p   <- .vip_for_scores(Tp, y, X_orig_blocks[[i]], p_b)
    ort_P_b <- orthoP[k_ort:(k_ort + p_b - 1L), , drop = FALSE]
    vip_o   <- .vip_ort_for_loadings(Q_ort, ort_P_b)
    vip_t   <- sqrt((vip_p^2 + vip_o^2) / 2)
    VIP_blocks[[TableLabels[i]]] <- data.frame(p = vip_p, o = vip_o, tot = vip_t,
                                               row.names = MB@Variables[[i]])
    k_ort <- k_ort + p_b
  }
  VIP_global <- setNames(unlist(lapply(VIP_blocks, `[[`, "tot")), varnames)

  # --- variable.block ---
  belong_block <- character(sum(variable_number))
  k <- 1L
  for (i in seq_len(ntable)) {
    belong_block[k:(k + variable_number[i] - 1L)] <- TableLabels[i]
    k <- k + variable_number[i]
  }

  # --- Training-set Y prediction ---
  Ypred <- Xnorm_mat %*% B_kopls +
    matrix(as.vector(B0_kopls), nrow = nrowMB, ncol = ncol(y), byrow = TRUE)
  rownames(Ypred) <- MB@Samples
  if (!is.null(colnames(y))) colnames(Ypred) <- colnames(y)

  # --- Class prediction, sensitivity, specificity, confusion matrix ---
  if (method == "OPLS-DA") {
    predClass <- rep(NA_character_, nrow(y))
    if (decisionRule == "max") {
      for (i in seq_len(nrow(Ypred))) {
        xx <- which(Ypred[i, ] == max(Ypred[i, ], na.rm = TRUE))
        if (length(xx) == 1) predClass[i] <- colnames(y)[xx]
        else { predClass[i] <- NaN; warning(sprintf("Class for sample %s could not be predicted.", i)) }
      }
    } else if (decisionRule == "fixed") {
      for (i in seq_len(nrow(Ypred))) {
        xx <- which(Ypred[, i] > 1 / ncol(y))
        if (length(xx) == 1) predClass[i] <- colnames(y)[xx]
        else { predClass[i] <- NaN; warning(sprintf("Class for sample %s could not be predicted.", i)) }
      }
    }

    classy <- colnames(y)
    Q2 <- DQ2 <- specvec <- sensvec <- setNames(rep(NA_real_, length(classy)), classy)
    confusionMatrix <- list()

    for (i in seq_along(classy)) {
      pos <- which(classVect == classy[i])
      neg <- which(classVect != classy[i])
      tp  <- length(which(classVect == classy[i] & predClass == classy[i]))
      tn  <- length(which(classVect != classy[i] & predClass != classy[i]))
      if (length(tp) == 0) tp <- 0
      if (length(tn) == 0) tn <- 0

      sensvec[i] <- tp / length(pos)
      specvec[i] <- tn / length(neg)

      cm <- matrix(c(tp,
                     length(which(classVect != classy[i] & predClass == classy[i])),
                     length(which(classVect == classy[i] & predClass != classy[i])),
                     tn), 2, 2)
      rownames(cm) <- c("predClass1", "predClass0")
      colnames(cm) <- c("trueClass1", "trueClass0")
      confusionMatrix[[classy[i]]] <- cm

      E0   <- Ypred[neg, i] - y[neg, i]
      E1   <- Ypred[pos, i] - y[pos, i]
      Ym   <- y[, i] - mean(y[, i])
      TSS  <- sum(Ym^2)
      DQ2[i] <- 1 - (sum(E0[E0 > 0]^2) + sum(E1[E1 < 0]^2)) / TSS
      Q2[i]  <- 1 - (sum(E0^2) + sum(E1^2)) / TSS
    }
  } else {
    Ym <- y - mean(y)
    Q2 <- 1 - sum((Ypred - y)^2) / sum(Ym^2)
  }

  # Override training-set Q2/DQ2 with cross-validated values when available
  if (cv.k >= 2) {
    q2_cv  <- cOpls@Q2
    dq2_cv <- if (method == "OPLS-DA") cOpls@DQ2 else NULL
    if (!is.null(q2_cv)  && length(q2_cv)  > 0) Q2  <- q2_cv
    if (!is.null(dq2_cv) && length(dq2_cv) > 0) DQ2 <- dq2_cv
  }

  # --- CV result slot ---
  if (cv.k >= 2) {
    Ypred_cv <- NULL
    cv_obj   <- cOpls@cv
    if (!is.null(cv_obj) && !is.null(cv_obj$AllYhat)) {
      yhat_list <- cv_obj$AllYhat
      last_yhat <- if (is.list(yhat_list)) yhat_list[[length(yhat_list)]] else yhat_list
      if (!is.null(cv_obj$cvTestIndex) && is.matrix(last_yhat)) {
        test_order <- unlist(cv_obj$cvTestIndex)
        if (length(test_order) == nrowMB) {
          Ypred_cv <- matrix(NA_real_, nrow = nrowMB, ncol = ncol(y),
                             dimnames = list(MB@Samples, colnames(y)))
          Ypred_cv[test_order, ] <- last_yhat
        }
      }
      if (is.null(Ypred_cv)) Ypred_cv <- last_yhat
    }
    cv_result <- list(
      k     = cv.k,
      Ypred = Ypred_cv,
      Q2    = Q2,
      DQ2   = if (method == "OPLS-DA") DQ2 else NULL
    )
  } else {
    if (cv.k != 0)
      warning("'cv.k' must be >= 2 or 0 (skip CV). No cross-validation performed.")
    cv_result <- list()
  }

  # --- Assemble ComDim object ---
  end_output   <- Sys.time()
  running_time <- as.numeric(end_output - start_ini_time)
  if (loquace) message(sprintf("Analysis finished after : %s millisecs", running_time * 1000))

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
  utils::setTxtProgressBar(progress_bar, value = 80)
  close(progress_bar)

  metadata <- list()
  if (length(MB@Metadata) != 0) metadata[[1]] <- MB@Metadata

  new("ComDim",
    Method     = method,
    ndim       = ndim,
    Q.scores   = Tp,
    T.scores   = T_Loc,
    P.loadings = L_CD_Vec,
    Saliences  = saliences,
    Orthogonal = list(
      nort            = nort,
      Q.scores        = Q_ort,
      T.scores        = T_Loc_ort,
      P.loadings.ort  = orthoP,
      Saliences.ort   = saliences_ort
    ),
    VIP        = VIP_global,
    VIP.block  = VIP_blocks,
    R2X        = r2x,
    R2Y        = r2y,
    Q2         = Q2,
    DQ2        = if (method == "OPLS-DA") DQ2 else vector(),
    Singular   = SingVal,
    Mean       = list(MeanMB = MeanMB, MeanY = meanY),
    Norm       = list(NormMB = NormMB, FrobNorms = frob_norms, RVweights = rv_weights),
    PLS.model  = list(W = W_pred, B = B_kopls, B0 = B0_kopls, Y = y),
    cv         = cv_result,
    Prediction = if (method == "OPLS-DA") {
      list(
        Y.pred          = Ypred,
        decisionRule    = decisionRule,
        trueClass       = classVect,
        predClass       = data.frame(predClass = predClass, row.names = MB@Samples),
        Sensitivity     = sensvec,
        Specificity     = specvec,
        confusionMatrix = confusionMatrix
      )
    } else {
      list(Y.pred = Ypred)
    },
    Metadata       = metadata,
    variable.block = belong_block,
    runtime        = running_time
  )
}


# ==============================================================================
# ARCHIVED DIRECT IMPLEMENTATION
# The code below is the original self-contained KOPLS implementation of
# ComDim_OPLS, kept here in case ConsensusOPLS becomes unavailable.
# To restore it: uncomment the function definition and comment out the wrapper
# above.
# ==============================================================================

# ComDim_OPLS <- function(MB = MB, y = y, ndim = 1, nort = 1,
#                         method = c("OPLS-DA", "OPLS-R"),
#                         decisionRule = c("fixed", "max")[2],
#                         normalise = FALSE, threshold = 1e-10,
#                         loquace = FALSE,
#                         cv.k = 7) {
#   # INITIALISATION
#
#   if (!(method == "OPLS-DA" || method == "OPLS-R")) {
#     stop("'method' must be 'OPLS-DA' or 'OPLS-R'.")
#   }
#
#   if (nort < 1) {
#     stop("'nort' must be at least 1.")
#   }
#
#   if (ndim < 1) {
#     stop("'ndim' must be at least 1.")
#   }
#
#   progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
#
#   start_ini_time <- Sys.time() # To start counting calculation time for the initialization
#
#   if (!inherits(MB, "MultiBlock")) {
#     stop("'MB' is  not a MultiBlock.")
#   }
#
#   ntable <- length(blockNames(MB)) # number of blocks
#   nrowMB <- length(sampleNames(MB)) # number of samples
#   variable_number <- ncol(MB) # Number of variables per block
#
#   give_error <- 0
#
#   # Remove all variables with 0 variance. If there are no remaining variables, give_error
#   numvar <- 0
#   numblock0 <- vector()
#   for (i in 1:ntable) {
#     var0 <- apply(MB@Data[[i]], 2, function(x) {
#       var(x)
#     })
#     var0 <- which(var0 == 0)
#
#     if (length(var0) != 0) {
#       numvar <- numvar + length(var0)
#       if (length(var0) == length(MB@Variables[[i]])) {
#         numblock0 <- append(numblock0, i)
#       }
#       MB@Data[[i]] <- MB@Data[[i]][, setdiff(1:variable_number[i], var0)]
#       MB@Variables[[i]] <- MB@Variables[[i]][setdiff(1:variable_number[i], var0)]
#       variable_number[i] <- length(MB@Variables[[i]])
#     }
#   }
#   if (numvar != 0) {
#     warning(sprintf("Number of variables excluded from the analysis because of their zero variance: %d", numvar))
#   }
#   if (length(numblock0) != 0) {
#     numblock0 <- rev(numblock0) # To sort in decreasing order.
#     for (i in seq_along(numblock0)) {
#       MB@Data[[numblock0[i]]] <- NULL
#       MB@Variables[[numblock0[i]]] <- NULL
#       if (names(MB@Data)[numblock0[i]] %in% names(MB@Batch)) {
#         MB@Batch[[numblock0[i]]] <- NULL
#       }
#       if (names(MB@Data)[numblock0[i]] %in% names(MB@Metadata)) {
#         MB@Metadata[[numblock0[i]]] <- NULL
#       }
#     }
#     warning(sprintf("Number of blocks excluded from the analysis because of their zero variance: %d", length(numblock0)))
#   }
#
#   ntable <- length(blockNames(MB)) # number of blocks
#   variable_number <- ncol(MB) # In case the number of blocks has changed.
#
#   if (length(MB@Data) == 0) { # In case all blocks had 0 variance...
#     warning("All blocks had 0 variance")
#     give_error <- 1
#   }
#
#   # Check for infinite or missing values (they cannot be handled with svd)
#   for (i in 1:ntable) {
#     if (any(is.na(MB@Data[[i]]))) {
#       stop("The MB contains NA values. They must be removed first for ComDim analysis.")
#     }
#     if (any(is.infinite(MB@Data[[i]]))) {
#       stop("The MB contains infinite values. They must be removed first for ComDim analysis.")
#     }
#   }
#
#   if (give_error) {
#     stop("The data is not ready for ComDim.")
#   } else {
#     message("The data can be used for ComDim.")
#   }
#
#   ## Check y matrix (convert to dummy matrix if it is not already.)
#   if (method == "OPLS-DA") {
#     tmp <- unique(as.vector(y))
#     if (all(tmp %in% c(0, 1))) { # Is it dummy?
#
#       if (!is.matrix(y)) { # If it's a vector and the method is OPLS-R
#         y <- as.matrix(y)
#       }
#
#       classVect <- rep(NA, nrow(y))
#
#       if (ncol(y) == 1) {
#         classVect <- as.vector(y)
#       } else {
#         if (any(rowSums(y) != 1)) {
#           if (any(rowSums(y) == 0)) {
#             stop("At least one sample was not assigned to a class.")
#           } else if (any(colSums(y) > 1)) {
#             stop("At least one sample was assigned to more than one class.")
#           }
#         }
#       }
#       if (is.null(colnames(y))) {
#         colnames(y) <- paste0("Class", seq_len(ncol(y)))
#       }
#       for (i in seq_len(ncol(y))) {
#         classVect[which(y[, i] == 1)] <- colnames(y)[i]
#       }
#     } else { # If not dummy
#       if ((is.matrix(y) || is.data.frame(y)) && ncol(y) > 1) {
#         stop("ComDim-OPLS can only be applied if Y is a dummy matrix or a class vector")
#       }
#       classVect <- as.vector(y)
#       classVect_sorted <- sort(unique(classVect))
#
#       # Now generate the dummy matrix from y
#       y <- matrix(rep(0, length(classVect_sorted) * length(classVect)),
#         ncol = length(classVect_sorted), nrow = length(classVect)
#       )
#       for (i in seq_along(classVect_sorted)) {
#         y[which(classVect == classVect_sorted[i]), i] <- 1
#       }
#       colnames(y) <- classVect_sorted
#       rm(classVect_sorted)
#     }
#   }
#
#   if (!is.matrix(y)) { # If it's a vector and the method is OPLS-R
#     y <- as.matrix(y)
#   }
#
#   pieceBar <- 4 + 2 * ntable + ndim + nort # Number of updates in the progress bar.
#   pieceBar <- 80 / pieceBar
#   total_progress <- pieceBar
#
#   DimLabels <- paste0("CC", 1:ndim) # One label per predictive component.
#   OrtLabels <- paste0("ort", 1:nort) # One label per orthogonal component.
#   TableLabels <- blockNames(MB) # One label per block.
#
#   end_ini_time <- Sys.time() # To end the count of the calculation time.
#
#   if (loquace) {
#     message(sprintf("Initialisation finished after : %s millisecs", (end_ini_time - start_ini_time) * 1000))
#   }
#
#   utils::setTxtProgressBar(progress_bar, value = total_progress)
#
#   # NORMALISATION
#
#   X_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))
#   Xnorm_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))
#
#   res_calib <- list()
#   temp_tabCalib <- list()
#   s_n <- list()
#   res_calib$SingVal <- as.vector(NULL)
#   res_calib$NormMB <- as.vector(NULL)
#   res_calib$MeanMB <- list()
#
#   for (i in 1:ntable) {
#     res_calib$MeanMB[[TableLabels[i]]] <- colMeans(MB@Data[[i]])
#     names(res_calib$MeanMB[[TableLabels[i]]]) <- MB@Variables[[i]]
#
#     if (normalise) {
#       # Normalise original blocks
#       X_mean <- MB@Data[[i]] - matrix(data = rep(1, nrowMB), ncol = 1, nrow = nrowMB) %*% res_calib$MeanMB[[i]]
#       XX <- X_mean * X_mean
#       Norm_X <- sqrt(sum(XX))
#       X_Normed <- X_mean / Norm_X
#       res_calib$NormMB[i] <- Norm_X
#       temp_tabCalib[[i]] <- X_Normed
#       s_n[[i]] <- X_Normed
#     } else {
#       res_calib$NormMB[[TableLabels[i]]] <- rep(1, length(MB@Variables[[i]]))
#       names(res_calib$NormMB[[TableLabels[i]]]) <- MB@Variables[[i]]
#       temp_tabCalib[[i]] <- MB@Data[[i]]
#       s_n[[i]] <- MB@Data[[i]]
#     }
#
#     if (i == 1) {
#       X_mat[, 1:variable_number[1]] <- MB@Data[[i]]
#       Xnorm_mat[, 1:variable_number[1]] <- temp_tabCalib[[i]]
#     } else {
#       beg <- sum(variable_number[1:(i - 1)]) + 1
#       ending <- sum(variable_number[1:i])
#       X_mat[, beg:ending] <- MB@Data[[i]]
#       Xnorm_mat[, beg:ending] <- temp_tabCalib[[i]]
#     }
#
#     # PLS addition
#     meanX <- colMeans(Xnorm_mat)
#     meanY <- colMeans(y)
#     # End PLS addition
#
#     norm_comdim <- Sys.time()
#     if (loquace) {
#       message(sprintf("Normalization of block %s finished after : %s millisecs", i, (norm_comdim - start_ini_time) * 1000))
#     }
#     total_progress <- total_progress + pieceBar
#     utils::setTxtProgressBar(progress_bar, value = total_progress)
#   }
#
#   names(res_calib$NormMB) <- TableLabels
#
#   nR <- nrow(Xnorm_mat)
#   nC <- ncol(Xnorm_mat)
#
#   # =========================================================================
#   # KOPLS KERNEL-BASED EXTRACTION
#   # =========================================================================
#
#   end_comp_time <- Sys.time()
#   if (loquace) {
#     message(sprintf(
#       "Kernel preparation finished after : %s millisecs",
#       (end_comp_time - start_ini_time) * 1000
#     ))
#   }
#   total_progress <- total_progress + pieceBar * ntable
#   utils::setTxtProgressBar(progress_bar, value = total_progress)
#
#   # Helper: modified RV coefficient (zeros diagonal before computing)
#   .RVmod <- function(X, Y) {
#     AA <- tcrossprod(X)
#     diag(AA) <- 0
#     BB <- tcrossprod(Y)
#     diag(BB) <- 0
#     denom <- sqrt(sum(AA^2)) * sqrt(sum(BB^2))
#     if (denom < .Machine$double.eps) {
#       return(0)
#     }
#     sum(diag(crossprod(AA, BB))) / denom
#   }
#
#   K_blocks <- vector("list", ntable)
#   frob_norms <- numeric(ntable) # Frobenius norm of each raw block kernel
#   rv_weights <- numeric(ntable) # RV weight for each block
#
#   for (i in seq_len(ntable)) {
#     Xi <- temp_tabCalib[[i]]
#     if (normalise) {
#       ss <- sum(Xi^2)
#       if (ss > 1e-14) Xi <- Xi / sqrt(ss)
#     }
#     Ki <- tcrossprod(Xi)
#     frob_Ki <- sqrt(sum(Ki^2))
#     frob_norms[i] <- if (frob_Ki > 1e-14) frob_Ki else 1
#     K_norm_i <- Ki / frob_norms[i]
#     rv_weights[i] <- (.RVmod(K_norm_i, y) + 1) / 2
#     K_blocks[[i]] <- K_norm_i
#   }
#   names(rv_weights) <- TableLabels
#
#   # RV-weighted combined kernel
#   W_mat <- Reduce("+", lapply(
#     seq_len(ntable),
#     function(i) rv_weights[i] * K_blocks[[i]]
#   ))
#
#   # Double-centre: K_combined = H * W_mat * H  (H = I - 1/n * J)
#   K_combined <- scale(t(scale(W_mat, scale = FALSE)), scale = FALSE)
#
#   rm(s_n)
#
#   # --- 2. Centre Y; SVD of Y' K Y -------------------------------------------
#   Y_c <- sweep(y, 2, colMeans(y)) # n x ncol(y)
#   YKY <- crossprod(Y_c, K_combined %*% Y_c) # ncol(y) x ncol(y)
#
#   A <- ndim
#   CSV <- svd(YKY, nu = A, nv = A)
#   Cp <- CSV$u[, 1:A, drop = FALSE] # ncol(y) x ndim  Y loadings
#   Sp <- diag(CSV$d[1:A]^(-0.5), nrow = A) # ndim x ndim     scale
#   Up <- Y_c %*% Cp # n x ndim        Y scores
#
#   colnames(Cp) <- DimLabels
#   rownames(Cp) <- colnames(y)
#   colnames(Up) <- DimLabels
#   rownames(Up) <- MB@Samples
#
#   # --- 3. Orthogonal extraction (KOPLS loop) --------------------------------
#   Q_ort <- matrix(NA_real_, nrow = nrowMB, ncol = nort)
#   saliences_ort <- matrix(NA_real_, nrow = ntable, ncol = nort)
#
#   to_list <- vector("list", nort) # unit-norm ort scores
#   co_list <- vector("list", nort) # ort direction in Y-score space
#   so_list <- vector("list", nort) # singular value
#   toNorm_list <- vector("list", nort) # norm before unit-normalisation
#
#   # Per-component storage for R2X and R2Y
#   # Output will have ndim + nort values: CC1..CC{ndim}, ort1..ort{nort}
#   sstot_K      <- sum(diag(K_combined))    # trace(K_original) = SS(X)
#   Kdiag_levels <- numeric(nort + 1)        # trace(K_both) before/after each ort deflation
#   Kdiag_levels[1] <- sstot_K              # before any ort deflation
#   Tp_levels    <- vector("list", nort + 1) # predictive scores at each ort level
#
#   K_right <- K_combined # n x n  right-only deflated kernel
#   K_both <- K_combined # n x n  both-sides deflated kernel
#
#   for (jj in seq_len(nort)) {
#     Tp_curr <- crossprod(K_right, Up %*% Sp) # = t(K_right) %*% Up %*% Sp
#     Tp_levels[[jj]] <- Tp_curr              # Tp using K deflated by (jj-1) ort components
#     K_resid <- K_both - tcrossprod(Tp_curr) # K - Tp Tp'
#
#     sv_tmp <- svd(crossprod(Tp_curr, K_resid %*% Tp_curr), nu = 1, nv = 1)
#     co_j <- sv_tmp$u[, 1, drop = FALSE] # ndim x 1
#     so_j <- sv_tmp$d[1]
#
#     to_raw <- K_resid %*% Tp_curr %*% co_j /
#       sqrt(max(so_j, .Machine$double.eps)) # n x 1
#     toNorm_j <- sqrt(sum(to_raw^2))
#     if (toNorm_j < 1e-14) toNorm_j <- 1
#     to_j <- to_raw / toNorm_j # unit-norm
#
#     to_list[[jj]] <- to_j
#     co_list[[jj]] <- co_j
#     so_list[[jj]] <- so_j
#     toNorm_list[[jj]] <- toNorm_j
#     Q_ort[, jj] <- as.vector(to_j)
#
#     for (i in seq_len(ntable)) {
#       saliences_ort[i, jj] <-
#         as.vector(crossprod(to_j, K_blocks[[i]]) %*% to_j)
#     }
#
#     # Kernel deflation
#     I_to <- diag(nrowMB) - tcrossprod(to_j)
#     K_right <- K_right %*% I_to # right only
#     K_both <- I_to %*% K_both %*% I_to # both sides
#     Kdiag_levels[jj + 1] <- sum(diag(K_both)) # trace after jj ort deflations
#
#     ort_comdim <- Sys.time()
#     if (loquace) {
#       message(sprintf(
#         "Orthogonal component %s determined after : %s millisecs",
#         jj, (ort_comdim - start_ini_time) * 1000
#       ))
#     }
#     total_progress <- total_progress + pieceBar
#     utils::setTxtProgressBar(progress_bar, value = total_progress)
#   }
#
#   colnames(Q_ort) <- OrtLabels
#   rownames(Q_ort) <- MB@Samples
#   colnames(saliences_ort) <- OrtLabels
#   rownames(saliences_ort) <- TableLabels
#
#   # --- 4. Final predictive scores -------------------------------------------
#   Tp <- crossprod(K_right, Up %*% Sp) # n x ndim
#   Tp_levels[[nort + 1]] <- Tp         # final predictive scores (all ort deflations applied)
#
#   # Inner regression coefficient
#   Bt <- pracma::pinv(crossprod(Tp)) %*% crossprod(Tp, Up) # ndim x ndim
#
#   # Predictive block saliences
#   saliences <- matrix(NA_real_, nrow = ntable, ncol = ndim)
#   for (i in seq_len(ntable)) {
#     for (j in seq_len(ndim)) {
#       tp_j <- Tp[, j]
#       saliences[i, j] <-
#         as.vector(crossprod(tp_j, K_blocks[[i]]) %*% tp_j)
#     }
#   }
#   colnames(saliences) <- DimLabels
#   rownames(saliences) <- TableLabels
#
#   # SingVal: SS of predictive scores
#   res_calib$SingVal <- setNames(colSums(Tp^2), DimLabels)
#
#   # R2X per component (non-cumulative)
#   r2x_pred        <- setNames(colSums(Tp^2) / sstot_K, DimLabels)
#   r2x_ort         <- setNames(-diff(Kdiag_levels) / sstot_K, OrtLabels)
#   res_calib$explained <- c(r2x_pred, r2x_ort)
#
#   # Store predictive scores as Q
#   Q <- Tp
#   colnames(Q) <- DimLabels
#   rownames(Q) <- MB@Samples
#   res_calib$Q <- Q
#
#   ## Adding metadata
#   res_calib$metadata <- list()
#   if (length(MB@Metadata) != 0) {
#     res_calib$metadata[[1]] <- MB@Metadata
#   }
#
#   iter_comdim <- Sys.time()
#   if (loquace) {
#     message(sprintf(
#       "KOPLS scores finished after : %s millisecs",
#       (iter_comdim - start_ini_time) * 1000
#     ))
#   }
#   total_progress <- total_progress + pieceBar * (1 + ndim)
#   utils::setTxtProgressBar(progress_bar, value = min(total_progress, 79))
#
#   rm(K_combined, K_right, K_both)
#
#   # --- 5. R2Y per component --------------------------------------------------
#   sstot_Y_r2 <- sum(Y_c^2)
#
#   Tp0 <- Tp_levels[[1]]  # pred scores before any ort deflation
#   r2y_pred_cum <- numeric(ndim)
#   for (.j in seq_len(ndim)) {
#     Tp_j   <- Tp0[, seq_len(.j), drop = FALSE]
#     Bt_j   <- pracma::pinv(crossprod(Tp_j)) %*% crossprod(Tp_j, Up[, seq_len(.j), drop = FALSE])
#     Yhat_j <- Tp_j %*% Bt_j %*% t(Cp[, seq_len(.j), drop = FALSE])
#     r2y_pred_cum[.j] <- if (sstot_Y_r2 < .Machine$double.eps) 0 else
#       1 - sum((Yhat_j - Y_c)^2) / sstot_Y_r2
#   }
#   r2y_pred <- setNames(c(r2y_pred_cum[1], diff(r2y_pred_cum)), DimLabels)
#
#   r2y_lev <- numeric(nort + 1)
#   for (.i in seq_len(nort + 1)) {
#     Tp_i      <- Tp_levels[[.i]]
#     Bt_i      <- pracma::pinv(crossprod(Tp_i)) %*% crossprod(Tp_i, Up)
#     Yhat_i    <- Tp_i %*% Bt_i %*% t(Cp)
#     r2y_lev[.i] <- if (sstot_Y_r2 < .Machine$double.eps) 0 else
#       1 - sum((Yhat_i - Y_c)^2) / sstot_Y_r2
#   }
#   r2y_ort <- setNames(diff(r2y_lev), OrtLabels)
#   r2y <- c(r2y_pred, r2y_ort)
#
#   ## Calculate Sums of saliences for each Dim
#   Sum_saliences_Dim <- colSums(saliences)
#   names(Sum_saliences_Dim) <- DimLabels
#   res_calib$Sum_saliences_Dim <- Sum_saliences_Dim
#   rm(Sum_saliences_Dim)
#
#   # Calculate Sums of saliences for each Table
#   Sum_saliences_Tab <- rowSums(saliences)
#   names(Sum_saliences_Tab) <- TableLabels
#   res_calib$Sum_saliences_Tab <- Sum_saliences_Tab
#   rm(Sum_saliences_Tab)
#
#   res_calib$saliences <- saliences
#   colnames(res_calib$saliences) <- DimLabels
#   rownames(res_calib$saliences) <- TableLabels
#
#   ## LOADING COMPUTATION
#
#   L_X <- list()
#   T_Loc_ort <- list()
#   T_Loc <- list()
#   for (i in 1:ntable) { # Prepare lists
#     T_Loc_ort[[TableLabels[i]]] <- matrix(, ncol = nort, nrow = nrowMB)
#     T_Loc[[TableLabels[i]]] <- matrix(, ncol = ndim, nrow = nrowMB)
#   }
#   b <- matrix(, ncol = ndim, nrow = ntable)
#
#   if (!requireNamespace("pracma", quietly = TRUE)) {
#     stop("The package pracma is needed.")
#   }
#
#   # Save pre-deflation blocks for VIP computation
#   X_orig_blocks <- temp_tabCalib
#
#   # Orthogonal loadings
#   orthoP <- NULL
#   for (jj in 1:nort) {
#     for (i in 1:ntable) {
#       tempo <- t(temp_tabCalib[[i]]) %*% Q_ort[, jj] # Scaled CD 'local' Loadings
#       T_Loc_ort[[TableLabels[i]]][, jj] <- temp_tabCalib[[i]] %*% (tempo %*% pracma::pinv(t(tempo) %*% tempo)) # local Scores
#       # Deflate each temp_tabCalib by orthogonal component
#       temp_tabCalib[[i]] <- temp_tabCalib[[i]] - Q_ort[, jj] %*% t(tempo)
#       if (i == 1) {
#         tempo2 <- tempo
#       } else {
#         tempo2 <- append(tempo2, tempo)
#       }
#     }
#     if (jj == 1) {
#       orthoP <- tempo2
#     } else {
#       orthoP <- cbind(orthoP, tempo2)
#     }
#   }
#   orthoP <- as.matrix(orthoP)
#   colnames(orthoP) <- OrtLabels
#
#   for (i in 1:ntable) {
#     rownames(T_Loc_ort[[TableLabels[i]]]) <- MB@Samples
#     colnames(T_Loc_ort[[TableLabels[i]]]) <- OrtLabels
#   }
#
#   # Rebuild Xnorm_mat from orthogonally-deflated temp_tabCalib.
#   for (i in 1:ntable) {
#     if (i == 1) {
#       Xnorm_mat <- temp_tabCalib[[i]]
#     } else {
#       Xnorm_mat <- cbind(Xnorm_mat, temp_tabCalib[[i]])
#     }
#   }
#
#   # Predictive loadings
#   for (j in 1:ndim) {
#     T_mat <- matrix(, nrow = nrowMB, ncol = 0)
#
#     q_j <- res_calib$Q[, j]
#     q_j_ss <- sum(q_j^2)
#     q_j_unit <- if (q_j_ss > 1e-14) q_j / sqrt(q_j_ss) else q_j
#
#     for (i in 1:ntable) {
#       temp <- t(temp_tabCalib[[i]]) %*% q_j_unit # Scaled CD 'local' Loadings
#
#       T_mat <- cbind(T_mat, temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp))) # local Scores
#
#       T_Loc[[TableLabels[i]]][, j] <- temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp)) # local Scores
#
#       # Deflate each temp_tabCalib
#       temp_tabCalib[[i]] <- temp_tabCalib[[i]] - q_j_unit %*% t(temp)
#     }
#
#     # For each CC
#     # MLR b-coefficients between Local and Global Scores
#     b[, j] <- pracma::pinv(t(T_mat) %*% T_mat) %*% t(T_mat) %*% res_calib$Q[, j]
#   }
#
#   for (i in 1:ntable) {
#     rownames(T_Loc[[TableLabels[i]]]) <- rownames(res_calib$Q)
#     colnames(T_Loc[[TableLabels[i]]]) <- colnames(res_calib$Q)
#   }
#
#   # Calculate Global Loadings from orthogonally-deflated X
#   L_CD_Vec <- t(Xnorm_mat) %*% res_calib$Q # Scaled CD 'global' Loadings
#
#   load_comdim <- Sys.time()
#   if (loquace) {
#     message(sprintf("Loadings finished after : %s millisecs",
#                     (load_comdim - start_ini_time) * 1000))
#   }
#
#   total_progress <- total_progress + pieceBar
#   utils::setTxtProgressBar(progress_bar, value = total_progress)
#
#   rm(X_mat)
#   rm(T_mat)
#   rm(temp_tabCalib)
#   rm(temp)
#
#   ## Output
#   end_function <- Sys.time()
#
#   ## KOPLS B matrix (variable-space regression coefficient)
#   UpSp <- Up %*% Sp # n x ndim
#   W_pred <- matrix(NA_real_, nrow = sum(variable_number), ncol = ndim)
#   k_idx <- 1
#   for (i in seq_len(ntable)) {
#     vi <- variable_number[i]
#     Xb_ort <- Xnorm_mat[, k_idx:(k_idx + vi - 1), drop = FALSE]
#     W_pred[k_idx:(k_idx + vi - 1), ] <- t(Xb_ort) %*% UpSp * rv_weights[i] / frob_norms[i]
#     k_idx <- k_idx + vi
#   }
#   B_kopls <- W_pred %*% Bt %*% t(Cp) # p x ncol(y)
#   B0_kopls <- colMeans(y) - colMeans(Xnorm_mat) %*% B_kopls # 1 x ncol(y)
#   colnames(W_pred) <- DimLabels
#   rm(UpSp)
#
#   res_calib$b <- b
#   colnames(res_calib$b) <- DimLabels
#   rownames(res_calib$b) <- TableLabels
#
#   res_calib$T_Loc <- T_Loc
#   rm(T_Loc)
#
#   varnames <- vector()
#   for (i in 1:ntable) {
#     varnames <- append(varnames, MB@Variables[[i]])
#   }
#
#   res_calib$P <- L_CD_Vec
#   colnames(res_calib$P) <- DimLabels
#   rownames(res_calib$P) <- varnames
#   rownames(orthoP) <- varnames
#
#   rm(L_CD_Vec)
#
#   ## VIP
#   .vip_for_scores <- function(scores_mat, Y_raw, X_b, p_b) {
#     ncomp <- ncol(scores_mat)
#     ss_T <- colSums(scores_mat^2)
#     if (any(ss_T < .Machine$double.eps)) {
#       return(rep(NA_real_, p_b))
#     }
#     Qs <- crossprod(Y_raw, scores_mat) %*% diag(1 / ss_T, nrow = ncomp)
#     ss_Q <- colSums(Qs^2)
#     if (any(ss_Q < .Machine$double.eps)) {
#       return(rep(NA_real_, p_b))
#     }
#     Us <- Y_raw %*% Qs %*% diag(1 / ss_Q, nrow = ncomp)
#     ss_U <- colSums(Us^2)
#     if (any(ss_U < .Machine$double.eps)) {
#       return(rep(NA_real_, p_b))
#     }
#     Ws <- crossprod(X_b, Us) %*% diag(1 / ss_U, nrow = ncomp)
#     Ws <- apply(Ws, 2, function(w) {
#       nrm <- sqrt(sum(w^2))
#       if (nrm > 1e-14) w / nrm else w
#     })
#     s_a <- ss_T * ss_Q
#     sum_s <- sum(s_a)
#     if (sum_s < .Machine$double.eps) {
#       return(rep(NA_real_, p_b))
#     }
#     as.vector(sqrt(p_b * (Ws^2 %*% s_a) / sum_s))
#   }
#
#   .vip_ort_for_loadings <- function(ort_scores_mat, ort_loadings_b) {
#     ncomp <- ncol(ort_scores_mat)
#     p_b   <- nrow(ort_loadings_b)
#     ss_T  <- colSums(ort_scores_mat^2)
#     if (any(ss_T < .Machine$double.eps)) return(rep(NA_real_, p_b))
#     P_norm <- apply(ort_loadings_b, 2, function(pp) {
#       nrm <- sqrt(sum(pp^2)); if (nrm > 1e-14) pp / nrm else pp
#     })
#     if (is.null(dim(P_norm))) P_norm <- matrix(P_norm, ncol = 1)
#     sum_s <- sum(ss_T)
#     if (sum_s < .Machine$double.eps) return(rep(NA_real_, p_b))
#     as.vector(sqrt(p_b * (P_norm^2 %*% ss_T) / sum_s))
#   }
#
#   VIP_blocks <- setNames(vector("list", ntable), TableLabels)
#   k_ort <- 1L  # running row index into orthoP for the current block
#   for (i in seq_len(ntable)) {
#     X_b   <- X_orig_blocks[[i]]
#     p_b   <- variable_number[i]
#     vip_p <- .vip_for_scores(Tp, y, X_b, p_b)
#     ort_P_b <- orthoP[k_ort:(k_ort + p_b - 1L), , drop = FALSE]
#     vip_o <- .vip_ort_for_loadings(Q_ort, ort_P_b)
#     vip_t <- sqrt((vip_p^2 + vip_o^2) / 2)
#     vdf <- data.frame(
#       p = vip_p, o = vip_o, tot = vip_t,
#       row.names = MB@Variables[[i]]
#     )
#     VIP_blocks[[TableLabels[i]]] <- vdf
#     k_ort <- k_ort + p_b
#   }
#
#   VIP_global <- as.vector(unlist(lapply(VIP_blocks, function(b) b$tot)))
#   names(VIP_global) <- varnames
#
#   # SingVal and R2X are computed from scores (set above).
#
#   rm(saliences)
#
#   # Define block length
#   belong_block <- rep(0, nrow(res_calib$P))
#   k <- 1
#   for (i in 1:ntable) {
#     belong_block[k:(k + variable_number[i] - 1)] <- TableLabels[i]
#     k <- k + variable_number[i]
#   }
#
#   rm(Q)
#
#   ## Y prediction using the orthogonally-deflated X and KOPLS B matrix
#   Ypred <- Xnorm_mat %*% B_kopls +
#     matrix(as.vector(B0_kopls), nrow = nrowMB, ncol = ncol(y), byrow = TRUE)
#   rownames(Ypred) <- MB@Samples
#   if (!is.null(colnames(y))) colnames(Ypred) <- colnames(y)
#
#   ## Predict classes
#   if (method == "OPLS-DA") {
#     predClass <- rep(NA, nrow(y))
#
#     if (decisionRule == "max") {
#       for (i in seq_len(nrow(Ypred))) {
#         xx <- which(Ypred[i, ] == max(Ypred[i, ], na.rm = TRUE))
#         if (length(xx) == 1) {
#           predClass[i] <- colnames(y)[xx] # Y stores the Class names
#         } else {
#           predClass[i] <- NaN # There is a tie.
#           warning(sprintf("Class for sample %s could not be predicted.", i))
#         }
#       }
#     } else if (decisionRule == "fixed") {
#       for (i in seq_len(nrow(Ypred))) {
#         xx <- which(Ypred[, i] > 1 / ncol(y))
#         if (length(xx) == 1) {
#           predClass[i] <- colnames(y)[xx] # Y stores the Class names
#         } else {
#           predClass[i] <- NaN # There is a tie.
#           warning(sprintf("Class for sample %s could not be predicted.", i))
#         }
#       }
#     }
#
#     if (is.matrix(y)) {
#       meanY <- colMeans(y)
#     } else {
#       meanY <- mean(y)
#     }
#
#     classy <- colnames(y)
#     Q2 <- DQ2 <- specvec <- sensvec <- rep(NA, length(classy))
#     confusionMatrix <- list()
#
#     for (i in seq_along(classy)) {
#       pos <- which(classVect == classy[i])
#       neg <- which(classVect != classy[i])
#       tp <- length(which(classVect == classy[i] & predClass == classy[i]))
#       tn <- length(which(classVect != classy[i] & predClass != classy[i]))
#       fp <- length(which(classVect != classy[i] & predClass == classy[i]))
#       fn <- length(which(classVect == classy[i] & predClass != classy[i]))
#
#       if (length(tp) == 0) {
#         tp <- 0
#       }
#
#       if (length(tn) == 0) {
#         tn <- 0
#       }
#
#       if (length(fp) == 0) {
#         fp <- 0
#       }
#
#       if (length(fn) == 0) {
#         fn <- 0
#       }
#
#       sensvec[i] <- tp / length(pos)
#       specvec[i] <- tn / length(neg)
#
#       cm <- matrix(c(tp, fp, fn, tn), ncol = 2, nrow = 2)
#       rownames(cm) <- c("predClass1", "predClass0")
#       colnames(cm) <- c("trueClass1", "trueClass0")
#       confusionMatrix[[classy[i]]] <- cm
#
#       ## DQ2
#       E0 <- Ypred[neg, i] - y[neg, i] # Calculate Residuals of Class0 samples
#       E1 <- Ypred[pos, i] - y[pos, i] # Calculate Residuals of Class1 samples
#
#       E0count <- which(E0 > 0) # Find predictions for Class0 samples larger than 0
#       E1count <- which(E1 < 0) # Find predictions for Class1 samples larger than 1
#
#       SSE0_all <- t(E0) %*% E0
#       SSE1_all <- t(E1) %*% E1
#
#       SSE0 <- t(E0[E0count]) %*% E0[E0count]
#       SSE1 <- t(E1[E1count]) %*% E1[E1count]
#
#       PRESSD <- SSE0 + SSE1
#       PRESS <- SSE0_all + SSE1_all
#
#       Ym <- y[, i] - mean(y[, i])
#       TSS <- t(Ym) %*% Ym
#
#       DQ2[i] <- 1 - (PRESSD / TSS)
#       Q2[i] <- 1 - PRESS / TSS
#     }
#     names(Q2)      <- classy
#     names(DQ2)     <- classy
#     names(sensvec) <- classy
#     names(specvec) <- classy
#   } else if (method == "OPLS-R") {
#     PRESS <- t(Ypred - y) %*% (Ypred - y)
#     Ym <- y - mean(y)
#     TSS <- t(Ym) %*% Ym
#     Q2 <- 1 - PRESS / TSS
#   }
#
#
#   # --- Internal kernel centering helpers ---
#   .cv_center_KtrTr <- function(K) scale(t(scale(K, scale = FALSE)), scale = FALSE)
#
#   .cv_center_KteTr <- function(KteTr, KtrTr) {
#     ntr <- nrow(KtrTr)
#     nte <- nrow(KteTr)
#     sc <- matrix(1 / ntr, nte, ntr)
#     (KteTr - sc %*% KtrTr) %*% (diag(ntr) - (1 / ntr) * matrix(1, ntr, ntr))
#   }
#
#   # --- Internal KOPLS model fit on a centred kernel ---
#   .cv_kopls_fit <- function(K_cc, Yc, a, nox) {
#     n <- nrow(K_cc)
#     CSV <- svd(crossprod(Yc, K_cc %*% Yc), nu = a, nv = a)
#     Cp <- CSV$u[, seq_len(a), drop = FALSE]
#     Sp <- diag(CSV$d[seq_len(a)]^(-0.5), nrow = a)
#     Up <- Yc %*% Cp
#     UpSp <- Up %*% Sp
#
#     K_right <- K_cc
#     K_both <- K_cc
#     to_l <- co_l <- so_l <- toNorm_l <- Tp_tr_l <- list()
#     K_right_l <- list()
#     K_right_l[[1]] <- K_cc
#
#     for (jj in seq_len(nox)) {
#       Tp_j <- crossprod(K_right, UpSp)
#       K_resid <- K_both - tcrossprod(Tp_j)
#       sv <- svd(crossprod(Tp_j, K_resid %*% Tp_j), nu = 1, nv = 1)
#       co_j <- sv$u[, 1, drop = FALSE]
#       so_j <- sv$d[1]
#       to_raw <- K_resid %*% Tp_j %*% co_j / sqrt(max(so_j, .Machine$double.eps))
#       tn <- sqrt(sum(to_raw^2))
#       if (tn < 1e-14) tn <- 1
#       to_j <- to_raw / tn
#
#       Tp_tr_l[[jj]] <- Tp_j
#       to_l[[jj]] <- to_j
#       co_l[[jj]] <- co_j
#       so_l[[jj]] <- so_j
#       toNorm_l[[jj]] <- tn
#
#       I_to <- diag(n) - tcrossprod(to_j)
#       K_right <- K_right %*% I_to
#       K_both <- I_to %*% K_both %*% I_to
#       K_right_l[[jj + 1]] <- K_right
#     }
#
#     Tp_f <- crossprod(K_right, UpSp)
#     Bt_f <- pracma::pinv(crossprod(Tp_f)) %*% crossprod(Tp_f, Up)
#
#     list(
#       Cp = Cp, Sp = Sp, Up = Up, UpSp = UpSp, Bt = Bt_f,
#       to = to_l, co = co_l, so = so_l, toNorm = toNorm_l,
#       Tp_tr = Tp_tr_l, K_right = K_right_l
#     )
#   }
#
#   # --- Internal KOPLS predict: returns Yhat in centred Y space ---
#   .cv_kopls_predict <- function(KteTr_c, m, nox) {
#     KteTr_curr <- KteTr_c
#
#     for (i in seq_len(nox)) {
#       Tp_te_i <- KteTr_curr %*% m$UpSp
#       K_res_tetr <- KteTr_curr - tcrossprod(Tp_te_i, m$Tp_tr[[i]])
#       to_te_i <- (K_res_tetr %*% m$Tp_tr[[i]] %*% m$co[[i]]) /
#         sqrt(max(m$so[[i]], .Machine$double.eps))
#       to_te_i <- to_te_i / m$toNorm[[i]]
#       KteTr_curr <- KteTr_curr -
#         tcrossprod(to_te_i, m$K_right[[i]] %*% m$to[[i]])
#     }
#
#     Tp_te <- KteTr_curr %*% m$UpSp
#     Tp_te %*% m$Bt %*% t(m$Cp)
#   }
#
#   if (cv.k >= 2) {
#     n_cv   <- nrowMB
#     n_lev  <- nort + 1L   # levels: nox = 0, 1, ..., nort
#
#     Ypred_cv_all <- vector("list", n_lev)
#     for (.lev in seq_len(n_lev))
#       Ypred_cv_all[[.lev]] <- matrix(NA_real_, nrow = n_cv, ncol = ncol(y),
#                                      dimnames = list(NULL, colnames(y)))
#     pressy_cls_lev <- matrix(0, nrow = n_lev, ncol = ncol(y))
#     TSS_cv_cls     <- numeric(ncol(y))
#     cv_test_order  <- integer(0)
#
#     for (fold in seq_len(cv.k)) {
#       pred_idx  <- seq.int(fold, n_cv, by = cv.k)
#       train_idx <- setdiff(seq_len(n_cv), pred_idx)
#
#       KtrTr_raw <- W_mat[train_idx, train_idx, drop = FALSE]
#       KteTr_raw <- W_mat[pred_idx,  train_idx, drop = FALSE]
#
#       KtrTr_c <- .cv_center_KtrTr(KtrTr_raw)
#       KteTr_c <- .cv_center_KteTr(KteTr_raw, KtrTr_raw)
#
#       Y_tr_mean <- colMeans(y[train_idx, , drop = FALSE])
#       Y_train_c <- sweep(y[train_idx, , drop = FALSE], 2, Y_tr_mean)
#       Y_test_c  <- sweep(y[pred_idx,  , drop = FALSE], 2, Y_tr_mean)
#
#       for (nox in 0L:nort) {
#         .lev  <- nox + 1L
#         m_lev <- .cv_kopls_fit(KtrTr_c, Y_train_c, ndim, nox)
#         Yhat  <- .cv_kopls_predict(KteTr_c, m_lev, nox)
#         Ypred_cv_all[[.lev]][pred_idx, ] <- sweep(Yhat, 2, Y_tr_mean, "+")
#         pressy_cls_lev[.lev, ] <- pressy_cls_lev[.lev, ] + colSums((Y_test_c - Yhat)^2)
#       }
#
#       TSS_cv_cls    <- TSS_cv_cls + colSums(Y_test_c^2)
#       cv_test_order <- c(cv_test_order, pred_idx)
#     }
#
#     Q2_cv <- setNames(1 - pressy_cls_lev[n_lev, ] / TSS_cv_cls, colnames(y))
#     q2_names  <- c("p", paste0("po", seq_len(nort)))
#     Q2_levels <- setNames(1 - rowSums(pressy_cls_lev) / sum(TSS_cv_cls), q2_names)
#
#     if (method == "OPLS-DA") {
#       DQ2_cv      <- setNames(rep(NA_real_, ncol(y)), colnames(y))
#       Ypred_final <- Ypred_cv_all[[n_lev]]
#       for (i in seq_len(ncol(y))) {
#         y_obs  <- y[cv_test_order, i]
#         y_pred <- Ypred_final[cv_test_order, i]
#         E0     <- y_pred[y_obs == 0]
#         E1     <- y_pred[y_obs == 1] - 1
#         PRESSD <- sum(E0[E0 > 0]^2) + sum(E1[E1 < 0]^2)
#         TSS_i  <- sum((y_obs - mean(y_obs))^2)
#         DQ2_cv[i] <- 1 - PRESSD / TSS_i
#       }
#     } else {
#       DQ2_cv <- vector()
#     }
#
#     Q2 <- Q2_cv
#     if (method == "OPLS-DA") DQ2 <- DQ2_cv
#
#     Ypred_cv   <- Ypred_cv_all[[n_lev]]
#     Ypred_cv_p <- Ypred_cv_all[[1L]]
#
#     cv_result <- list(
#       k          = cv.k,
#       Ypred      = Ypred_cv,
#       Ypred.p    = Ypred_cv_p,
#       Q2         = Q2_cv,
#       Q2.levels  = Q2_levels,
#       DQ2        = if (method == "OPLS-DA") DQ2_cv else NULL
#     )
#   } else {
#     if (cv.k != 0) {
#       warning("'cv.k' must be >= 2 or 0 (skip CV). No cross-validation performed.")
#     }
#     cv_result <- list()
#   }
#
#   end_output <- Sys.time()
#   running_time <- (end_output - start_ini_time)
#   if (loquace) {
#     message(sprintf("Analysis finished after : %s millisecs", running_time * 1000))
#   }
#
#   progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
#   utils::setTxtProgressBar(progress_bar, value = 80)
#
#   close(progress_bar)
#
#   res_calib <- new("ComDim",
#     Method = method,
#     ndim = ndim,
#     Q.scores = res_calib$Q,
#     T.scores = res_calib$T_Loc,
#     P.loadings = res_calib$P,
#     Saliences = res_calib$saliences,
#     Orthogonal = list(
#       nort = nort,
#       Q.scores = Q_ort,
#       T.scores = T_Loc_ort,
#       P.loadings.ort = orthoP,
#       Saliences.ort = saliences_ort
#     ),
#     VIP = VIP_global,
#     VIP.block = VIP_blocks,
#     R2X = res_calib$explained,
#     R2Y = r2y,
#     Q2 = Q2,
#     DQ2 = if (method == "OPLS-DA") {
#       DQ2
#     } else {
#       vector()
#     },
#     Singular = res_calib$SingVal,
#     Mean = list(
#       MeanMB = res_calib$MeanMB,
#       MeanY = meanY
#     ),
#     Norm = list(
#       NormMB = res_calib$NormMB,
#       FrobNorms = frob_norms,
#       RVweights = rv_weights
#     ),
#     PLS.model = list(
#       W = W_pred,
#       B = B_kopls,
#       B0 = B0_kopls,
#       Y = y
#     ),
#     cv = cv_result,
#     Prediction = if (method == "OPLS-DA") {
#       list(
#         Y.pred = Ypred,
#         decisionRule = decisionRule,
#         trueClass = classVect,
#         predClass = data.frame(predClass = predClass, row.names = MB@Samples),
#         Sensitivity = sensvec,
#         Specificity = specvec,
#         confusionMatrix = confusionMatrix
#       )
#     } else {
#       list(Y.pred = Ypred)
#     },
#     Metadata = res_calib$metadata,
#     variable.block = belong_block,
#     runtime = as.numeric(running_time)
#   )
#
#   return(res_calib)
# }


#' OPLS_NIPALS_DNR
#'
#' One NIPALS OPLS step on a (lambda-weighted) concatenated block matrix.
#' Computes one predictive component and one orthogonal component in a single
#' pass. This function is the recommended building block for constructing an
#' OPLS-based \code{FUN} argument for \code{ComDim_y()} when \code{nort > 0}.
#' For \code{nort = 0} (plain PLS) a simpler PLS wrapper is sufficient.
#' @param W Numeric matrix (n x p): the concatenated, lambda-weighted blocks
#'   as passed by \code{ComDim_y()}.
#' @param y Numeric matrix (n x q): the response block (dummy matrix for
#'   discriminant analysis, numeric matrix for regression). Only the first
#'   column drives the NIPALS u-score iteration; all columns are used to
#'   compute the Y-loading \code{q}.
#' @param threshold Convergence threshold for the u-score update (default
#'   \code{1e-10}).
#' @return A named list:
#' \describe{
#'   \item{t_pred}{Predictive X-score (length n).}
#'   \item{w_pred}{Predictive X-weight (length p), L2-normalised.}
#'   \item{p}{X-loading (length p).}
#'   \item{q}{Y-loading (length q = ncol(y)).}
#'   \item{u}{Y-score (length n).}
#'   \item{t_ort}{Orthogonal X-score (length n).}
#'   \item{w_ort}{Orthogonal X-weight (length p), L2-normalised.}
#'   \item{p_ort}{Orthogonal X-loading (length p).}
#' }
#' @seealso \code{\link{ComDim_y}} for the multi-block OPLS wrapper that uses
#'   this function.
#' @export
OPLS_NIPALS_DNR <- function(W = W, y = y, threshold = 1e-10) {
  u <- y[, 1]
  u_old <- u + threshold * 10 # Ensures the while loop runs at least once
  iters <- 0

  while (norm(u - u_old, "2") > threshold && iters < 100) {
    iters <- iters + 1
    u_old <- u

    # Predictive X-weight and score
    w_pred <- t(W) %*% u
    w_pred_len <- norm(w_pred, "2")
    if (w_pred_len > 1e-12) w_pred <- w_pred / w_pred_len
    t_pred <- W %*% w_pred

    # Y-loading and score
    q_y <- t(y) %*% t_pred / as.vector(t(t_pred) %*% t_pred)
    u <- y %*% q_y / as.vector(t(q_y) %*% q_y)
  }

  # X-loading for predictive direction
  p <- t(W) %*% t_pred / as.vector(t(t_pred) %*% t_pred)

  # Orthogonal weight: component of p orthogonal to w_pred
  w_ort <- p - w_pred * as.vector(t(w_pred) %*% p) / as.vector(t(w_pred) %*% w_pred)
  w_ort_len <- norm(w_ort, "2")
  if (w_ort_len > 1e-12) w_ort <- w_ort / w_ort_len

  # Orthogonal score and loading
  t_ort <- W %*% w_ort
  p_ort <- t(W) %*% t_ort / as.vector(t(t_ort) %*% t_ort)

  return(list(
    t_pred = t_pred,
    w_pred = w_pred,
    p = p,
    q = q_y,
    u = u,
    t_ort = t_ort,
    w_ort = w_ort,
    p_ort = p_ort
  ))
}

#' MakeComDimScoresTable
#'
#' Creates a long (tidy) data frame with the global and/or local scores from a ComDim model,
#' suitable for use with \pkg{ggplot2}.
#' @param model The output from a ComDim analysis (a \code{ComDim} object).
#' @param blocks The blocks from which local scores will be extracted. A vector of integers (block
#'   indices) or block names. When omitted, all blocks are included. Only relevant when
#'   \code{"T.scores"} or \code{"T.scores.ort"} are in \code{include}.
#' @param dim Integer vector of predictive component indices to include. When omitted, all
#'   predictive components in the model are included.
#' @param dim.ort Integer vector of orthogonal component indices to include. When \code{NULL},
#'   all orthogonal components present in the model are included. When \code{0}, no orthogonal
#'   components are included.
#' @param include Character vector selecting which score types to include. Accepted values
#'   (case-insensitive) are \code{"Q.scores"} (global predictive scores),
#'   \code{"T.scores"} (local predictive scores), \code{"Q.scores.ort"} (global orthogonal
#'   scores), and \code{"T.scores.ort"} (local orthogonal scores). Defaults to
#'   \code{c("Q.scores", "T.scores")}.
#' @return A long data frame with one row per sample–component–score-type combination, containing
#'   the following columns:
#'   \describe{
#'     \item{\code{sample.id}}{Sample name (factor).}
#'     \item{\code{sample.id.number}}{Integer position of the sample (factor).}
#'     \item{\code{block.id}}{Block index, or \code{"Global"} / \code{"Global.ort"} for Q scores.}
#'     \item{\code{block.name}}{Block name, or \code{"Global"} / \code{"Global.ort"} for Q scores (factor).}
#'     \item{\code{dim}}{Component number.}
#'     \item{\code{scores.type}}{One of \code{"Global"}, \code{"Global.ort"}, \code{"Local"}, or \code{"Local.ort"} (factor).}
#'     \item{\code{scores.type.dim}}{Concatenation of scores type and component number, e.g. \code{"Q.scores1"} (factor).}
#'     \item{\code{value}}{Score value.}
#'   }
#' @seealso \code{\link{MakeComDimLoadingsTable}}, \code{\link{ComDim_PCA}}
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50) # 10 rows and 50 columns
#' b2 <- matrix(rnorm(800), 10, 80) # 10 rows and 80 columns
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' model <- ComDim_PCA(mb, ndim = 2)
#' tbl <- MakeComDimScoresTable(model)
#' @export
MakeComDimScoresTable <- function(model,
                                  blocks = NULL,
                                  dim = NULL,
                                  dim.ort = NULL,
                                  include = c("Q.scores", "T.scores", "Q.scores.ort", "T.scores.ort")[1:2]) {
  if (!inherits(model, "ComDim")) {
    stop("'model' is  not the output of a ComDim analysis.")
  }

  if (is.null(dim)) {
    dim <- 1:model@ndim
  }

  if (is.null(dim.ort) && length(model@Orthogonal) != 0) {
    if (model@Orthogonal$nort != 0) {
      dim.ort <- 1:model@Orthogonal$nort
    }
  } else if (is.null(dim.ort) && length(model@Orthogonal) == 0) {
    dim.ort <- 0
  }

  if (is.null(include)) {
    include <- c("Q.scores", "T.scores")
  } else { # Added to match words regardless upper or lowercase
    include <- toupper(include)
    include <- gsub("T.SCORES", "T.scores", include)
    include <- gsub("Q.SCORES", "Q.scores", include)
    include <- gsub("ORT", "ort", include)
  }

  if (any(!(include %in% c("Q.scores", "T.scores", "Q.scores.ort", "T.scores.ort")))) {
    stop("Some elements in 'include' were not recognized'.")
  }


  if (!(is.null(blocks))) {
    if (is.character(blocks)) {
      block.ids <- match(blocks, names(model@T.scores))
      if (any(is.na(block.ids))) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(blocks[which(is.na(block.ids))], sep = ", ")
        )
        block.ids <- block.ids[!is.na(block.ids)]
      }
    } else if (is.numeric(blocks)) {
      block.ids <- intersect(seq_along(names(model@T.scores)), blocks)
      if (length(blocks) != length(block.ids)) {
        warning(
          "Some block names were not found in the MultiBlock: %s",
          paste0(setdiff(seq_along(names(model@T.scores)), blocks), sep = ", ")
        )
      }
    }
  } else {
    block.ids <- seq(1, length(names(model@T.scores)))
  }

  if (length(block.ids) == 0) {
    stop("0 blocks were selected. The table cannot be built.")
  }

  # Check number of samples is always the same across all Q and T.
  nsamples <- nrow(model@Q.scores)
  for (i in block.ids) {
    if (nsamples != nrow(model@T.scores[[i]])) {
      stop("'model' is not an output from a ComDim analysis.")
    }
  }

  df <- NULL
  if ("Q.scores" %in% include) {
    for (v in dim) { # For each of the dimensions
      df2 <- data.frame(
        sample.id = rownames(model@Q.scores),
        sample.id.number = seq(1, nsamples),
        block.id = rep("Global", nsamples),
        block.name = rep("Global", nsamples),
        dim = rep(v, nsamples),
        scores.type = rep("Global", nsamples),
        scores.type.dim = paste0(rep("Q.scores", nsamples), rep(v, nsamples)),
        value = model@Q.scores[, v]
      )
      if (is.null(df)) {
        df <- df2
      } else {
        df <- rbind.data.frame(df, df2)
      }
    }
  }

  if (dim.ort != 0) {
    if ("Q.scores.ort" %in% include) {
      for (v in dim.ort) { # For each of the dimensions
        df2 <- data.frame(
          sample.id = rownames(model@Q.scores),
          sample.id.number = seq(1, nsamples),
          block.id = rep("Global.ort", nsamples),
          block.name = rep("Global.ort", nsamples),
          dim = rep(v, nsamples),
          scores.type = rep("Global.ort", nsamples),
          scores.type.dim = paste0(rep("Q.scores.ort", nsamples), rep(v, nsamples)),
          value = model@Orthogonal$Q.scores[, v]
        )
        if (is.null(df)) {
          df <- df2
        } else {
          df <- rbind.data.frame(df, df2)
        }
      }
    }
  }

  if ("T.scores" %in% include) {
    for (i in block.ids) { # For each of the dimensions
      for (v in dim) {
        df2 <- data.frame(
          sample.id = rownames(model@Q.scores),
          sample.id.number = seq(1, nsamples),
          block.id = rep(i, nsamples),
          block.name = rep(names(model@T.scores)[i], nsamples),
          dim = rep(v, nsamples),
          scores.type = rep("Local", nsamples),
          scores.type.dim = paste0(rep("T.scores", nsamples), rep(v, nsamples)),
          value = model@T.scores[[i]][, v]
        )
        if (is.null(df)) {
          df <- df2
        } else {
          df <- rbind.data.frame(df, df2)
        }
      }
    }
  }

  if (dim.ort != 0) {
    if ("T.scores.ort" %in% include) {
      for (i in block.ids) { # For each of the dimensions
        for (v in dim.ort) {
          df2 <- data.frame(
            sample.id = rownames(model@Q.scores),
            sample.id.number = seq(1, nsamples),
            block.id = rep(i, nsamples),
            block.name = rep(names(model@Orthogonal$T.scores)[i], nsamples),
            dim = rep(v, nsamples),
            scores.type = rep("Local.ort", nsamples),
            scores.type.dim = paste0(rep("T.scores.ort", nsamples), rep(v, nsamples)),
            value = model@Orthogonal$T.scores[[i]][, v]
          )
          if (is.null(df)) {
            df <- df2
          } else {
            df <- rbind.data.frame(df, df2)
          }
        }
      }
    }
  }
  rownames(df) <- NULL

  # Add factors to data
  df$sample.id <- factor(df$sample.id, levels = df$sample.id[seq(1, nsamples)])
  df$sample.id.number <- factor(df$sample.id.number, levels = c("Global", seq(1, nsamples)))
  df$block.name <- factor(df$block.name, levels = c(names(model@T.scores), "Global", "Global.ort"))
  df$scores.type <- factor(df$scores.type, levels = unique(df$scores.type))
  df$scores.type.dim <- factor(df$scores.type.dim, levels = unique(df$scores.type.dim))

  return(df)
}

#' Metadata for the metabolites in growth medium
#'
#'
#' @name KEGG_table_metabolites
#'
#' @docType data
#'
#' @usage data(mouse_ds)
#'
#' @keywords dataset
#'
#' @author  Radic Shechter et al. (2021) Molecular Systems Biology 17:e10141
#' (\doi{10.15252/msb.202010141})
#'
#' @source \doi{10.15252/msb.202010141}
#'
#' @examples
#' data(mouse_ds)
#' #' intraMB <- ExpandMB(data = intra, metadata = KEGG_table_metabolites,
#' #'   minblock = 10, loquace = FALSE)
"KEGG_table_metabolites"

#' SplitRW
#'
#' Splits a multi-block into a replicate-wise (RW) structure by expanding each block along its
#' batch dimension. Each batch within each original block becomes a separate block in the output,
#' enabling replicate-wise ComDim analysis.
#' @param MB A \code{MultiBlock} object built with \code{\link{MultiBlock}()}. Must contain
#'   \code{Batch} information; blocks without batch labels are treated as a single batch.
#' @param checkSampleCorrespondence Logical. If \code{FALSE} (default), identical sample order
#'   and count are assumed across all batches and only the minimum batch size is used. If
#'   \code{TRUE}, sample names are intersected across all batches and only common samples are
#'   retained, which is safer when batches have different sizes or orderings.
#' @param batchNormalisation Logical. If \code{TRUE} (default), each replicate block is divided
#'   by its Frobenius norm and by the square root of the number of replicates in its original
#'   block, to prevent blocks with more replicates from dominating the analysis.
#' @param showSampleCorrespondence Logical. If \code{TRUE} (default), the matrix of sample
#'   names assigned to each replicate block is printed to the console via \code{print()}.
#'   Set to \code{FALSE} to suppress this output.
#' @details
#'   Output block names follow the convention \code{<original_block>} when the original block has
#'   only one batch, or \code{<original_block>_<batch_label>} when it has multiple batches.
#'   The \code{Metadata} slot of each source block is also split and carried over to the
#'   corresponding replicate blocks. If the MultiBlock has no \code{Batch} information at all,
#'   the original object is returned unchanged with a warning.
#' @return A \code{MultiBlock} object in which each block corresponds to one batch of one
#'   original data block (a replicate-wise structure ready for \code{\link{ComDim_PCA}} or
#'   similar).
#' @seealso \code{\link{MultiBlock}}, \code{\link{ComDim_PCA}}, \code{\link{SelectFeaturesRW}}
#' @examples
#' b1 <- matrix(rnorm(1500), 30, 50)
#' b2 <- matrix(rnorm(2400), 30, 80)
#' batch_b <- c(rep(1, 10), rep(2, 10), rep(3, 10))
#' # Generate the multi-block (mb) with 3 batches of 10 samples each
#' mb <- MultiBlock(
#'   Data = list(b1 = b1, b2 = b2),
#'   Batch = list(b1 = batch_b, b2 = batch_b),
#'   ignore.names = TRUE
#' )
#' rw <- SplitRW(mb)
#' @export
SplitRW <- function(MB = MB, checkSampleCorrespondence = FALSE,
                    batchNormalisation = TRUE, showSampleCorrespondence = TRUE) {
  newMB <- list()

  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is  not a MultiBlock.")
  }

  if (!is.list(MB@Samples)) {
    samples <- MB@Samples
    MB@Samples <- list()
    xx <- names(MB@Data)
    for (i in seq_along(xx)) {
      MB@Samples[[xx[i]]] <- samples
    }
    rm(samples)
  }

  # Check that the batch information is correctly provided for all the blocks:
  give_error <- 0

  batch_names <- list()
  samples_per_batch <- list()

  if (length(MB@Batch) != 0) {
    for (i in names(MB@Data)) { # Check the batch information is consistent across RWs
      if (i %in% names(MB@Batch)) {
        batch_names[[i]] <- sort(unique(MB@Batch[[i]]))
        sampled <- as.vector(NULL)
        for (j in seq_along(batch_names[[i]])) {
          sampled[j] <- length(which(MB@Batch[[i]] == batch_names[[i]][j]))
        }
        samples_per_batch[[i]] <- sampled
        if (length(unique(sampled)) > 1) {
          warning(sprintf("The replicate blocks in %s have a different number of samples.", names(MB@Data)[i]))
        }
      } else {
        batch_names[[i]] <- "Batch 1"
        samples_per_batch[[i]] <- MB@Samples[[i]]
      }
    }
    if (length(unique(unlist(samples_per_batch))) > 1 && checkSampleCorrespondence == FALSE) {
      message("Information is missing regarding the splitting.")
      message("Using checkSampleCorrespondence as TRUE is recommended.")
      stop("The data cannot be split into replicate blocks.")
    }

    for (i in names(MB@Batch)) {
      if (checkSampleCorrespondence == TRUE) {
        for (j in seq_along(batch_names[[i]])) {
          batch_position <- which(MB@Batch[[i]] == batch_names[[i]][j])
          if (i == names(MB@Batch)[1] && j == 1) {
            replicate_names <- MB@Samples[[i]][batch_position]
          } else {
            replicate_names <- intersect(replicate_names, MB@Samples[[i]][batch_position])
          }
        }
      } else {
        replicate_names <- seq_len(min(unlist(samples_per_batch))) # number of samples in the smallest batch
      }
    }

    if (checkSampleCorrespondence == TRUE && length(replicate_names) == 0) {
      message("There are 0 samples in common across the replicate blocks")
      give_error <- 1
    }

    if (give_error) {
      stop("The data cannot be split into replicate blocks.")
    }
  } else {
    warning("The MultiBlock does not contain 'Batch' information. It cannot be split into replicate blocks.")
    return(MB)
  }


  ## PROCEED WITH THE RW SPLITTING.
  df_SampleNames <- matrix(, nrow = length(replicate_names), ncol = 0)

  k <- 1
  for (i in names(MB@Batch)) {
    for (j in seq_along(batch_names[[i]])) {
      batch_position <- which(MB@Batch[[i]] == batch_names[[i]][j])
      replicate_position <- as.vector(NULL)
      if (checkSampleCorrespondence == TRUE) {
        for (namei in replicate_names) {
          replicate_position <- c(
            replicate_position,
            intersect(which(MB@Samples[[i]] == namei), batch_position)
          )
        }
      } else {
        replicate_position <- batch_position
      }
      # Keep only the common samples across blocks
      if (length(replicate_position) != length(replicate_names)) {
        warning(sprintf("There are sample duplicates in batch %s from block %s.", as.character(batch_names[[i]][j]), as.character(i)))
        warning("Duplicate samples should be removed.")
        stop("Existence of duplicate samples within one or more batches.")
      }

      sorted <- order(MB@Samples[[i]][replicate_position])

      df_SampleNames <- cbind(df_SampleNames, MB@Samples[[i]][replicate_position[sorted]])

      growingMB <- MultiBlock(
        Samples = replicate_names,
        Data = list(s1 = MB@Data[[i]][replicate_position[sorted], ]),
        Variables = list(s1 = MB@Variables[[i]]),
        Batch = list(s1 = rep(batch_names[[i]][j], length(sorted)))
      )

      if (length(batch_names[[i]]) == 1) {
        blockNames(growingMB) <- i
      } else {
        blockNames(growingMB) <- paste(i, as.character(batch_names[[i]][j]), sep = "_")
      }

      if (batchNormalisation) { # Divide each RW block by its norm and the sqrt(number_of_replicates) to avoid overrepresentation of the data in the multi-set.
        normed <- growingMB@Data[[1]] / norm(growingMB@Data[[1]], type = "F")
        growingMB@Data[[1]] <- normed / sqrt(length(batch_names[[i]]))
      }

      # if(length(names(MB[[i]])) > 5){ # In case there are additional fields
      if (length(MB@Metadata[[i]]) != 0) { # In case there is metadata

        growingMB@Metadata[[blockNames(growingMB)]] <- MB@Metadata[[i]][replicate_position[sorted], ]
      }
      k <- k + 1

      if (j == 1) {
        mbb <- growingMB
      } else {
        mbb@Data <- c(mbb@Data, growingMB@Data)
        mbb@Variables <- c(mbb@Variables, growingMB@Variables)
        mbb@Batch <- c(mbb@Batch, growingMB@Batch)
        mbb@Metadata <- c(mbb@Metadata, growingMB@Metadata)
      }
    }
    if (i == names(MB@Batch)[1]) {
      newMB <- mbb
    } else {
      newMB@Data <- c(newMB@Data, mbb@Data)
      newMB@Variables <- c(newMB@Variables, mbb@Variables)
      newMB@Batch <- c(newMB@Batch, mbb@Batch)
      newMB@Metadata <- c(newMB@Metadata, mbb@Metadata)
    }
  }


  ## SHOW A DATA-FRAME WITH ALL THE SAMPLE NAMES FOR ALL THE (REPLICATE) BLOCKS.
  ## It is wise to use it when checkSampleCorrespondence is set to FALSE.
  if (showSampleCorrespondence) {
    colnames(df_SampleNames) <- names(newMB@Data)
    message("The sample names are:")
    print(df_SampleNames)
  }

  return(newMB)
}

#' Intracellular metabolites in growth medium
#'
#'
#' @name intra
#'
#' @docType data
#'
#' @usage data(mouse_ds)
#'
#' @keywords dataset
#'
#' @author  Radic Shechter et al. (2021) Molecular Systems Biology 17:e10141
#' (\doi{10.15252/msb.202010141})
#'
#' @source \doi{10.15252/msb.202010141}
#'
#' @examples
#' data(mouse_ds)
#' # MB <- MultiBlock(Data = list(RNAseq3 = RNAseq3,
#' #   lipids = lipids, intra = intra, extra = extra))
"intra"

#' SummarizedExperiment2MultiBlock
#'
#' Converts a \code{SummarizedExperiment} into a MultiBlock. Each assay in the
#' \code{SummarizedExperiment} becomes one block (rows = samples, columns = features).
#' @param se A \code{SummarizedExperiment} object.
#' @param colData_samplenames Character string giving the name of the column in \code{colData(se)}
#'   that holds the sample names used for matching. When \code{NULL}, \code{colData} metadata is
#'   not stored in the resulting MultiBlock.
#' @param Batch Character string giving the name of the column in \code{colData(se)} to use as
#'   the Batch variable. When supplied, the column is extracted from \code{colData}, removed from
#'   the metadata, and stored in the \code{Batch} slot of the MultiBlock. Requires at least one
#'   of \code{colData_samplenames} or a named \code{colData}; an error is raised if the column
#'   is not found.
#' @details
#'   Each assay is transposed so that samples are in rows and features in columns. Sample order
#'   is aligned to \code{colData(se)} row names; samples not present in \code{colData} are
#'   removed. If an assay has no name, the block is labelled \code{"X"}. Metadata from
#'   \code{colData} (excluding the Batch column if specified) is stored only for the first block;
#'   subsequent assays are appended as additional blocks with no extra metadata.
#' @return A \code{MultiBlock} object with one block per assay in \code{se}.
#' @seealso \code{\link{MultiBlock}}, \code{\link{MultiAssayExperiment2MultiBlock}},
#'   \code{\link{MultiBlock2MultiAssayExperiment}}
#' @examples
#' \donttest{
#' if (requireNamespace("SummarizedExperiment", quietly = TRUE)) {
#' library(SummarizedExperiment)
#' nrows <- 20; ncols <- 6
#' counts <- matrix(runif(nrows * ncols, 1, 1e4), nrows)
#' rownames(counts) <- paste0("feature", seq_len(nrows))
#' colnames(counts) <- paste0("sample",  seq_len(ncols))
#' se <- SummarizedExperiment(assays = list(counts = counts))
#' mb <- SummarizedExperiment2MultiBlock(se)
#' }
#' }
#' @export
SummarizedExperiment2MultiBlock <- function(se,
                                            colData_samplenames = NULL, # Character string with the column metadata.
                                            Batch = NULL) { # Batch is the name of the column that will be considered as Batch

  # Batch is the name inlistData with the replicate information.
  numassays <- length(assays(se))

  for (i in 1:numassays) {
    Data <- t(as.matrix(assays(se)[[i]]))
    Variables <- list(colnames(Data))
    Data <- list(Data)
    block_name <- names(assays(se))[[i]]
    if (length(block_name) > 1) {
      block_name <- "X"
    }
    names(Variables) <- names(Data) <- block_name

    Data[[block_name]] <- Data[[block_name]][match(rownames(Data[[block_name]]), se@colData@rownames), ]
    if (any(is.na(match(rownames(Data[[block_name]]), se@colData@rownames)))) {
      Data[[block_name]] <- Data[[block_name]][!is.na(match(rownames(Data[[block_name]]), se@colData@rownames)), ]
    }
    rownames(Data[[block_name]]) <- NULL
    colnames(Data[[block_name]]) <- NULL


    if (i == 1) { # First MB

      if (!is.null(Batch)) {
        if (length(Batch) > 1) {
          stop("'Batch' must be a character string of length 1.")
        }
        batch_col <- Batch
        Batch <- se@colData@listData[[batch_col]]
        noBatch <- setdiff(names(se@colData@listData), batch_col)
        if (is.null(colData_samplenames)) {
          MB <- MultiBlock(
            Samples = se@colData@rownames,
            Data = Data,
            Variables = Variables,
            Batch = setNames(list(Batch), block_name)
          )
        } else {
          Metadata <- as.data.frame(se@colData@listData)
          Metadata <- Metadata[, noBatch, drop = FALSE]
          if (!(colData_samplenames %in% colnames(Metadata))) {
            stop("The string in 'colData_samplenames' is not a column name of colData.")
          }
          Metadata <- Metadata[match(
            Metadata[[colData_samplenames]],
            se@colData@rownames
          ), ]
          if (any(is.na(match(
            Metadata[[colData_samplenames]],
            se@colData@rownames
          )))) {
            Metadata <- Metadata[!is.na(match(
              Metadata[[colData_samplenames]],
              se@colData@rownames
            )), ]
          }
          Metadata <- list(Metadata)
          names(Metadata) <- block_name
          MB <- MultiBlock(
            Samples = as.character(Metadata[[block_name]][[colData_samplenames]]),
            Data = Data,
            Variables = Variables,
            Batch = setNames(list(Batch), block_name),
            Metadata = Metadata
          )
        }
      } else { # If Batch is NULL
        if (!is.null(colData_samplenames)) {
          Metadata <- as.data.frame(se@colData@listData)
          Metadata <- Metadata[match(
            Metadata[[colData_samplenames]],
            se@colData@rownames
          ), ]
          if (any(is.na(match(
            Metadata[[colData_samplenames]],
            se@colData@rownames
          )))) {
            Metadata <- Metadata[!is.na(match(
              Metadata[[colData_samplenames]],
              se@colData@rownames
            )), ]
          }
          Metadata <- list(Metadata)
          names(Metadata) <- block_name
          MB <- MultiBlock(
            Samples = as.character(Metadata[[block_name]][[colData_samplenames]]),
            Data = Data,
            Variables = Variables,
            Metadata = Metadata
          )
        } else {
          MB <- MultiBlock(
            Samples = se@colData@rownames,
            Data = Data,
            Variables = Variables
          )
        }
      }
    } else { # Iterations i > 1
      MB@Data <- c(MB@Data, Data)
      MB@Variables <- c(MB@Variables, Variables)
    }
  }

  return(MB)
}

#' ComDim_Exploratory - Extending matrix decomposition methods to multi-block data.
#'
#' Extends any matrix decomposition method used for exploratory purposes to the multi-block
#' field. The user provides a function (\code{FUN}) to compute the scores from the
#' salience-weighted concatenated blocks; these scores are then used to derive the global
#' scores, local scores, and loadings following the traditional ComDim-PCA framework.
#' @param MB A MultiBlock object.
#' @param ndim Number of Common Dimensions.
#' @param FUN The function used as the core of the ComDim analysis. It must accept a matrix
#'   \code{W} (salience-weighted concatenated blocks) and \code{ndim} as its first two
#'   arguments, and return a matrix of scores with one column per component.
#' @param normalise To apply block normalisation. FALSE == no (default), TRUE == yes.
#' @param threshold The threshold limit to stop the iterations. Iterations stop when the
#'   change in the global score vector is below this value (1e-10 as default).
#' @param loquace To display the calculation times. TRUE == yes, FALSE == no (default).
#' @param method A string label identifying the decomposition method used (default: 'FUN').
#' @param ... Additional arguments passed to \code{FUN}.
#' @return A \code{ComDim} object. Slots for supervised analysis
#'   (\code{R2Y}, \code{Q2}, \code{DQ2}, \code{VIP}, \code{VIP.block},
#'   \code{PLS.model}, \code{cv}, \code{Prediction}) are empty. The
#'   populated slots are:
#' \describe{
#'   \item{\code{Method}}{The label supplied via the \code{method} argument.}
#'   \item{\code{ndim}}{Number of Common Dimensions extracted.}
#'   \item{\code{Q.scores}}{Global scores matrix (\eqn{n \times ndim}).
#'     Column names are \code{CC1}, \code{CC2}, etc.; row names are sample
#'     names.  Each column \eqn{\mathbf{q}_a} is a unit-norm consensus score
#'     derived from the dominant left direction of FUN applied to the
#'     salience-weighted concatenated blocks
#'     \eqn{\mathbf{W} = [\sqrt{\lambda_1}\mathbf{X}_1 \mid \cdots \mid
#'     \sqrt{\lambda_B}\mathbf{X}_B]}.}
#'   \item{\code{T.scores}}{Named list of block-specific local scores matrices
#'     (\eqn{n \times ndim} each).  For block \eqn{b} and component \eqn{a}:
#'     local loading \eqn{\mathbf{p}_{ba} = \mathbf{X}_b'\mathbf{q}_a} and
#'     local score
#'     \eqn{\mathbf{t}_{ba} =
#'     \mathbf{X}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
#'   \item{\code{P.loadings}}{Global loadings matrix (\eqn{p_{tot} \times
#'     ndim}).  Column \eqn{a} is
#'     \eqn{\mathbf{P}_a = \mathbf{X}'\mathbf{q}_a},
#'     where \eqn{\mathbf{X}} is the mean-centred (and optionally normalised)
#'     concatenated blocks.}
#'   \item{\code{Saliences}}{Block salience (weight) matrix (\eqn{ntable
#'     \times ndim}, row names = block names).  Entry \eqn{(b,a)} is
#'     \eqn{\lambda_{ba} =
#'     \mathbf{q}_a'\mathbf{X}_b\mathbf{X}_b'\mathbf{q}_a},
#'     the variance of block \eqn{b} captured by global score \eqn{a}.}
#'   \item{\code{R2X}}{Proportion of multi-block variance captured by each
#'     component (named vector, length \eqn{ndim}).  Let \eqn{\mathbf{s}_a}
#'     be the score vector returned by FUN for component \eqn{a} (before
#'     unit-normalisation to obtain \eqn{\mathbf{q}_a}), so that
#'     \eqn{sv_a = \|\mathbf{s}_a\|}; then
#'     \deqn{R2X_a = sv_a^4 \big/ \sum_k sv_k^4 =
#'     Singular_a^2 \big/ \sum_k Singular_k^2.}}
#'   \item{\code{Singular}}{Squared L2 norms of the FUN score vectors:
#'     \eqn{Singular_a = sv_a^2 = \|\mathbf{s}_a\|^2}, used to derive
#'     \code{R2X}.}
#'   \item{\code{Mean}}{List with \code{MeanMB}: named list of column-mean
#'     vectors per block, used for mean-centring.}
#'   \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms used for
#'     block normalisation (all ones when \code{normalise = FALSE}).}
#'   \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
#'     indicating the block name of each row in \code{P.loadings}.}
#'   \item{\code{runtime}}{Total computation time in seconds.}
#' }
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
#' b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#'
#' if (requireNamespace("ica", quietly = TRUE)) {
#'   fun.ICA <- function(W, ndim, ...) {
#'     # W is the concatenated MB.
#'     # ndim is the number of components.
#'     result <- ica::ica(W, ndim)
#'     # The function must return the source estimates (analogous to PCA scores).
#'     return(result$S)
#'   }
#'   resultsICA <- ComDim_Exploratory(mb,
#'     ndim = 2,
#'     FUN = fun.ICA,
#'     method = "ICA"
#'   )
#' }
#' @references Jouan-Rimbaud Bouveresse D, Rutledge DN (2024).
#'   A synthetic review of some recent extensions of ComDim.
#'   \emph{Journal of Chemometrics}, 38(5), e3454.
#'   \doi{10.1002/cem.3454}
#' @export

ComDim_Exploratory <- function(MB = MB, ndim = NULL,
                               FUN = FUN,
                               normalise = FALSE, threshold = 1e-10,
                               loquace = FALSE,
                               method = "FUN",
                               ...) {
  # INITIALISATION

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")

  start_ini_time <- Sys.time() # To start counting calculation time for the initialization


  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is  not a MultiBlock.")
  }

  ntable <- length(blockNames(MB)) # number of blocks
  nrowMB <- length(sampleNames(MB)) # number of samples
  variable_number <- ncol(MB) # Number of variables per block

  give_error <- 0

  # Remove all variables with 0 variance. If there are no remaining variables, give_error
  numvar <- 0
  numblock0 <- vector()
  for (i in 1:ntable) {
    var0 <- apply(MB@Data[[i]], 2, function(x) {
      var(x)
    })
    var0 <- which(var0 == 0)

    if (length(var0) != 0) {
      numvar <- numvar + length(var0)
      if (length(var0) == length(MB@Variables[[i]])) {
        numblock0 <- append(numblock0, i)
      }
      MB@Data[[i]] <- MB@Data[[i]][, setdiff(1:variable_number[i], var0)]
      MB@Variables[[i]] <- MB@Variables[[i]][setdiff(1:variable_number[i], var0)]
      variable_number[i] <- length(MB@Variables[[i]])
    }
  }
  if (numvar != 0) {
    warning(sprintf("Number of variables excluded from the analysis because of their zero variance: %d", numvar))
  }
  if (length(numblock0) != 0) {
    numblock0 <- rev(numblock0) # To sort in decreasing order.
    for (i in seq_along(numblock0)) {
      MB@Data[[numblock0[i]]] <- NULL
      MB@Variables[[numblock0[i]]] <- NULL
      if (numblock0[[i]] %in% MB@Batch) {
        MB@Batch[[numblock0[i]]] <- NULL
      }
      if (numblock0[[i]] %in% MB@Metadata) {
        MB@Metadata[[numblock0[i]]] <- NULL
      }
    }
    warning(sprintf("Number of blocks excluded from the analysis because of their zero variance: %d", length(numblock0)))
  }

  ntable <- length(blockNames(MB)) # number of blocks
  variable_number <- ncol(MB) # In case the number of blocks has changed.

  if (length(MB@Data) == 0) { # In case all blocks had 0 variance...
    warning("All blocks had 0 variance")
    give_error <- 1
  }

  # Check for infinite or missing values (they cannot be handled with svd)
  for (i in 1:ntable) {
    if (any(is.na(MB@Data[[i]]))) {
      stop("The MB contains NA values. They must be removed first for ComDim analysis.")
    }
    if (any(is.infinite(MB@Data[[i]]))) {
      stop("The MB contains infinite values. They must be removed first for ComDim analysis.")
    }
  }

  if (give_error) {
    stop("The data is not ready for ComDim.")
  } else {
    message("The data can be used for ComDim.")
  }


  if (is.null(ndim)) {
    ndim <- ntable # If the number of components is not defined,
    # the number of components to extract is equal to the number of blocks.
  }

  pieceBar <- 4 + 2 * ntable + ndim # Number of updates in the progress bar.
  pieceBar <- 80 / pieceBar
  total_progress <- pieceBar

  DimLabels <- paste0("CC", 1:ndim) # One label per component.
  TableLabels <- blockNames(MB) # One label per block.

  end_ini_time <- Sys.time() # To end the count of the calculation time.

  if (loquace) {
    message(sprintf("Initialisation finished after : %s millisecs", (end_ini_time - start_ini_time) * 1000))
  }

  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # NORMALISATION

  X_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))
  Xnorm_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))

  res_calib <- list()
  temp_tabCalib <- list()
  s_r <- list()
  res_calib$SingVal <- as.vector(NULL)
  res_calib$NormMB <- as.vector(NULL)
  res_calib$MeanMB <- list()

  for (i in 1:ntable) {
    res_calib$MeanMB[[TableLabels[i]]] <- colMeans(MB@Data[[i]])
    names(res_calib$MeanMB[[TableLabels[i]]]) <- MB@Variables[[i]]

    if (normalise) {
      # Normalise original blocks

      X_mean <- MB@Data[[i]] - matrix(data = rep(1, nrowMB), ncol = 1, nrow = nrowMB) %*% res_calib$MeanMB[[i]]
      XX <- X_mean * X_mean
      Norm_X <- sqrt(sum(XX))
      X_Normed <- X_mean / Norm_X
      res_calib$NormMB[i] <- Norm_X
      temp_tabCalib[[i]] <- X_Normed
      s_r[[i]] <- X_Normed
    } else {
      res_calib$NormMB[[TableLabels[i]]] <- rep(1, length(MB@Variables[[i]]))
      names(res_calib$NormMB[[TableLabels[i]]]) <- MB@Variables[[i]]
      temp_tabCalib[[i]] <- MB@Data[[i]]
      s_r[[i]] <- MB@Data[[i]]
    }

    if (i == 1) {
      X_mat[, 1:variable_number[1]] <- MB@Data[[i]]
      Xnorm_mat[, 1:variable_number[1]] <- temp_tabCalib[[i]]
    } else {
      beg <- sum(variable_number[1:(i - 1)]) + 1
      ending <- sum(variable_number[1:i])
      X_mat[, beg:ending] <- MB@Data[[i]]
      Xnorm_mat[, beg:ending] <- temp_tabCalib[[i]]
    }

    norm_comdim <- Sys.time()
    if (loquace) {
      message(sprintf("Normalization of block %s finished after : %s millisecs", i, (norm_comdim - start_ini_time) * 1000))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  names(res_calib$NormMB) <- TableLabels

  nR <- nrow(Xnorm_mat)
  nC <- ncol(Xnorm_mat)

  total_progress <- total_progress + pieceBar * ntable
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  ## DO ComDim

  saliences <- matrix(, ncol = ndim, nrow = ntable)
  Q <- matrix(, ncol = ndim, nrow = nrowMB)
  unexplained <- ntable # Warning!: true only if the tables were set to norm=1
  varexp <- as.vector(NULL)

  for (dims in 1:ndim) {
    previousfit <- 10000
    deltafit <- 1000000
    lambda <- matrix(rep(1, ntable), nrow = ntable, ncol = 1)

    qini <- as.vector(s_r[[1]][, 1])
    qini <- qini / sqrt(as.vector(qini %*% qini)) # random initialization of t + set to unit length
    qold <- 100

    iters <- 0
    while (norm(qold - qini, "2") > threshold && iters < 100) {
      iters <- iters + 1
      qold <- qini
      q <- 0

      W <- matrix(, nrow = nrowMB, ncol = 0)
      for (j in 1:ntable) {
        W <- cbind(W, sqrt(lambda[j]) * s_r[[j]])
      }

      # HERE IS THE CORE COMDIM
      scores <- FUN(W = W, ndim = ndim, ...)

      sv <- sqrt(t(scores[, dims]) %*% scores[, dims]) # For R2X
      qtmp <- scores[, dims] / as.vector(sv)

      for (j in 1:ntable) {
        # takes each table into account for lambda after PCA
        lambda[j] <- t(qtmp) %*% (s_r[[j]] %*% t(s_r[[j]])) %*% qtmp
        q <- q + lambda[j] * qtmp
      }

      q <- q / sqrt(as.vector(t(q) %*% q)) # standardizes t
      if (abs(min(q)) > abs(max(q))) {
        q <- -q
      }
      qini <- q
    } # deltafit>threshold

    saliences[, dims] <- lambda
    Q[, dims] <- q


    res_calib$SingVal[dims] <- sv^2
    varexp[dims] <- res_calib$SingVal[dims]^2

    # Deflate blocks
    aux <- diag(nrowMB) - as.matrix(q) %*% t(as.matrix(q))
    for (j in 1:ntable) {
      s_r[[j]] <- aux %*% s_r[[j]]
    }

    iter_comdim <- Sys.time()
    if (loquace) {
      message(sprintf("Component %s determined after : %s millisecs", dims, (iter_comdim - start_ini_time) * 1000))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  # res_calib$SingVal$i <- 'Singular value'
  names(res_calib$SingVal) <- DimLabels # Dimensions

  # Q$i <- data[[1]]$Samples
  # Q$v <- DimLabels
  colnames(Q) <- DimLabels
  rownames(Q) <- MB@Samples
  res_calib$Q <- Q
  rm(Q)

  ## Adding metadata. Metadata extracted from the first block
  res_calib$metadata <- list()
  if (length(MB@Metadata) != 0) {
    res_calib$metadata[[1]] <- MB@Metadata
  }


  end_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Scores finished after : %s millisecs", (end_comdim - start_ini_time) * 1000))
  }
  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  meanW <- as.vector(NULL)
  for (j in 1:ntable) {
    meanW <- c(meanW, mean(s_r[[j]]))
  }
  # res_calib$s_n <- s_n
  # res_calib$s_r <- s_r
  rm(s_r)

  ##
  # explained <- list()
  explained <- varexp / sum(varexp)
  names(explained) <- DimLabels
  # explained$i <-'% explained'
  # explained$v <- DimLabels #Dimensions
  res_calib$explained <- explained
  rm(varexp)
  rm(explained)

  ## Calculate Sums of saliences for each Dim
  # Sum_saliences_Dim <- list()
  Sum_saliences_Dim <- colSums(saliences)
  names(Sum_saliences_Dim) <- DimLabels
  # Sum_saliences_Dim$i <-'Sum Dim Saliences'
  # Sum_saliences_Dim$v <- DimLabels #Dimensions

  res_calib$Sum_saliences_Dim <- Sum_saliences_Dim
  rm(Sum_saliences_Dim)

  # Calculate Sums of saliences for each Table
  # Sum_saliences_Tab <- list()
  Sum_saliences_Tab <- rowSums(saliences)
  # Sum_saliences_Tab$i <- 'Sum Tab Saliences'
  names(Sum_saliences_Tab) <- TableLabels
  # Sum_saliences_Tab$v <- TableLabels #tables

  res_calib$Sum_saliences_Tab <- Sum_saliences_Tab
  rm(Sum_saliences_Tab)

  # saliences$i <- TableLabels #tables
  # saliences$v <- DimLabels #Dimensions
  res_calib$saliences <- saliences
  colnames(res_calib$saliences) <- DimLabels
  rownames(res_calib$saliences) <- TableLabels

  ## Calculate Normalised concatenated Xs ('Calib') from col
  # Calculate concatenated CD loadings
  # Reorganise Loadings - 1 matrix / LV

  ## If Output = NULL, nothing else is calculated
  # if(!(is.null(output))){
  L_CD_Vec <- NULL
  L_X_Vec <- NULL
  Calib <- NULL

  nCalib <- nrowMB
  # }

  ## Already defined in during ComDim initiallization
  # isT <- grepl(output, 'T')
  # isL <- grepl(output, 'L')
  # isP <- grepl(output, 'P')
  # isLP <- c(isP,isL)
  # isTLP <- c(isT,isP,isL)

  # tic
  # Calculate concatenated CD loadings
  # Reorganise Loadings - 1 matrix / LV
  # L_CD <- list()
  # L_X <- list()
  T_Loc <- list()
  for (i in 1:ntable) { # Prepare lists
    # L_CD[[TableLabels[i]]] <- matrix(,ncol = ndim, nrow = ncol(temp_tabCalib[[i]]))
    # L_X[[TableLabels[i]]] <- matrix(,ncol = ndim, nrow = ncol(temp_tabCalib[[i]]))
    T_Loc[[TableLabels[i]]] <- matrix(, ncol = ndim, nrow = nrowMB)
  }
  # b <- matrix(,ncol = ndim, nrow = ntable)

  if (!requireNamespace("pracma", quietly = TRUE)) {
    stop("The package pracma is needed.")
  }

  for (j in 1:ndim) {

    for (i in 1:ntable) {

      temp <- t(temp_tabCalib[[i]]) %*% res_calib$Q[, j] # Scaled CD 'local' Loadings

      T_Loc[[TableLabels[i]]][, j] <- temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp)) # local Scores

      # Deflate each temp_tabCalib
      temp_tabCalib[[i]] <- temp_tabCalib[[i]] - res_calib$Q[, j] %*% t(temp)
    }

  }

  for (i in 1:ntable) {
    rownames(T_Loc[[TableLabels[i]]]) <- rownames(res_calib$Q)
    colnames(T_Loc[[TableLabels[i]]]) <- colnames(res_calib$Q)
  }

  # If Output==[], nothing else is calculated
  # if(!(is.null(output))){
  # Calculate Global Loadings
  # if(!(is.null(isP))){
  L_CD_Vec <- t(Xnorm_mat) %*% res_calib$Q # Scaled CD 'global' Loadings
  # }

  # if(!(is.null(isL))){
  #  L_X_Vec <- t(X_mat)%*%Q$Data # Unscaled X 'global' Loadings
  # }
  # }

  #### If Output==[], nothing else is calculated
  load_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Loadings finished after : %s millisecs", (load_comdim - start_ini_time) * 1000))
  }

  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  rm(X_mat)
  rm(temp_tabCalib)
  rm(temp)

  ## Output
  end_function <- Sys.time()


  # if(!(is.null(isT))){
  res_calib$T_Loc <- T_Loc

  # T_Loc is no longer needed
  rm(T_Loc)


  # if(!(is.null(isP))){

  varnames <- vector()
  for (i in 1:ntable) {
    varnames <- append(varnames, MB@Variables[[i]])
  }
  # if(length(which(duplicated(varnames))) != 0){
  #  warning('There are at least two variables with the same name. Their name in the rownames of P has been altered.\n')
  # }

  res_calib$P <- L_CD_Vec
  # res_calib$P$i <- varnames # numbers of all variables;
  # res_calib$P$v <- DimLabels # Dimensions
  colnames(res_calib$P) <- DimLabels
  rownames(res_calib$P) <- varnames


  rm(L_CD_Vec)
  # rm(L_CD)

  ##
  rm(saliences)

  # Define block length
  belong_block <- rep(0, nrow(res_calib$P))
  k <- 1
  for (i in 1:ntable) {
    belong_block[k:(k + variable_number[i] - 1)] <- TableLabels[i]
    k <- k + variable_number[i]
  }


  end_output <- Sys.time()
  running_time <- (end_output - start_ini_time)
  if (loquace) {
    message(sprintf("Analysis finished after : %s seconds", running_time))
  }
  res_calib$runtime <- running_time # Total time of analysis

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
  utils::setTxtProgressBar(progress_bar, value = 80)

  close(progress_bar)

  # Save data in a ComDim structure
  res_calib <- new("ComDim",
    Method = method,
    ndim = ndim,
    Q.scores = res_calib$Q,
    T.scores = res_calib$T_Loc,
    P.loadings = res_calib$P,
    Saliences = res_calib$saliences,
    Orthogonal = list(),
    R2X = res_calib$explained,
    R2Y = vector(),
    Q2 = vector(),
    DQ2 = vector(),
    Singular = res_calib$SingVal,
    Mean = list(MeanMB = res_calib$MeanMB, MeanY = NULL),
    Norm = list(NormMB = res_calib$NormMB),
    PLS.model = list(),
    cv = list(),
    Prediction = list(),
    Metadata = res_calib$metadata,
    variable.block = belong_block,
    runtime = as.numeric(res_calib$runtime)
  )
  return(res_calib)
}

#' FilterSamplesMultiBlock
#'
#' Retain a subset of samples in a MultiBlock object.
#' @param MB A MultiBlock object.
#' @param samples A vector of sample names to keep. Names not found in the
#'   MultiBlock are silently ignored; the order of the retained samples
#'   follows the order given here.
#' @return The MultiBlock object restricted to the requested samples. The \code{Batch} and
#'   \code{Metadata} slots are also subsetted and reordered to match.
#' @seealso \code{\link{MultiBlock}}, \code{\link{AddMetadata}}, \code{\link{ProcessMultiBlock}}
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' rownames(b1) <- rownames(b2) <- paste0("sample_", 1:10)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' mb <- FilterSamplesMultiBlock(mb, samples = paste0("sample_", 1:5))
#' @export
FilterSamplesMultiBlock <- function(MB, samples = sampleNames(MB)) {
  if (!("MultiBlock" %in% class(MB))) {
    stop("'MB' is not a MultiBlock object.")
  }

  all_samples <- sampleNames(MB)
  samples <- intersect(samples, all_samples)
  block_names <- blockNames(MB)

  if (length(samples) == 0) {
    stop("None of the provided sample names was found in MB.")
  }

  pos <- match(samples, all_samples)
  ntable <- length(block_names)

  MB@Samples <- MB@Samples[pos]

  for (i in seq_len(ntable)) {
    MB@Data[[i]] <- MB@Data[[i]][pos, , drop = FALSE]
    if (block_names[i] %in% names(MB@Batch)) {
      MB@Batch[[block_names[i]]] <- MB@Batch[[block_names[i]]][pos]
    }
    if (block_names[i] %in% names(MB@Metadata)) {
      MB@Metadata[[block_names[i]]] <- MB@Metadata[[block_names[i]]][pos, , drop = FALSE]
    }
  }

  return(MB)
}

#' ComDim_y - Extending PLS-like supervised methods to multi-block data.
#'
#' Extends any PLS-like method used for regression or discriminant purposes to
#' the multi-block field. The user provides a function (\code{FUN}) that
#' computes one predictive component from the salience-weighted concatenated
#' blocks; global scores, local scores, and loadings are then derived following
#' the traditional ComDim-PLS framework. Optionally, orthogonal components
#' returned by FUN (e.g. from an O-PLS wrapper) are captured. VIP scores and
#' k-fold cross-validation are also supported.
#'
#' @param MB A MultiBlock object.
#' @param y The response: a numeric vector or matrix for regression
#'   (\code{type = 'regression'}), or a class vector or dummy matrix for
#'   discriminant analysis (\code{type = 'discriminant'}).
#' @param ndim Number of predictive Common Dimensions. If \code{NULL},
#'   defaults to the number of blocks.
#' @param FUN The function used as the core of the ComDim analysis. It must
#'   accept \code{W} (salience-weighted concatenated blocks, n x p matrix),
#'   \code{y} (response), and \code{ndim} (number of components to compute)
#'   as its first three named arguments, and return a named list with at least:
#'   \describe{
#'     \item{scores}{X scores vector (length n) for the current component.}
#'     \item{P}{X loadings vector (length p) for the current component.}
#'     \item{W}{X weights vector (length p) for the current component.}
#'     \item{Q}{Y loadings vector for the current component.}
#'     \item{U}{Y scores vector (length n) for the current component.}
#'   }
#'   Optional return fields:
#'   \describe{
#'     \item{y}{y as used internally by FUN (e.g. after centring/scaling),
#'       so that \code{ComDim_y} can detect automatic y-transformation and
#'       back-transform the Q loadings for correct B and B0 computation.}
#'     \item{orthoscores}{A numeric vector (length n) or single-column matrix
#'       of orthogonal X-scores. \strong{Required} when \code{nort = 1}.
#'       Only the first column is used. FUN does not need to change its
#'       behaviour between ort and predictive phases; if it always returns
#'       \code{orthoscores}, the framework uses it only during the ort phase
#'       and ignores it during the predictive phase.}
#'   }
#' @param nort Number of orthogonal Common Dimensions to extract before the
#'   predictive loop. Default \code{0} (no orthogonalization, pure PLS-like).
#'   Only \code{nort = 0} or \code{nort = 1} are accepted; for multiple
#'   orthogonal components use \code{ComDim_OPLS()}.
#'   When \code{nort = 1} the function follows the same two-phase architecture
#'   as \code{ComDim_OPLS}: the orthogonal component is removed from the
#'   (lambda-weighted) concatenated blocks first, then predictive components are
#'   computed from the orthogonally-deflated blocks. \code{FUN} must return
#'   \code{output$orthoscores} (a numeric vector of length n, or a matrix whose
#'   first column is used) when called with \code{ndim = 1}.
#'   Cross-validation is automatically skipped when \code{nort > 0}.
#' @param type \code{'regression'} (default) or \code{'discriminant'}.
#' @param decisionRule Only used when \code{type = 'discriminant'}. If
#'   \code{'fixed'}, samples are assigned to the class whose predicted score
#'   exceeds \code{1/nclasses}; if \code{'max'}, the class with the highest
#'   predicted score is chosen. Default \code{'max'}.
#' @param normalise To apply block normalisation. \code{FALSE} == no
#'   (default), \code{TRUE} == yes.
#' @param scale.y Logical (default \code{FALSE}). When \code{TRUE} and
#'   \code{type = 'regression'}, each column of Y is mean-centred and scaled to
#'   unit variance before being passed to \code{FUN}. The stored B, B0, and
#'   Ypred are back-transformed to the original Y scale, so downstream outputs
#'   (R2Y, Q2, predictions) are always in the original units. Ignored when
#'   \code{type = 'discriminant'} (dummy Y should not be scaled).
#' @param threshold Convergence threshold: iterations stop when the change
#'   in the global score vector falls below this value (default \code{1e-10}).
#' @param loquace Display computation time at each step. \code{TRUE} == yes,
#'   \code{FALSE} == no (default).
#' @param method A string label identifying the method (default: \code{'FUN'}).
#' @param cv.k Number of folds for k-fold cross-validation (default 7). Set
#'   to 0 to skip CV. When \code{cv.k >= 2}, the \code{Q2} and \code{DQ2}
#'   slots in the output reflect cross-validated predictive ability; otherwise
#'   they reflect training-set fit. CV is skipped when \code{nort > 0}.
#' @param ... Additional arguments passed to \code{FUN}.
#' @return A \code{ComDim} object with the following slots:
#' \describe{
#'   \item{\code{Method}}{The label supplied via the \code{method} argument.}
#'   \item{\code{ndim}}{Number of predictive Common Dimensions extracted.}
#'   \item{\code{Q.scores}}{Global consensus scores matrix (\eqn{n \times
#'     ndim}).  Each column \eqn{\mathbf{q}_a} (unit-norm) is derived from
#'     the dominant left direction of FUN applied to the salience-weighted
#'     concatenated blocks.}
#'   \item{\code{T.scores}}{Named list of block-specific local scores
#'     (\eqn{n \times ndim} each).  Local loading
#'     \eqn{\mathbf{p}_{ba} = \tilde{\mathbf{X}}_b'\mathbf{q}_a} (computed
#'     on the ort-deflated block when \code{nort > 0}); local score
#'     \eqn{\mathbf{t}_{ba} =
#'     \tilde{\mathbf{X}}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
#'   \item{\code{P.loadings}}{Global loadings (\eqn{p_{tot} \times ndim}):
#'     \eqn{\mathbf{P} = \tilde{\mathbf{X}}'\mathbf{Q}}, where
#'     \eqn{\tilde{\mathbf{X}}} is the (optionally ort-deflated) mean-centred
#'     concatenated blocks.}
#'   \item{\code{Saliences}}{Block salience matrix (\eqn{ntable \times ndim}):
#'     \eqn{\lambda_{ba} =
#'     \mathbf{q}_a'\tilde{\mathbf{X}}_b\tilde{\mathbf{X}}_b'\mathbf{q}_a}.}
#'   \item{\code{R2X}}{Proportion of X variance captured by each predictive
#'     component (named vector, length \eqn{ndim}).  Let \eqn{\mathbf{t}_a}
#'     be the X-score vector returned by FUN for component \eqn{a}:
#'     \deqn{R2X_a = \|\mathbf{t}_a\|^4 \big/ \sum_k \|\mathbf{t}_k\|^4.}
#'     When \code{nort > 0}, the denominator also includes the orthogonal
#'     \eqn{\|\mathbf{t}_{ort,k}\|^4} terms, and the orthogonal R2X fractions
#'     are stored separately in \code{Orthogonal$R2X}.}
#'   \item{\code{R2Y}}{Cumulative Y-variance explained (named vector, length
#'     \eqn{ndim}):
#'     \deqn{R2Y_a = 1 - RSS_a / TSS_Y,}
#'     where \eqn{RSS_a} is the residual SS from an OLS regression of
#'     \eqn{\mathbf{Y}} on \eqn{[1, \mathbf{q}_1, \ldots, \mathbf{q}_a]}.
#'     \strong{Note:} \eqn{R2Y_a} is cumulative -- the total Y-variance
#'     explained by the first \eqn{a} components together, not the marginal
#'     contribution of component \eqn{a} alone.}
#'   \item{\code{Q2}}{Predictive Q2 per response column (regression) or per
#'     class (discriminant), named accordingly:
#'     \deqn{Q2 = 1 - PRESS / TSS_Y,}
#'     where \eqn{PRESS = \sum_i (\hat{y}_i - y_i)^2}.
#'     When \code{cv.k >= 2} and \code{nort = 0}: cross-validated (out-of-
#'     sample) predictions are used; otherwise training-set predictions.
#'     CV is automatically skipped when \code{nort > 0}.}
#'   \item{\code{DQ2}}{(Discriminant mode only) Discriminant Q2 per class,
#'     using only penalising residuals:
#'     \deqn{DQ2 = 1 - PRESSD / TSS_Y,}
#'     where \eqn{PRESSD} sums \eqn{\hat{y}_i^2} for class-0 samples with
#'     \eqn{\hat{y}_i > 0}, and \eqn{(\hat{y}_i - 1)^2} for class-1 samples
#'     with \eqn{\hat{y}_i < 1}.  Same cross-validation logic as \code{Q2}.}
#'   \item{\code{Singular}}{Squared L2 norm of the FUN X-score vector per
#'     component (\eqn{\|\mathbf{t}_a\|^2}), used to derive \code{R2X}.}
#'   \item{\code{VIP}}{Global total VIP (named vector, length \eqn{p_{tot}}):
#'     concatenation of \code{VIP.block[[b]]$tot} across blocks.  When
#'     \code{nort = 0}, uses the Wold formula; when \code{nort = 1}, tot
#'     combines predictive and orthogonal VIPs (see \code{VIP.block}).}
#'   \item{\code{VIP.block}}{Named list (one \code{data.frame} per block).
#'     When \code{nort = 0}: columns \code{p} and \code{tot} (= \code{p}),
#'     using the Wold formula:
#'     \deqn{VIPp_j = \sqrt{p_b \cdot
#'       \frac{\sum_a s_a \tilde{w}_{j,a}^2}{\sum_a s_a}},}
#'     where \eqn{s_a = \|\mathbf{t}_a\|^2\|\mathbf{q}_a\|^2} and
#'     \eqn{\tilde{w}_{j,a} = w_{j,a}/\|\mathbf{w}_a\|} is the L2-normalised
#'     \eqn{j}-th element of the \eqn{a}-th weight vector.
#'     When \code{nort = 1}: columns \code{p} (Wold, same as above),
#'     \code{o} (orthogonal VIP, loadings-based:
#'     \eqn{VIPo_j = \sqrt{p_b \cdot \sum_a s_{oa}\tilde{P}_{o,j,a}^2 /
#'     \sum_a s_{oa}}},
#'     where \eqn{s_{oa} = \|\mathbf{q}_{ort}[,a]\|^2} and
#'     \eqn{\tilde{\mathbf{P}}_o} is the column-L2-normalised block-slice of
#'     the ort loadings), and \code{tot}
#'     (\eqn{VIPtot_j = \sqrt{(VIPp_j^2 + VIPo_j^2)/2}}).
#'     Row names are variable names.}
#'   \item{\code{PLS.model}}{List with: \code{W} (X weight matrix collected
#'     from FUN, \eqn{p_{tot} \times ndim}); \code{B} (regression
#'     coefficients,
#'     \eqn{\mathbf{B} = \mathbf{W}(\mathbf{P}'\mathbf{W})^{-1}\mathbf{Q}'},
#'     in original Y units); \code{B0} (intercept,
#'     \eqn{\mathbf{B}_0 = \bar{\mathbf{y}} -
#'     \overline{\tilde{\mathbf{x}}}\mathbf{B}}); \code{Y} (original
#'     response matrix as supplied).
#'     Training-set predictions:
#'     \eqn{\hat{\mathbf{Y}} =
#'     \tilde{\mathbf{X}}\mathbf{B} + \mathbf{B}_0}.}
#'   \item{\code{cv}}{Cross-validation results when \code{cv.k >= 2} and
#'     \code{nort = 0} (empty list otherwise): \code{k}, \code{fold}
#'     (sample-to-fold vector), \code{Ypred} (\eqn{n \times ncol(Y)}
#'     out-of-sample predictions), \code{Q2} (CV Q2 per class/response),
#'     \code{DQ2} (mean CV DQ2, discriminant only),
#'     \code{DQ2.perclass} (CV DQ2 per class, discriminant only).}
#'   \item{\code{Orthogonal}}{When \code{nort > 0}: list with \code{nort},
#'     \code{Q.scores} (global ort scores, \eqn{n \times nort}, unit-norm),
#'     \code{T.scores} (block ort local scores, \eqn{n \times nort} each),
#'     \code{P.loadings.ort} (ort loadings, \eqn{p_{tot} \times nort}),
#'     \code{Saliences.ort} (\eqn{ntable \times nort}), and \code{R2X}
#'     (orthogonal X-variance fractions,
#'     \eqn{R2X_{ort,a} = \|\mathbf{t}_{ort,a}\|^4 / total}).
#'     Empty list when \code{nort = 0}.}
#'   \item{\code{Prediction}}{Training-set predictions: \code{Y.pred}
#'     (\eqn{n \times ncol(Y)}); for discriminant analysis also
#'     \code{decisionRule}, \code{trueClass}, \code{predClass} (data.frame),
#'     \code{Sensitivity} and \code{Specificity} (per class),
#'     \code{confusionMatrix} (named list of 2x2 matrices).}
#'   \item{\code{Mean}}{List with \code{MeanMB} (column means per block),
#'     \code{MeanY} (column means of Y), and \code{ScaleY} (column SDs of Y;
#'     all ones when \code{scale.y = FALSE}).}
#'   \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms for block
#'     normalisation.}
#'   \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
#'     mapping each row of \code{P.loadings} and each element of \code{VIP}
#'     to its block.}
#'   \item{\code{runtime}}{Total computation time in seconds.}
#' }
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
#' b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#'
#' ## Example 1: ComDim-PLS (regression) ---------------------------------------
#' # Single-step NIPALS PLS wrapper (one predictive component per call).
#' # Note: 'tx' is used instead of 't' to avoid shadowing base::t().
#' fun.PLS <- function(W, y, ndim, ...) {
#'   output <- list()
#'   w <- t(W) %*% y / as.numeric(t(y) %*% y) # X weight (u = y, 1 step)
#'   w <- w / sqrt(sum(w^2)) # L2 normalise
#'   tx <- W %*% w # X score
#'   p <- t(W) %*% tx / as.numeric(t(tx) %*% tx) # X loading
#'   q <- t(y) %*% tx / as.numeric(t(tx) %*% tx) # Y loading
#'   u <- y %*% q / as.numeric(t(q) %*% q) # Y score
#'   output$scores <- as.vector(tx)
#'   output$P <- as.vector(p)
#'   output$W <- as.vector(w)
#'   output$Q <- as.vector(q)
#'   output$U <- as.vector(u)
#'   return(output)
#' }
#'
#' y <- c(1, 1, 1, 1, 1, 5, 5, 5, 10, 10)
#' resultsPLS <- ComDim_y(mb,
#'   y = y, ndim = 2,
#'   type = "regression",
#'   FUN = fun.PLS,
#'   method = "PLS",
#'   cv.k = 0
#' )
#'
#' ## Example 2: ComDim-OPLS-DA (discriminant, nort = 1) ----------------------
#' # Thin wrapper around OPLS_NIPALS_DNR(), the package's NIPALS OPLS engine.
#' # All inputs (W, y, and any extra args such as 'threshold') are forwarded
#' # directly via '...'. Use this pattern when nort > 0; for nort = 0 the
#' # simpler PLS wrapper in Example 1 is sufficient (no orthoscores needed).
#' fun.OPLS <- function(W, y, ndim, ...) {
#'   res <- OPLS_NIPALS_DNR(W = W, y = y, ...)
#'   list(
#'     scores      = as.vector(res$t_pred),
#'     P           = as.vector(res$p),
#'     W           = as.vector(res$w_pred),
#'     Q           = as.vector(res$q),
#'     U           = as.vector(res$u),
#'     orthoscores = matrix(res$t_ort, ncol = 1)
#'   )
#' }
#'
#' groups <- c(rep("A", 5), rep("B", 5))
#' resultsOPLS <- ComDim_y(mb,
#'   y = groups, ndim = 1,
#'   nort = 1,
#'   type = "discriminant",
#'   FUN = fun.OPLS,
#'   method = "OPLS-DA",
#'   cv.k = 0
#' )
#'
#' ## Example 3 (not run): ComDim-OPLS-DA via ropls ---------------------------
#' # Wrapping ropls::opls is also possible. Key points:
#' #   - Use orthoI = 1 (fixed) instead of NA so the output is predictable.
#' #   - Always return output$orthoscores; ComDim_y ignores it in phases
#' #     where ort has already been removed.
#' #   - Expand the single ropls Q loading to match the ncol(y_dummy) width.
#' \donttest{
#' if (requireNamespace("ropls", quietly = TRUE)) {
#' fun.OPLSDA.ropls <- function(W, y, ndim, ...) {
#'   output <- list()
#'   # Convert dummy matrix to ropls-compatible -1/+1 vector
#'   Y <- c(-1, 1)[apply(y, 1, function(x) match(1, x))]
#'   result <- tryCatch(
#'     ropls::opls(
#'       x = W, y = Y, predI = 1, orthoI = 1,
#'       fig.pdfC = "none", info.txtC = "none"
#'     ),
#'     error = function(e) {
#'       ropls::opls(
#'         x = W, y = Y, predI = 1, orthoI = 0,
#'         fig.pdfC = "none", info.txtC = "none"
#'       )
#'     }
#'   )
#'   output$scores <- result@scoreMN[, 1]
#'   output$P <- result@loadingMN[, 1]
#'   output$W <- result@weightMN[, 1]
#'   output$U <- result@uMN[, 1]
#'   # Expand the single ropls Q loading to match the 2-column dummy matrix:
#'   # loadings for class1 and class2 are antisymmetric in binary PLS-DA.
#'   output$Q <- c(-result@cMN[, 1], result@cMN[, 1])
#'   output$y <- result@suppLs$yModelMN # internal y (for scaling detection)
#'   # Orthogonal scores (used during the ort pre-loop when nort > 0)
#'   if (!is.null(result@orthoScoreMN) && ncol(result@orthoScoreMN) > 0) {
#'     output$orthoscores <- result@orthoScoreMN   # n x k matrix; col jj used for jj-th ort
#'   } else {
#'     output$orthoscores <- matrix(0, nrow = nrow(W), ncol = 1)
#'   }
#'   return(output)
#' }
#'
#' b1_r <- matrix(rnorm(8 * 30), 8, 30)
#' b2_r <- matrix(rnorm(8 * 20), 8, 20)
#' mb_r <- MultiBlock(Data = list(b1 = b1_r, b2 = b2_r))
#' resultsOPLSDA <- ComDim_y(mb_r,
#'   y = c(rep("NI", 4), rep("OFF", 4)),
#'   ndim = 1,
#'   nort = 1,
#'   type = "discriminant",
#'   FUN = fun.OPLSDA.ropls,
#'   method = "OPLS-DA(ropls)",
#'   cv.k = 0
#' )
#' }
#' }
#' @export

ComDim_y <- function(MB = MB, y = y,
                     ndim = NULL, FUN = FUN,
                     nort = 0L,
                     type = c("regression", "discriminant")[1],
                     decisionRule = c("fixed", "max")[2],
                     normalise = FALSE, scale.y = FALSE,
                     threshold = 1e-10,
                     loquace = FALSE,
                     method = "FUN",
                     cv.k = 7,
                     ...) {
  # INITIALISATION

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")

  start_ini_time <- Sys.time() # To start counting calculation time for the initialization


  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is  not a MultiBlock.")
  }

  if (!(type %in% c("regression", "discriminant"))) {
    stop("'type' must be either 'regression' or 'discriminant'.")
  }

  if (type == "discriminant" && !(decisionRule %in% c("fixed", "max"))) {
    stop("'decisionRule' must be either 'fixed' or 'max'.")
  }

  nort <- as.integer(nort)
  if (nort < 0L) {
    stop("'nort' must be a non-negative integer.")
  }
  if (nort > 1L) {
    stop("'nort > 1' is not supported in ComDim_y(). For multiple orthogonal components, use ComDim_OPLS().")
  }

  ntable <- length(blockNames(MB)) # number of blocks
  nrowMB <- length(sampleNames(MB)) # number of samples
  variable_number <- ncol(MB) # Number of variables per block

  give_error <- 0

  # Remove all variables with 0 variance. If there are no remaining variables, give_error
  numvar <- 0
  numblock0 <- vector()
  for (i in 1:ntable) {
    var0 <- apply(MB@Data[[i]], 2, function(x) {
      var(x)
    })
    var0 <- which(var0 == 0)

    if (length(var0) != 0) {
      numvar <- numvar + length(var0)
      if (length(var0) == length(MB@Variables[[i]])) {
        numblock0 <- append(numblock0, i)
      }
      MB@Data[[i]] <- MB@Data[[i]][, setdiff(1:variable_number[i], var0)]
      MB@Variables[[i]] <- MB@Variables[[i]][setdiff(1:variable_number[i], var0)]
      variable_number[i] <- length(MB@Variables[[i]])
    }
  }
  if (numvar != 0) {
    warning(sprintf("Number of variables excluded from the analysis because of their zero variance: %d", numvar))
  }
  if (length(numblock0) != 0) {
    numblock0 <- rev(numblock0) # To sort in decreasing order.
    for (i in seq_along(numblock0)) {
      MB@Data[[numblock0[i]]] <- NULL
      MB@Variables[[numblock0[i]]] <- NULL
      if (numblock0[[i]] %in% MB@Batch) {
        MB@Batch[[numblock0[i]]] <- NULL
      }
      if (numblock0[[i]] %in% MB@Metadata) {
        MB@Metadata[[numblock0[i]]] <- NULL
      }
    }
    warning(sprintf("Number of blocks excluded from the analysis because of their zero variance: %d", length(numblock0)))
  }

  ntable <- length(blockNames(MB)) # number of blocks
  variable_number <- ncol(MB) # In case the number of blocks has changed.

  if (length(MB@Data) == 0) { # In case all blocks had 0 variance...
    warning("All blocks had 0 variance")
    give_error <- 1
  }

  # Check for infinite or missing values (they cannot be handled with svd)
  for (i in 1:ntable) {
    if (any(is.na(MB@Data[[i]]))) {
      stop("The MB contains NA values. They must be removed first for ComDim analysis.")
    }
    if (any(is.infinite(MB@Data[[i]]))) {
      stop("The MB contains infinite values. They must be removed first for ComDim analysis.")
    }
  }


  ## Check y matrix (convert to dummy matrix if it is not already.)
  if (type == "discriminant") {
    tmp <- unique(as.vector(y))
    if (all(tmp %in% c(0, 1))) { # Is it dummy?

      if (!is.matrix(y)) {
        y <- as.matrix(y)
      }

      classVect <- rep(NA, nrow(y))

      if (ncol(y) == 1) {
        classVect <- as.vector(y)
      } else {
        if (any(rowSums(y) != 1)) {
          if (any(rowSums(y) == 0)) {
            stop("At least one sample was not assigned to a class.")
          } else if (any(colSums(y) > 1)) {
            stop("At least one sample was assigned to more than one class.")
          }
        }
      }
      if (is.null(colnames(y))) {
        colnames(y) <- paste0("Class", seq_len(ncol(y)))
      }
      for (i in seq_len(ncol(y))) {
        classVect[which(y[, i] == 1)] <- colnames(y)[i]
      }
    } else { # If not dummy
      if ((is.matrix(y) || is.data.frame(y)) && ncol(y) > 1) {
        stop("Predictive analysis can only be performed if 'y' is a dummy matrix or a class vector.")
      }
      classVect <- as.vector(y)
      classVect_sorted <- sort(unique(classVect))

      # Now generate the dummy matrix from y
      y <- matrix(rep(0, length(classVect_sorted) * length(classVect)),
        ncol = length(classVect_sorted), nrow = length(classVect)
      )
      for (i in seq_along(classVect_sorted)) {
        y[which(classVect == classVect_sorted[i]), i] <- 1
      }
      colnames(y) <- classVect_sorted
      rm(classVect_sorted)
    }
  }

  if (!is.matrix(y)) { # If it's a vector and the method is PLS-R
    y <- as.matrix(y)
  }

  # Store original y (after validation / dummy conversion) for use in R2Y, Q2,
  # and as the response saved in PLS.model$Y.
  y_raw <- y

  # Scale Y for regression if requested.
  # For discriminant analysis the dummy matrix must not be scaled; emit a
  # warning and skip.
  if (scale.y && type == "discriminant") {
    warning("'scale.y = TRUE' is ignored when type = 'discriminant': dummy Y matrices should not be centred or scaled.")
    scale.y <- FALSE
  }
  if (scale.y) {
    y_center <- colMeans(y)
    y_sd <- apply(y, 2, sd)
    y_sd[y_sd < .Machine$double.eps] <- 1 # guard against constant columns
    y <- sweep(sweep(y, 2, y_center, "-"), 2, y_sd, "/")
  } else {
    y_center <- colMeans(y)
    y_sd <- rep(1, ncol(y))
  }

  if (give_error) {
    stop("The data is not ready for ComDim.")
  } else {
    message("The data can be used for ComDim.")
  }


  if (is.null(ndim)) {
    ndim <- ntable # If the number of components is not defined,
    # the number of components to extract is equal to the number of blocks.
  }

  pieceBar <- 4 + 2 * ntable + ndim + nort # Number of updates in the progress bar.
  pieceBar <- 80 / pieceBar
  total_progress <- pieceBar

  DimLabels <- paste0("CC", 1:ndim) # One label per predictive component.
  OrtLabels <- if (nort > 0) paste0("ort", 1:nort) else character(0)
  TableLabels <- blockNames(MB) # One label per block.

  end_ini_time <- Sys.time() # To end the count of the calculation time.

  if (loquace) {
    message(sprintf("Initialisation finished after : %s millisecs", (end_ini_time - start_ini_time) * 1000))
  }

  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # NORMALISATION

  X_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))
  Xnorm_mat <- matrix(, nrow = nrowMB, ncol = sum(variable_number))

  res_calib <- list()
  temp_tabCalib <- list()
  s_r <- list()
  res_calib$SingVal <- as.vector(NULL)
  res_calib$NormMB <- as.vector(NULL)
  res_calib$MeanMB <- list()

  for (i in 1:ntable) {
    res_calib$MeanMB[[TableLabels[i]]] <- colMeans(MB@Data[[i]])
    names(res_calib$MeanMB[[TableLabels[i]]]) <- MB@Variables[[i]]

    if (normalise) {
      # Normalise original blocks

      X_mean <- MB@Data[[i]] - matrix(data = rep(1, nrowMB), ncol = 1, nrow = nrowMB) %*% res_calib$MeanMB[[i]]
      XX <- X_mean * X_mean
      Norm_X <- sqrt(sum(XX))
      X_Normed <- X_mean / Norm_X
      res_calib$NormMB[i] <- Norm_X
      temp_tabCalib[[i]] <- X_Normed
      s_r[[i]] <- X_Normed
    } else {
      res_calib$NormMB[[TableLabels[i]]] <- rep(1, length(MB@Variables[[i]]))
      names(res_calib$NormMB[[TableLabels[i]]]) <- MB@Variables[[i]]
      temp_tabCalib[[i]] <- MB@Data[[i]]
      s_r[[i]] <- MB@Data[[i]]
    }
    if (i == 1) {
      X_mat[, 1:variable_number[1]] <- MB@Data[[i]]
      Xnorm_mat[, 1:variable_number[1]] <- temp_tabCalib[[i]]
    } else {
      beg <- sum(variable_number[1:(i - 1)]) + 1
      ending <- sum(variable_number[1:i])
      X_mat[, beg:ending] <- MB@Data[[i]]
      Xnorm_mat[, beg:ending] <- temp_tabCalib[[i]]
    }

    # Compute means (used for B0 and Y-scaling check)
    meanY <- colMeans(y)

    norm_comdim <- Sys.time()
    if (loquace) {
      message(sprintf("Normalization of block %s finished after : %s millisecs", i, (norm_comdim - start_ini_time) * 1000))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  }

  names(res_calib$NormMB) <- TableLabels

  nR <- nrow(Xnorm_mat)
  nC <- ncol(Xnorm_mat)

  total_progress <- total_progress + pieceBar * ntable
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  # ICs decrements from ndim to 1 across outer loop (consistent with ComDim_PLS)
  ICs <- ndim

  ## PRE-LOOP: orthogonal component extraction
  ##
  ## When nort > 0, extract nort orthogonal components from the
  ## (lambda-weighted) concatenated blocks using FUN. Each block is deflated by
  ## its orthogonal consensus score before the predictive loop runs. This two-
  ## phase structure is identical to ComDim_OPLS. FUN must return
  ## output$orthoscores (an n x 1 matrix or vector) when called with ndim = 1.

  Q_ort <- NULL
  saliences_ort <- NULL
  varexp_ort <- NULL

  if (nort > 0) {
    Q_ort <- matrix(, ncol = nort, nrow = nrowMB)
    saliences_ort <- matrix(, ncol = nort, nrow = ntable)
    varexp_ort <- as.vector(NULL)

    for (jj in 1:nort) {
      lambda_ort <- matrix(rep(1, ntable), nrow = ntable, ncol = 1)
      qini_ort <- as.vector(s_r[[1]][, 1])
      qini_ort <- qini_ort / sqrt(as.vector(qini_ort %*% qini_ort))
      qold_ort <- 100

      iters <- 0
      while (norm(qold_ort - qini_ort, "2") > threshold && iters < 100) {
        iters <- iters + 1
        qold_ort <- qini_ort
        q_ort_sum <- 0

        W <- matrix(, nrow = nrowMB, ncol = 0)
        for (j in 1:ntable) {
          W <- cbind(W, sqrt(lambda_ort[j]) * s_r[[j]])
        }

        output_ort <- FUN(W = W, y = y, ndim = 1, ...)

        if (is.null(output_ort$orthoscores)) {
          stop(
            "'FUN' must return 'orthoscores' (a numeric matrix n x k, k >= 1, ",
            "or a length-n vector) when 'nort' > 0."
          )
        }

        ort_mat <- as.matrix(output_ort$orthoscores)
        if (ncol(ort_mat) < 1L) {
          stop("'FUN$orthoscores' must have at least 1 column when 'nort' > 0.")
        }
        ort_col <- 1L
        ort_vec <- ort_mat[, ort_col]
        sv_ort  <- sqrt(sum(ort_vec^2))
        qtmp_ort <- ort_vec / sv_ort

        for (j in 1:ntable) {
          lambda_ort[j] <- t(qtmp_ort) %*% (s_r[[j]] %*% t(s_r[[j]])) %*% qtmp_ort
          q_ort_sum <- q_ort_sum + lambda_ort[j] * qtmp_ort
        }

        q_ort_sum <- q_ort_sum / sqrt(as.vector(t(q_ort_sum) %*% q_ort_sum))
        if (abs(min(q_ort_sum)) > abs(max(q_ort_sum))) q_ort_sum <- -q_ort_sum
        qini_ort <- q_ort_sum
      } # convergence loop

      saliences_ort[, jj] <- lambda_ort
      Q_ort[, jj] <- qini_ort
      varexp_ort[jj] <- sv_ort^4

      # Deflate each block by the orthogonal consensus score (same as OPLS_MB)
      for (j in 1:ntable) {
        p_j_ort <- t(s_r[[j]]) %*% Q_ort[, jj] /
          as.vector(t(Q_ort[, jj]) %*% Q_ort[, jj])
        s_r[[j]] <- s_r[[j]] - Q_ort[, jj] %*% t(p_j_ort)
      }

      ort_comdim <- Sys.time()
      if (loquace) {
        message(sprintf(
          "Orthogonal component %s determined after : %s millisecs",
          jj, (ort_comdim - start_ini_time) * 1000
        ))
      }
      total_progress <- total_progress + pieceBar
      utils::setTxtProgressBar(progress_bar, value = total_progress)
    } # jj loop

    colnames(Q_ort) <- OrtLabels
    rownames(Q_ort) <- MB@Samples
    colnames(saliences_ort) <- OrtLabels
    rownames(saliences_ort) <- TableLabels
  } # nort > 0

  ## PREDICTIVE EXTRACTION
  ## Run ComDim-PLS on the (orthogonally-deflated) s_r blocks.

  saliences <- matrix(, ncol = ndim, nrow = ntable)
  Q <- matrix(, ncol = ndim, nrow = nrowMB)
  varexp <- as.vector(NULL)
  PLS2_Scores <- matrix(, ncol = ndim, nrow = nrowMB)
  PLS2_P <- matrix(, ncol = ndim, nrow = sum(variable_number))
  PLS2_W <- matrix(, ncol = ndim, nrow = sum(variable_number))
  PLS2_Q <- matrix(, ncol = ndim, nrow = ncol(as.matrix(y)))
  PLS2_U <- matrix(, ncol = ndim, nrow = nrowMB)

  for (dims in 1:ndim) {
    lambda <- matrix(rep(1, ntable), nrow = ntable, ncol = 1)

    qini <- as.vector(s_r[[1]][, 1])
    qini <- qini / sqrt(as.vector(qini %*% qini)) # random initialisation + unit length
    qold <- 100

    iters <- 0
    while (norm(qold - qini, "2") > threshold && iters < 100) {
      iters <- iters + 1
      qold <- qini
      q <- 0

      W <- matrix(, nrow = nrowMB, ncol = 0)
      for (j in 1:ntable) {
        W <- cbind(W, sqrt(lambda[j]) * s_r[[j]])
      }

      # HERE IS THE CORE COMDIM
      # FUN receives ICs (current remaining components), consistent with
      # ComDim_PLS approach.
      output <- FUN(W = W, y = y, ndim = ICs, ...)

      PLS2_Scores[, dims] <- output$scores
      PLS2_P[, dims] <- output$P
      PLS2_W[, dims] <- output$W
      PLS2_Q[, dims] <- output$Q
      PLS2_U[, dims] <- output$U

      sv <- sqrt(t(PLS2_Scores[, dims]) %*% PLS2_Scores[, dims]) # For R2X
      qtmp <- PLS2_Scores[, dims] / as.vector(sv)

      for (j in 1:ntable) {
        lambda[j] <- t(qtmp) %*% (s_r[[j]] %*% t(s_r[[j]])) %*% qtmp
        q <- q + lambda[j] * qtmp
      }

      q <- q / sqrt(as.vector(t(q) %*% q)) # standardise
      if (abs(min(q)) > abs(max(q))) q <- -q
      qini <- q
    } # convergence loop

    saliences[, dims] <- lambda
    Q[, dims] <- q

    res_calib$SingVal[dims] <- sv^2 # Calculated from the scores
    varexp[dims] <- res_calib$SingVal[dims]^2

    # Deflate blocks by predictive consensus score
    aux <- diag(nrowMB) - as.matrix(q) %*% t(as.matrix(q))
    for (j in 1:ntable) {
      s_r[[j]] <- aux %*% s_r[[j]]
    }

    # Decrement ICs for the next component (consistent with ComDim_PLS)
    ICs <- ICs - 1

    iter_comdim <- Sys.time()
    if (loquace) {
      message(sprintf(
        "Component %s determined after : %s millisecs",
        dims, (iter_comdim - start_ini_time) * 1000
      ))
    }
    total_progress <- total_progress + pieceBar
    utils::setTxtProgressBar(progress_bar, value = total_progress)
  } # dims loop

  # Strip dim names from the internally-transformed y (if FUN returned it)
  if (!is.null(output$y)) {
    rownames(output$y) <- NULL
    colnames(output$y) <- NULL
  }

  names(res_calib$SingVal) <- DimLabels

  colnames(Q) <- DimLabels
  rownames(Q) <- MB@Samples
  res_calib$Q <- Q
  rm(Q)

  ## Adding metadata. Metadata extracted from the first block
  res_calib$metadata <- list()
  if (length(MB@Metadata) != 0) {
    res_calib$metadata[[1]] <- MB@Metadata
  }


  end_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Scores finished after : %s millisecs", (end_comdim - start_ini_time) * 1000))
  }
  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  rm(s_r)

  ##
  explained <- varexp / sum(varexp)
  explained.ort <- NULL
  if (nort > 0 && !is.null(varexp_ort)) {
    total.varexp <- sum(varexp) + sum(varexp_ort)
    explained <- varexp / total.varexp
    explained.ort <- varexp_ort / total.varexp
    names(explained.ort) <- OrtLabels
  }
  names(explained) <- DimLabels
  res_calib$explained <- explained
  rm(varexp)
  rm(explained)

  ## Calculate Sums of saliences for each Dim
  Sum_saliences_Dim <- colSums(saliences)
  names(Sum_saliences_Dim) <- DimLabels

  res_calib$Sum_saliences_Dim <- Sum_saliences_Dim
  rm(Sum_saliences_Dim)

  # Calculate Sums of saliences for each Table
  Sum_saliences_Tab <- rowSums(saliences)
  names(Sum_saliences_Tab) <- TableLabels

  res_calib$Sum_saliences_Tab <- Sum_saliences_Tab
  rm(Sum_saliences_Tab)

  res_calib$saliences <- saliences
  colnames(res_calib$saliences) <- DimLabels
  rownames(res_calib$saliences) <- TableLabels

  ## Calculate Normalised concatenated Xs ('Calib') from col
  # Calculate concatenated CD loadings
  # Reorganise Loadings - 1 matrix / LV

  L_CD_Vec <- NULL
  L_X_Vec <- NULL
  Calib <- NULL
  nCalib <- nrowMB

  # Calculate concatenated CD loadings / Local scores
  L_X <- list()
  T_Loc_ort <- T_Loc <- list()
  for (i in 1:ntable) { # Prepare lists
    T_Loc[[TableLabels[i]]] <- matrix(, ncol = ndim, nrow = nrowMB)
    if (nort > 0) {
      T_Loc_ort[[TableLabels[i]]] <- matrix(, ncol = nort, nrow = nrowMB)
    }
  }
  b <- matrix(, ncol = ndim, nrow = ntable)

  if (!requireNamespace("pracma", quietly = TRUE)) {
    stop("The package pracma is needed.")
  }

  # Orthogonal loadings: deflate temp_tabCalib block by block, compute local
  # orthogonal scores and the concatenated orthogonal loading matrix (orthoP).
  # Rebuild Xnorm_mat from the orthogonally-deflated temp_tabCalib so that
  # the global loadings and B0 are computed in the correct subspace.
  orthoP <- NULL
  if (nort > 0) {
    for (j in 1:nort) {
      for (i in 1:ntable) {
        tempo <- t(temp_tabCalib[[i]]) %*% Q_ort[, j] # local ort loadings
        T_Loc_ort[[TableLabels[i]]][, j] <-
          temp_tabCalib[[i]] %*% (tempo %*% pracma::pinv(t(tempo) %*% tempo))
        # Deflate each temp_tabCalib by the orthogonal component
        temp_tabCalib[[i]] <- temp_tabCalib[[i]] - Q_ort[, j] %*% t(tempo)
        if (i == 1) {
          tempo2 <- tempo
        } else {
          tempo2 <- append(tempo2, tempo)
        }
      }
      if (j == 1) {
        orthoP <- tempo2
      } else {
        orthoP <- cbind(orthoP, tempo2)
      }
    }
    orthoP <- as.matrix(orthoP)
    colnames(orthoP) <- OrtLabels
    # Rebuild Xnorm_mat from orthogonally-deflated temp_tabCalib
    for (i in 1:ntable) {
      if (i == 1) {
        Xnorm_mat <- temp_tabCalib[[i]]
      } else {
        Xnorm_mat <- cbind(Xnorm_mat, temp_tabCalib[[i]])
      }
    }
  }

  # Predictive loadings: compute local scores and b-coefficients from the
  # (orthogonally-deflated) temp_tabCalib.
  for (j in 1:ndim) {
    T_mat <- matrix(, nrow = nrowMB, ncol = 0)

    for (i in 1:ntable) {
      temp <- t(temp_tabCalib[[i]]) %*% res_calib$Q[, j] # local loadings

      T_mat <- cbind(
        T_mat,
        temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp))
      )

      T_Loc[[TableLabels[i]]][, j] <-
        temp_tabCalib[[i]] %*% (temp %*% pracma::pinv(t(temp) %*% temp))

      # Deflate each temp_tabCalib
      temp_tabCalib[[i]] <- temp_tabCalib[[i]] - res_calib$Q[, j] %*% t(temp)
    }

    # MLR b-coefficients between Local and Global Scores
    b[, j] <- pracma::pinv(t(T_mat) %*% T_mat) %*% t(T_mat) %*% res_calib$Q[, j]
  }

  for (i in 1:ntable) {
    rownames(T_Loc[[TableLabels[i]]]) <- rownames(res_calib$Q)
    colnames(T_Loc[[TableLabels[i]]]) <- colnames(res_calib$Q)
    if (nort > 0) {
      rownames(T_Loc_ort[[TableLabels[i]]]) <- MB@Samples
      colnames(T_Loc_ort[[TableLabels[i]]]) <- OrtLabels
    }
  }

  # Calculate Global Loadings
  L_CD_Vec <- t(Xnorm_mat) %*% res_calib$Q # Scaled CD 'global' Loadings

  load_comdim <- Sys.time()
  if (loquace) {
    message(sprintf("Loadings finished after : %s millisecs", (load_comdim - start_ini_time) * 1000))
  }

  total_progress <- total_progress + pieceBar
  utils::setTxtProgressBar(progress_bar, value = total_progress)

  rm(X_mat)
  rm(temp_tabCalib)
  rm(temp)

  ## Output
  end_function <- Sys.time()

  ## Regression coefficients B and intercept B0
  ##
  ## Two back-transformation paths:
  ## (a) scale.y = TRUE: the framework scaled y before passing it to FUN; B and
  ##     B0 are in scaled-y space and must be back-transformed to original units.
  ## (b) scale.y = FALSE but FUN internally scaled y (e.g. ropls): detected by
  ##     comparing output$y to standardised y and back-transforming Q loadings.
  ##     Falls back to un-transformed Q when output$y is NULL or comparison fails.
  if (scale.y) {
    ComDimPLS2_B <- PLS2_W %*% pracma::pinv(t(PLS2_P) %*% PLS2_W) %*% t(PLS2_Q)
  } else {
    y_was_scaled <- FALSE
    if (!is.null(output$y) && ncol(y) == 1) {
      y_was_scaled <- isTRUE(all.equal(
        as.vector(output$y[, 1]),
        as.vector((y[, 1] - mean(y[, 1])) / sd(y[, 1]))
      ))
    }
    if (y_was_scaled) {
      ComDimPLS2_B <- PLS2_W %*% pracma::pinv(t(PLS2_P) %*% PLS2_W) %*% t(PLS2_Q * sd(y) + mean(y))
    } else {
      ComDimPLS2_B <- PLS2_W %*% pracma::pinv(t(PLS2_P) %*% PLS2_W) %*% t(PLS2_Q)
    }
  }
  # B0: use colMeans(Xnorm_mat) here. When nort > 0, Xnorm_mat has been
  # overwritten with the orthogonally-deflated version (lines above), so
  # colMeans(Xnorm_mat) gives the ort-deflated means — consistent with Ypred
  # which is also computed from the deflated Xnorm_mat. When nort == 0,
  # Xnorm_mat is unchanged, so the result equals the original column means.
  ComDimPLS2_B0 <- colMeans(y) - colMeans(Xnorm_mat) %*% ComDimPLS2_B

  # Back-transform B and B0 to the original Y scale when scale.y = TRUE, then
  # restore y to original units for all downstream computations.
  if (scale.y) {
    ComDimPLS2_B <- sweep(ComDimPLS2_B, 2, y_sd, "*")
    ComDimPLS2_B0 <- ComDimPLS2_B0 * y_sd + y_center
    y <- y_raw # restore original y for R2Y, Q2, Ypred, etc.
  }

  ## R2Y — cumulative OLS of Y on [1 | Q_1..j] for j = 1..ndim.
  ## r2y[j] reports the fraction of Y variance explained by the first j global
  ## ComDim scores. An explicit intercept is included so that Y with non-zero
  ## mean is handled correctly.
  Y_c_r2   <- sweep(y, 2, colMeans(y))
  sstot_Y_r2 <- sum(Y_c_r2^2)
  if (sstot_Y_r2 < .Machine$double.eps) {
    r2y <- setNames(rep(0, ndim), paste0("CC", seq_len(ndim)))
  } else {
    r2y <- setNames(numeric(ndim), paste0("CC", seq_len(ndim)))
    for (.j in seq_len(ndim)) {
      T_r2_j     <- cbind(1, res_calib$Q[, seq_len(.j), drop = FALSE])
      B_r2_j     <- pracma::pinv(crossprod(T_r2_j)) %*% crossprod(T_r2_j, y)
      Ypred_r2_j <- T_r2_j %*% B_r2_j
      r2y[.j]    <- 1 - sum((Ypred_r2_j - y)^2) / sstot_Y_r2
    }
  }
  rm(Y_c_r2)

  res_calib$b <- b
  colnames(res_calib$b) <- DimLabels
  rownames(res_calib$b) <- TableLabels

  res_calib$T_Loc <- T_Loc

  # T_Loc is no longer needed
  rm(T_Loc)

  varnames <- vector()
  for (i in 1:ntable) {
    varnames <- append(varnames, MB@Variables[[i]])
  }

  res_calib$P <- L_CD_Vec
  colnames(res_calib$P) <- DimLabels
  rownames(res_calib$P) <- varnames
  rownames(ComDimPLS2_B) <- varnames
  if (!is.null(colnames(y_raw))) colnames(ComDimPLS2_B) <- colnames(y_raw)

  rm(L_CD_Vec)

  ## -------------------------------------------------------------------------
  ## VIP (Variable Importance in Projection)
  ##
  ## Predictive VIP ("p") — Wold formula using PLS2_W, PLS2_Scores, PLS2_Q:
  ##   W_j  = L2-normalised column of the weight matrix (per block)
  ##   s_a  = colSums(T^2) * colSums(Q^2)   [length ndim]
  ##   VIPp_j = sqrt( p_b * sum_a(W_ja^2 * s_a) / sum_a(s_a) )
  ##
  ## Orthogonal VIP ("o", only when nort > 0) — loadings-based formula:
  ##   VIPo_j = sqrt( p_b * sum(ss_to * P_norm_j^2) / sum(ss_to) )
  ## (ort components have no W/Q analogue, so their weight is measured via
  ##  the magnitude of the ort loadings in X-space.)
  ##
  ## Per-block output: data.frame with "p", "o" (if nort>0), "tot" columns.
  ## Global VIP (@VIP): "tot" concatenated across blocks.
  ## -------------------------------------------------------------------------

  s_vip     <- colSums(PLS2_Scores^2) * colSums(PLS2_Q^2)
  sum_s_vip <- sum(s_vip)

  # Loadings-based orthogonal VIP helper (unchanged, nort>0 only)
  .vip_ort_for_loadings_y <- function(ort_scores_mat, ort_loadings_b) {
    ncomp  <- ncol(ort_scores_mat)
    p_b    <- nrow(ort_loadings_b)
    ss_T   <- colSums(ort_scores_mat^2)
    if (any(ss_T < .Machine$double.eps)) return(rep(NA_real_, p_b))
    P_norm <- apply(ort_loadings_b, 2, function(pp) {
      nrm <- sqrt(sum(pp^2)); if (nrm > 1e-14) pp / nrm else pp
    })
    if (is.null(dim(P_norm))) P_norm <- matrix(P_norm, ncol = 1)
    sum_s <- sum(ss_T)
    if (sum_s < .Machine$double.eps) return(rep(NA_real_, p_b))
    as.vector(sqrt(p_b * (P_norm^2 %*% ss_T) / sum_s))
  }

  has_ort <- !is.null(Q_ort) && !is.null(orthoP)

  if (sum_s_vip < .Machine$double.eps) {
    warning("Sum of explained SS is zero; VIP scores cannot be computed.")
    VIP_global <- setNames(rep(NA_real_, length(varnames)), varnames)
    VIP_blocks <- setNames(
      lapply(seq_len(ntable), function(i) {
        v <- rep(NA_real_, variable_number[i])
        if (has_ort) data.frame(p=v, o=v, tot=v, row.names=MB@Variables[[i]])
        else         data.frame(p=v,     tot=v, row.names=MB@Variables[[i]])
      }),
      TableLabels
    )
  } else {
    # Global VIP: L2-normalise W columns across all variables
    W_norm_all <- apply(PLS2_W, 2, function(w) {
      nrm <- sqrt(sum(w^2)); if (nrm > 1e-14) w / nrm else w
    })
    p_total    <- nrow(PLS2_W)
    VIP_global <- setNames(
      as.vector(sqrt(p_total * (W_norm_all^2 %*% s_vip) / sum_s_vip)),
      varnames
    )

    # Per-block VIP
    VIP_blocks <- setNames(vector("list", ntable), TableLabels)
    k_vip     <- 1L
    k_ort_vip <- 1L
    for (i in seq_len(ntable)) {
      p_b <- variable_number[i]
      idx <- k_vip:(k_vip + p_b - 1L)

      # Predictive VIP (Wold): L2-normalise W using block rows only
      W_b_norm <- apply(PLS2_W[idx, , drop = FALSE], 2, function(w) {
        nrm <- sqrt(sum(w^2)); if (nrm > 1e-14) w / nrm else w
      })
      vip_p <- as.vector(sqrt(p_b * (W_b_norm^2 %*% s_vip) / sum_s_vip))

      if (has_ort) {
        ort_P_b <- orthoP[k_ort_vip:(k_ort_vip + p_b - 1L), , drop = FALSE]
        vip_o   <- .vip_ort_for_loadings_y(Q_ort, ort_P_b)
        vip_t   <- sqrt((vip_p^2 + vip_o^2) / 2)
        VIP_blocks[[TableLabels[i]]] <- data.frame(
          p = vip_p, o = vip_o, tot = vip_t, row.names = MB@Variables[[i]]
        )
        k_ort_vip <- k_ort_vip + p_b
      } else {
        VIP_blocks[[TableLabels[i]]] <- data.frame(
          p = vip_p, tot = vip_p, row.names = MB@Variables[[i]]
        )
      }
      k_vip <- k_vip + p_b
    }

    # Global VIP overwritten with "tot" to be consistent with per-block
    VIP_global <- setNames(
      as.vector(unlist(lapply(VIP_blocks, function(b) b$tot))),
      varnames
    )
  }
  ## -------------------------------------------------------------------------

  rm(saliences)

  # Define block membership of each variable in P.loadings
  belong_block <- rep(0, nrow(res_calib$P))
  k <- 1
  for (i in 1:ntable) {
    belong_block[k:(k + variable_number[i] - 1)] <- TableLabels[i]
    k <- k + variable_number[i]
  }

  ## Calculate Y predictions
  Ypred <- Xnorm_mat %*% ComDimPLS2_B

  for (i in seq_len(ncol(y))) {
    Ypred[, i] <- unlist(Ypred[, i]) + ComDimPLS2_B0[i]
  }
  rownames(Ypred) <- MB@Samples
  if (!is.null(colnames(y_raw))) colnames(Ypred) <- colnames(y_raw)

  ## Predict classes / compute Q2
  if (type == "discriminant") {
    predClass <- rep(NA, nrow(y))

    if (decisionRule == "max") {
      for (i in seq_len(nrow(Ypred))) {
        xx <- which(Ypred[i, ] == max(Ypred[i, ], na.rm = TRUE))
        if (length(xx) == 1) {
          predClass[i] <- colnames(y)[xx] # Y stores the Class names
        } else {
          predClass[i] <- NaN # There is a tie.
          warning(sprintf("Class for sample %s could not be predicted.", i))
        }
      }
    } else if (decisionRule == "fixed") {
      for (i in seq_len(nrow(Ypred))) {
        xx <- which(Ypred[, i] > 1 / ncol(y))
        if (length(xx) == 1) {
          predClass[i] <- colnames(y)[xx] # Y stores the Class names
        } else {
          predClass[i] <- NaN # There is a tie.
          warning(sprintf("Class for sample %s could not be predicted.", i))
        }
      }
    }

    if (is.matrix(y)) {
      meanY <- colMeans(y)
    } else {
      meanY <- mean(y)
    }

    classy <- colnames(y)
    Q2 <- DQ2 <- specvec <- sensvec <- rep(NA, length(classy))
    confusionMatrix <- list()

    for (i in seq_along(classy)) {
      pos <- which(classVect == classy[i])
      neg <- which(classVect != classy[i])
      tp <- length(which(classVect == classy[i] & predClass == classy[i]))
      tn <- length(which(classVect != classy[i] & predClass != classy[i]))
      fp <- length(which(classVect != classy[i] & predClass == classy[i]))
      fn <- length(which(classVect == classy[i] & predClass != classy[i]))

      if (length(tp) == 0) {
        tp <- 0
      }
      if (length(tn) == 0) {
        tn <- 0
      }
      if (length(fp) == 0) {
        fp <- 0
      }
      if (length(fn) == 0) {
        fn <- 0
      }

      sensvec[i] <- tp / length(pos)
      specvec[i] <- tn / length(neg)

      cm <- matrix(c(tp, fp, fn, tn), ncol = 2, nrow = 2)
      rownames(cm) <- c("predClass1", "predClass0")
      colnames(cm) <- c("trueClass1", "trueClass0")
      confusionMatrix[[classy[i]]] <- cm

      ## DQ2
      E0 <- Ypred[neg, i] - y[neg, i] # Residuals of Class0 samples
      E1 <- Ypred[pos, i] - y[pos, i] # Residuals of Class1 samples

      E0count <- which(E0 > 0) # Class0 predictions above 0
      E1count <- which(E1 < 0) # Class1 predictions below 1

      SSE0_all <- t(E0) %*% E0
      SSE1_all <- t(E1) %*% E1

      SSE0 <- t(E0[E0count]) %*% E0[E0count]
      SSE1 <- t(E1[E1count]) %*% E1[E1count]

      PRESSD <- SSE0 + SSE1
      PRESS <- SSE0_all + SSE1_all

      Ym <- y[, i] - mean(y[, i])
      TSS <- t(Ym) %*% Ym

      DQ2[i] <- 1 - (PRESSD / TSS)
      Q2[i] <- 1 - PRESS / TSS
    }
    names(Q2)      <- classy
    names(DQ2)     <- classy
    names(sensvec) <- classy
    names(specvec) <- classy
  } else if (type == "regression") {
    PRESS_r <- colSums((Ypred - y)^2)
    TSS_r <- colSums(sweep(y, 2, colMeans(y))^2)
    Q2 <- setNames(1 - PRESS_r / TSS_r, colnames(y))
  }

  ## -------------------------------------------------------------------------
  ## k-fold cross-validation: Q2 and DQ2
  ## Skipped when nort > 0 (CV with user-defined orthogonal FUNs is not
  ## generically supported) or when cv.k < 2.
  ## -------------------------------------------------------------------------
  if (cv.k >= 2 && nort == 0) {
    # Sequential fold assignment
    fold_ids <- integer(nrowMB)
    for (.f in seq_len(cv.k)) fold_ids[seq.int(.f, nrowMB, by = cv.k)] <- .f

    Ypred_cv <- matrix(NA_real_, nrow = nrowMB, ncol = ncol(y))
    colnames(Ypred_cv) <- colnames(y)

    for (fold in seq_len(cv.k)) {
      test_idx  <- seq.int(fold, nrowMB, by = cv.k)
      train_idx <- setdiff(seq_len(nrowMB), test_idx)

      # Subset MB to training samples
      MB_cv <- MB
      MB_cv@Data <- lapply(MB@Data, function(x) x[train_idx, , drop = FALSE])
      MB_cv@Samples <- MB@Samples[train_idx]

      # Fit inner model on training fold (cv.k = 0 prevents nested CV)
      invisible(utils::capture.output(
        res_cv_fold <- ComDim_y(
          MB = MB_cv,
          y = y_raw[train_idx, , drop = FALSE],
          ndim = ndim,
          FUN = FUN,
          nort = 0L,
          type = type,
          decisionRule = decisionRule,
          normalise = normalise,
          scale.y = scale.y,
          threshold = threshold,
          loquace = FALSE,
          method = method,
          cv.k = 0,
          ...
        )
      ))

      # Build raw test X (concatenated blocks, same order as Xnorm_mat)
      X_test <- matrix(0, nrow = length(test_idx), ncol = sum(variable_number))
      k_idx <- 1
      for (i in seq_len(ntable)) {
        vi <- variable_number[i]
        X_test[, k_idx:(k_idx + vi - 1)] <- MB@Data[[i]][test_idx, , drop = FALSE]
        k_idx <- k_idx + vi
      }

      # Apply training normalisation to test set (only when normalise = TRUE)
      if (normalise) {
        k_idx <- 1
        for (i in seq_len(ntable)) {
          vi <- variable_number[i]
          mn <- res_cv_fold@Mean$MeanMB[[i]] # training col means
          nrm <- res_cv_fold@Norm$NormMB[i] # training Frobenius norm
          X_test[, k_idx:(k_idx + vi - 1)] <-
            (X_test[, k_idx:(k_idx + vi - 1)] -
              matrix(mn, nrow = length(test_idx), ncol = vi, byrow = TRUE)) / nrm
          k_idx <- k_idx + vi
        }
      }

      # Predict Y for the test fold using the training model
      B_cv <- res_cv_fold@PLS.model$B
      B0_cv <- res_cv_fold@PLS.model$B0
      Ypred_cv[test_idx, ] <-
        X_test %*% B_cv +
        matrix(B0_cv, nrow = length(test_idx), ncol = ncol(y), byrow = TRUE)
    }

    # Q2 per response column (or class)
    TSS_cv <- colSums(sweep(y, 2, colMeans(y))^2)
    PRESS_cv <- colSums((Ypred_cv - y)^2)
    Q2_cv <- setNames(1 - PRESS_cv / TSS_cv, colnames(y))

    # DQ2 per class then averaged across classes (discriminant only).
    # Per-class values are stored in cv$DQ2.perclass; @DQ2 holds the mean,
    # consistent with ComDim_OPLS / ConsensusOPLS convention.
    if (type == "discriminant") {
      DQ2_per_class <- setNames(rep(NA_real_, ncol(y)), colnames(y))
      for (i in seq_len(ncol(y))) {
        pos <- which(classVect == colnames(y)[i])
        neg <- which(classVect != colnames(y)[i])
        E0 <- Ypred_cv[neg, i] - y[neg, i]
        E1 <- Ypred_cv[pos, i] - y[pos, i]
        PRESSD <- sum(E0[E0 > 0]^2) + sum(E1[E1 < 0]^2)
        DQ2_per_class[i] <- 1 - PRESSD / TSS_cv[i]
      }
      DQ2_cv <- mean(DQ2_per_class)   # single averaged value (matches ComDim_OPLS)
    } else {
      DQ2_per_class <- vector()
      DQ2_cv <- vector()
    }

    # Override training R2 with CV-based Q2/DQ2
    Q2 <- Q2_cv
    if (type == "discriminant") DQ2 <- DQ2_cv

    cv_result <- list(
      k            = cv.k,
      fold         = fold_ids,
      Ypred        = Ypred_cv,
      Q2           = Q2_cv,
      DQ2          = if (type == "discriminant") DQ2_cv        else NULL,
      DQ2.perclass = if (type == "discriminant") DQ2_per_class else NULL
    )
  } else {
    if (cv.k != 0 && nort == 0) {
      warning("'cv.k' must be >= 2 or 0 (skip CV). No cross-validation performed.")
    }
    cv_result <- list()
  }
  ## -------------------------------------------------------------------------

  end_output <- Sys.time()
  running_time <- (end_output - start_ini_time)
  if (loquace) {
    message(sprintf("Analysis finished after : %s seconds", running_time))
  }
  res_calib$runtime <- running_time # Total time of analysis

  progress_bar <- utils::txtProgressBar(min = 0, max = 80, style = 3, char = "=")
  utils::setTxtProgressBar(progress_bar, value = 80)

  close(progress_bar)

  # Save data in a ComDim structure
  res_calib <- new("ComDim",
    Method = method,
    ndim = ndim,
    Q.scores = res_calib$Q,
    T.scores = res_calib$T_Loc,
    P.loadings = res_calib$P,
    Saliences = res_calib$saliences,
    Orthogonal = if (nort > 0) {
      list(
        nort = nort,
        Q.scores = Q_ort,
        T.scores = T_Loc_ort,
        P.loadings.ort = orthoP,
        Saliences.ort = saliences_ort,
        R2X = explained.ort
      )
    } else {
      list()
    },
    VIP = VIP_global,
    VIP.block = VIP_blocks,
    R2X = res_calib$explained,
    R2Y = r2y,
    Q2 = Q2,
    DQ2 = if (type == "discriminant") {
      DQ2
    } else {
      vector()
    },
    Singular = res_calib$SingVal,
    Mean = list(
      MeanMB = res_calib$MeanMB,
      MeanY = y_center, # original Y column means
      ScaleY = y_sd
    ), # original Y column SDs (1 if scale.y=FALSE)
    Norm = list(NormMB = res_calib$NormMB),
    PLS.model = list(
      W = PLS2_W,
      B = ComDimPLS2_B, # in original Y units
      B0 = ComDimPLS2_B0, # in original Y units
      Y = y_raw
    ), # original (unscaled) Y
    cv = cv_result,
    Prediction = if (type == "discriminant") {
      list(
        Y.pred = Ypred,
        decisionRule = decisionRule,
        trueClass = classVect,
        predClass = data.frame(predClass = predClass, row.names = MB@Samples),
        Sensitivity = sensvec,
        Specificity = specvec,
        confusionMatrix = confusionMatrix
      )
    } else {
      list(Y.pred = Ypred)
    },
    Metadata = res_calib$metadata,
    variable.block = belong_block,
    runtime = as.numeric(res_calib$runtime)
  )

  return(res_calib)
}

#' PredictMultiBlock
#'
#' Projects a new MultiBlock dataset into an existing ComDim model.
#' Works with models produced by ComDim_PCA, ComDim_PLS,
#' ComDim_OPLS, ComDim_y, and ComDim_Exploratory. The projection
#' type (PCA-like, PLS-like, OPLS-like) is determined from the model
#' structure rather than the method string, so custom method labels from
#' ComDim_y are handled automatically.
#'
#' @param MB A MultiBlock object containing the new samples to project.
#' @param y Response vector or dummy matrix (optional). When supplied for a
#'   supervised model, Q2, DQ2, and classification statistics are computed for
#'   the new samples.
#' @param model A ComDim object (the calibration model).
#' @param normalise If TRUE, each block is mean-centred using the training
#'   column means and divided by the training Frobenius norm. Must match the
#'   \code{normalise} setting used during calibration. Default FALSE.
#' @param loquace If TRUE, print a message for each set of model elements
#'   that were projected. Default TRUE.
#' @return The \code{model} ComDim object with updated slots:
#'   \itemize{
#'     \item \code{Q.scores} — projected global scores (new samples x ndim).
#'     \item \code{T.scores} — projected local scores (per block).
#'     \item \code{Orthogonal$Q.scores} — projected global ort scores (if
#'       model has orthogonal components).
#'     \item \code{Orthogonal$T.scores} — projected local ort scores.
#'     \item \code{Prediction$Y.pred} — predicted Y (supervised models).
#'     \item \code{Q2}, \code{DQ2}, classification slots — when \code{y}
#'       is supplied for a supervised model.
#'   }
#' @export
PredictMultiBlock <- function(MB = MB, y, model = model,
                              normalise = FALSE, loquace = TRUE) {
  # ---------------------------------------------------------------------------
  # Input validation
  # ---------------------------------------------------------------------------
  if (!inherits(MB, "MultiBlock")) {
    stop("'MB' is not a MultiBlock.")
  }

  if (!inherits(model, "ComDim")) {
    stop("'model' is not class ComDim.")
  }

  if (!(all(names(model@T.scores) %in% blockNames(MB)))) {
    stop("Not all the model blocks are included in MB.")
  }

  # ---------------------------------------------------------------------------
  # Block alignment: remove extra blocks; reorder/subset variables to match model
  # ---------------------------------------------------------------------------
  if (any(!(blockNames(MB) %in% names(model@T.scores)))) {
    dif_names <- setdiff(blockNames(MB), names(model@T.scores))
    for (i in rev(dif_names)) {
      MB@Data[[i]] <- NULL
      MB@Variables[[i]] <- NULL
    }
  }

  for (i in names(model@T.scores)) {
    model_vars_i <- rownames(model@P.loadings)[model@variable.block == i]
    if (all(model_vars_i %in% MB@Variables[[i]])) {
      idx <- match(model_vars_i, MB@Variables[[i]])
      MB@Data[[i]] <- MB@Data[[i]][, idx, drop = FALSE]
      MB@Variables[[i]] <- MB@Variables[[i]][idx]
    } else {
      stop(sprintf("Some variables required by the model are missing in block '%s'.", i))
    }
  }

  # ---------------------------------------------------------------------------
  # Determine model type from structure (works for any Method string)
  # ---------------------------------------------------------------------------
  is_supervised <- length(model@PLS.model) > 0 && !is.null(model@PLS.model$B)
  has_ort <- length(model@Orthogonal) > 0 &&
    !is.null(model@Orthogonal$P.loadings.ort)
  is_discriminant <- is_supervised && !is.null(model@Prediction$decisionRule)

  # ---------------------------------------------------------------------------
  # Process y (when supplied for a supervised model)
  # ---------------------------------------------------------------------------
  if (hasArg(y) && is_supervised) {
    if (is.vector(y) || is.matrix(y) || is.data.frame(y)) {
      n_new <- length(sampleNames(MB))
      if ((is.vector(y) && length(y) != n_new) ||
        (!is.vector(y) && nrow(y) != n_new)) {
        stop("'y' does not have the same number of rows as MB.")
      }
    }

    if (is_discriminant) {
      tmp <- unique(as.vector(y))
      if (all(tmp %in% c(0, 1))) { # already a dummy matrix
        if (!is.matrix(y)) y <- as.matrix(y)
        classVect <- rep(NA_character_, nrow(y))
        if (ncol(y) == 1) {
          classVect <- as.character(as.vector(y))
        } else {
          if (is.null(colnames(y))) {
            stop("Column names for 'y' must be provided when it is a dummy matrix.")
          }
          if (any(rowSums(y) != 1)) {
            if (any(rowSums(y) == 0)) stop("At least one sample was not assigned to a class.")
            if (any(colSums(y) > 1)) stop("At least one sample was assigned to more than one class.")
          }
          for (i in seq_len(ncol(y))) {
            classVect[which(y[, i] == 1)] <- colnames(y)[i]
          }
        }
      } else { # class vector → convert to dummy
        if ((is.matrix(y) || is.data.frame(y)) && ncol(y) > 1) {
          stop("For a discriminant model, 'y' must be a dummy matrix or a class vector.")
        }
        classVect <- as.character(as.vector(y))
        classy <- sort(unique(classVect))
        if (any(!(classy %in% colnames(model@PLS.model$Y)))) {
          stop("Not all classes in 'y' exist in the calibration model.")
        }
        y_dm <- matrix(0, nrow = length(classVect), ncol = length(classy))
        for (i in seq_along(classy)) {
          y_dm[classVect == classy[i], i] <- 1
        }
        colnames(y_dm) <- classy
        y <- y_dm
        rm(y_dm, classy)
      }
    } else { # regression
      if ((is.matrix(y) || is.data.frame(y)) && ncol(y) > 1) {
        stop("For a regression model, 'y' must be a vector or single-column matrix.")
      }
      if (is.vector(y)) y <- as.matrix(y)
    }
  }

  # ---------------------------------------------------------------------------
  # Build normalised Xnorm list and Xnorm_mat
  # ---------------------------------------------------------------------------
  names_table <- names(model@T.scores)
  nrowMB <- length(sampleNames(MB))
  p_total <- nrow(model@P.loadings)

  Xnorm <- list()
  Xnorm_mat <- matrix(0, nrow = nrowMB, ncol = p_total)

  for (i in names_table) {
    blk_idx <- which(model@variable.block == i) # column range in the global matrix

    if (normalise) {
      # Apply training mean-centring and training Frobenius norm
      trn_mean <- model@Mean$MeanMB[[i]]
      # NormMB can be a named list (normalise=FALSE training) or unnamed numeric
      # vector (normalise=TRUE training). Use block position as fallback index.
      blk_pos <- match(i, names_table)
      trn_norm <- tryCatch(
        as.numeric(model@Norm$NormMB[[i]]), # named access
        error = function(e) as.numeric(model@Norm$NormMB[blk_pos]) # positional
      )
      if (length(trn_norm) > 1) trn_norm <- trn_norm[1] # should be scalar
      X_mean <- MB@Data[[i]] -
        matrix(trn_mean, nrow = nrowMB, ncol = length(trn_mean), byrow = TRUE)
      Xnorm[[i]] <- X_mean / trn_norm
    } else {
      Xnorm[[i]] <- MB@Data[[i]]
    }
    Xnorm_mat[, blk_idx] <- Xnorm[[i]]
  }

  # ---------------------------------------------------------------------------
  # STEP 1 — Orthogonal deflation (only when model has ort components)
  # ---------------------------------------------------------------------------
  if (has_ort) {
    nort <- model@Orthogonal$nort
    ort_labs <- colnames(model@Orthogonal$P.loadings.ort)
    if (is.null(ort_labs)) ort_labs <- paste0("ort", seq_len(nort))

    Q_ort <- matrix(NA_real_,
      nrow = nrowMB, ncol = nort,
      dimnames = list(MB@Samples, ort_labs)
    )
    T_Loc_ort <- setNames(
      lapply(names_table, function(i) {
        matrix(NA_real_,
          nrow = nrowMB, ncol = nort,
          dimnames = list(MB@Samples, ort_labs)
        )
      }),
      names_table
    )

    for (jj in seq_len(nort)) {
      p_ort_global <- model@Orthogonal$P.loadings.ort[, jj] # p × 1
      norm2_ort <- sum(p_ort_global^2)

      # Global ort score: project full (current) Xnorm_mat onto ort loading
      Q_ort[, jj] <- as.vector(Xnorm_mat %*% p_ort_global) / norm2_ort

      # Local ort scores + block-level deflation
      for (i in names_table) {
        blk_idx <- which(model@variable.block == i)
        p_ort_loc <- p_ort_global[blk_idx] # p_i × 1
        norm2_loc <- sum(p_ort_loc^2)
        T_Loc_ort[[i]][, jj] <- as.vector(Xnorm[[i]] %*% p_ort_loc) / norm2_loc
        Xnorm[[i]] <- Xnorm[[i]] - Q_ort[, jj] %*% t(p_ort_loc)
      }

      # Rebuild Xnorm_mat from ort-deflated blocks
      Xnorm_mat <- do.call(
        cbind,
        lapply(names_table, function(i) Xnorm[[i]])
      )
    }

    model@Orthogonal$Q.scores <- Q_ort
    model@Orthogonal$T.scores <- T_Loc_ort
    if (loquace) {
      message("Orthogonal Q.scores and T.scores were projected from the ComDim model.")
    }
  }

  # Capture Xnorm_mat here for Y prediction:
  #   - PCA/PLS: original (not ort-deflated) normalised X
  #   - OPLS: ort-deflated X  (B was fitted on this subspace)
  Xnorm_mat_for_ypred <- Xnorm_mat

  # ---------------------------------------------------------------------------
  # STEP 2 — Predictive score projection (sequential deflation)
  # ---------------------------------------------------------------------------
  ndim <- model@ndim
  dim_labs <- colnames(model@Q.scores)
  if (is.null(dim_labs)) dim_labs <- paste0("CC", seq_len(ndim))

  Q <- matrix(NA_real_,
    nrow = nrowMB, ncol = ndim,
    dimnames = list(MB@Samples, dim_labs)
  )
  T_Loc <- setNames(
    lapply(names_table, function(i) {
      matrix(NA_real_,
        nrow = nrowMB, ncol = ndim,
        dimnames = list(MB@Samples, dim_labs)
      )
    }),
    names_table
  )

  for (j in seq_len(ndim)) {
    p_global <- model@P.loadings[, j] # p × 1
    norm2_g <- sum(p_global^2)

    # Global score: project current Xnorm_mat onto the j-th global loading
    Q[, j] <- as.vector(Xnorm_mat %*% p_global) / norm2_g

    # Local scores + block-level deflation
    for (i in names_table) {
      blk_idx <- which(model@variable.block == i)
      p_local <- p_global[blk_idx] # p_i × 1
      norm2_l <- sum(p_local^2)
      T_Loc[[i]][, j] <- as.vector(Xnorm[[i]] %*% p_local) / norm2_l
      Xnorm[[i]] <- Xnorm[[i]] - Q[, j] %*% t(p_local)
    }

    # Deflate Xnorm_mat so the next component operates on the residual
    Xnorm_mat <- Xnorm_mat - Q[, j] %*% t(p_global)
  }

  model@Q.scores <- Q
  model@T.scores <- T_Loc
  if (loquace) {
    message("Predictive Q.scores and T.scores were projected from the ComDim model.")
  }

  # ---------------------------------------------------------------------------
  # STEP 3 — Y prediction (supervised models only)
  # ---------------------------------------------------------------------------
  if (is_supervised) {
    Ypred <- Xnorm_mat_for_ypred %*% model@PLS.model$B
    for (i in seq_len(ncol(Ypred))) {
      Ypred[, i] <- Ypred[, i] + model@PLS.model$B0[i]
    }
    model@Prediction$Y.pred <- Ypred
    if (loquace) message("Y.pred was predicted from the ComDim model.")

    # -----------------------------------------------------------------------
    # STEP 4 — Classification stats (discriminant + y supplied)
    # -----------------------------------------------------------------------
    if (hasArg(y) && is_discriminant) {
      predClass <- rep(NA_character_, nrow(y))
      dr <- model@Prediction$decisionRule

      if (dr == "max") {
        for (i in seq_len(nrow(Ypred))) {
          xx <- which(Ypred[i, ] == max(Ypred[i, ], na.rm = TRUE))
          if (length(xx) == 1) {
            predClass[i] <- colnames(y)[xx]
          } else {
            predClass[i] <- NA_character_
            warning(sprintf("Class for sample %d could not be predicted (tie).", i))
          }
        }
      } else if (dr == "fixed") {
        for (i in seq_len(nrow(Ypred))) {
          xx <- which(Ypred[i, ] > 1 / ncol(y))
          if (length(xx) == 1) {
            predClass[i] <- colnames(y)[xx]
          } else {
            predClass[i] <- NA_character_
            warning(sprintf("Class for sample %d could not be predicted (tie/no class).", i))
          }
        }
      }

      classy <- colnames(y)
      Q2 <- setNames(rep(NA_real_, length(classy)), classy)
      DQ2 <- setNames(rep(NA_real_, length(classy)), classy)
      specvec <- setNames(rep(NA_real_, length(classy)), classy)
      sensvec <- setNames(rep(NA_real_, length(classy)), classy)
      confusionMatrix <- list()

      for (i in seq_along(classy)) {
        pos <- which(classVect == classy[i])
        neg <- which(classVect != classy[i])
        tp <- sum(classVect == classy[i] & predClass == classy[i], na.rm = TRUE)
        tn <- sum(classVect != classy[i] & predClass != classy[i], na.rm = TRUE)
        fp <- sum(classVect != classy[i] & predClass == classy[i], na.rm = TRUE)
        fn <- sum(classVect == classy[i] & predClass != classy[i], na.rm = TRUE)

        sensvec[i] <- if (length(pos) > 0) tp / length(pos) else NA_real_
        specvec[i] <- if (length(neg) > 0) tn / length(neg) else NA_real_

        cm <- matrix(c(tp, fp, fn, tn), ncol = 2, nrow = 2)
        rownames(cm) <- c("predClass1", "predClass0")
        colnames(cm) <- c("trueClass1", "trueClass0")
        confusionMatrix[[classy[i]]] <- cm

        E0 <- Ypred[neg, i] - y[neg, i]
        E1 <- Ypred[pos, i] - y[pos, i]

        SSE0_all <- sum(E0^2)
        SSE1_all <- sum(E1^2)
        SSE0 <- sum(E0[E0 > 0]^2)
        SSE1 <- sum(E1[E1 < 0]^2)

        PRESSD <- SSE0 + SSE1
        PRESS <- SSE0_all + SSE1_all
        Ym <- y[, i] - mean(y[, i])
        TSS <- sum(Ym^2)

        DQ2[i] <- 1 - PRESSD / TSS
        Q2[i] <- 1 - PRESS / TSS
      }

      model@Q2 <- Q2
      model@DQ2 <- DQ2
      model@Prediction$trueClass <- classVect
      model@Prediction$predClass <- data.frame(predClass = predClass, row.names = MB@Samples)
      model@Prediction$Sensitivity <- sensvec
      model@Prediction$Specificity <- specvec
      model@Prediction$confusionMatrix <- confusionMatrix
      if (loquace) {
        message(
          "Q2, DQ2, predClass, Sensitivity, Specificity, and confusionMatrix",
          "were computed for the new samples."
        )
      }
    } else if (hasArg(y) && !is_discriminant) {
      # -----------------------------------------------------------------------
      # STEP 5 — Q2 for regression (y supplied, not discriminant)
      # -----------------------------------------------------------------------
      PRESS_r <- colSums((Ypred - y)^2)
      TSS_r <- colSums(sweep(y, 2, colMeans(y))^2)
      model@Q2 <- setNames(1 - PRESS_r / TSS_r, colnames(as.matrix(y)))
      if (loquace) message("Q2 was computed for the new samples.")
    }
  }

  return(model)
}

# Pareto, mean-center, auto-scale, geometric mean, ranknorm

#' NormalizeMultiBlock
#'
#' Normalize all blocks from a MultiBlock object. The ranknorm transform is
#' based on that from the RNOmni package (doi:10.1111/biom.13214).
#' @param MB The MultiBlock object.
#' @param blocks Blocks to normalize. A vector of integers or block names
#'   (optional; all blocks are processed when omitted).
#' @param method Normalization method. One of: \code{'none'} (no
#'   transformation beyond optional Inf-to-NA conversion), \code{'auto'}
#'   (auto-scaling), \code{'mean'} (mean-centering), \code{'pareto'}
#'   (mean-centered and divided by the square root of the SD), \code{'norm'}
#'   (mean-centered and divided by its Frobenius norm), \code{'geometric'}
#'   (geometric mean scaling), \code{'ranknorm'} (rank-based inverse normal
#'   transform). Default: \code{'norm'}.
#' @param infinite.as.NA If \code{TRUE}, Infinite values are converted to NA
#'   before normalization. The presence of NA or Inf values will cause ComDim
#'   to stop; use \code{\link{NAInfRemoveMultiBlock}} to handle them.
#' @param constant For \code{'geometric'}: value added to each element before
#'   computing the geometric mean (useful when data contain zeros). Default: 0.
#' @param offset For \code{'ranknorm'}: offset in the Blom-type transform.
#'   Default: 3/8 (Blom transform).
#' @param showWarning If \code{TRUE}, warn when all-NA variables are
#'   discarded. Default: \code{TRUE}.
#' @return The normalized MultiBlock object.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' b2[c(2, 3, 5), c(1, 2, 3)] <- NA
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' mb <- NormalizeMultiBlock(mb, method = "auto")
#' @export
NormalizeMultiBlock <- function(MB,
                                blocks = NULL,
                                method = c(
                                  "none", "auto", "mean", "pareto",
                                  "norm", "geometric", "ranknorm"
                                )[5],
                                infinite.as.NA = FALSE,
                                constant = 0,
                                offset = 3 / 8,
                                showWarning = TRUE) {
  if (is.null(method) || length(method) > 1) {
    stop("A method should be provided. Try 'none', 'auto', 'mean', 'pareto', 'norm', 'geometric', or 'ranknorm'.")
  }

  if (!("MultiBlock" %in% class(MB))) {
    stop("'MB' is not a MultiBlock object.")
  }

  method <- tolower(method)
  if (!(method %in% c("none", "auto", "mean", "pareto", "norm", "geometric", "ranknorm"))) {
    stop("The provided method is not in the list of available methods. Try 'none', 'auto', 'mean', 'norm', 'pareto', 'geometric', or 'ranknorm'.")
  }

  if (!is.null(blocks)) {
    if (is.character(blocks)) {
      block.ids <- match(blocks, names(MB@Data))
      if (any(is.na(block.ids))) {
        warning(sprintf(
          "Some block names were not found in the MultiBlock: %s",
          paste(blocks[is.na(block.ids)], collapse = ", ")
        ))
        block.ids <- block.ids[!is.na(block.ids)]
      }
    } else if (is.numeric(blocks)) {
      block.ids <- intersect(seq_along(MB@Data), blocks)
      if (length(blocks) != length(block.ids)) {
        warning(sprintf(
          "Some block indices were not found in the MultiBlock: %s",
          paste(setdiff(blocks, seq_along(MB@Data)), collapse = ", ")
        ))
      }
    }
  } else {
    block.ids <- seq_along(MB@Data)
  }

  if (length(block.ids) == 0) {
    warning("0 blocks were selected. Data was not modified.")
    return(MB)
  }

  to_ignore <- 0
  to_delete <- integer(0)

  for (i in block.ids) {
    if (any(is.complex(MB@Data[[i]]))) {
      stop("The multi-block contains complex values that cannot be processed with this function.")
    }

    if (any(is.logical(MB@Data[[i]]))) {
      stop("The multi-block contains logical values that cannot be processed with this function.")
    }

    if (any(is.infinite(MB@Data[[i]]))) {
      if (infinite.as.NA) {
        MB@Data[[i]][is.infinite(MB@Data[[i]])] <- NA
      } else {
        warning(sprintf(
          "Block '%s' contains Infinite values. Convert them to NA with infinite.as.NA = TRUE or using NAInfRemoveMultiBlock().",
          names(MB@Data)[i]
        ))
      }
    }

    # Discard variables that are entirely NA
    numna <- apply(MB@Data[[i]], 2, function(x) sum(is.na(x)))
    colna <- which(numna >= nrow(MB@Data[[i]]))
    if (length(colna) > 0) {
      keep_idx <- setdiff(seq_len(ncol(MB@Data[[i]])), colna)
      MB@Data[[i]] <- MB@Data[[i]][, keep_idx, drop = FALSE]
      MB@Variables[[i]] <- MB@Variables[[i]][keep_idx]
      to_ignore <- to_ignore + length(colna)
    }
    if (ncol(MB@Data[[i]]) == 0) {
      to_delete <- c(to_delete, i)
    }
  }

  # Remove blocks that became empty
  if (length(to_delete) > 0) {
    to_delete <- sort(to_delete, decreasing = TRUE)
    names_to_delete <- names(MB@Data)[to_delete]
    old_names <- names(MB@Data)
    for (i in to_delete) {
      nm <- names(MB@Data)[i]
      if (nm %in% names(MB@Batch)) MB@Batch[[nm]] <- NULL
      if (nm %in% names(MB@Metadata)) MB@Metadata[[nm]] <- NULL
      MB@Data[[i]] <- NULL
      MB@Variables[[i]] <- NULL
    }
    kept_names <- setdiff(old_names[block.ids], names_to_delete)
    if (length(kept_names) == 0) {
      stop("The selected blocks contained only NA values. They cannot be used in ComDim.")
    }
    block.ids <- match(kept_names, names(MB@Data))
  }

  for (i in block.ids) {
    if (method == "none") {
      next # Can be used solely to convert Inf to NA via infinite.as.NA.
    } else if (method == "auto") {
      MB@Data[[i]] <- scale(MB@Data[[i]], center = TRUE, scale = TRUE)
    } else if (method == "mean") {
      MB@Data[[i]] <- scale(MB@Data[[i]], center = TRUE, scale = FALSE)
    } else if (method == "pareto") {
      MB@Data[[i]] <- scale(MB@Data[[i]],
        center = TRUE,
        scale = apply(
          MB@Data[[i]], 2,
          function(x) sqrt(sd(x, na.rm = TRUE))
        )
      )
    } else if (method == "norm") {
      Xmean <- scale(MB@Data[[i]], center = TRUE, scale = FALSE)
      Norm_X <- sqrt(sum(Xmean * Xmean, na.rm = TRUE))
      MB@Data[[i]] <- Xmean / Norm_X
    } else if (method == "geometric") {
      for (nc in seq_len(ncol(MB@Data[[i]]))) {
        notna <- which(!is.na(MB@Data[[i]][, nc]))
        if (length(notna) == 0) next
        vals <- MB@Data[[i]][notna, nc] + constant
        if (any(vals <= 0)) {
          stop(sprintf(
            "Block '%s', column %d: values must be positive for the geometric mean. Use the 'constant' argument to shift them.",
            names(MB@Data)[i], nc
          ))
        }
        MB@Data[[i]][notna, nc] <- vals / prod(vals)^(1 / length(vals))
      }
    } else if (method == "ranknorm") {
      for (nc in seq_len(ncol(MB@Data[[i]]))) {
        notna <- which(!is.na(MB@Data[[i]][, nc]))
        n <- length(notna)
        r <- rank(MB@Data[[i]][notna, nc])
        MB@Data[[i]][notna, nc] <- qnorm((r - offset) / (n - 2 * offset + 1))
      }
    }
  }

  if (to_ignore != 0 && showWarning) {
    warning(sprintf("%d variable(s) contained only NAs and were deleted.", to_ignore))
  }

  return(MB)
}

## All class definitions should go in here.

############################################################
## Class unions
setClassUnion("data.frameNULL", c("data.frame", "NULL"))
setClassUnion("matrixNULL", c("matrix", "NULL"))
setClassUnion("vectorNULL", c("vector", "NULL", "character"))
setClassUnion("vectorlist", c("vector", "list"))
setClassUnion("listNULL", c("list", "NULL"))

############################################################

## Validate MultiBlock class
#' checkMultiBlock
#'
#' Validation function used during MultiBlock building.
#' @param object A MultiBlock object.
#' @return TRUE if valid, or a character vector of error messages.
#' @keywords internal
checkMultiBlock <- function(object) {
  errors <- character()

  if (any(!(names(object@Batch)) %in% names(object@Data))) {
    msg <- "
    There is at least a Batch vector with no corresponding Data.
    Check Batch and Data names."
    errors <- c(msg, errors)
  }
  if (any(!(names(object@Metadata)) %in% names(object@Data))) {
    msg <- "
    There is at least a Metadata data.frame with no corresponding Data.
    Check Batch and Data names."
    errors <- c(msg, errors)
  }
  if (any(!(names(object@Variables)) %in% names(object@Data))) {
    msg <- "
    There is at least an element in the list of Variables with no
    corresponding Data. Check Batch and variable block names."
    errors <- c(msg, errors)
  }
  samples_is_list <- is.list(object@Samples)

  for (i in seq_along(object@Data)) {
    block_name <- names(object@Data)[i]

    # Check sample count — handle both vector Samples and per-block list Samples
    if (samples_is_list) {
      if (!(block_name %in% names(object@Samples))) {
        msg <- sprintf("
      Block '%s' has no corresponding entry in the Samples list.", block_name)
        errors <- c(msg, errors)
      } else if (nrow(object@Data[[i]]) != length(object@Samples[[block_name]])) {
        msg <- "
      The number of samples is not correct for at least one block."
        errors <- c(msg, errors)
      }
    } else {
      if (nrow(object@Data[[i]]) != length(object@Samples)) {
        msg <- "
      The number of samples is not correct for at least one block. Use
      Use MultiBlock() with 'ignore.size = TRUE' to build
      a MultiBlock with blocks of different sample number."
        errors <- c(msg, errors)
      }
    }

    if (ncol(object@Data[[i]]) != length(object@Variables[[i]])) {
      msg <- "
      The number of variables is not correct for at least one block."
      errors <- c(msg, errors)
    }

    if (names(object@Data)[i] %in% names(object@Batch)) {
      if (nrow(object@Data[[i]]) != length(object@Batch[[i]])) {
        msg <- "
        The length of Batch is not correct for at least one block."
        errors <- c(msg, errors)
      }
    }
    if (names(object@Data)[i] %in% names(object@Metadata)) {
      if (nrow(object@Data[[i]]) != nrow(object@Metadata[[i]])) {
        msg <- "
        The length of Metadata is not correct for at least one block."
        errors <- c(msg, errors)
      }
    }

    if (length(errors) != 0) {
      break
    }
  }

  if (length(errors) == 0) TRUE else errors
}


############################################################

## MultiBlock
#'
#' @title MultiBlock object
#'
#' @description Object of the type \code{MultiBlock}, to use as input for ComDim
#' analyses.
#'
#' @slot Samples vector with the sample names. If this data is not available,
#' the slot will be filled with integers.
#'
#' @slot Data A list with the data-blocks.
#'
#' @slot Variables A character vector with the variable names. If this data is
#' not available, the slot will be filled with integers.
#'
#' @slot Batch A list with the vectors with the batch information for each
#' data-block. Optional.
#'
#' @slot Metadata A list with the samples metadata.
#'
#' @name ComDim-classes
#' @importFrom methods new setClass setClassUnion setGeneric setMethod validObject is

setClass("MultiBlock",
  representation(
    Samples = "vectorlist",
    Data = "list",
    Variables = "list", # List of vectors
    Batch = "list",
    Metadata = "list"
  ),
  prototype(
    Samples = vector(),
    Data = list(),
    Variables = list(),
    Batch = list(),
    Metadata = list()
  ),
  validity = checkMultiBlock
)


############################################################
## ComDim
#' @aliases ComDim
#'
#' @title ComDim object
#'
#' @description The output of a ComDim analysis.
#'
#' @slot Method The algorithm used in the core of the ComDim analysis (ex. PCA,
#' PLS,...)
#'
#' @slot ndim The number of components.
#'
#' @slot Q.scores The Global scores.
#'
#' @slot T.scores The Local scores.
#'
#' @slot P.loadings The Loadings
#'
#' @slot Saliences The Saliences
#'
#' @slot R2X The explained variance of the MultiBlock.
#'
#' @slot R2Y For regression or discriminant models, the explained variance of
#' the Y-block.
#'
#' @slot Q2 For regression or discriminant models, the predicted variance of
#' the Y-block.
#'
#' @slot DQ2 For discriminant models, the predicted discriminant variance of the
#'  Y-block.
#'
#' @slot Singular The singular values.
#'
#' @slot Mean The mean values for each variable in the MultiBlock.
#'
#' @slot Norm The norm values for each variable in the MultiBlock.
#'
#' @slot PLS.model For ComDim analyses using PLS as the core algorithm, contains
#' the W, B, B0 and Y matrices.
#'
#' @slot cv For ComDim_KOPLS, it contains the index of the samples used during
#' the cross-validation.
#'
#' @slot Prediction A list with the predicted Y, the decision rule used for
#' sample classification, the sensitivity, the specificity, and a confusion
#' matrix.
#'
#' @slot Metadata A list with the per-sample metadata for each block.
#'
#' @slot variable.block A vector with the same length as the P.loadings,
#' indicating the block each variable belongs to.
#'
#' @slot runtime The used running time.
#'
#' @name ComDim-classes
#' @importFrom methods new

setClass(
  "ComDim",
  representation(
    Method = "character",
    ndim = "numeric",
    Q.scores = "matrix",
    T.scores = "list",
    P.loadings = "matrix", # List of P
    Saliences = "matrix",
    Orthogonal = "list",
    VIP = "vector",
    VIP.block = "list",
    R2X = "vector",
    R2Y = "vector",
    Q2 = "vector",
    DQ2 = "vector",
    Singular = "vector",
    Mean = "list",
    Norm = "list",
    PLS.model = "list", # W, U, B, B0, Y
    cv = "list", # cvTrainIndex & cvTestIndex
    Prediction = "list", #  Y.pred, trueClass, sensSpec, confusionMatrix, nclasses, decisionRule
    Metadata = "listNULL",
    variable.block = "vector",
    runtime = "numeric"
  ),
  prototype(
    Method = NULL,
    ndim = NULL,
    Q.scores = NULL,
    T.scores = list(),
    P.loadings = matrix(),
    Saliences = matrix(),
    Orthogonal = list(),
    VIP = vector(),
    VIP.block = list(),
    R2X = vector(),
    R2Y = vector(),
    Q2 = vector(),
    DQ2 = vector(),
    Singular = vector(),
    Mean = list(),
    Norm = list(),
    PLS.model = NULL, # W, U, B, B0, Y, Y.Pred,
    cv = NULL, # cvTrainIndex & cvTestIndex
    Prediction = NULL, #  Y.pred, decisionRule, trueClass, sensSpec, confusionMatrix, nclasses, decisionRule
    Metadata = NULL,
    variable.block = NULL,
    runtime = NULL
  )
) # It would be nice to add a checkComDim like checkMultiBlock

############################################################
## S4 Generics for MultiBlock accessors

#' @rdname blockNames-MultiBlock-method
#' @export
setGeneric("blockNames", function(x, ...) methods::standardGeneric("blockNames"))
#' @rdname blockNames-set-MultiBlock-method
#' @export
setGeneric("blockNames<-", function(x, value) methods::standardGeneric("blockNames<-"))

#' @rdname sampleNames-MultiBlock-method
#' @export
setGeneric("sampleNames", function(x) methods::standardGeneric("sampleNames"))
#' @rdname sampleNames-set-MultiBlock-method
#' @export
setGeneric("sampleNames<-", function(x, value) methods::standardGeneric("sampleNames<-"))

#' @rdname variableNames-MultiBlock-method
#' @export
setGeneric("variableNames", function(x, ...) methods::standardGeneric("variableNames"))
#' @rdname variableNames-set-MultiBlock-method
#' @export
setGeneric("variableNames<-", function(x, value) methods::standardGeneric("variableNames<-"))

############################################################
## S4 Methods for MultiBlock

# ---- blockNames ----

#' blockNames
#'
#' Return the block names of a MultiBlock object.
#' @param x A MultiBlock object.
#' @param ... Not used. Present for S4 generic dispatch compatibility.
#' @param slot A string: \code{"Data"} (default), \code{"Batch"}, or
#'   \code{"Metadata"}, indicating which slot to retrieve names from.
#' @return A character vector with the block names of the requested slot.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' blockNames(mb) # c("b1", "b2")
#' blockNames(mb, "Data") # same
#' @export
setMethod("blockNames", "MultiBlock", function(x, slot = "Data") {
  if (length(slot) > 1) slot <- "Data"
  switch(tolower(slot),
    data     = names(x@Data),
    batch    = names(x@Batch),
    metadata = names(x@Metadata),
    stop("'slot' must be 'Data', 'Batch', or 'Metadata'.")
  )
})

#' blockNames<-
#'
#' Set the block names of a MultiBlock object.
#' Renames the Data and Variables slots, and updates Batch and Metadata
#' names to stay consistent.
#' @param x A MultiBlock object.
#' @param value A character vector of new block names (same length as the
#'   number of blocks).
#' @return The updated MultiBlock object.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' blockNames(mb) <- c("block1", "block2")
#' blockNames(mb) # c("block1", "block2")
#' @export
setMethod("blockNames<-", "MultiBlock", function(x, value) {
  if (length(names(x@Batch)) != 0) {
    names(x@Batch) <- value[match(names(x@Batch), names(x@Data))]
  }
  if (length(names(x@Metadata)) != 0) {
    names(x@Metadata) <- value[match(names(x@Metadata), names(x@Data))]
  }
  names(x@Data) <- value
  names(x@Variables) <- value
  x
})

# ---- sampleNames ----

#' sampleNames
#'
#' Return the sample names of a MultiBlock object.
#' @param x A MultiBlock object.
#' @return A vector with the sample names.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' rownames(b1) <- paste0("s", 1:10)
#' mb <- MultiBlock(Data = list(b1 = b1))
#' sampleNames(mb) # "s1" ... "s10"
#' @export
setMethod("sampleNames", "MultiBlock", function(x) {
  x@Samples
})

#' sampleNames<-
#'
#' Set the sample names of a MultiBlock object.
#' @param x A MultiBlock object.
#' @param value A vector of new sample names. Must have the same length as
#'   the current number of samples.
#' @return The updated MultiBlock object.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' mb <- MultiBlock(Data = list(b1 = b1))
#' sampleNames(mb) <- paste0("patient_", 1:10)
#' sampleNames(mb)
#' @export
setMethod("sampleNames<-", "MultiBlock", function(x, value) {
  if (!is.vector(value)) {
    warning("The provided sample names are not a vector. Sample names were not changed.")
    return(x)
  }
  if (length(x@Samples) != length(value)) {
    warning("The length of the provided vector does not match. Sample names were not changed.")
    return(x)
  }
  x@Samples <- value
  x
})

# ---- variableNames ----

#' variableNames
#'
#' Return the variable names of a MultiBlock object.
#' @param x A MultiBlock object.
#' @param ... Not used. Present for S4 generic dispatch compatibility.
#' @param block Optional. A vector of block names or indices to retrieve.
#'   When omitted, variable names for all blocks are returned as a named list.
#'   When a single block is specified, a plain vector is returned.
#' @return A named list of variable-name vectors (all blocks), or a single
#'   vector when exactly one block is requested.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' variableNames(mb) # named list: b1 = 1:50, b2 = 1:80
#' variableNames(mb, "b1") # 1:50
#' variableNames(mb, 1:2) # same as variableNames(mb)
#' @export
setMethod("variableNames", "MultiBlock", function(x, block) {
  if (missing(block)) block <- seq_along(x@Data)
  if (is.list(block)) block <- unlist(block)

  if (length(block) == 1) {
    x@Variables[[block]]
  } else {
    thevars <- list()
    if (is.character(block)) {
      for (i in block) thevars[[i]] <- x@Variables[[i]]
    } else {
      bn <- blockNames(x)
      for (i in block) thevars[[bn[i]]] <- x@Variables[[bn[i]]]
    }
    thevars
  }
})

#' variableNames<-
#'
#' Set the variable names of a MultiBlock object.
#' \code{value} must be a named list with one entry per block, where each
#' entry is a vector of variable names whose length matches the number of
#' columns in that block.
#'
#' To update a single block, use standard list-replacement chaining:
#' \code{variableNames(mb)[["b1"]] <- newNames}
#'
#' @param x A MultiBlock object.
#' @param value A named list of variable-name vectors, one per block.
#' @return The updated MultiBlock object.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#'
#' # Replace all at once:
#' variableNames(mb) <- list(
#'   b1 = paste0("v", 1:50),
#'   b2 = paste0("v", 1:80)
#' )
#'
#' # Replace a single block using chaining:
#' variableNames(mb)[["b1"]] <- paste0("feat_", 1:50)
#' @export
setMethod("variableNames<-", "MultiBlock", function(x, value) {
  if (!is.list(value)) {
    stop("'value' must be a named list of variable-name vectors, one per block.")
  }
  if (is.null(names(value)) || any(names(value) == "")) {
    stop("All elements of 'value' must be named.")
  }
  for (nm in names(value)) {
    if (!(nm %in% names(x@Data))) {
      stop(sprintf("Block '%s' not found in the MultiBlock.", nm))
    }
    if (length(value[[nm]]) != ncol(x@Data[[nm]])) {
      stop(sprintf(
        "The length of variable names for block '%s' (%d) does not match ncol (%d).",
        nm, length(value[[nm]]), ncol(x@Data[[nm]])
      ))
    }
    x@Variables[[nm]] <- value[[nm]]
  }
  x
})

# ---- ncol / nrow ----

#' ncol for MultiBlock
#'
#' Return the number of columns (variables) in each block of a MultiBlock.
#' @param x A MultiBlock object.
#' @return A named integer vector with the number of columns per block.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' ncol(mb) # c(b1 = 50, b2 = 80)
#' @export
setMethod("ncol", "MultiBlock", function(x) {
  sapply(x@Data, ncol)
})

#' nrow for MultiBlock
#'
#' Return the number of rows (samples) in each block of a MultiBlock.
#' For a valid MultiBlock all blocks share the same number of rows, so the
#' result is a named integer vector with one (identical) value per block.
#' @param x A MultiBlock object.
#' @return A named integer vector with the number of rows per block.
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50)
#' b2 <- matrix(rnorm(800), 10, 80)
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' nrow(mb) # c(b1 = 10, b2 = 10)
#' @export
setMethod("nrow", "MultiBlock", function(x) {
  sapply(x@Data, nrow)
})

#' AddMetadata
#'
#' Adds or overwrites batch and/or metadata information for one or more blocks
#' of an existing MultiBlock object.
#' @param MB A MultiBlock object.
#' @param block The name of the block to modify (character). If \code{NULL} and
#'   the MultiBlock contains a single block, that block is used automatically.
#'   To modify several blocks at once, pass a character vector of block names.
#' @param batches A numeric vector of batch labels, one per sample in the
#'   target block. Replaces any existing Batch entry for that block.
#' @param metadata A data.frame with one row per sample in the target block.
#'   Replaces any existing Metadata entry for that block.
#' @return The updated MultiBlock.
#' @seealso \code{\link{MultiBlock}}, \code{\link{FilterSamplesMultiBlock}}
#' @examples
#' b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
#' b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
#' batch_b1 <- rep(1, 10)
#' meta_b1 <- data.frame(condition = rep(c("A", "B"), 5))
#' mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
#' # Add batch information to block 'b1':
#' mb <- AddMetadata(mb, block = "b1", batches = batch_b1)
#' # Add (or overwrite) metadata for block 'b1':
#' mb <- AddMetadata(mb, block = "b1", metadata = meta_b1)
#' @export
AddMetadata <- function(MB, block = NULL, batches = NULL, metadata = NULL) {
  if (!is(MB, "MultiBlock")) {
    stop("'MB' must be a MultiBlock object.")
  }

  if (is.null(block)) {
    if (length(MB@Data) == 1) {
      block <- names(MB@Data)
    } else {
      stop("'block' must be specified when the MultiBlock contains more than one block.")
    }
  }

  if (!all(block %in% names(MB@Data))) {
    missing_blocks <- block[!block %in% names(MB@Data)]
    stop(sprintf(
      "Block(s) not found in MultiBlock: %s",
      paste(missing_blocks, collapse = ", ")
    ))
  }

  for (blk in block) {
    n <- nrow(MB@Data[[blk]])

    if (!is.null(batches)) {
      if (!is.numeric(batches)) {
        stop("'batches' must be a numeric vector.")
      }
      if (length(batches) != n) {
        stop(sprintf(
          "'batches' must have %d elements (samples in block '%s').", n, blk
        ))
      }
      MB@Batch[[blk]] <- batches
    }

    if (!is.null(metadata)) {
      if (!is.data.frame(metadata)) {
        stop("'metadata' must be a data.frame.")
      }
      if (nrow(metadata) != n) {
        stop(sprintf(
          "'metadata' must have %d rows (samples in block '%s').", n, blk
        ))
      }
      MB@Metadata[[blk]] <- metadata
    }
  }

  return(MB)
}

#' MultiAssayExperiment2MultiBlock
#'
#' Converts a MultiAssayExperiment into a MultiBlock. Samples are first intersected across all
#' experiments so that only common samples are retained.
#' @param se A \code{MultiAssayExperiment} object.
#' @param colData_samplenames Character string giving the name of the column in \code{colData(se)}
#'   that holds the sample names used for matching. When \code{NULL}, \code{colData} metadata is
#'   not stored in the resulting MultiBlock.
#' @param Batch Character string giving the name of the column in \code{colData(se)} to use as
#'   the Batch variable. Requires \code{colData_samplenames} to be set; an error is raised if
#'   \code{Batch} is supplied but \code{colData_samplenames} is \code{NULL} or the column is not
#'   found in \code{colData}.
#' @details
#'   Columns (samples) are intersected across all experiments via \code{intersectColumns} before
#'   conversion. Each experiment in the \code{MultiAssayExperiment} becomes one block in the
#'   MultiBlock (rows = samples, columns = features). Sample names are taken from the
#'   \code{colData} row names of \code{se}. Metadata from \code{colData} is stored only for the
#'   first block; subsequent blocks share the same sample order but carry no additional metadata.
#'   If \code{Batch} is specified, the corresponding column is extracted from \code{colData},
#'   removed from the metadata, and stored as the \code{Batch} slot of the MultiBlock.
#' @return A \code{MultiBlock} object with one block per experiment in \code{se}.
#' @seealso \code{\link{MultiBlock}}, \code{\link{MultiBlock2MultiAssayExperiment}}
#' @examples
#' \donttest{
#' if (requireNamespace("MultiAssayExperiment", quietly = TRUE)) {
#' library(MultiAssayExperiment)
#' mae <- MultiAssayExperiment(
#'   experiments = ExperimentList(
#'     block1 = matrix(rnorm(50), nrow = 5, dimnames = list(paste0("s", 1:5), paste0("v", 1:10))),
#'     block2 = matrix(rnorm(30), nrow = 5, dimnames = list(paste0("s", 1:5), paste0("w", 1:6)))
#'   )
#' )
#' mb <- MultiAssayExperiment2MultiBlock(mae)
#' }
#' }
#' @export
MultiAssayExperiment2MultiBlock <- function(se,
                                            colData_samplenames = NULL, # Character string with the column metadata.
                                            Batch = NULL) { # Batch is the name of the column that will be considered as Batch

  # Here the metadata of the MultiAssayExperiment2MultiBlock is stored in the first MultiBlock.

  se <- intersectColumns(se[, , names(experiments(se))])

  simplelist <- assays(se)

  samplemap <- sampleMap(se)

  for (i in seq_along(simplelist@listData)) {
    Data <- list(t(as.matrix(simplelist@listData[[i]])))
    names(Data) <- names(simplelist@listData)[i]
    Variables <- list(rownames(simplelist@listData[[i]]))
    names(Variables) <- names(Data)
    eq.names <- samplemap@listData$primary[match(rownames(Data[[1]]), samplemap@listData$colname)]
    Data[[1]] <- Data[[1]][match(se@colData@rownames, eq.names), ]
    rownames(Data[[1]]) <- NULL
    colnames(Data[[1]]) <- NULL

    if (i == 1) {
      Samples <- se@colData@rownames
      if (is.null(colData_samplenames)) {
        Metadata <- NULL
      } else {
        Metadata <- as.data.frame(colData(se))
      }
      if (is.null(Metadata) && is.null(Batch)) {
        MB <- MultiBlock(
          Samples = Samples,
          Data = Data,
          Variables = Variables
        )
      } else if (is.null(Metadata) && !is.null(Batch)) {
        stop("There is no metadata in the MultiAssayExperiment to extract the Batch")
      } else if (!is.null(Metadata) && is.null(Batch)) {
        if (!(colData_samplenames %in% colnames(Metadata))) {
          stop("The string in 'colData_samplenames' is not a column name of colData.")
        }
        Metadata <- Metadata[match(
          Metadata[[colData_samplenames]],
          se@colData@rownames
        ), ]
        if (any(is.na(match(
          Metadata[[colData_samplenames]],
          se@colData@rownames
        )))) {
          Metadata <- Metadata[!is.na(match(
            Metadata[[colData_samplenames]],
            se@colData@rownames
          )), ]
        }
        Metadata <- list(x = Metadata)
        names(Metadata) <- names(Data)

        MB <- MultiBlock(
          Samples = Samples,
          Data = Data,
          Variables = Variables,
          Metadata = Metadata
        )
      } else {
        if (Batch %in% colnames(Metadata)) {
          batch_col <- Batch
          Batch <- Metadata[[batch_col]]
          noBatch <- setdiff(colnames(Metadata), batch_col)
          Metadata <- Metadata[, noBatch, drop = FALSE]
          Metadata <- Metadata[match(
            Metadata[[colData_samplenames]],
            se@colData@rownames
          ), ]
          if (any(is.na(match(
            Metadata[[colData_samplenames]],
            se@colData@rownames
          )))) {
            Metadata <- Metadata[!is.na(match(
              Metadata[[colData_samplenames]],
              se@colData@rownames
            )), ]
          }
          Metadata <- list(x = Metadata)
          names(Metadata) <- names(Data)
          Batch <- setNames(list(Batch), names(Data))
          MB <- MultiBlock(
            Samples = Samples,
            Data = Data,
            Variables = Variables,
            Batch = Batch,
            Metadata = Metadata
          )
        } else {
          stop("The string in batch was not found in the metadata.")
        }
      }
    } else {
      MB@Data <- c(MB@Data, Data)
      MB@Variables <- c(MB@Variables, Variables)
    }
  }

  return(MB)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SplitRW.R
\name{SplitRW}
\alias{SplitRW}
\title{SplitRW}
\usage{
SplitRW(
  MB = MB,
  checkSampleCorrespondence = FALSE,
  batchNormalisation = TRUE,
  showSampleCorrespondence = TRUE
)
}
\arguments{
\item{MB}{A \code{MultiBlock} object built with \code{\link{MultiBlock}()}. Must contain
\code{Batch} information; blocks without batch labels are treated as a single batch.}

\item{checkSampleCorrespondence}{Logical. If \code{FALSE} (default), identical sample order
and count are assumed across all batches and only the minimum batch size is used. If
\code{TRUE}, sample names are intersected across all batches and only common samples are
retained, which is safer when batches have different sizes or orderings.}

\item{batchNormalisation}{Logical. If \code{TRUE} (default), each replicate block is divided
by its Frobenius norm and by the square root of the number of replicates in its original
block, to prevent blocks with more replicates from dominating the analysis.}

\item{showSampleCorrespondence}{Logical. If \code{TRUE} (default), the matrix of sample
names assigned to each replicate block is printed to the console via \code{print()}.
Set to \code{FALSE} to suppress this output.}
}
\value{
A \code{MultiBlock} object in which each block corresponds to one batch of one
  original data block (a replicate-wise structure ready for \code{\link{ComDim_PCA}} or
  similar).
}
\description{
Splits a multi-block into a replicate-wise (RW) structure by expanding each block along its
batch dimension. Each batch within each original block becomes a separate block in the output,
enabling replicate-wise ComDim analysis.
}
\details{
Output block names follow the convention \code{<original_block>} when the original block has
  only one batch, or \code{<original_block>_<batch_label>} when it has multiple batches.
  The \code{Metadata} slot of each source block is also split and carried over to the
  corresponding replicate blocks. If the MultiBlock has no \code{Batch} information at all,
  the original object is returned unchanged with a warning.
}
\examples{
b1 <- matrix(rnorm(1500), 30, 50)
b2 <- matrix(rnorm(2400), 30, 80)
batch_b <- c(rep(1, 10), rep(2, 10), rep(3, 10))
# Generate the multi-block (mb) with 3 batches of 10 samples each
mb <- MultiBlock(
  Data = list(b1 = b1, b2 = b2),
  Batch = list(b1 = batch_b, b2 = batch_b),
  ignore.names = TRUE
)
rw <- SplitRW(mb)
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{ComDim_PCA}}, \code{\link{SelectFeaturesRW}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_PCA.R, R/ComDim_PLS.R
\name{Compress_Data_2020}
\alias{Compress_Data_2020}
\title{Compress_Data_2020}
\usage{
Compress_Data_2020(s_n = s_n, CompMethod = CompMethod, Partitions = Partitions)

Compress_Data_2020(s_n = s_n, CompMethod = CompMethod, Partitions = Partitions)
}
\arguments{
\item{s_n}{The multi-block object.}

\item{CompMethod}{It can be 'Normal' (default), 'Kernel', 'PCT', 'Tall' or 'Wide'.}

\item{Partitions}{The number of partitions.}
}
\value{
The compressed multi-block.

The compressed multi-block.
}
\description{
Compress large multi-block objects. Internal function of ComDim_PCA().

Compress large multi-block objects. Internal function of ComDim_PLS().
}
\references{
Original MATLAB implementation:
  \url{https://github.com/DNRutledge/ComDim/blob/main/Compress_Data_2020.m}

Original MATLAB implementation:
  \url{https://github.com/DNRutledge/ComDim/blob/main/Compress_Data_2020.m}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MultiBlock2Matrix.R
\name{MultiBlock2Matrix}
\alias{MultiBlock2Matrix}
\title{MultiBlock2Matrix}
\usage{
MultiBlock2Matrix(MB = MB, blocks = NULL, vars = NULL)
}
\arguments{
\item{MB}{A \code{MultiBlock} object.}

\item{blocks}{The blocks to combine. A vector of integers (block indices) or a vector of block
names. When omitted, all blocks are included.}

\item{vars}{The variables to keep. A list of the same length as \code{blocks}. Each element
contains the indices (integer) or names (character) of the variables to retain from the
corresponding block. When omitted, all variables in each block are included.}
}
\value{
A numeric matrix with rows corresponding to samples and columns corresponding to the
  variables of the selected blocks concatenated in order.
}
\description{
Combines the blocks of a MultiBlock into a single matrix by column-binding the selected blocks.
Useful for computing summary statistics across the whole MultiBlock (e.g. maximum value).
}
\details{
Blocks are column-bound in the order given by \code{blocks}. Row names of the output matrix
  are the sample names of \code{MB}. Column names are the variable names; if the same variable
  name appears in more than one block, duplicates are disambiguated by appending
  \code{.<block_name>} to all occurrences of the repeated name, and a warning is issued.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
# Combine all blocks into a matrix and compute a summary statistic:
mat <- MultiBlock2Matrix(mb)
max(mat)
# Combine only the first block, keeping variables 1-10:
mat <- MultiBlock2Matrix(mb, blocks = 1, vars = list(1:10))
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{ProcessMultiBlock}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_OPLS.R
\name{ComDim_OPLS}
\alias{ComDim_OPLS}
\title{ComDim_OPLS}
\usage{
ComDim_OPLS(
  MB = MB,
  y = y,
  ndim = 1,
  nort = 1,
  method = c("OPLS-DA", "OPLS-R"),
  decisionRule = c("fixed", "max")[2],
  normalise = FALSE,
  loquace = FALSE,
  cv.k = 7
)
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{y}{The Y-block. A class vector or dummy matrix for OPLS-DA, or a
numeric matrix/vector for OPLS-R.}

\item{ndim}{Number of predictive Common Dimensions. Default is 1.}

\item{nort}{Maximum number of orthogonal Common Dimensions. Default is 1.
The actual number used is determined by ConsensusOPLS cross-validation and
may be less than this value.}

\item{method}{'OPLS-DA' for discriminant analysis or 'OPLS-R' for regression.}

\item{decisionRule}{Only used if method is 'OPLS-DA'. If 'fixed', samples are
assigned to the class with Y-hat above 1/nclasses. If 'max', samples are
assigned to the class with the highest Y-hat.}

\item{normalise}{To apply block normalisation. FALSE == no (default), TRUE == yes.}

\item{loquace}{To display the calculation times. TRUE == yes, FALSE == no (default).}

\item{cv.k}{Number of folds for k-fold cross-validation (default 7). Set
to 0 to skip CV output. ConsensusOPLS always performs internal CV to select
the optimal number of orthogonal components; when \code{cv.k >= 2} the
resulting \code{Q2} and \code{DQ2} reflect that cross-validation.}
}
\value{
A \code{ComDim} object.  All slots are populated.  Key slots:
\describe{
  \item{\code{Method}}{\code{"OPLS-DA"} or \code{"OPLS-R"}.}
  \item{\code{ndim}}{Number of predictive Common Dimensions.}
  \item{\code{Q.scores}}{Predictive global scores matrix (\eqn{n \times ndim}).}
  \item{\code{T.scores}}{Named list of block-specific predictive local scores.}
  \item{\code{P.loadings}}{Global predictive loadings.}
  \item{\code{Saliences}}{Predictive block salience matrix (\eqn{ntable \times ndim}).}
  \item{\code{Orthogonal}}{List with orthogonal component outputs: \code{nort},
    \code{Q.scores}, \code{T.scores}, \code{P.loadings.ort}, \code{Saliences.ort}.}
  \item{\code{R2X}}{Named vector (length \eqn{ndim + nort}) of X-variance fractions.}
  \item{\code{R2Y}}{Named vector (length \eqn{ndim + nort}) of Y-variance fractions.}
  \item{\code{Q2}}{Cross-validated Q2 per class/response (when \code{cv.k >= 2};
    otherwise training-set fit).}
  \item{\code{DQ2}}{(OPLS-DA only) Cross-validated discriminant Q2 per class.}
  \item{\code{VIP}}{Global total VIP (named vector, length \eqn{p_{tot}}).}
  \item{\code{VIP.block}}{Named list (one data.frame per block) with columns
    \code{p}, \code{o}, \code{tot}.}
  \item{\code{PLS.model}}{KOPLS regression objects: \code{W}, \code{B}, \code{B0}, \code{Y}.}
  \item{\code{cv}}{Cross-validation results when \code{cv.k >= 2}: \code{k},
    \code{Ypred}, \code{Q2}, \code{DQ2}.}
  \item{\code{Prediction}}{Training-set predictions: \code{Y.pred}; for OPLS-DA
    also \code{decisionRule}, \code{trueClass}, \code{predClass},
    \code{Sensitivity}, \code{Specificity}, \code{confusionMatrix}.}
  \item{\code{Mean}}{List with \code{MeanMB} and \code{MeanY}.}
  \item{\code{Norm}}{List with \code{NormMB}, \code{FrobNorms}, \code{RVweights}.}
  \item{\code{variable.block}}{Block membership of each variable.}
  \item{\code{runtime}}{Total computation time in seconds.}
}
}
\description{
Finding common dimensions in multi-block datasets using OPLS. Also known as ConsensusOPLS
(ComDim-OPLS) for multiblock structures: orthogonal components uncorrelated
with Y are extracted from all blocks simultaneously before the predictive
components are computed.
}
\details{
This function is a wrapper around \code{\link[ConsensusOPLS]{ConsensusOPLS}}.
The core kernel-OPLS extraction is delegated to that package; all ComDim
output slots (local scores, loadings, VIP, sensitivity, confusion matrix,
etc.) are computed from the returned model objects.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
y <- rep(c("A", "B"), 5)
results <- ComDim_OPLS(mb, y, ndim = 1, nort = 1, method = "OPLS-DA")
}
\references{
Boccard J, Rutledge DN (2013).
  A consensus OPLS-DA strategy for multiblock Omics data fusion.
  \emph{Analytica Chimica Acta}, 769, 30--39.
  \doi{10.1016/j.aca.2013.01.022}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NormalizeMultiBlock.R
\name{NormalizeMultiBlock}
\alias{NormalizeMultiBlock}
\title{NormalizeMultiBlock}
\usage{
NormalizeMultiBlock(
  MB,
  blocks = NULL,
  method = c("none", "auto", "mean", "pareto", "norm", "geometric", "ranknorm")[5],
  infinite.as.NA = FALSE,
  constant = 0,
  offset = 3/8,
  showWarning = TRUE
)
}
\arguments{
\item{MB}{The MultiBlock object.}

\item{blocks}{Blocks to normalize. A vector of integers or block names
(optional; all blocks are processed when omitted).}

\item{method}{Normalization method. One of: \code{'none'} (no
transformation beyond optional Inf-to-NA conversion), \code{'auto'}
(auto-scaling), \code{'mean'} (mean-centering), \code{'pareto'}
(mean-centered and divided by the square root of the SD), \code{'norm'}
(mean-centered and divided by its Frobenius norm), \code{'geometric'}
(geometric mean scaling), \code{'ranknorm'} (rank-based inverse normal
transform). Default: \code{'norm'}.}

\item{infinite.as.NA}{If \code{TRUE}, Infinite values are converted to NA
before normalization. The presence of NA or Inf values will cause ComDim
to stop; use \code{\link{NAInfRemoveMultiBlock}} to handle them.}

\item{constant}{For \code{'geometric'}: value added to each element before
computing the geometric mean (useful when data contain zeros). Default: 0.}

\item{offset}{For \code{'ranknorm'}: offset in the Blom-type transform.
Default: 3/8 (Blom transform).}

\item{showWarning}{If \code{TRUE}, warn when all-NA variables are
discarded. Default: \code{TRUE}.}
}
\value{
The normalized MultiBlock object.
}
\description{
Normalize all blocks from a MultiBlock object. The ranknorm transform is
based on that from the RNOmni package (doi:10.1111/biom.13214).
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
b2[c(2, 3, 5), c(1, 2, 3)] <- NA
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
mb <- NormalizeMultiBlock(mb, method = "auto")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{blockNames<-}
\alias{blockNames<-}
\alias{blockNames<-,MultiBlock-method}
\title{blockNames<-}
\usage{
blockNames(x) <- value

\S4method{blockNames}{MultiBlock}(x) <- value
}
\arguments{
\item{x}{A MultiBlock object.}

\item{value}{A character vector of new block names (same length as the
number of blocks).}
}
\value{
The updated MultiBlock object.
}
\description{
Set the block names of a MultiBlock object.
Renames the Data and Variables slots, and updates Batch and Metadata
names to stay consistent.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
blockNames(mb) <- c("block1", "block2")
blockNames(mb) # c("block1", "block2")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_OPLS.R
\name{OPLS_NIPALS_DNR}
\alias{OPLS_NIPALS_DNR}
\title{OPLS_NIPALS_DNR}
\usage{
OPLS_NIPALS_DNR(W = W, y = y, threshold = 1e-10)
}
\arguments{
\item{W}{Numeric matrix (n x p): the concatenated, lambda-weighted blocks
as passed by \code{ComDim_y()}.}

\item{y}{Numeric matrix (n x q): the response block (dummy matrix for
discriminant analysis, numeric matrix for regression). Only the first
column drives the NIPALS u-score iteration; all columns are used to
compute the Y-loading \code{q}.}

\item{threshold}{Convergence threshold for the u-score update (default
\code{1e-10}).}
}
\value{
A named list:
\describe{
  \item{t_pred}{Predictive X-score (length n).}
  \item{w_pred}{Predictive X-weight (length p), L2-normalised.}
  \item{p}{X-loading (length p).}
  \item{q}{Y-loading (length q = ncol(y)).}
  \item{u}{Y-score (length n).}
  \item{t_ort}{Orthogonal X-score (length n).}
  \item{w_ort}{Orthogonal X-weight (length p), L2-normalised.}
  \item{p_ort}{Orthogonal X-loading (length p).}
}
}
\description{
One NIPALS OPLS step on a (lambda-weighted) concatenated block matrix.
Computes one predictive component and one orthogonal component in a single
pass. This function is the recommended building block for constructing an
OPLS-based \code{FUN} argument for \code{ComDim_y()} when \code{nort > 0}.
For \code{nort = 0} (plain PLS) a simpler PLS wrapper is sufficient.
}
\seealso{
\code{\link{ComDim_y}} for the multi-block OPLS wrapper that uses
  this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SimulateMultiBlock.R
\name{SimulateMultiBlock}
\alias{SimulateMultiBlock}
\title{SimulateMultiBlock}
\usage{
SimulateMultiBlock(
  n = 500L,
  p = 2000L,
  n_sources = 4L,
  noise = 0.05,
  n_blocks = 2L
)
}
\arguments{
\item{n}{Number of samples. Default: \code{500}.}

\item{p}{Total number of variables (split evenly across blocks). Must be
divisible by \code{n_blocks} and by \code{n_sources}. Default: \code{2000}.}

\item{n_sources}{Number of orthogonal latent sources. Default: \code{4}.}

\item{noise}{Fraction of total variance attributed to noise, in (0, 1).
Default: \code{0.05} (5 \% noise).}

\item{n_blocks}{Number of blocks to split the variables into. Default:
\code{2}.}
}
\value{
A \code{\link{MultiBlock}} object with \code{n_blocks} blocks, each
  of size \eqn{n \times (p / \text{n\_blocks})}, named \code{"Block1"},
  \code{"Block2"}, etc.
}
\description{
Generate a synthetic \code{MultiBlock} dataset built from a known number of
orthogonal latent sources plus Gaussian noise.  Useful for benchmarking and
testing ComDim functions.
}
\details{
The dataset is constructed as follows:
\enumerate{
  \item \code{n_sources} score vectors (\eqn{n \times \text{n\_sources}}) are
    drawn from a standard normal distribution and orthonormalised by QR
    decomposition.
  \item Loading vectors (\eqn{\text{n\_sources} \times p}) are built so that
    each source loads primarily (SD = 1) on one equal-sized variable segment,
    with small cross-loadings (SD = 0.10) on the remaining variables.
  \item The true signal \eqn{X = TP} is computed.
  \item Gaussian noise is added such that
    \eqn{\text{noise\_var} = \text{signal\_var} \times
    \text{noise} / (1 - \text{noise})}.
  \item The \eqn{p} variables are split into \code{n_blocks} equal-width
    blocks, each assembled as a named element of the returned
    \code{MultiBlock}.
}
}
\examples{
mb <- SimulateMultiBlock(n = 100, p = 200, n_sources = 4,
                         noise = 0.05, n_blocks = 2)
mb <- NormalizeMultiBlock(mb, method = 'norm')
res <- ComDim_PCA(mb, ndim = 4)
}
\seealso{
\code{\link{ComDim_PCA}}, \code{\link{MultiBlock}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MakeComDimScoresTable.R
\name{MakeComDimScoresTable}
\alias{MakeComDimScoresTable}
\title{MakeComDimScoresTable}
\usage{
MakeComDimScoresTable(
  model,
  blocks = NULL,
  dim = NULL,
  dim.ort = NULL,
  include = c("Q.scores", "T.scores", "Q.scores.ort", "T.scores.ort")[1:2]
)
}
\arguments{
\item{model}{The output from a ComDim analysis (a \code{ComDim} object).}

\item{blocks}{The blocks from which local scores will be extracted. A vector of integers (block
indices) or block names. When omitted, all blocks are included. Only relevant when
\code{"T.scores"} or \code{"T.scores.ort"} are in \code{include}.}

\item{dim}{Integer vector of predictive component indices to include. When omitted, all
predictive components in the model are included.}

\item{dim.ort}{Integer vector of orthogonal component indices to include. When \code{NULL},
all orthogonal components present in the model are included. When \code{0}, no orthogonal
components are included.}

\item{include}{Character vector selecting which score types to include. Accepted values
(case-insensitive) are \code{"Q.scores"} (global predictive scores),
\code{"T.scores"} (local predictive scores), \code{"Q.scores.ort"} (global orthogonal
scores), and \code{"T.scores.ort"} (local orthogonal scores). Defaults to
\code{c("Q.scores", "T.scores")}.}
}
\value{
A long data frame with one row per sample–component–score-type combination, containing
  the following columns:
  \describe{
    \item{\code{sample.id}}{Sample name (factor).}
    \item{\code{sample.id.number}}{Integer position of the sample (factor).}
    \item{\code{block.id}}{Block index, or \code{"Global"} / \code{"Global.ort"} for Q scores.}
    \item{\code{block.name}}{Block name, or \code{"Global"} / \code{"Global.ort"} for Q scores (factor).}
    \item{\code{dim}}{Component number.}
    \item{\code{scores.type}}{One of \code{"Global"}, \code{"Global.ort"}, \code{"Local"}, or \code{"Local.ort"} (factor).}
    \item{\code{scores.type.dim}}{Concatenation of scores type and component number, e.g. \code{"Q.scores1"} (factor).}
    \item{\code{value}}{Score value.}
  }
}
\description{
Creates a long (tidy) data frame with the global and/or local scores from a ComDim model,
suitable for use with \pkg{ggplot2}.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50) # 10 rows and 50 columns
b2 <- matrix(rnorm(800), 10, 80) # 10 rows and 80 columns
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
model <- ComDim_PCA(mb, ndim = 2)
tbl <- MakeComDimScoresTable(model)
}
\seealso{
\code{\link{MakeComDimLoadingsTable}}, \code{\link{ComDim_PCA}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_PCA.R, R/ComDim_PLS.R
\name{PCA_Tall_PCT_DNR}
\alias{PCA_Tall_PCT_DNR}
\title{PCA_Tall_PCT_DNR}
\usage{
PCA_Tall_PCT_DNR(Xn = Xn, PCs = PCs, Partitions = Partitions)

PCA_Tall_PCT_DNR(Xn = Xn, PCs = PCs, Partitions = Partitions)
}
\arguments{
\item{Xn}{The block to compress.}

\item{PCs}{Number of PCs to keep.}

\item{Partitions}{The number of partitions.}
}
\value{
The compressed block.

The compressed block.
}
\description{
Compress a block using the PCT method. Internal function of ComDim_PCA().

Compress a block using the PCT method. Internal function of ComDim_PLS().
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{variableNames<-}
\alias{variableNames<-}
\alias{variableNames<-,MultiBlock-method}
\title{variableNames<-}
\usage{
variableNames(x) <- value

\S4method{variableNames}{MultiBlock}(x) <- value
}
\arguments{
\item{x}{A MultiBlock object.}

\item{value}{A named list of variable-name vectors, one per block.}
}
\value{
The updated MultiBlock object.
}
\description{
Set the variable names of a MultiBlock object.
\code{value} must be a named list with one entry per block, where each
entry is a vector of variable names whose length matches the number of
columns in that block.
}
\details{
To update a single block, use standard list-replacement chaining:
\code{variableNames(mb)[["b1"]] <- newNames}
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))

# Replace all at once:
variableNames(mb) <- list(
  b1 = paste0("v", 1:50),
  b2 = paste0("v", 1:80)
)

# Replace a single block using chaining:
variableNames(mb)[["b1"]] <- paste0("feat_", 1:50)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{nrow,MultiBlock-method}
\alias{nrow,MultiBlock-method}
\title{nrow for MultiBlock}
\usage{
\S4method{nrow}{MultiBlock}(x)
}
\arguments{
\item{x}{A MultiBlock object.}
}
\value{
A named integer vector with the number of rows per block.
}
\description{
Return the number of rows (samples) in each block of a MultiBlock.
For a valid MultiBlock all blocks share the same number of rows, so the
result is a named integer vector with one (identical) value per block.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
nrow(mb) # c(b1 = 10, b2 = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MakeComDimLoadingsTable.R
\name{MakeComDimLoadingsTable}
\alias{MakeComDimLoadingsTable}
\title{MakeComDimLoadingsTable}
\usage{
MakeComDimLoadingsTable(model, blocks = NULL, dim = NULL, dim.ort = NULL)
}
\arguments{
\item{model}{The output from a ComDim analysis (a \code{ComDim} object).}

\item{blocks}{The blocks from which loadings will be extracted. A vector of integers (block
indices) or block names. When omitted, all blocks are included.}

\item{dim}{Integer vector of predictive component indices to include. When omitted, all
predictive components in the model are included.}

\item{dim.ort}{Integer vector of orthogonal component indices to include. When \code{NULL},
all orthogonal components present in the model are included. When \code{0}, no orthogonal
components are included.}
}
\value{
A long data frame with one row per variable–component combination, containing the
  following columns:
  \describe{
    \item{\code{variable.id}}{Variable name (factor).}
    \item{\code{variable.id.number}}{Position of the variable across all blocks (factor).}
    \item{\code{block.id}}{Integer index of the block.}
    \item{\code{block.name}}{Name of the block (factor).}
    \item{\code{dim}}{Component number.}
    \item{\code{value}}{Loading value (from \code{P.loadings} or \code{Orthogonal$P.loadings.ort}).}
  }
}
\description{
Creates a long (tidy) data frame with the local P-loadings from a ComDim model, suitable for
use with \pkg{ggplot2}.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50) # 10 rows and 50 columns
b2 <- matrix(rnorm(800), 10, 80) # 10 rows and 80 columns
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
model <- ComDim_PCA(mb, ndim = 2)
tbl <- MakeComDimLoadingsTable(model)
}
\seealso{
\code{\link{MakeComDimScoresTable}}, \code{\link{ComDim_PCA}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-metadata_lipids.R
\docType{data}
\name{metadata_lipids}
\alias{metadata_lipids}
\title{Metadata for the lipid profiles of cell extracts}
\format{
An object of class \code{data.frame} with 437 rows and 2 columns.
}
\source{
\doi{10.15252/msb.202010141}
}
\usage{
data(mouse_ds)
}
\description{
Metadata for the lipid profiles of cell extracts
}
\examples{
data(mouse_ds)
#' lipidsMB <- ExpandMB(data = lipids, metadata = metadata_lipids,
#'   minblock = 0, loquace = FALSE)
}
\author{
Radic Shechter et al. (2021) Molecular Systems Biology 17:e10141
(\doi{10.15252/msb.202010141})
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{sampleNames<-}
\alias{sampleNames<-}
\alias{sampleNames<-,MultiBlock-method}
\title{sampleNames<-}
\usage{
sampleNames(x) <- value

\S4method{sampleNames}{MultiBlock}(x) <- value
}
\arguments{
\item{x}{A MultiBlock object.}

\item{value}{A vector of new sample names. Must have the same length as
the current number of samples.}
}
\value{
The updated MultiBlock object.
}
\description{
Set the sample names of a MultiBlock object.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
mb <- MultiBlock(Data = list(b1 = b1))
sampleNames(mb) <- paste0("patient_", 1:10)
sampleNames(mb)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ExpandMultiBlock.R
\name{ExpandMultiBlock}
\alias{ExpandMultiBlock}
\title{ExpandMultiBlock}
\usage{
ExpandMultiBlock(data = NULL, metadata = NULL, minblock = 0, loquace = TRUE)
}
\arguments{
\item{data}{A data.frame or matrix with samples in rows and variables in columns.}

\item{metadata}{A 2-column data.frame describing how variables are assigned to blocks. The
first column gives the block name; the second column gives the variable name, and must match
the column names of \code{data}. A variable may appear in multiple rows (and therefore in
multiple blocks).}

\item{minblock}{Integer. Blocks with fewer than \code{minblock} variables are discarded. Use
\code{0} (default) to keep all blocks regardless of size.}

\item{loquace}{Logical. If \code{TRUE} (default), a warning is printed listing column names in
\code{data} that have no matching entry in the second column of \code{metadata}.}
}
\value{
A \code{MultiBlock} object whose blocks are defined by the first column of
  \code{metadata}, or \code{NULL} if no valid blocks could be constructed.
}
\description{
Splits data into several blocks, allowing variables to appear in more than one block
simultaneously. Each variable is duplicated into every block to which it is assigned according
to the \code{metadata} mapping table.
}
\details{
For each row in \code{metadata} that matches a variable in \code{data}, the variable's
  values are copied into the corresponding block column. Column names in the resulting expanded
  matrix are formed as \code{<block>.<variable>}. Variables with all-\code{NA} values after
  expansion are removed. If no matches exist between \code{data} column names and
  \code{metadata}, \code{NULL} is returned with a warning. If \code{minblock} filtering
  removes all blocks, \code{NULL} is returned.
}
\examples{
data(mouse_ds)
lipidsMB <- ExpandMultiBlock(data = lipids, metadata = metadata_lipids,
  minblock = 0, loquace = FALSE)
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{ProcessMultiBlock}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-lcms.R
\docType{data}
\name{lcms}
\alias{lcms}
\title{LCMS data on cell differentiation}
\format{
An object of class \code{matrix} (inherits from \code{array}) with 36 rows and 44 columns.
}
\source{
\href{https://www.ebi.ac.uk/metabolights/MTBLS283}{Metabolights}
}
\usage{
data(dataset3)
}
\description{
LCMS data on cell differentiation
}
\examples{
data(dataset3)
#' mb_d3 <- MultiBlock(Data = list(lcms = lcms))
}
\author{
Cabrero et al. (2019) Scientific data 6:256
(\doi{10.1038/s41597-019-0202-7})
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MultiBlock.R
\name{MultiBlock}
\alias{MultiBlock}
\title{MultiBlock}
\usage{
MultiBlock(
  Samples = NULL,
  Data,
  Variables = NULL,
  Batch = NULL,
  Metadata = NULL,
  ignore.names = FALSE,
  ignore.size = FALSE
)
}
\arguments{
\item{Samples}{A vector of sample names shared across all blocks (optional).
When omitted, sample names are taken from the row names of each block.
If no row names exist and all blocks have the same number of rows,
samples are numbered as integers. Use \code{ignore.names} or
\code{ignore.size} for more flexible behaviour.}

\item{Data}{A named list of matrices or data.frames (one entry per block).}

\item{Variables}{A named list of variable-name vectors, one per block
(optional). When omitted, column names are taken from each block; if
absent, variables are numbered as integers.}

\item{Batch}{A named list of batch vectors, one per block (optional).}

\item{Metadata}{A named list of metadata data.frames, one per block
(optional).}

\item{ignore.names}{If TRUE, sample names are not checked across blocks.
All blocks must have the same number of rows unless \code{ignore.size}
is also TRUE.}

\item{ignore.size}{If TRUE (only meaningful when \code{ignore.names = TRUE}),
blocks with different row counts are accepted. The resulting MultiBlock
stores a per-block Samples list and is not directly suitable for ComDim
(use \code{SplitRW()} first).}
}
\value{
A MultiBlock object.
}
\description{
Creates a MultiBlock object from a named list of data blocks.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
# Minimal call: Samples and Variables are filled in automatically.
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))

# With explicit sample names (enables cross-block alignment):
rownames(b1) <- paste0("s", 1:10)
rownames(b2) <- paste0("s", 1:10)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))

# Blocks with different row counts (replicate design):
b3 <- matrix(rnorm(800), 30, 80)
batch_b3 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
mb3 <- MultiBlock(
  Data = list(b3 = b3), Batch = list(b3 = batch_b3),
  ignore.names = TRUE, ignore.size = TRUE
)
}
\references{
Puig-Castellví F, Jouan-Rimbaud Bouveresse D, Mazéas L, Chapleur O, Rutledge DN (2021).
  Rearrangement of incomplete multi-omics datasets combined with ComDim for evaluating replicate
  cross-platform variability and batch influence.
  \emph{Chemometrics and Intelligent Laboratory Systems}, 218, 104422.
  \doi{10.1016/j.chemolab.2021.104422}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_PCA.R
\name{ComDim_PCA}
\alias{ComDim_PCA}
\title{ComDim_PCA}
\usage{
ComDim_PCA(
  MB = MB,
  ndim = NULL,
  normalise = FALSE,
  threshold = 1e-10,
  loquace = FALSE,
  CompMethod = "Normal",
  Partitions = 1
)
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{ndim}{Number of Common Dimensions.}

\item{normalise}{To apply normalisation. FALSE == no (default), TRUE == yes.}

\item{threshold}{The threshold limit to stop the iterations. If the "difference of fit" < threshold (1e-10 as default).}

\item{loquace}{To display the calculation times. TRUE == yes, FALSE == no (default).}

\item{CompMethod}{To speed-up the analysis for really big MultiBlocks. 'Normal' (default), 'Kernel', 'PCT', 'Tall' or 'Wide'.}

\item{Partitions}{To speed-up the analysis for really big MultiBlocks. This parameter is used if CompMethod is 'Tall' or 'Wide'.}
}
\value{
A \code{ComDim} object. Slots for supervised analysis
  (\code{R2Y}, \code{Q2}, \code{DQ2}, \code{VIP}, \code{VIP.block},
  \code{PLS.model}, \code{cv}, \code{Prediction}) are empty. The
  populated slots are:
\describe{
  \item{\code{Method}}{\code{"PCA"}.}
  \item{\code{ndim}}{Number of Common Dimensions extracted.}
  \item{\code{Q.scores}}{Global scores matrix (\eqn{n \times ndim}).
    Column names are \code{CC1}, \code{CC2}, etc.; row names are sample
    names.  Each column \eqn{\mathbf{q}_a} is a unit-norm consensus score,
    the dominant left singular vector of the salience-weighted concatenated
    blocks
    \eqn{\mathbf{W} = [\sqrt{\lambda_1}\mathbf{X}_1 \mid \cdots \mid
    \sqrt{\lambda_B}\mathbf{X}_B]}.}
  \item{\code{T.scores}}{Named list of block-specific local scores matrices
    (\eqn{n \times ndim} each).  For block \eqn{b} and component \eqn{a}:
    local loading \eqn{\mathbf{p}_{ba} = \mathbf{X}_b'\mathbf{q}_a} and
    local score
    \eqn{\mathbf{t}_{ba} =
    \mathbf{X}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
  \item{\code{P.loadings}}{Global loadings matrix (\eqn{p_{tot} \times
    ndim}).  Column \eqn{a} is
    \eqn{\mathbf{P}_a = \mathbf{X}'\mathbf{q}_a},
    where \eqn{\mathbf{X}} is the mean-centred (and optionally normalised)
    concatenated blocks.}
  \item{\code{Saliences}}{Block salience (weight) matrix (\eqn{ntable
    \times ndim}, row names = block names).  Entry \eqn{(b,a)} is
    \eqn{\lambda_{ba} =
    \mathbf{q}_a'\mathbf{X}_b\mathbf{X}_b'\mathbf{q}_a},
    the variance of block \eqn{b} captured by global score \eqn{a}.}
  \item{\code{R2X}}{Proportion of multi-block inertia captured by each
    component (named vector, length \eqn{ndim}).  Let
    \eqn{d_a} be the leading singular value of \eqn{\mathbf{W}} for
    component \eqn{a} (stored as \eqn{Singular_a = d_a^2}); then
    \deqn{R2X_a = Singular_a^2 \big/ \sum_k Singular_k^2 = d_a^4 \big/
    \sum_k d_k^4.}}
  \item{\code{Singular}}{Squared leading singular values of
    \eqn{\mathbf{W}}, one per component:
    \eqn{Singular_a = d_a^2}.}
  \item{\code{Mean}}{List with \code{MeanMB}: named list of column-mean
    vectors per block, used for mean-centring.}
  \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms used for
    block normalisation (all ones when \code{normalise = FALSE}).}
  \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
    indicating the block name of each row in \code{P.loadings}.}
  \item{\code{runtime}}{Total computation time in seconds.}
}
}
\description{
Finding common dimensions in multi-block datasets.
}
\examples{
# Example 1: two data blocks.
b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
results <- ComDim_PCA(mb, 2)

# Example 2: two data blocks, each with different replicate number
b1 <- matrix(rnorm(500), 10, 50)
batch_b1 <- rep(1, 10)
b2 <- matrix(rnorm(2400), 30, 80)
batch_b2 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
mb <- MultiBlock(
  Samples = list(
    b1 = paste0("samples_", 1:10),
    b2 = rep(paste0("samples_", 1:10), 3)
  ),
  Data = list(b1 = b1, b2 = b2),
  Batch = list(b1 = batch_b1, b2 = batch_b2),
  ignore.size = TRUE
)
rw <- SplitRW(mb)
results <- ComDim_PCA(rw, 2)
}
\references{
Jouan-Rimbaud Bouveresse D, Rutledge DN (2024).
  A synthetic review of some recent extensions of ComDim.
  \emph{Journal of Chemometrics}, 38(5), e3454.
  \doi{10.1002/cem.3454}

  Original MATLAB implementation: \url{https://github.com/DNRutledge/ComDim/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\docType{class}
\name{ComDim-classes}
\alias{ComDim-classes}
\alias{ComDim}
\title{MultiBlock object}
\description{
Object of the type \code{MultiBlock}, to use as input for ComDim
analyses.

The output of a ComDim analysis.
}
\section{Slots}{

\describe{
\item{\code{Samples}}{vector with the sample names. If this data is not available,
the slot will be filled with integers.}

\item{\code{Data}}{A list with the data-blocks.}

\item{\code{Variables}}{A character vector with the variable names. If this data is
not available, the slot will be filled with integers.}

\item{\code{Batch}}{A list with the vectors with the batch information for each
data-block. Optional.}

\item{\code{Metadata}}{A list with the samples metadata.}

\item{\code{Method}}{The algorithm used in the core of the ComDim analysis (ex. PCA,
PLS,...)}

\item{\code{ndim}}{The number of components.}

\item{\code{Q.scores}}{The Global scores.}

\item{\code{T.scores}}{The Local scores.}

\item{\code{P.loadings}}{The Loadings}

\item{\code{Saliences}}{The Saliences}

\item{\code{R2X}}{The explained variance of the MultiBlock.}

\item{\code{R2Y}}{For regression or discriminant models, the explained variance of
the Y-block.}

\item{\code{Q2}}{For regression or discriminant models, the predicted variance of
the Y-block.}

\item{\code{DQ2}}{For discriminant models, the predicted discriminant variance of the
Y-block.}

\item{\code{Singular}}{The singular values.}

\item{\code{Mean}}{The mean values for each variable in the MultiBlock.}

\item{\code{Norm}}{The norm values for each variable in the MultiBlock.}

\item{\code{PLS.model}}{For ComDim analyses using PLS as the core algorithm, contains
the W, B, B0 and Y matrices.}

\item{\code{cv}}{For ComDim_KOPLS, it contains the index of the samples used during
the cross-validation.}

\item{\code{Prediction}}{A list with the predicted Y, the decision rule used for
sample classification, the sensitivity, the specificity, and a confusion
matrix.}

\item{\code{Metadata}}{A list with the per-sample metadata for each block.}

\item{\code{variable.block}}{A vector with the same length as the P.loadings,
indicating the block each variable belongs to.}

\item{\code{runtime}}{The used running time.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SummarizedExperiment2MultiBlock.R
\name{SummarizedExperiment2MultiBlock}
\alias{SummarizedExperiment2MultiBlock}
\title{SummarizedExperiment2MultiBlock}
\usage{
SummarizedExperiment2MultiBlock(se, colData_samplenames = NULL, Batch = NULL)
}
\arguments{
\item{se}{A \code{SummarizedExperiment} object.}

\item{colData_samplenames}{Character string giving the name of the column in \code{colData(se)}
that holds the sample names used for matching. When \code{NULL}, \code{colData} metadata is
not stored in the resulting MultiBlock.}

\item{Batch}{Character string giving the name of the column in \code{colData(se)} to use as
the Batch variable. When supplied, the column is extracted from \code{colData}, removed from
the metadata, and stored in the \code{Batch} slot of the MultiBlock. Requires at least one
of \code{colData_samplenames} or a named \code{colData}; an error is raised if the column
is not found.}
}
\value{
A \code{MultiBlock} object with one block per assay in \code{se}.
}
\description{
Converts a \code{SummarizedExperiment} into a MultiBlock. Each assay in the
\code{SummarizedExperiment} becomes one block (rows = samples, columns = features).
}
\details{
Each assay is transposed so that samples are in rows and features in columns. Sample order
  is aligned to \code{colData(se)} row names; samples not present in \code{colData} are
  removed. If an assay has no name, the block is labelled \code{"X"}. Metadata from
  \code{colData} (excluding the Batch column if specified) is stored only for the first block;
  subsequent assays are appended as additional blocks with no extra metadata.
}
\examples{
\donttest{
if (requireNamespace("SummarizedExperiment", quietly = TRUE)) {
library(SummarizedExperiment)
nrows <- 20; ncols <- 6
counts <- matrix(runif(nrows * ncols, 1, 1e4), nrows)
rownames(counts) <- paste0("feature", seq_len(nrows))
colnames(counts) <- paste0("sample",  seq_len(ncols))
se <- SummarizedExperiment(assays = list(counts = counts))
mb <- SummarizedExperiment2MultiBlock(se)
}
}
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{MultiAssayExperiment2MultiBlock}},
  \code{\link{MultiBlock2MultiAssayExperiment}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_PCA.R, R/ComDim_PLS.R
\name{RowsPartition}
\alias{RowsPartition}
\title{RowsPartition}
\usage{
RowsPartition(Xn = Xn, Partitions = Partitions)

RowsPartition(Xn = Xn, Partitions = Partitions)
}
\arguments{
\item{Xn}{A block.}

\item{Partitions}{The number of partitions.}
}
\value{
A list of row-partitioned sub-matrices.

A list of row-partitioned sub-matrices.
}
\description{
Calculate horizontal partitions in a block. Internal function of ComDim_PCA().

Calculate horizontal partitions in a block. Internal function of ComDim_PLS().
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{ncol,MultiBlock-method}
\alias{ncol,MultiBlock-method}
\title{ncol for MultiBlock}
\usage{
\S4method{ncol}{MultiBlock}(x)
}
\arguments{
\item{x}{A MultiBlock object.}
}
\value{
A named integer vector with the number of columns per block.
}
\description{
Return the number of columns (variables) in each block of a MultiBlock.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
ncol(mb) # c(b1 = 50, b2 = 80)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FilterSamplesMultiBlock.R
\name{FilterSamplesMultiBlock}
\alias{FilterSamplesMultiBlock}
\title{FilterSamplesMultiBlock}
\usage{
FilterSamplesMultiBlock(MB, samples = sampleNames(MB))
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{samples}{A vector of sample names to keep. Names not found in the
MultiBlock are silently ignored; the order of the retained samples
follows the order given here.}
}
\value{
The MultiBlock object restricted to the requested samples. The \code{Batch} and
  \code{Metadata} slots are also subsetted and reordered to match.
}
\description{
Retain a subset of samples in a MultiBlock object.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
rownames(b1) <- rownames(b2) <- paste0("sample_", 1:10)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
mb <- FilterSamplesMultiBlock(mb, samples = paste0("sample_", 1:5))
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{AddMetadata}}, \code{\link{ProcessMultiBlock}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/R.ComDim-package.R
\docType{package}
\name{R.ComDim-package}
\alias{R.ComDim}
\alias{R.ComDim-package}
\title{R.ComDim: ComDim Analysis in R}
\description{
Provides functions to run ComDim analysis. ComDim is an unsupervised
multi-block method that simultaneously considers multiple data tables to
find latent components common to all tables as well as those specific to
each table.
}
\author{
\strong{Maintainer}: Francesc Puig-Castellví \email{puig.francesc@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AddMetadata.R
\name{AddMetadata}
\alias{AddMetadata}
\title{AddMetadata}
\usage{
AddMetadata(MB, block = NULL, batches = NULL, metadata = NULL)
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{block}{The name of the block to modify (character). If \code{NULL} and
the MultiBlock contains a single block, that block is used automatically.
To modify several blocks at once, pass a character vector of block names.}

\item{batches}{A numeric vector of batch labels, one per sample in the
target block. Replaces any existing Batch entry for that block.}

\item{metadata}{A data.frame with one row per sample in the target block.
Replaces any existing Metadata entry for that block.}
}
\value{
The updated MultiBlock.
}
\description{
Adds or overwrites batch and/or metadata information for one or more blocks
of an existing MultiBlock object.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
batch_b1 <- rep(1, 10)
meta_b1 <- data.frame(condition = rep(c("A", "B"), 5))
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
# Add batch information to block 'b1':
mb <- AddMetadata(mb, block = "b1", batches = batch_b1)
# Add (or overwrite) metadata for block 'b1':
mb <- AddMetadata(mb, block = "b1", metadata = meta_b1)
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{FilterSamplesMultiBlock}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_Exploratory.R
\name{ComDim_Exploratory}
\alias{ComDim_Exploratory}
\title{ComDim_Exploratory - Extending matrix decomposition methods to multi-block data.}
\usage{
ComDim_Exploratory(
  MB = MB,
  ndim = NULL,
  FUN = FUN,
  normalise = FALSE,
  threshold = 1e-10,
  loquace = FALSE,
  method = "FUN",
  ...
)
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{ndim}{Number of Common Dimensions.}

\item{FUN}{The function used as the core of the ComDim analysis. It must accept a matrix
\code{W} (salience-weighted concatenated blocks) and \code{ndim} as its first two
arguments, and return a matrix of scores with one column per component.}

\item{normalise}{To apply block normalisation. FALSE == no (default), TRUE == yes.}

\item{threshold}{The threshold limit to stop the iterations. Iterations stop when the
change in the global score vector is below this value (1e-10 as default).}

\item{loquace}{To display the calculation times. TRUE == yes, FALSE == no (default).}

\item{method}{A string label identifying the decomposition method used (default: 'FUN').}

\item{...}{Additional arguments passed to \code{FUN}.}
}
\value{
A \code{ComDim} object. Slots for supervised analysis
  (\code{R2Y}, \code{Q2}, \code{DQ2}, \code{VIP}, \code{VIP.block},
  \code{PLS.model}, \code{cv}, \code{Prediction}) are empty. The
  populated slots are:
\describe{
  \item{\code{Method}}{The label supplied via the \code{method} argument.}
  \item{\code{ndim}}{Number of Common Dimensions extracted.}
  \item{\code{Q.scores}}{Global scores matrix (\eqn{n \times ndim}).
    Column names are \code{CC1}, \code{CC2}, etc.; row names are sample
    names.  Each column \eqn{\mathbf{q}_a} is a unit-norm consensus score
    derived from the dominant left direction of FUN applied to the
    salience-weighted concatenated blocks
    \eqn{\mathbf{W} = [\sqrt{\lambda_1}\mathbf{X}_1 \mid \cdots \mid
    \sqrt{\lambda_B}\mathbf{X}_B]}.}
  \item{\code{T.scores}}{Named list of block-specific local scores matrices
    (\eqn{n \times ndim} each).  For block \eqn{b} and component \eqn{a}:
    local loading \eqn{\mathbf{p}_{ba} = \mathbf{X}_b'\mathbf{q}_a} and
    local score
    \eqn{\mathbf{t}_{ba} =
    \mathbf{X}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
  \item{\code{P.loadings}}{Global loadings matrix (\eqn{p_{tot} \times
    ndim}).  Column \eqn{a} is
    \eqn{\mathbf{P}_a = \mathbf{X}'\mathbf{q}_a},
    where \eqn{\mathbf{X}} is the mean-centred (and optionally normalised)
    concatenated blocks.}
  \item{\code{Saliences}}{Block salience (weight) matrix (\eqn{ntable
    \times ndim}, row names = block names).  Entry \eqn{(b,a)} is
    \eqn{\lambda_{ba} =
    \mathbf{q}_a'\mathbf{X}_b\mathbf{X}_b'\mathbf{q}_a},
    the variance of block \eqn{b} captured by global score \eqn{a}.}
  \item{\code{R2X}}{Proportion of multi-block variance captured by each
    component (named vector, length \eqn{ndim}).  Let \eqn{\mathbf{s}_a}
    be the score vector returned by FUN for component \eqn{a} (before
    unit-normalisation to obtain \eqn{\mathbf{q}_a}), so that
    \eqn{sv_a = \|\mathbf{s}_a\|}; then
    \deqn{R2X_a = sv_a^4 \big/ \sum_k sv_k^4 =
    Singular_a^2 \big/ \sum_k Singular_k^2.}}
  \item{\code{Singular}}{Squared L2 norms of the FUN score vectors:
    \eqn{Singular_a = sv_a^2 = \|\mathbf{s}_a\|^2}, used to derive
    \code{R2X}.}
  \item{\code{Mean}}{List with \code{MeanMB}: named list of column-mean
    vectors per block, used for mean-centring.}
  \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms used for
    block normalisation (all ones when \code{normalise = FALSE}).}
  \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
    indicating the block name of each row in \code{P.loadings}.}
  \item{\code{runtime}}{Total computation time in seconds.}
}
}
\description{
Extends any matrix decomposition method used for exploratory purposes to the multi-block
field. The user provides a function (\code{FUN}) to compute the scores from the
salience-weighted concatenated blocks; these scores are then used to derive the global
scores, local scores, and loadings following the traditional ComDim-PCA framework.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))

if (requireNamespace("ica", quietly = TRUE)) {
  fun.ICA <- function(W, ndim, ...) {
    # W is the concatenated MB.
    # ndim is the number of components.
    result <- ica::ica(W, ndim)
    # The function must return the source estimates (analogous to PCA scores).
    return(result$S)
  }
  resultsICA <- ComDim_Exploratory(mb,
    ndim = 2,
    FUN = fun.ICA,
    method = "ICA"
  )
}
}
\references{
Jouan-Rimbaud Bouveresse D, Rutledge DN (2024).
  A synthetic review of some recent extensions of ComDim.
  \emph{Journal of Chemometrics}, 38(5), e3454.
  \doi{10.1002/cem.3454}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PredictMultiBlock.R
\name{PredictMultiBlock}
\alias{PredictMultiBlock}
\title{PredictMultiBlock}
\usage{
PredictMultiBlock(MB = MB, y, model = model, normalise = FALSE, loquace = TRUE)
}
\arguments{
\item{MB}{A MultiBlock object containing the new samples to project.}

\item{y}{Response vector or dummy matrix (optional). When supplied for a
supervised model, Q2, DQ2, and classification statistics are computed for
the new samples.}

\item{model}{A ComDim object (the calibration model).}

\item{normalise}{If TRUE, each block is mean-centred using the training
column means and divided by the training Frobenius norm. Must match the
\code{normalise} setting used during calibration. Default FALSE.}

\item{loquace}{If TRUE, print a message for each set of model elements
that were projected. Default TRUE.}
}
\value{
The \code{model} ComDim object with updated slots:
  \itemize{
    \item \code{Q.scores} — projected global scores (new samples x ndim).
    \item \code{T.scores} — projected local scores (per block).
    \item \code{Orthogonal$Q.scores} — projected global ort scores (if
      model has orthogonal components).
    \item \code{Orthogonal$T.scores} — projected local ort scores.
    \item \code{Prediction$Y.pred} — predicted Y (supervised models).
    \item \code{Q2}, \code{DQ2}, classification slots — when \code{y}
      is supplied for a supervised model.
  }
}
\description{
Projects a new MultiBlock dataset into an existing ComDim model.
Works with models produced by ComDim_PCA, ComDim_PLS,
ComDim_OPLS, ComDim_y, and ComDim_Exploratory. The projection
type (PCA-like, PLS-like, OPLS-like) is determined from the model
structure rather than the method string, so custom method labels from
ComDim_y are handled automatically.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ProcessMultiBlock.R
\name{ProcessMultiBlock}
\alias{ProcessMultiBlock}
\title{ProcessMultiBlock}
\usage{
ProcessMultiBlock(
  MB = MB,
  blocks = NULL,
  vars = NULL,
  FUN = NULL,
  FUN.SelectVars = NULL,
  FUN.SelectBlocks = NULL
)
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{blocks}{The blocks to process. A vector of integers or block names. When omitted, all
blocks are processed.}

\item{vars}{The variables to keep. A list with the same length as \code{blocks}. Each element
contains the indices (integer) or names (character) of the variables to retain in that block.
Subsetting is applied before any of the \code{FUN} arguments.}

\item{FUN}{A function applied to each selected block's data matrix (samples x variables). It
receives the matrix as its sole argument and must return a matrix of the same dimensions.}

\item{FUN.SelectVars}{A function applied to each selected block's data matrix to determine which
variables to keep. It receives the matrix as its sole argument and must return a logical vector
of length equal to \code{ncol} of the matrix (\code{TRUE} = keep).}

\item{FUN.SelectBlocks}{A function applied to each selected block's data matrix to determine
whether the block should be retained. It receives the matrix as its sole argument and must
return a single \code{TRUE} or \code{FALSE}.}
}
\value{
The processed MultiBlock object, with data matrices transformed and/or blocks/variables
  removed according to the supplied arguments. Blocks not listed in \code{blocks} are left
  unchanged.
}
\description{
Apply a custom function to transform a MultiBlock and/or select variables or blocks.
When multiple operations are supplied, the order of execution is:
\code{vars} subsetting, then \code{FUN}, then \code{FUN.SelectVars}, then \code{FUN.SelectBlocks}.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
# Normalize each block to 0-100 range
BY100 <- function(x) 100 * (x - min(x)) / (max(x) - min(x))
mb <- ProcessMultiBlock(mb, FUN = BY100)
# Keep only variables with non-zero variance
mb <- ProcessMultiBlock(mb, FUN.SelectVars = function(x) apply(x, 2, var) > 0)
# Remove blocks where all values are below 0.5
mb <- ProcessMultiBlock(mb, FUN.SelectBlocks = function(x) max(x) >= 0.5)
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{MultiBlock2Matrix}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-intra.R
\docType{data}
\name{intra}
\alias{intra}
\title{Intracellular metabolites in growth medium}
\format{
An object of class \code{matrix} (inherits from \code{array}) with 12 rows and 230 columns.
}
\source{
\doi{10.15252/msb.202010141}
}
\usage{
data(mouse_ds)
}
\description{
Intracellular metabolites in growth medium
}
\examples{
data(mouse_ds)
# MB <- MultiBlock(Data = list(RNAseq3 = RNAseq3,
#   lipids = lipids, intra = intra, extra = extra))
}
\author{
Radic Shechter et al. (2021) Molecular Systems Biology 17:e10141
(\doi{10.15252/msb.202010141})
}
\keyword{dataset}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{checkMultiBlock}
\alias{checkMultiBlock}
\title{checkMultiBlock}
\usage{
checkMultiBlock(object)
}
\arguments{
\item{object}{A MultiBlock object.}
}
\value{
TRUE if valid, or a character vector of error messages.
}
\description{
Validation function used during MultiBlock building.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_y.R
\name{ComDim_y}
\alias{ComDim_y}
\title{ComDim_y - Extending PLS-like supervised methods to multi-block data.}
\usage{
ComDim_y(
  MB = MB,
  y = y,
  ndim = NULL,
  FUN = FUN,
  nort = 0L,
  type = c("regression", "discriminant")[1],
  decisionRule = c("fixed", "max")[2],
  normalise = FALSE,
  scale.y = FALSE,
  threshold = 1e-10,
  loquace = FALSE,
  method = "FUN",
  cv.k = 7,
  ...
)
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{y}{The response: a numeric vector or matrix for regression
(\code{type = 'regression'}), or a class vector or dummy matrix for
discriminant analysis (\code{type = 'discriminant'}).}

\item{ndim}{Number of predictive Common Dimensions. If \code{NULL},
defaults to the number of blocks.}

\item{FUN}{The function used as the core of the ComDim analysis. It must
accept \code{W} (salience-weighted concatenated blocks, n x p matrix),
\code{y} (response), and \code{ndim} (number of components to compute)
as its first three named arguments, and return a named list with at least:
\describe{
  \item{scores}{X scores vector (length n) for the current component.}
  \item{P}{X loadings vector (length p) for the current component.}
  \item{W}{X weights vector (length p) for the current component.}
  \item{Q}{Y loadings vector for the current component.}
  \item{U}{Y scores vector (length n) for the current component.}
}
Optional return fields:
\describe{
  \item{y}{y as used internally by FUN (e.g. after centring/scaling),
    so that \code{ComDim_y} can detect automatic y-transformation and
    back-transform the Q loadings for correct B and B0 computation.}
  \item{orthoscores}{A numeric vector (length n) or single-column matrix
    of orthogonal X-scores. \strong{Required} when \code{nort = 1}.
    Only the first column is used. FUN does not need to change its
    behaviour between ort and predictive phases; if it always returns
    \code{orthoscores}, the framework uses it only during the ort phase
    and ignores it during the predictive phase.}
}}

\item{nort}{Number of orthogonal Common Dimensions to extract before the
predictive loop. Default \code{0} (no orthogonalization, pure PLS-like).
Only \code{nort = 0} or \code{nort = 1} are accepted; for multiple
orthogonal components use \code{ComDim_OPLS()}.
When \code{nort = 1} the function follows the same two-phase architecture
as \code{ComDim_OPLS}: the orthogonal component is removed from the
(lambda-weighted) concatenated blocks first, then predictive components are
computed from the orthogonally-deflated blocks. \code{FUN} must return
\code{output$orthoscores} (a numeric vector of length n, or a matrix whose
first column is used) when called with \code{ndim = 1}.
Cross-validation is automatically skipped when \code{nort > 0}.}

\item{type}{\code{'regression'} (default) or \code{'discriminant'}.}

\item{decisionRule}{Only used when \code{type = 'discriminant'}. If
\code{'fixed'}, samples are assigned to the class whose predicted score
exceeds \code{1/nclasses}; if \code{'max'}, the class with the highest
predicted score is chosen. Default \code{'max'}.}

\item{normalise}{To apply block normalisation. \code{FALSE} == no
(default), \code{TRUE} == yes.}

\item{scale.y}{Logical (default \code{FALSE}). When \code{TRUE} and
\code{type = 'regression'}, each column of Y is mean-centred and scaled to
unit variance before being passed to \code{FUN}. The stored B, B0, and
Ypred are back-transformed to the original Y scale, so downstream outputs
(R2Y, Q2, predictions) are always in the original units. Ignored when
\code{type = 'discriminant'} (dummy Y should not be scaled).}

\item{threshold}{Convergence threshold: iterations stop when the change
in the global score vector falls below this value (default \code{1e-10}).}

\item{loquace}{Display computation time at each step. \code{TRUE} == yes,
\code{FALSE} == no (default).}

\item{method}{A string label identifying the method (default: \code{'FUN'}).}

\item{cv.k}{Number of folds for k-fold cross-validation (default 7). Set
to 0 to skip CV. When \code{cv.k >= 2}, the \code{Q2} and \code{DQ2}
slots in the output reflect cross-validated predictive ability; otherwise
they reflect training-set fit. CV is skipped when \code{nort > 0}.}

\item{...}{Additional arguments passed to \code{FUN}.}
}
\value{
A \code{ComDim} object with the following slots:
\describe{
  \item{\code{Method}}{The label supplied via the \code{method} argument.}
  \item{\code{ndim}}{Number of predictive Common Dimensions extracted.}
  \item{\code{Q.scores}}{Global consensus scores matrix (\eqn{n \times
    ndim}).  Each column \eqn{\mathbf{q}_a} (unit-norm) is derived from
    the dominant left direction of FUN applied to the salience-weighted
    concatenated blocks.}
  \item{\code{T.scores}}{Named list of block-specific local scores
    (\eqn{n \times ndim} each).  Local loading
    \eqn{\mathbf{p}_{ba} = \tilde{\mathbf{X}}_b'\mathbf{q}_a} (computed
    on the ort-deflated block when \code{nort > 0}); local score
    \eqn{\mathbf{t}_{ba} =
    \tilde{\mathbf{X}}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
  \item{\code{P.loadings}}{Global loadings (\eqn{p_{tot} \times ndim}):
    \eqn{\mathbf{P} = \tilde{\mathbf{X}}'\mathbf{Q}}, where
    \eqn{\tilde{\mathbf{X}}} is the (optionally ort-deflated) mean-centred
    concatenated blocks.}
  \item{\code{Saliences}}{Block salience matrix (\eqn{ntable \times ndim}):
    \eqn{\lambda_{ba} =
    \mathbf{q}_a'\tilde{\mathbf{X}}_b\tilde{\mathbf{X}}_b'\mathbf{q}_a}.}
  \item{\code{R2X}}{Proportion of X variance captured by each predictive
    component (named vector, length \eqn{ndim}).  Let \eqn{\mathbf{t}_a}
    be the X-score vector returned by FUN for component \eqn{a}:
    \deqn{R2X_a = \|\mathbf{t}_a\|^4 \big/ \sum_k \|\mathbf{t}_k\|^4.}
    When \code{nort > 0}, the denominator also includes the orthogonal
    \eqn{\|\mathbf{t}_{ort,k}\|^4} terms, and the orthogonal R2X fractions
    are stored separately in \code{Orthogonal$R2X}.}
  \item{\code{R2Y}}{Cumulative Y-variance explained (named vector, length
    \eqn{ndim}):
    \deqn{R2Y_a = 1 - RSS_a / TSS_Y,}
    where \eqn{RSS_a} is the residual SS from an OLS regression of
    \eqn{\mathbf{Y}} on \eqn{[1, \mathbf{q}_1, \ldots, \mathbf{q}_a]}.
    \strong{Note:} \eqn{R2Y_a} is cumulative -- the total Y-variance
    explained by the first \eqn{a} components together, not the marginal
    contribution of component \eqn{a} alone.}
  \item{\code{Q2}}{Predictive Q2 per response column (regression) or per
    class (discriminant), named accordingly:
    \deqn{Q2 = 1 - PRESS / TSS_Y,}
    where \eqn{PRESS = \sum_i (\hat{y}_i - y_i)^2}.
    When \code{cv.k >= 2} and \code{nort = 0}: cross-validated (out-of-
    sample) predictions are used; otherwise training-set predictions.
    CV is automatically skipped when \code{nort > 0}.}
  \item{\code{DQ2}}{(Discriminant mode only) Discriminant Q2 per class,
    using only penalising residuals:
    \deqn{DQ2 = 1 - PRESSD / TSS_Y,}
    where \eqn{PRESSD} sums \eqn{\hat{y}_i^2} for class-0 samples with
    \eqn{\hat{y}_i > 0}, and \eqn{(\hat{y}_i - 1)^2} for class-1 samples
    with \eqn{\hat{y}_i < 1}.  Same cross-validation logic as \code{Q2}.}
  \item{\code{Singular}}{Squared L2 norm of the FUN X-score vector per
    component (\eqn{\|\mathbf{t}_a\|^2}), used to derive \code{R2X}.}
  \item{\code{VIP}}{Global total VIP (named vector, length \eqn{p_{tot}}):
    concatenation of \code{VIP.block[[b]]$tot} across blocks.  When
    \code{nort = 0}, uses the Wold formula; when \code{nort = 1}, tot
    combines predictive and orthogonal VIPs (see \code{VIP.block}).}
  \item{\code{VIP.block}}{Named list (one \code{data.frame} per block).
    When \code{nort = 0}: columns \code{p} and \code{tot} (= \code{p}),
    using the Wold formula:
    \deqn{VIPp_j = \sqrt{p_b \cdot
      \frac{\sum_a s_a \tilde{w}_{j,a}^2}{\sum_a s_a}},}
    where \eqn{s_a = \|\mathbf{t}_a\|^2\|\mathbf{q}_a\|^2} and
    \eqn{\tilde{w}_{j,a} = w_{j,a}/\|\mathbf{w}_a\|} is the L2-normalised
    \eqn{j}-th element of the \eqn{a}-th weight vector.
    When \code{nort = 1}: columns \code{p} (Wold, same as above),
    \code{o} (orthogonal VIP, loadings-based:
    \eqn{VIPo_j = \sqrt{p_b \cdot \sum_a s_{oa}\tilde{P}_{o,j,a}^2 /
    \sum_a s_{oa}}},
    where \eqn{s_{oa} = \|\mathbf{q}_{ort}[,a]\|^2} and
    \eqn{\tilde{\mathbf{P}}_o} is the column-L2-normalised block-slice of
    the ort loadings), and \code{tot}
    (\eqn{VIPtot_j = \sqrt{(VIPp_j^2 + VIPo_j^2)/2}}).
    Row names are variable names.}
  \item{\code{PLS.model}}{List with: \code{W} (X weight matrix collected
    from FUN, \eqn{p_{tot} \times ndim}); \code{B} (regression
    coefficients,
    \eqn{\mathbf{B} = \mathbf{W}(\mathbf{P}'\mathbf{W})^{-1}\mathbf{Q}'},
    in original Y units); \code{B0} (intercept,
    \eqn{\mathbf{B}_0 = \bar{\mathbf{y}} -
    \overline{\tilde{\mathbf{x}}}\mathbf{B}}); \code{Y} (original
    response matrix as supplied).
    Training-set predictions:
    \eqn{\hat{\mathbf{Y}} =
    \tilde{\mathbf{X}}\mathbf{B} + \mathbf{B}_0}.}
  \item{\code{cv}}{Cross-validation results when \code{cv.k >= 2} and
    \code{nort = 0} (empty list otherwise): \code{k}, \code{fold}
    (sample-to-fold vector), \code{Ypred} (\eqn{n \times ncol(Y)}
    out-of-sample predictions), \code{Q2} (CV Q2 per class/response),
    \code{DQ2} (mean CV DQ2, discriminant only),
    \code{DQ2.perclass} (CV DQ2 per class, discriminant only).}
  \item{\code{Orthogonal}}{When \code{nort > 0}: list with \code{nort},
    \code{Q.scores} (global ort scores, \eqn{n \times nort}, unit-norm),
    \code{T.scores} (block ort local scores, \eqn{n \times nort} each),
    \code{P.loadings.ort} (ort loadings, \eqn{p_{tot} \times nort}),
    \code{Saliences.ort} (\eqn{ntable \times nort}), and \code{R2X}
    (orthogonal X-variance fractions,
    \eqn{R2X_{ort,a} = \|\mathbf{t}_{ort,a}\|^4 / total}).
    Empty list when \code{nort = 0}.}
  \item{\code{Prediction}}{Training-set predictions: \code{Y.pred}
    (\eqn{n \times ncol(Y)}); for discriminant analysis also
    \code{decisionRule}, \code{trueClass}, \code{predClass} (data.frame),
    \code{Sensitivity} and \code{Specificity} (per class),
    \code{confusionMatrix} (named list of 2x2 matrices).}
  \item{\code{Mean}}{List with \code{MeanMB} (column means per block),
    \code{MeanY} (column means of Y), and \code{ScaleY} (column SDs of Y;
    all ones when \code{scale.y = FALSE}).}
  \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms for block
    normalisation.}
  \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
    mapping each row of \code{P.loadings} and each element of \code{VIP}
    to its block.}
  \item{\code{runtime}}{Total computation time in seconds.}
}
}
\description{
Extends any PLS-like method used for regression or discriminant purposes to
the multi-block field. The user provides a function (\code{FUN}) that
computes one predictive component from the salience-weighted concatenated
blocks; global scores, local scores, and loadings are then derived following
the traditional ComDim-PLS framework. Optionally, orthogonal components
returned by FUN (e.g. from an O-PLS wrapper) are captured. VIP scores and
k-fold cross-validation are also supported.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50) # 10 samples, 50 variables
b2 <- matrix(rnorm(800), 10, 80) # 10 samples, 80 variables
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))

## Example 1: ComDim-PLS (regression) ---------------------------------------
# Single-step NIPALS PLS wrapper (one predictive component per call).
# Note: 'tx' is used instead of 't' to avoid shadowing base::t().
fun.PLS <- function(W, y, ndim, ...) {
  output <- list()
  w <- t(W) \%*\% y / as.numeric(t(y) \%*\% y) # X weight (u = y, 1 step)
  w <- w / sqrt(sum(w^2)) # L2 normalise
  tx <- W \%*\% w # X score
  p <- t(W) \%*\% tx / as.numeric(t(tx) \%*\% tx) # X loading
  q <- t(y) \%*\% tx / as.numeric(t(tx) \%*\% tx) # Y loading
  u <- y \%*\% q / as.numeric(t(q) \%*\% q) # Y score
  output$scores <- as.vector(tx)
  output$P <- as.vector(p)
  output$W <- as.vector(w)
  output$Q <- as.vector(q)
  output$U <- as.vector(u)
  return(output)
}

y <- c(1, 1, 1, 1, 1, 5, 5, 5, 10, 10)
resultsPLS <- ComDim_y(mb,
  y = y, ndim = 2,
  type = "regression",
  FUN = fun.PLS,
  method = "PLS",
  cv.k = 0
)

## Example 2: ComDim-OPLS-DA (discriminant, nort = 1) ----------------------
# Thin wrapper around OPLS_NIPALS_DNR(), the package's NIPALS OPLS engine.
# All inputs (W, y, and any extra args such as 'threshold') are forwarded
# directly via '...'. Use this pattern when nort > 0; for nort = 0 the
# simpler PLS wrapper in Example 1 is sufficient (no orthoscores needed).
fun.OPLS <- function(W, y, ndim, ...) {
  res <- OPLS_NIPALS_DNR(W = W, y = y, ...)
  list(
    scores      = as.vector(res$t_pred),
    P           = as.vector(res$p),
    W           = as.vector(res$w_pred),
    Q           = as.vector(res$q),
    U           = as.vector(res$u),
    orthoscores = matrix(res$t_ort, ncol = 1)
  )
}

groups <- c(rep("A", 5), rep("B", 5))
resultsOPLS <- ComDim_y(mb,
  y = groups, ndim = 1,
  nort = 1,
  type = "discriminant",
  FUN = fun.OPLS,
  method = "OPLS-DA",
  cv.k = 0
)

## Example 3 (not run): ComDim-OPLS-DA via ropls ---------------------------
# Wrapping ropls::opls is also possible. Key points:
#   - Use orthoI = 1 (fixed) instead of NA so the output is predictable.
#   - Always return output$orthoscores; ComDim_y ignores it in phases
#     where ort has already been removed.
#   - Expand the single ropls Q loading to match the ncol(y_dummy) width.
\donttest{
if (requireNamespace("ropls", quietly = TRUE)) {
fun.OPLSDA.ropls <- function(W, y, ndim, ...) {
  output <- list()
  # Convert dummy matrix to ropls-compatible -1/+1 vector
  Y <- c(-1, 1)[apply(y, 1, function(x) match(1, x))]
  result <- tryCatch(
    ropls::opls(
      x = W, y = Y, predI = 1, orthoI = 1,
      fig.pdfC = "none", info.txtC = "none"
    ),
    error = function(e) {
      ropls::opls(
        x = W, y = Y, predI = 1, orthoI = 0,
        fig.pdfC = "none", info.txtC = "none"
      )
    }
  )
  output$scores <- result@scoreMN[, 1]
  output$P <- result@loadingMN[, 1]
  output$W <- result@weightMN[, 1]
  output$U <- result@uMN[, 1]
  # Expand the single ropls Q loading to match the 2-column dummy matrix:
  # loadings for class1 and class2 are antisymmetric in binary PLS-DA.
  output$Q <- c(-result@cMN[, 1], result@cMN[, 1])
  output$y <- result@suppLs$yModelMN # internal y (for scaling detection)
  # Orthogonal scores (used during the ort pre-loop when nort > 0)
  if (!is.null(result@orthoScoreMN) && ncol(result@orthoScoreMN) > 0) {
    output$orthoscores <- result@orthoScoreMN   # n x k matrix; col jj used for jj-th ort
  } else {
    output$orthoscores <- matrix(0, nrow = nrow(W), ncol = 1)
  }
  return(output)
}

b1_r <- matrix(rnorm(8 * 30), 8, 30)
b2_r <- matrix(rnorm(8 * 20), 8, 20)
mb_r <- MultiBlock(Data = list(b1 = b1_r, b2 = b2_r))
resultsOPLSDA <- ComDim_y(mb_r,
  y = c(rep("NI", 4), rep("OFF", 4)),
  ndim = 1,
  nort = 1,
  type = "discriminant",
  FUN = fun.OPLSDA.ropls,
  method = "OPLS-DA(ropls)",
  cv.k = 0
)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MultiAssayExperiment2MultiBlock.R
\name{MultiAssayExperiment2MultiBlock}
\alias{MultiAssayExperiment2MultiBlock}
\title{MultiAssayExperiment2MultiBlock}
\usage{
MultiAssayExperiment2MultiBlock(se, colData_samplenames = NULL, Batch = NULL)
}
\arguments{
\item{se}{A \code{MultiAssayExperiment} object.}

\item{colData_samplenames}{Character string giving the name of the column in \code{colData(se)}
that holds the sample names used for matching. When \code{NULL}, \code{colData} metadata is
not stored in the resulting MultiBlock.}

\item{Batch}{Character string giving the name of the column in \code{colData(se)} to use as
the Batch variable. Requires \code{colData_samplenames} to be set; an error is raised if
\code{Batch} is supplied but \code{colData_samplenames} is \code{NULL} or the column is not
found in \code{colData}.}
}
\value{
A \code{MultiBlock} object with one block per experiment in \code{se}.
}
\description{
Converts a MultiAssayExperiment into a MultiBlock. Samples are first intersected across all
experiments so that only common samples are retained.
}
\details{
Columns (samples) are intersected across all experiments via \code{intersectColumns} before
  conversion. Each experiment in the \code{MultiAssayExperiment} becomes one block in the
  MultiBlock (rows = samples, columns = features). Sample names are taken from the
  \code{colData} row names of \code{se}. Metadata from \code{colData} is stored only for the
  first block; subsequent blocks share the same sample order but carry no additional metadata.
  If \code{Batch} is specified, the corresponding column is extracted from \code{colData},
  removed from the metadata, and stored as the \code{Batch} slot of the MultiBlock.
}
\examples{
\donttest{
if (requireNamespace("MultiAssayExperiment", quietly = TRUE)) {
library(MultiAssayExperiment)
mae <- MultiAssayExperiment(
  experiments = ExperimentList(
    block1 = matrix(rnorm(50), nrow = 5, dimnames = list(paste0("s", 1:5), paste0("v", 1:10))),
    block2 = matrix(rnorm(30), nrow = 5, dimnames = list(paste0("s", 1:5), paste0("w", 1:6)))
  )
)
mb <- MultiAssayExperiment2MultiBlock(mae)
}
}
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{MultiBlock2MultiAssayExperiment}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_PCA.R, R/ComDim_PLS.R
\name{ColumnsPartition}
\alias{ColumnsPartition}
\title{ColumnsPartition}
\usage{
ColumnsPartition(Xn = Xn, Partitions = Partitions)

ColumnsPartition(Xn = Xn, Partitions = Partitions)
}
\arguments{
\item{Xn}{A block.}

\item{Partitions}{The number of partitions.}
}
\value{
A list of column-partitioned sub-matrices.

A list of column-partitioned sub-matrices.
}
\description{
Calculate vertical partitions in a block. Internal function of ComDim_PCA().

Calculate vertical partitions in a block. Internal function of ComDim_PLS().
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NAInfRemoveMultiBlock.R
\name{NAInfRemoveMultiBlock}
\alias{NAInfRemoveMultiBlock}
\title{NAInfRemoveMultiBlock}
\usage{
NAInfRemoveMultiBlock(
  MB,
  blocks = NULL,
  minfrac = 0.5,
  na.method = c("none", "zero", "median", "discard", "fixed.value", "fixed.value.all",
    "fixed.noise", "random.noise", "QRILC")[8],
  inf.method = c("none", "fixed.noise", "random.noise")[3],
  constant = 0,
  factor.NA = 0.5,
  sd.noise = 0.3,
  tune.sigma = 1,
  showWarning = TRUE
)
}
\arguments{
\item{MB}{The MultiBlock object.}

\item{blocks}{Blocks to process: a vector of integers or block names
(optional; all blocks are processed when omitted).}

\item{minfrac}{Minimum fraction of valid (non-NA, finite) values required
to retain a variable. Variables at or below this threshold are discarded.
Default: 0.5.}

\item{na.method}{Imputation method for remaining NA values after the
minfrac filter. One of: \code{'none'} (keep NAs), \code{'zero'},
\code{'median'}, \code{'discard'} (drop columns still containing NAs),
\code{'fixed.value'}, \code{'fixed.value.all'}, \code{'fixed.noise'},
\code{'random.noise'}, \code{'QRILC'}. Default: \code{'random.noise'}.}

\item{inf.method}{Replacement method for remaining Infinite values after
the minfrac filter. One of: \code{'none'} (keep Inf),
\code{'fixed.noise'}, \code{'random.noise'}.
Default: \code{'random.noise'}.}

\item{constant}{For \code{na.method = 'fixed.value'}: value to replace NAs.
For \code{na.method = 'fixed.value.all'}: value added to every element
(NAs are then set to this value). Default: 0.}

\item{factor.NA}{Noise factor used by \code{'fixed.noise'} and
\code{'random.noise'} for both NA and Inf imputation. For NA and -Inf:
the replacement mean equals \code{col_min * factor.NA}. For +Inf:
\code{col_max / factor.NA}. Default: 0.5.}

\item{sd.noise}{Standard-deviation factor for \code{'random.noise'}:
\code{SD = abs(mean * sd.noise)}. Default: 0.3.}

\item{tune.sigma}{Tuning scalar for \code{na.method = 'QRILC'}. Default:
1 (Gaussian complete-data assumption).}

\item{showWarning}{If \code{TRUE}, emit a warning when variables are
discarded by the minfrac filter. Default: \code{TRUE}.}
}
\value{
The processed MultiBlock object.
}
\description{
Remove NA and Infinite values from a MultiBlock object in a single pass.
Variables whose combined count of NA and Infinite values meets or exceeds
\code{minfrac * nrow} are discarded first. Remaining Infinite values are
then replaced according to \code{inf.method}; remaining NA values are
imputed according to \code{na.method}.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
b2[c(2, 3, 5), c(1, 2, 3)] <- NA
b2[c(1, 4), c(4, 5)] <- Inf
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
mb <- NAInfRemoveMultiBlock(mb, na.method = "zero", inf.method = "fixed.noise")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{sampleNames}
\alias{sampleNames}
\alias{sampleNames,MultiBlock-method}
\title{sampleNames}
\usage{
sampleNames(x)

\S4method{sampleNames}{MultiBlock}(x)
}
\arguments{
\item{x}{A MultiBlock object.}
}
\value{
A vector with the sample names.
}
\description{
Return the sample names of a MultiBlock object.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
rownames(b1) <- paste0("s", 1:10)
mb <- MultiBlock(Data = list(b1 = b1))
sampleNames(mb) # "s1" ... "s10"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SelectFeaturesRW.R
\name{SelectFeaturesRW}
\alias{SelectFeaturesRW}
\title{SelectFeaturesRW}
\usage{
SelectFeaturesRW(
  RW = RW,
  results = results,
  ndim = NULL,
  blocks = NULL,
  threshold_cor = 1,
  threshold_cov = 1,
  mean.RW = TRUE,
  plots = "NO"
)
}
\arguments{
\item{RW}{The object used as input in the ComDim analysis.}

\item{results}{The output object obtained in the ComDim analysis.}

\item{ndim}{The number of the component for which the important variables are to be identified.}

\item{blocks}{A vector with the indices or the names for the replicate blocks of the same data type.}

\item{threshold_cor}{The "times" parameter used to calculate the threshold in the following formula: cor(variable) > times * sd(cor(variables)). Minimal value that can be assigned to threshold_cor is 1.}

\item{threshold_cov}{The "times" parameter used to calculate the threshold in the following formula: cov(variable) > times * sd(cov(variables)). Minimal value that can be assigned to threshold_cov is 1.}

\item{mean.RW}{Logical value to indicate whether the RW data must be mean-centered (TRUE) or not (FALSE).}

\item{plots}{Parameter to indicate whether S-plots (covariance vs. correlation with the Q scores)
must be produced. Possible values are \code{"NO"} for no plots, \code{"separated"} for displaying
one S-plot per block individually (pausing between plots), and \code{"together"} to arrange all
S-plots side by side in a single grid. In all plot variants, variables selected as important are
highlighted in red and labelled with their variable name. Requires the \code{ggplot2} package;
\code{"together"} additionally requires \code{gridExtra}.}
}
\value{
A named list with two elements:
  \describe{
    \item{\code{$positive}}{Integer indices (named by variable name) of the important variables
      presenting a positive relationship with the Q scores (positive covariance, positive
      correlation, and positive local P-loading) across all specified blocks.}
    \item{\code{$negative}}{Integer indices (named by variable name) of the important variables
      presenting a negative relationship with the Q scores (negative covariance, negative
      correlation, and negative local P-loading) across all specified blocks.}
  }
  When \code{plots} is not \code{"NO"}, S-plots are also displayed as a side effect: each plot
  shows covariance on the x-axis and correlation on the y-axis, with selected variables
  highlighted in red.
}
\description{
Finds the important variables presenting a coordinated response across all specified replicate-blocks
for a given ComDim component.
}
\details{
The function applies an S-plot approach to identify variables that are both strongly covarying
  and strongly correlated with the Q scores of the chosen component. For each block in
  \code{blocks}, covariance (\code{s1}) and correlation (\code{s2}) of every variable with the
  (pseudo-inverse-scaled) Q scores are computed. A variable is considered important in a block if
  its absolute covariance exceeds \code{threshold_cov * sd(s1)} \emph{and} its absolute correlation
  exceeds \code{threshold_cor * sd(s2)}. Only variables that satisfy both criteria in \emph{all}
  specified blocks simultaneously are returned. The sign of the local P-loadings is used to
  separate variables into positive and negative groups.
}
\examples{
\donttest{
b1 <- matrix(rnorm(500), 10, 50)
batch_b1 <- rep(1, 10)
b2 <- matrix(rnorm(800), 30, 80)
batch_b2 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
mb <- MultiBlock(
  Data = list(b1 = b1, b2 = b2),
  Batch = list(b1 = batch_b1, b2 = batch_b2),
  ignore.names = TRUE, ignore.size = TRUE
)
rw <- SplitRW(mb)
results <- ComDim_PCA(rw, 2)
# Identify important variables for component 1 across replicate blocks 2, 3, and 4
features <- SelectFeaturesRW(RW = rw, results = results, ndim = 1, blocks = c(2, 3, 4))
# Use stricter thresholds and display S-plots side by side
features <- SelectFeaturesRW(RW = rw, results = results, ndim = 1, blocks = c(2, 3, 4),
                             threshold_cor = 2, threshold_cov = 2, plots = "together")
}
}
\seealso{
\code{\link{SplitRW}}, \code{\link{ComDim_PCA}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MultiBlock2MultiAssayExperiment.R
\name{MultiBlock2MultiAssayExperiment}
\alias{MultiBlock2MultiAssayExperiment}
\title{MultiBlock2MultiAssayExperiment}
\usage{
MultiBlock2MultiAssayExperiment(MB, MSEmetadata = NULL)
}
\arguments{
\item{MB}{A \code{MultiBlock} object.}

\item{MSEmetadata}{An optional list of unstructured metadata describing the overall content of
the MultiAssayExperiment (stored in its \code{metadata} slot). Pass \code{NULL} (default) to
omit experiment-level metadata.}
}
\value{
A \code{MultiAssayExperiment} object with one experiment per block in \code{MB}.
}
\description{
Converts a MultiBlock into a MultiAssayExperiment. Each block becomes one experiment; sample
names, batch information, and metadata are carried over into the \code{colData} of the result.
}
\details{
Each block in \code{MB} is transposed (features x samples) and stored as a named matrix in
  the \code{ExperimentList}. Row names are set to the variable names and column names to the
  sample names of the MultiBlock. The \code{colData} is constructed from \code{MB@Samples};
  any \code{Metadata} and \code{Batch} information present in the MultiBlock is appended as
  additional columns. A \code{sampleMap} is generated mapping every sample to every experiment
  using the same primary and column names.
}
\examples{
\donttest{
if (requireNamespace("MultiAssayExperiment", quietly = TRUE)) {
library(MultiAssayExperiment)
b1 <- matrix(rnorm(50), 5, 10, dimnames = list(paste0("s", 1:5), paste0("v", 1:10)))
b2 <- matrix(rnorm(30), 5,  6, dimnames = list(paste0("s", 1:5), paste0("w", 1:6)))
mb <- MultiBlock(Data = list(block1 = b1, block2 = b2))
mae <- MultiBlock2MultiAssayExperiment(mb)
mae <- MultiBlock2MultiAssayExperiment(mb, MSEmetadata = list(study = "example"))
}
}
}
\seealso{
\code{\link{MultiBlock}}, \code{\link{MultiAssayExperiment2MultiBlock}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{variableNames}
\alias{variableNames}
\alias{variableNames,MultiBlock-method}
\title{variableNames}
\usage{
variableNames(x, ...)

\S4method{variableNames}{MultiBlock}(x, block)
}
\arguments{
\item{x}{A MultiBlock object.}

\item{...}{Not used. Present for S4 generic dispatch compatibility.}

\item{block}{Optional. A vector of block names or indices to retrieve.
When omitted, variable names for all blocks are returned as a named list.
When a single block is specified, a plain vector is returned.}
}
\value{
A named list of variable-name vectors (all blocks), or a single
  vector when exactly one block is requested.
}
\description{
Return the variable names of a MultiBlock object.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
variableNames(mb) # named list: b1 = 1:50, b2 = 1:80
variableNames(mb, "b1") # 1:50
variableNames(mb, 1:2) # same as variableNames(mb)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim-classes.R
\name{blockNames}
\alias{blockNames}
\alias{blockNames,MultiBlock-method}
\title{blockNames}
\usage{
blockNames(x, ...)

\S4method{blockNames}{MultiBlock}(x, slot = "Data")
}
\arguments{
\item{x}{A MultiBlock object.}

\item{...}{Not used. Present for S4 generic dispatch compatibility.}

\item{slot}{A string: \code{"Data"} (default), \code{"Batch"}, or
\code{"Metadata"}, indicating which slot to retrieve names from.}
}
\value{
A character vector with the block names of the requested slot.
}
\description{
Return the block names of a MultiBlock object.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
b2 <- matrix(rnorm(800), 10, 80)
mb <- MultiBlock(Data = list(b1 = b1, b2 = b2))
blockNames(mb) # c("b1", "b2")
blockNames(mb, "Data") # same
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ComDim_PLS.R
\name{ComDim_PLS}
\alias{ComDim_PLS}
\title{ComDim_PLS}
\usage{
ComDim_PLS(
  MB = MB,
  y = y,
  ndim = NULL,
  method = c("PLS-DA", "PLS-R"),
  decisionRule = c("fixed", "max")[2],
  normalise = FALSE,
  scale.y = FALSE,
  threshold = 1e-10,
  loquace = FALSE,
  CompMethod = "Normal",
  Partitions = 1,
  cv.k = 7
)
}
\arguments{
\item{MB}{A MultiBlock object.}

\item{y}{The Y-block to use in the PLS model as dependent data. A class vector or a dummy matrix.}

\item{ndim}{Number of Common Dimensions.}

\item{method}{PLS-DA or PLS-R.}

\item{decisionRule}{Only used if method is set to PLS-DA. If 'fixed', samples are assigned to the class with Y-hat above 1/nclasses. If 'max', samples are assigned to the class with the highest Y-hat.}

\item{normalise}{To apply block normalisation. FALSE == no (default), TRUE == yes.
When TRUE each block is mean-centred and then divided by its Frobenius norm so
that all blocks have unit total inertia entering the ComDim loop.
Has no effect on the Y-block; use \code{scale.y} for that.}

\item{scale.y}{Logical (default FALSE). When TRUE and \code{method = 'PLS-R'},
each column of Y is mean-centred and scaled to unit variance before the PLS
fit. The stored B, B0, and Ypred are back-transformed to the original Y scale,
so downstream outputs (R2Y, Q2, predictions) are always in the original units.
Ignored when \code{method = 'PLS-DA'} (dummy Y should not be scaled).}

\item{threshold}{The threshold limit to stop the iterations. If the "difference of fit" < threshold (1e-10 as default).}

\item{loquace}{To display the calculation times. TRUE == yes, FALSE == no (default).}

\item{CompMethod}{To speed up the analysis for really big multi-blocks. 'Normal' (default), 'Kernel', 'PCT', 'Tall' or 'Wide'.}

\item{Partitions}{To speed up the analysis for really big multi-blocks. This parameter is used if CompMethod is 'Tall' or 'Wide'.}

\item{cv.k}{Number of folds for k-fold cross-validation (default 7). Set to 0 to skip CV. When cv.k >= 2, Q2 and DQ2 in the output reflect cross-validated predictive ability; otherwise they reflect training-set fit (R2).}
}
\value{
A \code{ComDim} object. All slots are populated. Key slots:
\describe{
  \item{\code{Method}}{\code{"PLS-DA"} or \code{"PLS-R"}.}
  \item{\code{ndim}}{Number of Common Dimensions extracted.}
  \item{\code{Q.scores}}{Global consensus PLS scores (\eqn{n \times ndim}).
    Each column \eqn{\mathbf{q}_a} (unit-norm) is the dominant left
    singular vector from the NIPALS PLS applied to the salience-weighted
    concatenated blocks.}
  \item{\code{T.scores}}{Named list of block-specific local scores
    (\eqn{n \times ndim} each).  Local loading
    \eqn{\mathbf{p}_{ba} = \mathbf{X}_b'\mathbf{q}_a}; local score
    \eqn{\mathbf{t}_{ba} =
    \mathbf{X}_b\,\mathbf{p}_{ba}(\mathbf{p}_{ba}'\mathbf{p}_{ba})^{-1}}.}
  \item{\code{P.loadings}}{Global loadings (\eqn{p_{tot} \times ndim}):
    \eqn{\mathbf{P} = \mathbf{X}'\mathbf{Q}}, where \eqn{\mathbf{X}} is
    the mean-centred (and optionally normalised) concatenated blocks.}
  \item{\code{Saliences}}{Block salience matrix (\eqn{ntable \times ndim}):
    \eqn{\lambda_{ba} =
    \mathbf{q}_a'\mathbf{X}_b\mathbf{X}_b'\mathbf{q}_a}.}
  \item{\code{R2X}}{Proportion of X variance captured by each component
    (named vector, length \eqn{ndim}).  Let \eqn{\mathbf{t}_a} be the
    NIPALS PLS X-score vector for component \eqn{a} on the salience-
    weighted blocks; then
    \deqn{R2X_a = \|\mathbf{t}_a\|^4 \big/ \sum_k \|\mathbf{t}_k\|^4.}}
  \item{\code{R2Y}}{Cumulative Y-variance explained (named vector, length
    \eqn{ndim}).  \eqn{R2Y_a} is the \eqn{R^2} from an OLS regression of
    \eqn{\mathbf{Y}} on the first \eqn{a} global scores with an intercept:
    \deqn{R2Y_a = 1 - RSS_a / TSS_Y,}
    where \eqn{RSS_a} is the residual SS when predicting \eqn{\mathbf{Y}}
    from \eqn{[1, \mathbf{q}_1, \ldots, \mathbf{q}_a]}.
    \strong{Note:} \eqn{R2Y_a} is cumulative — it reflects the total
    Y-variance explained by the first \eqn{a} components together,
    not the marginal contribution of component \eqn{a} alone.}
  \item{\code{Q2}}{Predictive Q2 per response column (PLS-R) or per class
    (PLS-DA), named accordingly:
    \deqn{Q2 = 1 - PRESS / TSS_Y,}
    where \eqn{PRESS = \sum_i (\hat{y}_i - y_i)^2} and
    \eqn{TSS_Y = \sum_i (y_i - \bar{y})^2}.
    When \code{cv.k >= 2}, \eqn{\hat{y}_i} are out-of-sample
    cross-validated predictions; otherwise training-set predictions are
    used (i.e. Q2 = R2Y for the full model).}
  \item{\code{DQ2}}{(PLS-DA only) Discriminant Q2 per class.  Only
    penalising residuals contribute to the sum:
    \deqn{DQ2 = 1 - PRESSD / TSS_Y,}
    where \eqn{PRESSD} sums \eqn{\hat{y}_i^2} for class-0 samples with
    \eqn{\hat{y}_i > 0}, and \eqn{(\hat{y}_i - 1)^2} for class-1 samples
    with \eqn{\hat{y}_i < 1}.  Same cross-validation logic as \code{Q2}.}
  \item{\code{Singular}}{Squared L2 norm of the NIPALS PLS score vector
    per component (\eqn{\|\mathbf{t}_a\|^2}), used to derive
    \code{R2X}.}
  \item{\code{VIP}}{Global VIP scores (named numeric vector, length
    \eqn{p_{tot}}) using the Wold formula:
    \deqn{VIP_j = \sqrt{p_{tot}
      \cdot \frac{\sum_a s_a \tilde{w}_{j,a}^2}{\sum_a s_a}},}
    where \eqn{s_a = \|\mathbf{t}_a\|^2 \|\mathbf{q}_a\|^2},
    \eqn{\tilde{w}_{j,a} = w_{j,a} / \|\mathbf{w}_a\|} is the
    L2-normalised \eqn{j}-th element of the \eqn{a}-th NIPALS weight
    vector, and \eqn{p_{tot}} is the total number of variables.}
  \item{\code{VIP.block}}{Named list (one \code{data.frame} per block)
    with columns \code{p} (per-block predictive VIP computed with block
    size \eqn{p_b} instead of \eqn{p_{tot}}) and \code{tot} (= \code{p}
    for PLS; included for consistency with OPLS output).  Row names are
    variable names.}
  \item{\code{PLS.model}}{List with: \code{W} (NIPALS X weight matrix,
    \eqn{p_{tot} \times ndim}); \code{B} (regression coefficients,
    \eqn{p_{tot} \times ncol(Y)},
    \eqn{\mathbf{B} =
    \mathbf{W}(\mathbf{P}'\mathbf{W})^{-1}\mathbf{Q}'},
    in original Y units); \code{B0} (intercept vector, length
    \eqn{ncol(Y)},
    \eqn{\mathbf{B}_0 = \bar{\mathbf{y}} - \bar{\mathbf{x}}\mathbf{B}});
    \code{Y} (original response matrix as supplied).
    Training-set Y predictions:
    \eqn{\hat{\mathbf{Y}} = \mathbf{X}\mathbf{B} + \mathbf{B}_0}.}
  \item{\code{cv}}{Cross-validation results when \code{cv.k >= 2} (empty
    list otherwise): \code{k} (number of folds), \code{fold}
    (sample-to-fold vector), \code{Ypred} (\eqn{n \times ncol(Y)} matrix
    of out-of-sample predictions), \code{Q2} (CV Q2 per class/response),
    \code{DQ2} (mean CV DQ2 across classes, PLS-DA only),
    \code{DQ2.perclass} (CV DQ2 per class, PLS-DA only).}
  \item{\code{Prediction}}{Training-set predictions: \code{Y.pred}
    (\eqn{n \times ncol(Y)}); for PLS-DA also \code{decisionRule},
    \code{trueClass} (character vector), \code{predClass} (data.frame),
    \code{Sensitivity} and \code{Specificity} (named per class),
    \code{confusionMatrix} (named list of 2x2 matrices, one per class).}
  \item{\code{Mean}}{List with \code{MeanMB} (column means per block),
    \code{MeanY} (column means of Y before any scaling), and
    \code{ScaleY} (column SDs of Y; all ones when \code{scale.y = FALSE}).}
  \item{\code{Norm}}{List with \code{NormMB}: Frobenius norms for block
    normalisation.}
  \item{\code{variable.block}}{Character vector (length \eqn{p_{tot}})
    mapping each row of \code{P.loadings} and each element of \code{VIP}
    to its block.}
  \item{\code{runtime}}{Total computation time in seconds.}
}
}
\description{
Finding common dimensions in multi-block datasets using PLS.
}
\examples{
b1 <- matrix(rnorm(500), 10, 50)
batch_b1 <- rep(1, 10)
b2 <- matrix(rnorm(2400), 30, 80)
batch_b2 <- c(rep(1, 10), rep(2, 10), rep(3, 10))
mb <- MultiBlock(
  Samples = list(
    b1 = paste0("samples_", 1:10),
    b2 = rep(paste0("samples_", 1:10), 3)
  ),
  Data = list(b1 = b1, b2 = b2),
  Batch = list(b1 = batch_b1, b2 = batch_b2),
  ignore.size = TRUE
)
rw <- SplitRW(mb)
y <- scale(1:10, center = TRUE)
results.plsr <- ComDim_PLS(rw, y, 2, method = "PLS-R")
groups <- c(rep("A", 5), rep("B", 5))
results.plsda <- ComDim_PLS(rw, y = groups, 2, method = "PLS-DA")
}
\references{
Jouan-Rimbaud Bouveresse D, Rutledge DN (2024).
  A synthetic review of some recent extensions of ComDim.
  \emph{Journal of Chemometrics}, 38(5), e3454.
  \doi{10.1002/cem.3454}
}
